Package: scrapper
Version: 1.7.0
Date: 2026-04-24
Authors@R: person("Aaron", "Lun", role=c("cre", "aut"), email="infinite.monkeys.with.keyboards@gmail.com")
Title: Bindings to C++ Libraries for Single-Cell Analysis
Description: 
    Implements R bindings to C++ code for analyzing single-cell (expression) data, mostly from various libscran libraries. 
    Each function performs an individual step in the single-cell analysis workflow, ranging from quality control to clustering and marker detection.
    Additional wrappers are provided for easy construction of end-to-end workflows involving Bioconductor objects like SingleCellExperiments.
License: MIT + file LICENSE
Imports: methods, Rcpp, beachmat (>= 2.25.1), S4Vectors, SparseArray,
        DelayedArray, BiocNeighbors (>= 1.99.0), parallel
Suggests: testthat, knitr, rmarkdown, BiocStyle, Matrix, IRanges,
        SummarizedExperiment, SingleCellExperiment, scRNAseq,
        org.Mm.eg.db, scater, igraph
LinkingTo: Rcpp, assorthead (>= 1.5.16), beachmat, BiocNeighbors,
        Rigraphlib
biocViews: Normalization, RNASeq, Software, GeneExpression,
        Transcriptomics, SingleCell, BatchEffect, QualityControl,
        DifferentialExpression, FeatureExtraction, PrincipalComponent,
        Clustering
SystemRequirements: C++17, GNU make
URL: https://github.com/libscran/scrapper
BugReports: https://github.com/libscran/scrapper/issues
VignetteBuilder: knitr
Encoding: UTF-8
RoxygenNote: 7.3.3
git_url: https://git.bioconductor.org/packages/scrapper
git_branch: devel
git_last_commit: 8005249
git_last_commit_date: 2026-04-28
Repository: Bioconductor 3.24
Date/Publication: 2026-04-28
NeedsCompilation: yes
Packaged: 2026-04-29 01:07:39 UTC; biocbuild
Author: Aaron Lun [cre, aut]
Maintainer: Aaron Lun <infinite.monkeys.with.keyboards@gmail.com>

# Generated by roxygen2: do not edit by hand

export(LogNormalizedMatrix)
export(LogNormalizedMatrixSeed)
export(aggregateAcrossCells)
export(aggregateAcrossCells.se)
export(aggregateAcrossGenes)
export(aggregateAcrossGenes.se)
export(aggregateColData)
export(analyze)
export(analyze.se)
export(buildSnnGraph)
export(centerSizeFactors)
export(chooseHighlyVariableGenes)
export(choosePseudoCount)
export(chooseRnaHvgs.se)
export(clusterGraph)
export(clusterGraph.se)
export(clusterKmeans)
export(clusterKmeans.se)
export(combineFactors)
export(computeAdtQcMetrics)
export(computeBlockWeights)
export(computeClrm1Factors)
export(computeCrisprQcMetrics)
export(computeRnaQcMetrics)
export(computeRnaQcMetricsWithAltExps)
export(convertAnalyzeResults)
export(correctMnn)
export(correctMnn.se)
export(countGroupsByBlock)
export(filterAdtQcMetrics)
export(filterCrisprQcMetrics)
export(filterRnaQcMetrics)
export(fitVarianceTrend)
export(formatComputeAdtQcMetricsResult)
export(formatComputeCrisprQcMetricsResult)
export(formatComputeRnaQcMetricsResult)
export(formatModelGeneVariancesResult)
export(formatScoreMarkersResult)
export(getTestAdtData.se)
export(getTestCrisprData.se)
export(getTestRnaData.se)
export(modelGeneVariances)
export(normalizeAdtCounts.se)
export(normalizeCounts)
export(normalizeCrisprCounts.se)
export(normalizeRnaCounts.se)
export(previewMarkers)
export(quickAdtQc.se)
export(quickCrisprQc.se)
export(quickRnaQc.se)
export(reportGroupMarkerStatistics)
export(runAllNeighborSteps)
export(runAllNeighborSteps.se)
export(runPca)
export(runPca.se)
export(runTsne)
export(runTsne.se)
export(runUmap)
export(runUmap.se)
export(sanitizeSizeFactors)
export(scaleByNeighbors)
export(scaleByNeighbors.se)
export(scoreGeneSet)
export(scoreGeneSet.se)
export(scoreMarkers)
export(scoreMarkers.se)
export(subsampleByNeighbors)
export(suggestAdtQcThresholds)
export(suggestCrisprQcThresholds)
export(suggestRnaQcThresholds)
export(summarizeEffects)
export(testEnrichment)
export(tsnePerplexityToNeighbors)
exportClasses(LogNormalizedMatrix)
exportClasses(LogNormalizedMatrixSeed)
exportMethods(DelayedArray)
exportMethods(dim)
exportMethods(dimnames)
exportMethods(extract_array)
exportMethods(extract_sparse_array)
exportMethods(initializeCpp)
exportMethods(is_sparse)
exportMethods(type)
import(methods)
importClassesFrom(DelayedArray,DelayedArray)
importClassesFrom(S4Vectors,DataFrame)
importClassesFrom(S4Vectors,List)
importFrom(BiocNeighbors,AnnoyParam)
importFrom(BiocNeighbors,buildIndex)
importFrom(BiocNeighbors,defineBuilder)
importFrom(BiocNeighbors,findKNN)
importFrom(DelayedArray,DelayedArray)
importFrom(DelayedArray,extract_array)
importFrom(DelayedArray,new_DelayedArray)
importFrom(DelayedArray,t)
importFrom(DelayedArray,type)
importFrom(Rcpp,sourceCpp)
importFrom(S4Vectors,"metadata<-")
importFrom(S4Vectors,DataFrame)
importFrom(S4Vectors,I)
importFrom(S4Vectors,List)
importFrom(S4Vectors,cbind)
importFrom(S4Vectors,make_zero_col_DFrame)
importFrom(S4Vectors,mcols)
importFrom(S4Vectors,metadata)
importFrom(S4Vectors,setValidity2)
importFrom(SparseArray,extract_sparse_array)
importFrom(SparseArray,is_sparse)
importFrom(beachmat,getExecutor)
importFrom(beachmat,initializeCpp)
importFrom(beachmat,tatami.column.sums)
importFrom(beachmat,tatami.dim)
importFrom(beachmat,tatami.subset)
importFrom(methods,as)
importFrom(methods,is)
importFrom(methods,new)
importFrom(parallel,makeCluster)
importFrom(parallel,parLapply)
importFrom(parallel,stopCluster)
importFrom(utils,head)
useDynLib(scrapper)

# R bindings to C++ single-cell analysis libraries

|Environment|Status|
|---|---|
|[BioC-release](https://bioconductor.org/packages/release/bioc/html/scrapper.html)|[![Release OK](https://bioconductor.org/shields/build/release/bioc/scrapper.svg)](https://bioconductor.org/checkResults/release/bioc-LATEST/scrapper/)|
|[BioC-devel](https://bioconductor.org/packages/devel/bioc/html/scrapper.html)|[![Devel OK](https://bioconductor.org/shields/build/devel/bioc/scrapper.svg)](https://bioconductor.org/checkResults/devel/bioc-LATEST/scrapper/)|

The **scrapper** package implements R bindings to C++ code for analyzing single-cell (expression) data, mostly various [**libscran**](https://github.com/libscran) libraries. 
Each function performs an individual step in the single-cell analysis workflow, ranging from quality control to clustering and marker detection.
It is mostly intended for other Bioconductor package developers to build more user-friendly end-to-end workflows.

## ----echo=FALSE, results="hide", message=FALSE--------------------------------
require(knitr)
opts_chunk$set(error=FALSE, message=FALSE, warning=FALSE)

library(BiocStyle)
self <- Biocpkg("scrapper")

## -----------------------------------------------------------------------------
library(scRNAseq)
sce.z <- ZeiselBrainData()
sce.z

## -----------------------------------------------------------------------------
library(scrapper)
is.mito <- grep("^mt-", rownames(sce.z))
results <- analyze.se(
    sce.z,
    rna.qc.subsets=list(Mito=is.mito),
    num.threads=2 # limit number of threads to keep R CMD check happy.
)
results$x # most outputs are stored in a copy of the input SingleCellExperiment. 

## -----------------------------------------------------------------------------
library(scater)
plotReducedDim(results$x, "TSNE", colour_by="graph.cluster")

## -----------------------------------------------------------------------------
# Ordering by the median AUC across pairwise comparisons involving the cluster.
previewMarkers(results$markers$rna[[1]], order.by="auc.median")

## -----------------------------------------------------------------------------
test <- sce.z
test <- quickRnaQc.se(test, subsets=list(mt=is.mito), num.threads=2)
test <- test[,test$keep]
test <- normalizeRnaCounts.se(test, size.factors=test$sum)
test <- chooseRnaHvgs.se(test, num.threads=2)
test <- runPca.se(test, features=rowData(test)$hvg, num.threads=2)
test <- runAllNeighborSteps.se(test, num.threads=2)
markers <- scoreMarkers.se(test, groups=test$clusters, num.threads=2)

## -----------------------------------------------------------------------------
sce.t <- TasicBrainData()

# Only using the common genes in both datasets.
common <- intersect(rownames(sce.z), rownames(sce.t))
combined <- combineCols(sce.t[common,], sce.z[common,])
block <- rep(c("tasic", "zeisel"), c(ncol(sce.t), ncol(sce.z)))

## -----------------------------------------------------------------------------
# No mitochondrial genes in the combined set, so we'll just skip it.
blocked_res <- analyze.se(combined, block=block, num.threads=2)

## -----------------------------------------------------------------------------
plotReducedDim(blocked_res$x, "TSNE", colour_by="block")

## -----------------------------------------------------------------------------
table(blocked_res$x$graph.cluster, blocked_res$x$block)

## -----------------------------------------------------------------------------
previewMarkers(blocked_res$markers$rna[[1]])

## -----------------------------------------------------------------------------
aggregates <- aggregateAcrossCells.se(
    blocked_res$x,
    list(cluster=blocked_res$x$graph.cluster, dataset=blocked_res$x$block)
)
aggregates

## -----------------------------------------------------------------------------
sce.s <- StoeckiusHashingData(mode=c("human", "adt1", "adt2"))
sce.s <- sce.s[,1:5000]

# Combining the various ADT-related alternative experiments into a single object.
altExp(sce.s, "all.adts") <- rbind(altExp(sce.s, "adt1"), altExp(sce.s, "adt2"))
altExp(sce.s, "adt1") <- altExp(sce.s, "adt2") <- NULL

sce.s

## -----------------------------------------------------------------------------
is.mito <- grepl("^MT-", rownames(sce.s))
is.igg <- grepl("^IgG", rownames(altExp(sce.s, 'all.adts')))
multi_res <- analyze.se(
    sce.s,
    adt.altexp="all.adts",
    rna.qc.subsets=list(mito=is.mito),
    adt.qc.subsets=list(igg=is.igg),
    num.threads=2
)

## -----------------------------------------------------------------------------
plotReducedDim(multi_res$x, "UMAP", colour_by="graph.cluster")

## -----------------------------------------------------------------------------
# By default, markers are sorted by the mean Cohen's d.
# We just replace the column name when printing the dataframe here.
previewMarkers(multi_res$markers$adt[[5]], c(effect="cohens.d.mean"))

## -----------------------------------------------------------------------------
# Using the delta-detected to focus on silent-to-expressed genes.
previewMarkers(multi_res$markers$rna[[5]], order.by="delta.detected.mean")

## -----------------------------------------------------------------------------
sessionInfo()

#' Quick quality control for CRISPR data in a SummarizedExperiment
#'
#' Quickly compute quality control (QC) metrics, thresholds and filters from CRISPR data in a \link[SummarizedExperiment]{SummarizedExperiment}.
#' This calls \code{\link{computeCrisprQcMetrics}} on an assay in a SummarizedExperiment,
#' followed by \code{\link{suggestCrisprQcThresholds}} and \code{\link{filterCrisprQcMetrics}} to identify high-quality cells.
#' 
#' @param x A \link[SummarizedExperiment]{SummarizedExperiment} object or one of its subclasses.
#' Rows correspond to CRISPR guides and columns correspond to cells.
#' @param num.threads Number of threads, to pass to \code{\link{computeCrisprQcMetrics}}.
#' @param block Block assignment for each cell, to pass to \code{\link{suggestCrisprQcThresholds}} and \code{\link{filterCrisprQcMetrics}}.
#' @param thresholds List containing pre-defined thresholds for each QC metric,
#' see the return value of \code{\link{suggestRnaQcThresholds}} for the expected format.
#' @param more.suggest.args Named list of additional arguments to pass to \code{\link{suggestCrisprQcThresholds}}.
#' @param assay.type Integer or string specifying the assay of \code{x} containing the CRISPR count matrix.
#' @param output.prefix String containing a prefix to add to the names of the \code{link[SummarizedExperiment]{colData}} columns containing the output statistics.
#' @param meta.name String containing the name of the \code{\link[S4Vectors]{metadata}} entry containing additional outputs like the filtering thresholds.
#' If \code{NULL}, additional outputs are not reported. 
#' @param compute.res \link[S4Vectors]{DataFrame} returned by \code{\link[scrapper]{computeCrisprQcMetrics}}.
#' 
#' @return
#' For \code{quickCrisprQc.se}, \code{x} is returned with additional columns added to its \code{\link[SummarizedExperiment]{colData}}.
#' Each column contains per-cell values for one of the QC metrics, see \code{\link[scrapper]{computeCrisprQcMetrics}} for details.
#' The suggested thresholds are stored as a list in \code{\link[S4Vectors]{metadata}}.
#' The \code{colData} also contains a \code{keep} column, specifying which cells are to be retained.
#'
#' For \code{formatComputeCrisprQcMetricsResult}, a \link[S4Vectors]{DataFrame} is returned with the per-cell QC metrics.
#'
#' @author Aaron Lun
#'
#' @examples
#' library(SingleCellExperiment)
#' sce <- altExp(getTestCrisprData.se(), "CRISPR Guide Capture")
#' sce <- quickCrisprQc.se(sce)
#' colData(sce)[,c("sum", "detected", "max.value", "max.index")]
#' metadata(sce)$qc$thresholds
#' summary(sce$keep)
#' 
#' @export
#' @importFrom S4Vectors cbind metadata metadata<-
quickCrisprQc.se <- function( 
    x,
    num.threads = 1,
    thresholds = NULL,
    block = NULL,
    more.suggest.args = list(),
    assay.type = "counts",
    output.prefix = NULL,
    meta.name = "qc"
) {
    metrics <- computeCrisprQcMetrics(SummarizedExperiment::assay(x, assay.type, withDimnames=FALSE), num.threads=num.threads)

    if (is.null(thresholds)) {
        thresholds <- .call(
            suggestCrisprQcThresholds,
            list(metrics),
            list(block=block),
            more.suggest.args
        )
    }

    keep <- filterCrisprQcMetrics(thresholds, metrics, block=block)

    df <- formatComputeCrisprQcMetricsResult(metrics)
    df$keep <- keep
    colnames(df) <- paste0(output.prefix, colnames(df))
    SummarizedExperiment::colData(x) <- cbind(SummarizedExperiment::colData(x), df)

    if (!is.null(meta.name)) {
        metadata(x)[[meta.name]] <- list(thresholds=thresholds)
    }

    x
}

#' @export
#' @rdname quickCrisprQc.se
formatComputeCrisprQcMetricsResult <- function(compute.res) {
    # Maybe we'll add something more later.
    compute.res
}

# library(testthat); library(scrapper); source("test-se_quickCrisprQc.R")

library(SummarizedExperiment)
mat <- matrix(rpois(1000, 1), ncol=10)
se <- SummarizedExperiment(list(counts=mat))

test_that("quickCrisprQc.se works as expected", {
    out <- quickCrisprQc.se(se)
    expect_type(out$sum, "double")
    expect_type(out$detected, "integer")
    expect_type(out$max.value, "double")
    expect_type(out$max.index, "integer")
    expect_type(out$keep, "logical")
    expect_type(metadata(out)$qc$thresholds, "list")

    out2 <- quickCrisprQc.se(se, output.prefix="WHEE.")
    expect_identical(out2$WHEE.sum, out$sum)
    expect_identical(out2$WHEE.detected, out$detected)
    expect_identical(out2$WHEE.max.value, out$max.value)
    expect_identical(out2$WHEE.max.value, out$max.value)
    expect_identical(out2$WHEE.keep, out$keep)

    out4 <- quickCrisprQc.se(se, meta.name=NULL)
    expect_null(metadata(out4)$qc)
})

#' Summarize pairwise effect sizes for each group
#'
#' For each group, summarize the effect sizes for all pairwise comparisons to other groups.
#' This yields a set of summary statistics that can be used to rank marker genes for each group.
#'
#' @param effects A 3-dimensional numeric containing the effect sizes from each pairwise comparison between groups.
#' The extents of the first two dimensions are equal to the number of groups, while the extent of the final dimension is equal to the number of genes.
#' The entry \code{[i, j, k]} represents the effect size from the comparison of group \code{j} against group \code{i} for gene \code{k}.
#' See also the output of \code{\link{scoreMarkers}} with \code{all.pairwise=TRUE}.
#' @param compute.summary.min Boolean specifying whether to compute the minimum as a summary statistic.
#' @param compute.summary.mean Boolean specifying whether to compute the mean as a summary statistic. 
#' @param compute.summary.median Boolean specifying whether to compute the median as a summary statistic. 
#' @param compute.summary.max Boolean specifying whether to compute the maximum as a summary statistic. 
#' @param compute.summary.quantiles Numeric scalars containing the probabilities of quantiles to compute as summary statistics. 
#' If \code{NULL}, no quantiles are computed.
#' @param compute.summary.min.rank Boolean specifying whether to compute the mininum rank as a summary statistic. 
#' @param num.threads Integer scalar specifying the number of threads to use.
#'
#' @return List of \link[S4Vectors]{DataFrame}s containing summary statistics for the effect sizes.
#' Each DataFrame corresponds to a group, each row corresponds to a gene, and each column contains a summary statistic.
#' If \code{compute.summary.quantiles} is provided, the \code{"quantile"} column is a nested DataFrame where each column coresponds to a probability in \code{compute.summary.quantiles}.
#'
#' @details
#' Each summary statistic can be used to prioritize different sets of marker genes for the group of interest, by ranking them in decreasing order according to said statistic:
#' \itemize{
#' \item \code{min} contains the minimum effect size across all comparisons involving the group of interest.
#' Genes with large values are upregulated in all comparisons.
#' As such, it is the most stringent summary as markers will only have large values if they are uniquely upregulated in the group of interest compared to every other group.
#' \item \code{mean} contains the mean effect size across all comparisons involving the group of interest.
#' Genes with large values are upregulated on average compared to the other groups.
#' This is a good general-purpose summary statistic.
#' \item \code{median} contains the median effect size across all comparisons involving the group of interest.
#' Genes with large values are upregulated compared to most (i.e., at least 50% of) other groups.
#' Compared to the mean, this is more robust to outlier effects but less sensitive to strong effects in a minority of comparisons.
#' \item \code{max} contains the maximum effect size across all comparisons involving the group of interest.
#' Using this to define markers will focus on genes that are upregulated in at least one comparison.
#' As such, it is the least stringent summary as markers can achieve large values if they are upregulated in the group of interest compared to any one other group.
#' \item \code{quantile[[P]]} contains the quantile P across all comparisons involving the group of interest.
#' This is a generalization of the minimum, median and maximum for arbitrary quantile probabilities.
#' For example, a large \code{quantile[["20"]]} would mean that the gene is upregulated in the group of interest compared to 80% of other groups.
#' }
#'
#' The exact definition of \dQuote{large} depends on the choice of effect size.
#' For signed effects like Cohen's d, delta-mean and delta-detected, the value must be positive to be considered \dQuote{large}.
#' For the AUC, a value greater than 0.5 is considered \dQuote{large}.
#' This interpretation is also affected by the choice of \code{threshold=} used to compute each effect size in \code{\link{scoreMarkers}},
#' e.g., a negative Cohen's d cannot be interpreted as downregulation when the threshold is positive. 
#' 
#' The \code{min.rank} is a more exotic summary statistic, containing the minimum rank for each gene across all comparisons involving the group of interest.
#' This is defined by ranking the effect sizes across genes within each comparison, and then taking the minimum of these ranks across comparisons.
#' Taking all genes with \code{min.rank <= T} will yield a set containing the top \code{T} genes from each comparison.
#' The idea is to ensure that there are at least \code{T} genes that can distinguish the group of interest from any other group.
#'
#' \code{NaN} effect sizes are allowed, e.g., if two groups do not exist in the same block for a blocked analysis in \code{\link{scoreMarkers}} with \code{block=}. 
#' This function will ignore \code{NaN} values when computing each summary.
#' If all effects are NaN for a particular group, the summary statistic will also be \code{NaN}.
#'
#' @author Aaron Lun
#' 
#' @examples
#' # Mocking a matrix:
#' library(Matrix)
#' x <- round(abs(rsparsematrix(1000, 100, 0.1) * 100))
#' normed <- normalizeCounts(x, size.factors=centerSizeFactors(colSums(x)))
#'
#' g <- sample(letters[1:4], ncol(x), replace=TRUE)
#' effects <- scoreMarkers(normed, g, all.pairwise=TRUE)
#'
#' summarized <- summarizeEffects(effects$cohens.d)
#' str(summarized)
#'
#' @seealso
#' The \code{summarize_effects} function in \url{https://libscran.github.io/scran_markers/}.
#'
#' \code{\link{scoreMarkers}}, to compute the pairwise effects in the first place.
#'
#' @export
summarizeEffects <- function(
    effects,
    compute.summary.min=TRUE,
    compute.summary.mean=TRUE,
    compute.summary.median=TRUE,
    compute.summary.max=TRUE,
    compute.summary.quantiles=NULL,
    compute.summary.min.rank=TRUE,
    num.threads=1
) {
    dm <- dim(effects)
    ngroups <- dm[1]
    ngenes <- dm[3]

    out <- summarize_effects(
        ngenes,
        ngroups,
        effects,
        compute_summary_mean=compute.summary.mean,
        compute_summary_min=compute.summary.min,
        compute_summary_median=compute.summary.median,
        compute_summary_max=compute.summary.max,
        compute_summary_quantiles=compute.summary.quantiles,
        compute_summary_min_rank=compute.summary.min.rank,
        num_threads=num.threads
    )
    dmn <- dimnames(effects)
    names(out) <- dmn[[1]]

    format_summary_output(out, ngenes, dmn[[3]], compute.summary.quantiles)
}

#' @importFrom S4Vectors DataFrame
format_summary_output <- function(raw, ngenes, rownames, quantiles) {
    for (i in seq_along(raw)) {
        current <- raw[[i]]

        if ("quantile" %in% names(current)) {
            names(current$quantile) <- as.character(quantiles)
            to.add <- DataFrame(current$quantile, check.names=FALSE)
            current$quantile <- logical(ngenes)
        } else {
            to.add <- NULL
        }

        if (length(current)) {
            df <- DataFrame(current)
        } else {
            df <- DataFrame(matrix(0, ngenes, 0))
        }
        rownames(df) <- rownames

        if (!is.null(to.add)) {
            df$quantile <- to.add
        }

        raw[[i]] <- df
    }
    raw
}

# library(testthat); library(scrapper); source("test-summarizeEffects.R")

set.seed(999)
library(Matrix)
x <- abs(rsparsematrix(1000, 100, 0.1))
x@x <- jitter(x@x)

test_that("summarizeEffects works as expected", {
    g <- sample(4, ncol(x), replace=TRUE)
    summ <- scoreMarkers(x, g, min.rank.limit=nrow(x))
    full <- scoreMarkers(x, g, all.pairwise=TRUE)
    expect_equal(summarizeEffects(full$cohens.d), summ$cohens.d)
    expect_equal(summarizeEffects(full$auc), summ$auc)
})

test_that("summarizeEffects works with quantiles", {
    g <- sample(4, ncol(x), replace=TRUE)
    full <- scoreMarkers(x, g, all.pairwise=TRUE)
    res <- summarizeEffects(full$cohens.d, compute.summary.quantiles=c(0, 0.5, 1))

    for (i in seq_len(max(g))) {
        expect_identical(res[[i]]$median, res[[i]]$quantile[["0.5"]])
        expect_identical(res[[i]]$min, res[[i]]$quantile[["0"]])
        expect_identical(res[[i]]$max, res[[i]]$quantile[["1"]])
    }
})

test_that("summarizeEffects works with everything disabled", {
    g <- sample(4, ncol(x), replace=TRUE)
    full <- scoreMarkers(x, g, all.pairwise=TRUE)
    res <- summarizeEffects(
        full$cohens.d,
        compute.summary.min=FALSE,
        compute.summary.mean=FALSE,
        compute.summary.median=FALSE,
        compute.summary.max=FALSE,
        compute.summary.min.rank=FALSE
    )

    for (i in seq_len(max(g))) {
        expect_identical(ncol(res[[i]]), 0L)
        expect_identical(nrow(res[[i]]), nrow(x))
    }
})

#' K-means clustering
#'
#' Perform k-means clustering with a variety of different initialization and refinement algorithms.
#'
#' @param x Matrix-like object where rows are dimensions and columns are cells.
#' This is typically a dense double-precision matrix containing a low-dimensional representation from, e.g., \code{\link{runPca}}.
#' However, any matrix representation supported by \code{\link[beachmat]{initializeCpp}} can also be used.
#' @param k Integer scalar specifying the number of clusters.
#' @param init.method String specifying the initialization method for the centers:
#' \itemize{
#' \item \code{"var-part"} uses variance partitioning as described by Su and Dy (2007).
#' The dataset is repeatedly split along the dimension of greatest variance until \code{k} partitions are formed, the centroids of which form the initial clusters.
#' This approach is slower than the others but fully deterministic.
#' \item \code{"kmeans++"} uses the weighted sampling method described by Arthur and Vassilvitskii (2007).
#' \code{k} points are sampled with probability based on the smallest distance to any previously sampled point.
#' This improves the likelihood of choosing initial centroids that are far apart from each other.
#' \item \code{"random"} initialization involves choosing \code{k} random points as the initial centers.
#' This is the simplest and fastest method but may not yield good starting points. 
#' }
#' @param refine.method String specifying the refinement method.
#' \itemize{
#' \item \code{"lloyd"} uses Lloyd's algorithm, which performs a batch update in each iteration.
#' This is simple and amenable to parallelization but may not converge.
#' \item \code{"hartigan-wong"} uses the Hartigan-Wong algorithm, which transfers points between clusters to optimize the drop in the within-cluster sum of squares.
#' This is slower but has a greater chance of convergence.
#' }
#' @param var.part.optimize.partition Logical scalar indicating whether each partition boundary should be optimized to reduce the sum of squares in the child partitions.
#' This is slower but improves the quality of the partition.
#' Only used if \code{init.method = "var.part"}.
#' @param var.part.size.adjustment Numeric scalar between 0 and 1, specifying the adjustment to the cluster size when selecting the next cluster to partition.
#' Setting this to 0 or 1 will select the cluster with the highest variance or sum of squares, respectively, for partitioning.
#' In other words, a value of 0 will ignore the cluster size while setting a value of 1 will generally cause larger clusters to be selected. 
#' Only used if \code{init.method = "var.part"}.
#' @param lloyd.iterations Integer scalar specifying the maximum number of iterations for the Lloyd algorithm.
#' Larger values increase the chance of convergence at the cost of increasing compute time.
#' Only used if \code{refine.method = "lloyd"}.
#' @param hartigan.wong.iterations Integer scalar specifying the maximum number of iterations for the Hartigan-Wong algorithm.
#' Larger values increase the chance of convergence at the cost of increasing compute time.
#' Only used if \code{refine.method = "hartigan-wong"}.
#' @param hartigan.wong.quick.transfer.iterations Integer scalar specifying the maximum number of quick transfer iterations for the Hartigan-Wong algorithm.
#' Larger values increase the chance of convergence at the cost of increasing compute time.
#' Only used if \code{refine.method = "hartigan-wong"}.
#' @param hartigan.wong.quit.quick.transfer.failure Logical scalar indicating whether to quit the Hartigan-Wong algorithm upon convergence failure during quick transfer iterations.
#' Setting this to \code{FALSE} gives the algorithm another chance to converge by attempting another optimal transfer iteration, at the cost of more compute time.
#' If \code{TRUE}, the function follows the same behavior as R's \code{\link{kmeans}}.
#' Only used if \code{refine.method = "hartigan-wong"}.
#' @param seed Integer scalar specifying the seed for random number generation.
#' Only used if \code{init.method = "random"} or \code{"kmeans++"}.
#' @param warn Boolean specifying whether a warning should be emitted if the k-means algorithm failed to converge.
#' @param num.threads Integer scalar specifying the number of threads to use.
#' 
#' @return 
#' By default, a list is returned containing:
#' \itemize{
#' \item \code{clusters}, a factor containing the cluster assignment for each cell.
#' The number of levels is no greater than \code{k}, where each level is an integer that refer to a column of \code{centers}.
#' \item \code{centers}, a numeric matrix with the coordinates of the cluster centroids (dimensions in rows, centers in columns).
#' The number of columns is no greater than \code{k}.
#' Empty clusters are automatically removed.
#' \item \code{iterations}, an integer scalar specifying the number of refinement iterations that were performed.
#' \item \code{status}, an integer scalar specifying the completion status of the algorithm.
#' A value of zero indicates success while the meaning of any non-zero value depends on the choice of \code{refine.method}:
#' \itemize{
#' \item For Lloyd, a value of 2 indicates convergence failure.
#' \item For Hartigan-Wong, a value of 2 indicates convergence failure in the optimal transfer iterations.
#' A value of 4 indicates convergence failure in the quick transfer iterations when \code{hartigan.wong.quit.quick.transfer.failure = TRUE}.
#' }
#' }
#'
#' @seealso
#' \url{https://libscran.github.io/kmeans/}, which describes the various initialization and refinement algorithms in more detail.
#'
#' \code{\link{clusterKmeans.se}}, for k-means clustering on a \link[SingleCellExperiment]{SingleCellExperiment}.
#'
#' @references
#' Hartigan JA. and Wong MA (1979).
#' Algorithm AS 136: A K-means clustering algorithm.
#' \emph{Applied Statistics} 28, 100-108.
#'
#' Arthur D and Vassilvitskii S (2007).
#' k-means++: the advantages of careful seeding.
#' \emph{Proceedings of the eighteenth annual ACM-SIAM symposium on Discrete algorithms} 1027-1035.
#'
#' Su T and Dy JG (2007).
#' In Search of Deterministic Methods for Initializing K-Means and Gaussian Mixture Clustering.
#' \emph{Intelligent Data Analysis} 11, 319-338.
#'
#' @author Aaron Lun
#'
#' @examples
#' x <- t(as.matrix(iris[,1:4]))
#' clustering <- clusterKmeans(x, k=3)
#' table(clustering$clusters, iris[,"Species"])
#'
#' @export
#' @importFrom beachmat initializeCpp
clusterKmeans <- function(
    x,
    k,
    init.method = c("var-part", "kmeans++", "random"),
    refine.method = c("hartigan-wong", "lloyd"),
    var.part.optimize.partition = TRUE,
    var.part.size.adjustment = 1,
    lloyd.iterations = 100,
    hartigan.wong.iterations = 10,
    hartigan.wong.quick.transfer.iterations = 50,
    hartigan.wong.quit.quick.transfer.failure = FALSE,
    seed=5489L,
    warn=TRUE,
    num.threads=1
) {
    .checkSEX(x, "clusterKmeans.se")

    tatami <- FALSE
    x0 <- x
    if (!is.matrix(x) || !is.numeric(x)) {
        x0 <- initializeCpp(x)
        tatami <- TRUE
    }

    output <- cluster_kmeans(
        x0,
        k,
        tatami=tatami,
        init_method=match.arg(init.method),
        refine_method=match.arg(refine.method),
        var_part_optimize_partition=var.part.optimize.partition,
        var_part_size_adjustment=var.part.size.adjustment, 
        lloyd_iterations=lloyd.iterations,
        hartigan_wong_iterations=hartigan.wong.iterations,
        hartigan_wong_quick_transfer_iterations=hartigan.wong.quick.transfer.iterations,
        hartigan_wong_quit_quick_transfer_failure=hartigan.wong.quit.quick.transfer.failure,
        seed=seed, 
        nthreads=num.threads
    )

    if (output$status != 0 && warn) {
        warning("convergence failure for k-means")
    }

    output$clusters <- factor(output$clusters + 1L)
    rownames(output$centers) <- rownames(x)

    output 
}

# library(testthat); library(scrapper); source("test-clusterKmeans.R")

x <- t(as.matrix(iris[,1:4]))

test_that("clusterKmeans works in basic mode", {
    clustering <- clusterKmeans(x, 10)
    expect_identical(length(clustering$clusters), ncol(x))
    expect_identical(length(unique(clustering$clusters)), 10L)
    expect_identical(ncol(clustering$centers), 10L)

    # Randomness should be fully controlled.
    again <- clusterKmeans(x, 10)
    expect_identical(clustering, again)

    # same results when wrapped in a DelayedArray.
    library(DelayedArray)
    delayed <- clusterKmeans(DelayedArray(x), 10)
    expect_identical(clustering, delayed)

    expect_error(clusterKmeans(SummarizedExperiment::SummarizedExperiment(x)), "not supported")
})

test_that("clusterKmeans automatically ignores too many clusters", {
    clustering <- clusterKmeans(x, 10000)
    num.unique <- nrow(unique(iris))
    expect_identical(length(clustering$clusters), ncol(x))
    expect_identical(length(unique(clustering$clusters)), num.unique)
    expect_identical(ncol(clustering$centers), num.unique)
    expect_equal(clustering$centers[,clustering$clusters], x)
})

test_that("clusterKmeans works with different methods", {
    clustering <- clusterKmeans(x, 5, init.method="random")
    expect_identical(length(clustering$clusters), ncol(x))
    expect_identical(length(unique(clustering$clusters)), 5L)
    expect_identical(ncol(clustering$centers), 5L)

    clustering <- clusterKmeans(x, 5, init.method="kmeans++", refine.method="lloyd")
    expect_identical(length(clustering$clusters), ncol(x))
    expect_identical(length(unique(clustering$clusters)), 5L)
    expect_identical(ncol(clustering$centers), 5L)
})

#' Graph-based clustering of cells in a SingleCellExperiment 
#'
#' Construct a shared-nearest neighbor (SNN) graph from an existing low-dimensional embedding
#' by calling \code{\link{buildSnnGraph}} on a reduced dimension entry in a \link[SingleCellExperiment]{SingleCellExperiment}.
#' Then, apply community detection algorithms to obtain clusters of cells with \code{\link{clusterGraph}}.
#'
#' @param x A \link[SingleCellExperiment]{SingleCellExperiment} object or one of its subclasses.
#' Rows correspond to genomic features and columns correspond to cells.
#' @param num.neighbors Number of neighbors for constructing the graph, passed to \code{\link{buildSnnGraph}}. 
#' @param num.threads Number of threads for graph construction, passed to \code{\link{buildSnnGraph}}.
#' @param more.build.args Named list of further arguments to be passed to \code{\link{buildSnnGraph}}.
#' @param method Clustering method to use, passed to \code{\link{clusterGraph}}. 
#' @param resolution Resolution for the community detection method in \code{\link{clusterGraph}}.
#' This is either passed to \code{multilevel.resolution} or \code{leiden.resolution} depending on \code{method}.
#' @param more.cluster.args Named list of further arguments to be passed to \code{\link{clusterGraph}}.
#' @param reddim.type Integer or string specifying the existing embedding in the \code{\link[SingleCellExperiment]{reducedDim}} of \code{x}.
#' Alternatively, a named integer or character vector of length 1,
#' where the name specifies an alternative experiment of \code{x} and the value is the name/index of a \code{reducedDim} entry in that alternative experiment.
#' @param output.name String containing the name of the column of the \code{\link[SummarizedExperiment]{colData}} in which to store the cluster assignments.
#' @param meta.name String containing the name of the \code{\link[S4Vectors]{metadata}} entry in which to store extra clustering output.
#' If \code{NULL}, no extra clustering output is stored. 
#' @param graph.name String containing the name of the \code{\link[S4Vectors]{metadata}} entry in which to store the SNN graph.
#' If \code{NULL}, the SNN graph is not stored. 
#'
#' @return \code{x} is returned with the cluster assignment for each cell stored in the \code{colData}.
#' Additional clustering output is stored in the \code{metadata}.
#'
#' @author Aaron Lun
#' @examples
#' sce <- getTestRnaData.se("pca")
#' sce <- clusterGraph.se(sce)
#' table(sce$clusters)
#'
#' @export
clusterGraph.se <- function(
    x,
    num.neighbors = 10,
    num.threads = 1,
    more.build.args = list(),
    method = "multilevel",
    resolution = NULL,
    more.cluster.args = list(),
    reddim.type = "PCA",
    output.name = "clusters",
    meta.name = NULL,
    graph.name = NULL
) {
    graph.out <- .call(
        buildSnnGraph,
        list(.getTransposedReddim(x, reddim.type)),
        list(num.neighbors=num.neighbors, num.threads=num.threads, as.pointer=is.null(graph.name)),
        more.build.args
    )
    x <- .addBuildGraphResults(x, graph.out, graph.name=graph.name)

    res.args <- list(method=method)
    if (!is.null(resolution)) {
        res.args$multilevel.resolution <- resolution
        res.args$leiden.resolution <- resolution
    }

    clust.out <- .call(
        clusterGraph,
        list(graph.out),
        res.args,
        more.cluster.args
    )

    .addClusterGraphResults(x, clust.out, output.name=output.name, meta.name=meta.name)
}

#' @importFrom S4Vectors metadata metadata<-
.addBuildGraphResults <- function(x, graph, graph.name) {
    if (!is.null(graph.name)) {
        metadata(x)[[graph.name]] <- graph
    }
    x
}

#' @importFrom S4Vectors metadata metadata<-
.addClusterGraphResults <- function(x, res, output.name, meta.name) {
    SummarizedExperiment::colData(x)[[output.name]] <- res$membership
    if (!is.null(meta.name)) {
        res$membership <- NULL
        metadata(x)[[meta.name]] <- res
    }
    x
}

# library(testthat); library(scrapper); source("test-se_clusterGraph.R")

library(SingleCellExperiment)
set.seed(9999)
sce <- SingleCellExperiment(list(counts = matrix(0, 0, 200)))
reducedDim(sce, "PCA") <- matrix(runif(1000), nrow=200)

test_that("clusterGraph.se works as expected", {
    sce <- clusterGraph.se(sce)
    expect_true(is.factor(sce$clusters))

    sce2 <- clusterGraph.se(sce, resolution=2)
    expect_false(identical(sce$clusters, sce2$clusters))

    scel <- clusterGraph.se(sce, method="leiden")
    expect_false(identical(sce$clusters, scel$clusters))

    sce <- clusterGraph.se(sce, graph.name="graph", meta.name="cluster")
    expect_type(metadata(sce)$cluster, "list")
    expect_null(metadata(sce)$cluster$membership)
    expect_type(metadata(sce)$graph, "list")
})

#' Compute size factors for ADT counts
#'
#' Compute size factors from an ADT count matrix using the CLRm1 method.
#' This is a variant of the centered log-ratio (CLR) method, where the size factors are defined from the geometric mean of counts within each cell.
#'
#' @param x A matrix-like object containing ADT count data.
#' Rows correspond to tags and columns correspond to cells.
#' @param num.threads Number of threads to use.
#'
#' @return Numeric vector containing the CLRm1 size factor for each cell.
#' Note that these size factors are not centered and should be passed through, e.g., \code{\link{centerSizeFactors}} before normalization.
#'
#' @author Aaron Lun
#'
#' @seealso
#' \url{https://github.com/libscran/clrm1}, for a description of the CLRm1 method.
#'
#' \code{\link{normalizeAdtCounts.se}}, which computes CLRm1 factors prior to normalization.
#'
#' @examples
#' library(Matrix)
#' x <- abs(rsparsematrix(1000, 100, 0.1) * 10)
#' head(computeClrm1Factors(x))
#'
#' @export
#' @importFrom beachmat initializeCpp
computeClrm1Factors <- function(x, num.threads=1) {
    .checkSEX(x, "normalizeAdtCounts.se")
    ptr <- initializeCpp(x, .check.na=FALSE)
    compute_clrm1_factors(ptr, num_threads=num.threads)
}

# library(testthat); library(scrapper); source("test-computeClrm1Factors.R")

set.seed(10000)
library(Matrix)
x <- abs(rsparsematrix(20, 1000, 0.1) * 10)

test_that("computeClrm1Factors works against a reference", {
    ref <- expm1(Matrix::colMeans(log1p(x[Matrix::rowSums(x)>0,,drop=FALSE])))
    obs <- computeClrm1Factors(x)
    expect_equal(ref, obs)

    par <- computeClrm1Factors(x, num.threads=2)
    expect_equal(obs, par)

    expect_error(computeClrm1Factors(SummarizedExperiment::SummarizedExperiment(x)), "not supported")
})

#' Aggregate expression across gene sets in a SummarizedExperiment
#'
#' Aggregate expression values across sets of genes for each cell,
#' by calling \code{\link{aggregateAcrossGenes}} on an assay in a \link[SummarizedExperiment]{SummarizedExperiment}.
#'
#' @param x A \link[SummarizedExperiment]{SummarizedExperiment} object or one of its subclasses.
#' Rows correspond to genes and columns correspond to cells.
#' @param sets List of gene sets, see \code{\link{aggregateAcrossGenes}} for more details.
#'
#' Alternatively, \code{sets} may be a \link[S4Vectors]{List} subclass,
#' in which case the \code{\link[S4Vectors]{mcols}} are used as the \code{\link[SummarizedExperiment]{rowData}} of the output object.
#' Weighted gene sets can be represented by a list of \pkg{DataFrames} or a \link[IRanges]{DataFrameList},
#' where each DataFrame contains two columns, i.e., the gene identities and the associated weights.
#' @param num.threads Number of threads, passed to \code{\link{aggregateAcrossGenes}}.
#' @param more.aggr.args Named list of additional arguments to pass to \code{\link{aggregateAcrossGenes}}.
#' @param assay.type Integer or string specifying the assay of \code{x} to be aggregated.
#' @param output.name String specifying the assay name in the output object.
#' Defaults to \code{assay.type} if it is a string, otherwise \code{"aggregated"}.
#'
#' @return A \link[SummarizedExperiment]{SummarizedExperiment} with number of rows equal to the number of gene sets.
#' The lone assay contains the aggregated values for each gene set for all cells.
#' The \code{\link[SummarizedExperiment]{colData}} is the same as that of \code{x}.
#'
#' @author Aaron Lun
#' @examples
#' library(SingleCellExperiment)
#' sce <- getTestRnaData.se("norm")
#'
#' library(org.Mm.eg.db)
#' set.df <- select(
#'     org.Mm.eg.db,
#'     keytype="GO",
#'     keys=c(
#'         "GO:0048709", # oligodendrocyte differentiation 
#'         "GO:0048699", # neuron development
#'         "GO:0048143"  # astrocyte activation 
#'     ),
#'     columns="SYMBOL"
#' )
#' sets <- splitAsList(set.df$SYMBOL, set.df$GO)
#'
#' aggregated <- aggregateAcrossGenes.se(sce, sets)
#' aggregated
#' assay(aggregated)[,1:10]
#'
#' @export
#' @importFrom methods is
#' @importFrom S4Vectors mcols
aggregateAcrossGenes.se <- function(
    x,
    sets,
    num.threads = 1,
    more.aggr.args = list(),
    assay.type = "logcounts",
    output.name = NULL
) {
    rd <- NULL
    if (is(sets, "List")) {
        rd <- mcols(sets)
        if (!is.null(rd)) {
            rownames(rd) <- names(sets) # just in case
        }
        sets <- as.list(sets)
    }

    for (i in seq_along(sets)) {
        if (is(sets[[i]], "List")) {
            sets[[i]] <- as.list(sets[[i]])
        }
    }

    vecs <- .call(
        aggregateAcrossGenes,
        list(SummarizedExperiment::assay(x, assay.type)),
        list(sets=sets, num.threads=num.threads),
        more.aggr.args
    )
    if (length(vecs)) {
        mat <- do.call(rbind, vecs)
    } else {
        mat <- matrix(0, 0, ncol(x))
    }

    assays <- list(mat)
    if (!is.null(output.name)) {
        names(assays) <- output.name
    } else if (is.character(assay.type)) {
        names(assays) <- assay.type
    } else {
        names(assays) <- "aggregated"
    }

    out <- SummarizedExperiment::SummarizedExperiment(assays, colData=SummarizedExperiment::colData(x))
    if (!is.null(rd)) {
        SummarizedExperiment::rowData(out) <- rd
    }
    out
}

# library(testthat); library(scrapper); source("test-se_aggregateAcrossGenes.R")

library(SummarizedExperiment)
mat <- matrix(runif(1000), ncol=10)
se <- SummarizedExperiment(list(logcounts=mat))
se$stuff <- sample(LETTERS[1:5], ncol(se), replace=TRUE)
se$whee <- sample(c(TRUE, FALSE), ncol(se), replace=TRUE)

sets <- list(foo=sample(nrow(mat), 10), bar=sample(nrow(mat), 20), stuff=sample(nrow(mat), 30))

test_that("aggregateAcrossGenes.se works as expected", {
    out <- aggregateAcrossGenes.se(se, sets)
    expect_identical(assayNames(out), "logcounts")
    expect_identical(rownames(out), names(sets))
    expect_identical(colData(out), colData(se))

    # Checking the names.
    out2 <- aggregateAcrossGenes.se(se, sets, assay.type=1)
    expect_identical(assayNames(out2), "aggregated")
    out2 <- aggregateAcrossGenes.se(se, sets, assay.type=1, output.name="FOO")
    expect_identical(assayNames(out2), "FOO")

    # Works with empty inputs.
    zeroed <- aggregateAcrossGenes.se(se, sets[0])
    expect_identical(nrow(zeroed), 0L)
})

test_that("aggregateAcrossGenes.se works with Lists", {
    out <- aggregateAcrossGenes.se(se, sets)

    lsets <- List(sets)
    lout <- aggregateAcrossGenes.se(se, lsets)
    expect_identical(lout, out)

    mcols(lsets) <- DataFrame(description=c("fooish", "barish", "stuffish"))
    lout2 <- aggregateAcrossGenes.se(se, lsets)
    expect_identical(rowData(lout2), mcols(lsets))

    # Cleans up DataFrames.
    wsets <- lapply(sets, function(x) DataFrame(index=x, weight=runif(length(x))))
    wout <- aggregateAcrossGenes.se(se, wsets)
    wref <- aggregateAcrossGenes.se(se, lapply(wsets, as.list))
    expect_identical(wout, wref)
})

#' Fit a mean-variance trend
#'
#' Fit a trend to the per-gene variances with respect to their means, typically from normalized and log-transformed expression values.
#'
#' @param means Numeric vector containing the mean (log-)expression for each gene.
#' @param variances Numeric vector containing the variance in the (log-)expression for each gene.
#' @param mean.filter Logical scalar indicating whether to filter on the means before trend fitting.
#' The assumption is that there is a bulk of low-abundance genes that are uninteresting and should be removed to avoid skewing the windows of the LOWESS smoother.
#' @param min.mean Numeric scalar specifying the minimum mean of genes to use in trend fitting.
#' Genes with lower means do not participate in the LOWESS fit, to ensure that windows are not skewed towards the majority of low-abundance genes.
#' Instead, the fitted values for these genes are defined by extrapolating the left edge of the fitted trend is extrapolated to the origin.
#' The default value is chosen based on the typical distribution of means of log-expression values across genes.
#' Only used if \code{mean.filter=TRUE}.
#' @param transform Logical scalar indicating whether a quarter-root transformation should be applied before trend fitting.
#' This transformation is copied from \code{limma::voom} and shrinks all values towards 1, flattening any sharp gradients in the trend for an easier fit.
#' The default of \code{TRUE} assumes that the variances are computed from log-expression values, in which case there is typically a strong \dQuote{hump} in the mean-variance relationship.
#' @param span Numeric scalar specifying the span of the LOWESS smoother, as a proportion of the total number of points.
#' Larger values improve stability at the cost of sensitivity to changes in low-density regions.
#' Ignored if \code{use.min.width=TRUE}.
#' @param use.min.width Logical scalar indicating whether a minimum width constraint should be applied to the LOWESS smoother.
#' This replaces the proportion-based span for defining each window.
#' Instead, the window for each point must be of a minimum width and is extended until it contains a minimum number of points. 
#' Setting this to `TRUE` ensures that sensitivity is maintained in the trend fit at low-density regions for the distribution of means, e.g., at high abundances.
#' It also avoids overfitting from very small windows in high-density intervals. 
#' @param min.width Minimum width of the window to use when \code{use.min.width=TRUE}.
#' The default value is chosen based on the typical range of means in single-cell RNA-seq data.
#' @param min.window.count Minimum number of observations in each window.
#' This ensures that each window contains at least a given number of observations for a stable fit.
#' If the minimum width window contains fewer observations, it is extended using the standard LOWESS logic until the minimum number is achieved.
#' Only used if \code{use.min.width=TRUE}.
#' @param num.threads Number of threads to use.
#'
#' @return List containing \code{fitted}, a numeric vector containing the fitted values of the trend for each gene;
#' and \code{residuals}, a numeric vector containing the residuals from the trend.
#'
#' @author Aaron Lun
#'
#' @seealso
#' \code{\link{modelGeneVariances}}, to compute the means and variances on which the trend is fitted.
#'
#' The \code{fit_variance_trend} function in \url{https://libscran.github.io/scran_variances/}.
#'
#' @examples
#' # Setting up some single-cell-like data.
#' mu <- 2^runif(1000, -10, 10)
#' counts <- matrix(rpois(20 * length(mu), lambda=mu), ncol=20)
#'
#' sf <- centerSizeFactors(colSums(counts))
#' normalized <- normalizeCounts(counts, size.factors=sf)
#' stats <- modelGeneVariances(normalized)
#'
#' out <- fitVarianceTrend(stats$statistics$means, stats$statistics$variances)
#' plot(stats$statistics$means, stats$statistics$variances)
#' curve(approxfun(stats$statistics$means, out$fitted)(x), col="red", add=TRUE)
#'
#' @export
fitVarianceTrend <- function(means, variances, mean.filter=TRUE, min.mean=0.1, transform=TRUE, span=0.3, use.min.width=FALSE, min.width=1, min.window.count=200, num.threads=1) {
    fit_variance_trend(
        means,
        variances,
        mean_filter=mean.filter,
        min_mean=min.mean,
        transform=transform,
        span=span,
        use_min_width=use.min.width,
        min_width=min.width,
        min_window_count=min.window.count,
        num_threads=num.threads
    )
}

# library(testthat); library(scrapper); source("test-modelGeneVariances.R")

set.seed(121212)
library(Matrix)
x <- abs(rsparsematrix(1000, 100, 0.1) * 10)

test_that("modelGeneVariances works without blocking", {
    out <- modelGeneVariances(x)$statistics
    expect_equal(out$variances, apply(as.matrix(x), 1, var))
    expect_equal(out$means, Matrix::rowMeans(x))
    expect_equal(out$variances - out$fitted, out$residuals)

    fit <- fitVarianceTrend(out$means, out$variances)
    expect_identical(out$fitted, fit$fitted)
    expect_identical(out$residuals, fit$residuals)

    # Responds to trend-fitting options. 
    out2 <- modelGeneVariances(x, span=0.5)$statistics
    expect_identical(out$variances, out2$variances)
    expect_identical(out$means, out2$means)
    expect_false(identical(out$fitted, out2$fitted))

    fit2 <- fitVarianceTrend(out2$means, out2$variances, span=0.5)
    expect_equal(out2$fitted, fit2$fitted)
    expect_equal(out2$residuals, fit2$residuals)

    expect_error(modelGeneVariances(SummarizedExperiment::SummarizedExperiment(x)), "not supported")
})

test_that("modelGeneVariances works with blocking", {
    block <- sample(LETTERS[1:3], ncol(x), replace=TRUE)
    out <- modelGeneVariances(x, block, block.weight.policy="equal")

    expect_identical(out$block.ids, LETTERS[1:3])
    for (b in out$block.ids) {
        sub <- x[,block == b]
        current <- out$per.block[[b]]
        expect_equal(current$variances, apply(as.matrix(sub), 1, var))
        expect_equal(current$means, Matrix::rowMeans(sub))
        expect_equal(current$variances - current$fitted, current$residuals)
    }

    average.mean <- rowMeans(do.call(cbind, lapply(out$per.block, function(x) x$means)))
    expect_equal(average.mean, out$statistics$means)
    average.variance <- rowMeans(do.call(cbind, lapply(out$per.block, function(x) x$variances)))
    expect_equal(average.variance, out$statistics$variances)
    average.fitted <- rowMeans(do.call(cbind, lapply(out$per.block, function(x) x$fitted)))
    expect_equal(average.fitted, out$statistics$fitted)
    average.residuals <- rowMeans(do.call(cbind, lapply(out$per.block, function(x) x$residuals)))
    expect_equal(average.residuals, out$statistics$residuals)
})

# library(testthat); library(scrapper); source("test-fitVarianceTrend.R")

test_that("fitVarianceTrend works as expected", {
    x <- runif(1000)
    y <- 2^rnorm(1000)
    out <- fitVarianceTrend(x, y)
    expect_equal(y - out$fitted, out$residuals)

    out2 <- fitVarianceTrend(x, y, num.threads=2)
    expect_identical(out, out2)

    # Responds to the various options.
    out2 <- fitVarianceTrend(x, y, use.min.width=TRUE, min.width=0.5)
    expect_false(identical(out, out2))

    out2 <- fitVarianceTrend(x, y, transform=FALSE)
    expect_false(identical(out, out2))

    out2 <- fitVarianceTrend(x, y, span=0.5)
    expect_false(identical(out, out2))
})

#' Scale and combine multiple embeddings in a SingleCellExperiment
#'
#' Scale embeddings for different modalities to equalize their intra-population variance, and combine them into a single embedding for downstream analysis.
#' This calls \code{\link{scaleByNeighbors}} on the reduced dimensions of the main/alternative experiments in a \link[SingleCellExperiment]{SingleCellExperiment}.
#'
#' @param x A \link[SingleCellExperiment]{SingleCellExperiment} object or one of its subclasses.
#' Rows correspond to genomic features and columns correspond to cells.
#' @param altexp.reddims Named list of character or integer vectors.
#' Each entry is named after an alternative experiment.
#' Each vector contains the names/indices of the \code{\link[SingleCellExperiment]{reducedDim}} embeddings from that experiment to be combined.
#' @param main.reddims Character or integer vector specifying the names/indices of the \code{\link[SingleCellExperiment]{reducedDim}} entries from \code{x} to be combined.
#' @param num.neighbors Number of neighbors used to define the scaling factor, passed to \code{\link{scaleByNeighbors}}.
#' @param block Block assignment for each cell, passed to \code{\link{scaleByNeighbors}}.
#' @param num.threads Number of threads for the neighbor search, passed to \code{\link{scaleByNeighbors}}.
#' @param BNPARAM Algorithm for the nearest neighbor search, passed to \code{\link{scaleByNeighbors}}.
#' @param more.scale.args Named list of additional arguments to pass to \code{\link[scrapper]{scaleByNeighbors}}.
#' @param output.name String containing the name of the \code{\link[SingleCellExperiment]{reducedDim}} entry in which to store the combined embeddings.
#' @param meta.name String containing the name of the \code{\link[S4Vectors]{metadata}} entry in which to store additional metrics.
#' If \code{NULL}, additional metrics are not stored.
#' @param delayed.transpose Logical scalar indicating whether to delay the transposition when storing coordinates in the \code{\link[SingleCellExperiment]{reducedDims}}.
#'
#' @return \code{x} is returned with the combined embeddings stored in its \code{reducedDims}.
#' The scaling factors for all embeddings are stored in the \code{metadata}.
#'
#' @author Aaron Lun
#'
#' @examples
#' library(SingleCellExperiment)
#' sce <- getTestAdtData.se("pca")
#' sce <- scaleByNeighbors.se(sce, altexp.reddims=list(ADT="PCA"))
#' reducedDimNames(sce) 
#' metadata(sce)$combined
#'
#' @export
#' @importFrom BiocNeighbors AnnoyParam
#' @importFrom S4Vectors metadata metadata<-
scaleByNeighbors.se <- function(
    x,
    altexp.reddims,
    main.reddims = "PCA", 
    num.neighbors = 20,
    block = NULL,
    BNPARAM = AnnoyParam(),
    num.threads = 1,
    more.scale.args = list(),
    output.name = "combined",
    meta.name = "combined",
    delayed.transpose = FALSE
) {
    all.embeddings <- list()
    main.reddims <- .sanitizeReddims(main.reddims, SingleCellExperiment::reducedDimNames(x))
    for (r in main.reddims) {
        all.embeddings <- append(all.embeddings, list(.getTransposedReddim(x, r)))
    }

    altexp.reddims <- altexp.reddims[!duplicated(names(altexp.reddims))]
    for (ae in names(altexp.reddims)) {
        ae.se <- SingleCellExperiment::altExp(x, ae)
        cur.reddim <- .sanitizeReddims(altexp.reddims[[ae]], SingleCellExperiment::reducedDimNames(ae.se))
        for (r in cur.reddim) {
            all.embeddings <- append(all.embeddings, list(.getTransposedReddim(ae.se, r)))
        }
        altexp.reddims[[ae]] <- cur.reddim
    }

    out <- .call(
        scaleByNeighbors,
        list(all.embeddings),
        list(num.neighbors=num.neighbors, block=block, num.threads=num.threads, BNPARAM=BNPARAM),
        more.scale.args
    )

    x <- .addTransposedReddim(x, output.name, out$combined, delayed.transpose)
    if (!is.null(meta.name)) {
        out$combined <- NULL

        # Formatting it in the same manner as the arguments.
        counter <- 1L
        out$main.scaling <- numeric(0)
        for (r in main.reddims) {
            out$main.scaling[[r]] <- out$scaling[[counter]]
            counter <- counter + 1L
        }

        out$altexp.scaling <- list()
        for (ae in names(altexp.reddims)) {
            current <- numeric(0)
            for (r in altexp.reddims[[ae]]) {
                current[[r]] <- out$scaling[[counter]]
                counter <- counter + 1L
            }
            out$altexp.scaling[[ae]] <- current
        }

        out$scaling <- NULL
        metadata(x)[[meta.name]] <- out
    }

    x
}

.sanitizeReddims <- function(reddims, all.reddims) {
    reddims <- unique(reddims)
    if (is.integer(reddims)) {
        all.reddims[reddims]
    } else {
        reddims
    }
}

# library(testthat); library(scrapper); source("test-se_scaleByNeighbors.R")

library(SingleCellExperiment)
set.seed(1111)
sce <- SingleCellExperiment(list(counts = matrix(0, 0, 200)))
reducedDim(sce, "PCA1") <- matrix(runif(1000), nrow=200)
reducedDim(sce, "PCA2") <- matrix(runif(1000), nrow=200)
altExp(sce, "ADT") <- sce
altExp(sce, "other") <- sce

test_that("scaleByNeighbors.se works as expected", {
    out <- scaleByNeighbors.se(
        sce,
        altexp.reddims=list(
            ADT=c("PCA2", "PCA1"),
            other=c("PCA1", "PCA2")
        ),
        main.reddims=c("PCA1", "PCA2")
    )

    ref <- scaleByNeighbors(
        list(
            t(reducedDim(sce, "PCA1")),
            t(reducedDim(sce, "PCA2")),
            t(reducedDim(altExp(sce, "ADT"), "PCA2")),
            t(reducedDim(altExp(sce, "ADT"), "PCA1")),
            t(reducedDim(altExp(sce, "other"), "PCA1")),
            t(reducedDim(altExp(sce, "other"), "PCA2"))
        )
    )
    expect_identical(reducedDim(out, "combined"), t(ref$combined))

    meta <- metadata(out)$combined
    expect_identical(names(meta$main.scaling), c("PCA1", "PCA2"))
    expect_identical(names(meta$altexp.scaling), c("ADT", "other"))
    expect_identical(names(meta$altexp.scaling$ADT), c("PCA2", "PCA1"))
    expect_identical(names(meta$altexp.scaling$other), c("PCA1", "PCA2"))

    expect_identical(
        c(
            meta$main.scaling[["PCA1"]],
            meta$main.scaling[["PCA2"]],
            meta$altexp.scaling[["ADT"]][["PCA2"]],
            meta$altexp.scaling[["ADT"]][["PCA1"]],
            meta$altexp.scaling[["other"]][["PCA1"]],
            meta$altexp.scaling[["other"]][["PCA2"]]
        ),
        ref$scaling
    )
})

test_that("scaleByNeighbors.se works with edge cases", {
    # Main experiment only.
    {
        out <- scaleByNeighbors.se(sce, altexp.reddims=list(), main.reddims=c("PCA2", "PCA1"))
        ref <- scaleByNeighbors(list(t(reducedDim(sce, "PCA2")), t(reducedDim(sce, "PCA1"))))

        expect_identical(reducedDim(out, "combined"), t(ref$combined))
        expect_identical(names(metadata(out)$combined$main.scaling), c("PCA2", "PCA1"))
        expect_identical(unname(metadata(out)$combined$main.scaling), ref$scaling)
    }

    # Alternative experiments only.
    {
        out <- scaleByNeighbors.se(sce, altexp.reddims=list(ADT="PCA2", other="PCA2"), main.reddims=NULL)
        ref <- scaleByNeighbors(list(t(reducedDim(altExp(sce, "ADT"), "PCA2")), t(reducedDim(altExp(sce, "other"), "PCA2"))))
        expect_identical(reducedDim(out, "combined"), t(ref$combined))

        meta.alt <- metadata(out)$combined$altexp.scaling
        expect_identical(names(metadata(out)$combined$altexp.scaling), c("ADT", "other"))
        expect_identical(names(metadata(out)$combined$altexp.scaling$ADT), "PCA2")
        expect_identical(names(metadata(out)$combined$altexp.scaling$other), "PCA2")
        expect_identical(c(meta.alt[["ADT"]][["PCA2"]], meta.alt[["other"]][["PCA2"]]), ref$scaling)
    }

    # Skips the metadata if requested.
    out <- scaleByNeighbors.se(sce, altexp.reddims=list(ADT="PCA2", other="PCA2"), main.reddims="PCA1", meta.name=NULL)
    expect_null(metadata(out)$combined)
})

test_that("scaleByNeighbors.se works deduplicates correctly", {
    out <- scaleByNeighbors.se(sce, altexp.reddims=list(ADT=c("PCA1", "PCA1"), ADT=c("PCA2", "PCA2")), main.reddims=c("PCA2", "PCA2"))
    ref <- scaleByNeighbors.se(sce, altexp.reddims=list(ADT=c("PCA1", "PCA1")), main.reddims="PCA2")
    expect_identical(reducedDim(out, "combined"), reducedDim(ref, "combined"))
    expect_identical(metadata(out)$combined, metadata(ref)$combined)
})

#' t-stochastic neighbor embedding
#'
#' Compute t-SNE coordinates to visualize similarities between cells.
#'
#' @param x Numeric matrix where rows are dimensions and columns are cells,
#' typically containing a low-dimensional representation from, e.g., \code{\link{runPca}}.
#'
#' Alternatively, a named list of nearest-neighbor search results like that returned by \code{\link[BiocNeighbors]{findKNN}}.
#' This should contain \code{index}, an integer matrix where rows are neighbors and columns are cells;
#' and \code{distance}, a numeric matrix of the same dimensions containing the distances to each neighbor.
#' Each column contains 1-based indices for the nearest neighbors of the corresponding cell, ordered by increasing distance.
#' The number of neighbors should be the same as \code{num.neighbors}, otherwise a warning is raised.
#'
#' Alternatively, an index constructed by \code{\link[BiocNeighbors]{buildIndex}}.
#' @param perplexity Numeric scalar specifying the perplexity to use in the t-SNE algorithm.
#' Higher perplexities will focus on global structure, at the cost of increased runtime and decreased local resolution.
#' @param num.neighbors Integer scalar specifying the number of neighbors, typically derived from \code{perplexity}.
#' If \code{x} contains pre-computed neighbor search results with a different number of neighbors than \code{num.neighbors}, an error is thrown;
#' this can be suppressed by setting \code{num.neighbors = NULL}.
#' @param theta Numeric scalar specifying the approximation level for the Barnes-Hut calculation of repulsive forces.
#' Lower values increase accuracy at the cost of increased compute time.
#' All values should be non-negative.
#' @param early.exaggeration.iterations Integer scalar specifying the number of iterations of the early exaggeration phase,
#' where clusters are artificially compacted to leave more empty space so that cells can easily relocate to find a good global organization.
#' Larger values improve convergence within this phase at the cost of reducing the remaining iterations in \code{max.iterations}.
#' @param exaggeration.factor Numeric scalar containing the exaggeration factor for the early exaggeration phase (see \code{early.exaggeration.iterations}).
#' Larger values increase the attraction between nearest neighbors to favor local structure.
#' @param momentum.switch.iterations Integer scalar specifying the number of iterations to perform before switching from the starting momentum to the final momentum.
#' Higher momentums can improve convergence by increasing the step size and smoothing over local oscillations, at the risk of potentially skipping over relevant minima.
#' @param start.momentum Numeric scalar containing the starting momentum, to be used in the iterations before the momentum switch at \code{momentum.switch.iterations}.
#' This is usually lower than \code{final.momentum} to avoid skipping over suitable local minima. 
#' @param final.momentum Numeric scalar containing the final momentum, to be used in the iterations after the momentum switch at \code{momentum.switch.iterations}.
#' This is usually higher than \code{start.momentum} to accelerate convergence to the local minima once the observations are moderately well-organized.
#' @param eta Numeric scalar containing the learning rate, used to scale the updates for each cell.
#' Larger values can speed up convergence at the cost of skipping over local minima. 
#' @param max.depth Integer scalar specifying the maximum depth of the Barnes-Hut quadtree.
#' If neighboring cells cannot be separated before the maximum depth is reached, they will be assigned to the same leaf node of the quadtree.
#' Smaller values (7-10) improve speed by bounding the recursion depth at the cost of accuracy.
#' @param leaf.approximation Logical scalar indicating whether to use the \dQuote{leaf approximation}.
#' If \code{TRUE}, repulsive forces are computed between leaf nodes and re-used across all cells assigned to that leaf node. 
#' This sacrifices some accuracy for greater speed, assuming that \code{max.depth} is small enough for multiple cells to be assigned to the same leaf.
#' @param seed Integer scalar specifying the seed to use for generating the initial coordinates.
#' @param num.threads Integer scalar specifying the number of threads to use.
#' @param max.iterations Integer scalar specifying the maximum number of iterations to perform.
#' Larger values improve convergence at the cost of compute time.
#' @param BNPARAM A \link[BiocNeighbors]{BiocNeighborParam} object specifying the algorithm to use.
#' Only used if \code{x} is not a prebuilt index or a list of existing nearest-neighbor search results.
#' 
#' @return 
#' For \code{runTsne}, a numeric matrix where rows are cells and columns are the two dimensions of the embedding.
#'
#' For \code{tsnePerplexityToNeighbors}, an integer scalar specifying the number of neighbors to use for a given \code{perplexity}.
#'
#' @author Aaron Lun
#'
#' @seealso
#' \url{https://libscran.github.io/qdtsne/}, for an explanation of the approximations.
#'
#' \code{\link{runTsne.se}}, to run t-SNE on a \link[SingleCellExperiment]{SingleCellExperiment}.
#'
#' @examples
#' x <- t(as.matrix(iris[,1:4]))
#' embedding <- runTsne(x)
#' plot(embedding[,1], embedding[,2], col=iris[,5])
#' 
#' @references
#' van der Maaten LJP and Hinton GE (2008). 
#' Visualizing high-dimensional data using t-SNE. 
#' \emph{Journal of Machine Learning Research_} 9, 2579-2605.
#' 
#' van der Maaten LJP (2014). 
#' Accelerating t-SNE using tree-based algorithms. 
#' \emph{Journal of Machine Learning Research} 15, 3221-3245.
#' 
#' @export
#' @importFrom BiocNeighbors findKNN AnnoyParam
runTsne <- function(
    x,
    perplexity=30,
    num.neighbors=tsnePerplexityToNeighbors(perplexity),
    theta=1,
    early.exaggeration.iterations=250,
    exaggeration.factor=12,
    momentum.switch.iterations=250,
    start.momentum=0.5,
    final.momentum=0.8,
    eta=200,
    max.depth=7,
    leaf.approximation=FALSE,
    max.iterations=500,
    seed=42,
    num.threads=1,
    BNPARAM=AnnoyParam())
{
    .checkSEX(x, "runTsne.se")

    if (!is.list(x)) {
        x <- findKNN(x, k=num.neighbors, transposed=TRUE, get.index="transposed", get.distance="transposed", num.threads=num.threads, BNPARAM=BNPARAM)
    } else {
        .checkNeighborResults(x$index, x$distance)
        if (!is.null(num.neighbors) && nrow(x$index) != num.neighbors) {
            warning("'nrow(x$index)' is not consistent with 'num.neighbors'")
        }
    }

    output <- run_tsne(
        nnidx=x$index, 
        nndist=x$distance, 
        perplexity=perplexity,
        theta=theta,
        early_exaggeration_iterations=early.exaggeration.iterations,
        exaggeration_factor=exaggeration.factor,
        momentum_switch_iterations=momentum.switch.iterations,
        start_momentum=start.momentum,
        final_momentum=final.momentum,
        eta=eta,
        max_depth=max.depth,
        leaf_approx=leaf.approximation,
        max_iter=max.iterations,
        num_threads=num.threads,
        seed=seed
    )

    t(output)
}

#' @export
#' @rdname runTsne
tsnePerplexityToNeighbors <- function(perplexity) {
    perplexity_to_neighbors(perplexity)
}

# library(testthat); library(scrapper); source("test-runTsne.R")

library(BiocNeighbors)
x <- t(as.matrix(iris[,1:4]))

test_that("runTSNE works in basic mode", {
    embed <- runTsne(x)
    expect_identical(nrow(embed), ncol(x))
    expect_identical(ncol(embed), 2L)

    again <- runTsne(x)
    expect_identical(embed, again)  # check that it's reproducible.

    alt <- runTsne(x, perplexity=20)
    expect_identical(dim(alt), dim(embed))
    expect_false(identical(alt, embed)) # check that perplexity has an effect.

    res <- findKNN(x, transposed=TRUE, k=tsnePerplexityToNeighbors(30), get.distance="transposed", get.index="transposed", BNPARAM=AnnoyParam())
    nnin <- runTsne(res)
    expect_identical(nnin, embed)

    res <- findKNN(x, transposed=TRUE, k=30, get.distance="transposed", get.index="transposed", BNPARAM=AnnoyParam())
    expect_warning(runTsne(res), "not consistent with 'num.neighbors'")
    expect_warning(runTsne(res, num.neighbors=30), NA)
    expect_warning(runTsne(res, num.neighbors=NULL), NA)

    expect_error(runTsne(SummarizedExperiment::SummarizedExperiment(x)), "not supported")
})

# library(testthat); library(scrapper); source("test-runAllNeighborSteps.R")

x <- t(as.matrix(iris[,1:4]))

test_that("runAllNeighborSteps gives the same results as the reference", {
    res <- runAllNeighborSteps(x, num.threads=2, collapse.search=FALSE)

    umap.ref <- runUmap(x)
    expect_identical(res$runUmap, umap.ref)
    umap.single <- runAllNeighborSteps(x, runTsne.args=NULL, clusterGraph.args=NULL)
    expect_identical(umap.single$runUmap, umap.ref)

    tsne.ref <- runTsne(x)
    expect_identical(res$runTsne, tsne.ref)
    tsne.single <- runAllNeighborSteps(x, runUmap.args=NULL, clusterGraph.args=NULL)
    expect_identical(tsne.single$runTsne, tsne.ref)

    graph <- buildSnnGraph(x)
    clustering <- clusterGraph(graph)
    expect_identical(res$clusterGraph, clustering)
    expect_null(res$buildSnnGraph)
    cluster.single <- runAllNeighborSteps(x, runUmap.args=NULL, runTsne.args=NULL)
    expect_identical(cluster.single$clusterGraph, clustering)

    expect_error(runAllNeighborSteps(SummarizedExperiment::SummarizedExperiment(x)), "not supported")
})

test_that("runAllNeighborSteps works with collapsed neighbors", {
    BNPARAM <- BiocNeighbors::KmknnParam()
    res <- runAllNeighborSteps(x, BNPARAM=BNPARAM, num.threads=2)
    res.collapse <- runAllNeighborSteps(x, collapse.search=TRUE, BNPARAM=BNPARAM, num.threads=2)
    expect_identical(res, res.collapse)
})

test_that("runAllNeighborSteps works when returning the graph", {
    BNPARAM <- BiocNeighbors::KmknnParam()
    out <- runAllNeighborSteps(x, BNPARAM=BNPARAM, return.graph=TRUE, num.threads=2)
    expect_identical(ncol(x), out$buildSnnGraph$vertices)
    expect_identical(ncol(x), length(out$clusterGraph$membership))
})

test_that("runAllNeighborSteps returns sensibly when nothing is requested", {
    res <- runAllNeighborSteps(x, runUmap.args=NULL, runTsne.args=NULL, clusterGraph.args=NULL)
    expect_type(res, "list")
    expect_identical(length(res), 0L)
})

#' t-SNE on a SummarizedExperiment
#'
#' Generate a t-SNE visualization from an existing embedding,
#' by calling \code{\link{runUmap}} on a reduced dimension entry in \link[SingleCellExperiment]{SingleCellExperiment}.
#'
#' @param x A \link[SingleCellExperiment]{SingleCellExperiment} object or one of its subclasses.
#' Rows correspond to genomic features and columns correspond to cells.
#' @param perplexity Perplexity to use in the t-SNE algorithm, passed to \code{\link{runTsne}}.
#' @param num.threads Number of threads for the neighbor search and optimization, passed to \code{\link{runTsne}}.
#' @param more.tsne.args Named list of further arguments to pass to \code{\link{runTsne}}.
#' @param reddim.type Integer or string specifying the existing embedding in the \code{\link[SingleCellExperiment]{reducedDim}} of \code{x}.
#' Alternatively, a named integer or character vector of length 1,
#' where the name specifies an alternative experiment of \code{x} and the value is the name/index of a \code{reducedDim} entry in that alternative experiment.
#' @param output.name String containing the name of the output \code{\link[SingleCellExperiment]{reducedDim}}. 
#'
#' @return \code{x} is returned with the t-SNE coordinates stored in the \code{reducedDim}.
#'
#' @author Aaron Lun
#' @examples
#' library(SingleCellExperiment)
#' sce <- getTestRnaData.se("pca")
#' # Using fewer iterations for a faster-running example.
#' sce <- runTsne.se(sce, more.tsne.args=list(max.iterations=50))
#' head(reducedDim(sce, "TSNE"))
#'
#' @export
runTsne.se <- function(
    x,
    perplexity = 30,
    num.threads = 1,
    more.tsne.args = list(),
    reddim.type = "PCA", 
    output.name = "TSNE"
) {
    out <- .call(
        runTsne,
        list(.getTransposedReddim(x, reddim.type)),
        list(perplexity=perplexity, num.threads=num.threads),
        more.tsne.args
    )

    .addTsneResults(x, output.name, out)
}

.addTsneResults <- function(x, output.name, res) {
    SingleCellExperiment::reducedDim(x, output.name) <- res
    x
}

# library(testthat); library(scrapper); source("test-se_runTsne.R")

library(SingleCellExperiment)
set.seed(7777)
sce <- SingleCellExperiment(list(counts = matrix(0, 0, 200)))
reducedDim(sce, "PCA") <- matrix(runif(1000), nrow=200)

test_that("runTsne.se works as expected", {
    out <- runTsne.se(sce)
    expect_true("TSNE" %in% reducedDimNames(out))
})

#' Analyze single-cell data from a SummarizedExperiment
#'
#' Execute a simple single-cell analysis pipeline, starting from a count matrix and ending with clusters, visualizations and markers.
#' This also supports integration of multiple modalities and correction of batch effects.
#'
#' @param x A \link[SummarizedExperiment]{SummarizedExperiment} object or one of its subclasses.
#' Rows correspond to genomic features (genes, ADTs or CRISPR guides) and columns correspond to cells.
#' @param rna.altexp String or integer specifying the alternative experiment of \code{x} containing the RNA data.
#' If \code{NA}, the main experiment is assumed to contain the RNA data.
#' If \code{NULL}, it is assumed that no RNA data is available.
#' @param adt.altexp String or integer specifying the alternative experiment of \code{x} containing the ADT data.
#' If \code{NA}, the main experiment is assumed to contain the ADT data.
#' If \code{NULL}, it is assumed that no ADT data is available.
#' @param crispr.altexp String or integer specifying the alternative experiment of \code{x} containing the CRISPR data.
#' If \code{NA}, the main experiment is assumed to contain the CRISPR data.
#' If \code{NULL}, it is assumed that no CRISPR data is available.
#' @param rna.assay.type String or integer specifying the assay containing the RNA count data.
#' Only used if \code{rna.altexp} is not \code{NULL}.
#' @param adt.assay.type String or integer specifying the assay containing the ADT count data.
#' Only used if \code{adt.altexp} is not \code{NULL}.
#' @param crispr.assay.type String or integer specifying the assay containing the CRISPR count data.
#' Only used if \code{crispr.altexp} is not \code{NULL}.
#' @param block Vector or factor specifying the block of origin (e.g., batch, sample) for each cell in \code{x}.
#' Alternatively \code{NULL}, if all cells are from the same block.
#' @param block.name String specifying the name of the \code{\link[SummarizedExperiment]{colData}} column in which to store the blocking factor.
#' Only used if \code{block} is not \code{NULL}.
#' If \code{NULL}, the blocking factor is not stored in the \code{colData}.
#' @param rna.qc.subsets Passed to \code{\link{quickRnaQc.se}} as the \code{subsets} argument.
#' Only used if \code{rna.altexp} is not \code{NULL}.
#' @param rna.qc.output.prefix Passed to \code{\link{quickRnaQc.se}} as the \code{output.prefix} argument.
#' Only used if \code{rna.altexp} is not \code{NULL}.
#' @param more.rna.qc.args Named list of additional arguments to pass to \code{\link{quickRnaQc.se}}.
#' Only used if \code{rna.altexp} is not \code{NULL}.
#' @param adt.qc.subsets Passed to \code{\link{quickAdtQc.se}} as the \code{subsets} argument.
#' Only used if \code{adt.altexp} is not \code{NULL}.
#' @param adt.qc.output.prefix Passed to \code{\link{quickAdtQc.se}} as the \code{output.prefix} argument.
#' Only used if \code{adt.altexp} is not \code{NULL}.
#' @param more.adt.qc.args Named list of additional arguments to pass to \code{\link{quickAdtQc.se}}.
#' Only used if \code{adt.altexp} is not \code{NULL}.
#' @param crispr.qc.output.prefix Passed to \code{\link{quickCrisprQc.se}} as the \code{output.prefix} argument.
#' Only used if \code{crispr.altexp} is not \code{NULL}.
#' @param more.crispr.qc.args Named list of additional arguments to pass to \code{\link{quickCrisprQc.se}}.
#' Only used if \code{crispr.altexp} is not \code{NULL}.
#' @param filter.cells Logical scalar indicating whether to filter \code{x} to only retain high-quality cells in all modalities.
#' If \code{FALSE}, QC metrics and thresholds are still computed but are not used to filter the count matrices.
#' @param rna.norm.output.name Passed to \code{\link{normalizeRnaCounts.se}} as the \code{output.name} argument.
#' Only used if \code{rna.altexp} is not \code{NULL}.
#' @param more.rna.norm.args Named list of arguments to pass to \code{\link{normalizeRnaCounts.se}}.
#' Only used if \code{rna.altexp} is not \code{NULL}.
#' @param adt.norm.output.name Passed to \code{\link{normalizeAdtCounts.se}} as the \code{output.name} argument.
#' Only used if \code{adt.altexp} is not \code{NULL}.
#' @param more.adt.norm.args Named list of arguments to pass to \code{\link{normalizeAdtCounts.se}}.
#' Only used if \code{adt.altexp} is not \code{NULL}.
#' @param crispr.norm.output.name Passed to \code{\link{normalizeCrisprCounts.se}} as the \code{output.name} argument.
#' Only used if \code{crispr.altexp} is not \code{NULL}.
#' @param more.crispr.norm.args Named list of arguments to pass to \code{\link{normalizeCrisprCounts.se}}.
#' Only used if \code{crispr.altexp} is not \code{NULL}.
#' @param rna.hvg.output.prefix Passed to \code{\link{chooseRnaHvgs.se}} as the \code{output.prefix} argument.
#' Only used if \code{rna.altexp} is provided.
#' @param more.rna.hvg.args Named list of arguments to pass to \code{\link{chooseRnaHvgs.se}}.
#' Only used if \code{rna.altexp} is provided.
#' @param rna.pca.output.name Passed to \code{\link{runPca.se}} as the \code{output.name} argument.
#' Only used if \code{rna.altexp} is provided.
#' @param more.rna.pca.args Named list of arguments to pass to \code{\link{runPca.se}}.
#' Only used if \code{rna.altexp} is provided.
#' @param adt.pca.output.name Passed to \code{\link{runPca.se}} as the \code{output.name} argument.
#' Only used if \code{adt.altexp} is provided.
#' @param more.adt.pca.args Named list of arguments to pass to \code{\link{runPca.se}}.
#' Only used if \code{adt.altexp} is provided.
#' @param use.rna.pcs Logical scalar indicating whether to use the RNA-derived PCs for downstream steps (i.e., clustering, visualization).
#' Only used if \code{rna.altexp} is provided.
#' @param use.adt.pcs Logical scalar indicating whether to use the ADT-derived PCs for downstream steps (i.e., clustering, visualization).
#' Only used if \code{adt.altexp} is provided.
#' @param scale.output.name Passed to \code{\link{scaleByNeighbors.se}} as the \code{output.name} argument.
#' Only used if multiple modalities are available and their corresponding \code{use.*.pcs} arguments are \code{TRUE}.
#' @param more.scale.args Named list of arguments to pass to \code{\link{scaleByNeighbors.se}}.
#' Only used if multiple modalities are available and their corresponding \code{use.*.pcs} arguments are \code{TRUE}.
#' @param mnn.output.name Passed to \code{\link{correctMnn.se}} as the \code{output.name} argument.
#' Only used if \code{block} is supplied.
#' @param more.mnn.args Named list of arguments to pass to \code{\link{correctMnn.se}}.
#' Only used if \code{block} is supplied.
#' @param more.tsne.args Passed to \code{\link{runAllNeighborSteps.se}}.
#' @param more.umap.args Passed to \code{\link{runAllNeighborSteps.se}}.
#' @param more.build.graph.args Passed to \code{\link{runAllNeighborSteps.se}}. 
#' @param cluster.graph.output.name Passed to \code{\link{runAllNeighborSteps.se}} as \code{cluster.output.name}.
#' @param more.cluster.graph.args Passed to \code{\link{runAllNeighborSteps.se}}. 
#' @param more.neighbor.args Passed to \code{\link{runAllNeighborSteps.se}}.
#' @param kmeans.clusters Passed to \code{\link{clusterKmeans.se}} as the \code{k} argument.
#' If \code{NULL}, k-means clustering is not performed.
#' @param kmeans.clusters.output.name Passed to \code{\link{clusterKmeans.se}} as the \code{output.name} argument.
#' Ignored if \code{kmeans.clusters = NULL}.
#' @param more.kmeans.args Named list of arguments to pass to \code{\link{clusterKmeans.se}}.
#' Ignored if \code{kmeans.clusters = NULL}.
#' @param clusters.for.markers Character vector of clustering algorithms (either \code{"graph"} or \code{"kmeans"}), specifying the clustering to be used for marker detection.
#' The first available clustering will be chosen.
#' If no clustering is available from the list, markers will not be computed.
#' @param more.rna.marker.args Named list of arguments to pass to \code{\link{scoreMarkers.se}} for the RNA data.
#' Ignored if no suitable clusterings are available or if \code{rna.altexp=NULL}.
#' @param more.adt.marker.args Named list of arguments to pass to \code{\link{scoreMarkers.se}} for the ADT data.
#' Ignored if no suitable clusterings are available or if \code{adt.altexp=NULL}.
#' @param more.crispr.marker.args Named list of arguments to pass to \code{\link{scoreMarkers.se}} for the CRISPR data.
#' Ignored if no suitable clusterings are available or if \code{crispr.altexp=NULL}.
#' @param BNPARAM A \link[BiocNeighbors]{BiocNeighborParam} instance specifying the nearest-neighbor search algorithm to use.
#' @param num.threads Integer scalar specifying the number of threads to use in each step.
#'
#' @return List containing:
#' \itemize{
#' \item \code{x}, a \link[SingleCellExperiment]{SingleCellExperiment} that is a copy of the input \code{x}.
#' It is also decorated with the results of each analysis step - see Details.
#' \item \code{markers}, a list of list of \link[S4Vectors]{DataFrame}s containing the marker statistics for each modality.
#' Each inner list corresponds to a modality (RNA, ADT, etc.) while each DataFrame corresponds to a cluster.
#' If no clusterings were generated, this is set to \code{NULL}.
#' }
#'
#' @details
#' This function is equivalent to:
#' \itemize{
#' \item Running \code{\link{quickRnaQc.se}}, \code{\link{quickAdtQc.se}} and/or \code{\link{quickCrisprQc.se}}, for quality control.
#' \item Subsetting \code{x} to only retain the high-quality cells in all modalities, based on \code{filter.cells}.
#' \item Running \code{\link{normalizeRnaCounts.se}}, \code{\link{normalizeAdtCounts.se}} and/or \code{\link{normalizeCrisprCounts.se}}, for normalization.
#' \item Running \code{\link{chooseRnaHvgs.se}} to identify highly variable genes.
#' \item Running \code{\link{runPca.se}} on the RNA and/or ADT data.
#' \item Running \code{\link{scaleByNeighbors.se}} if multiple modalities are present.
#' \item Running \code{\link{correctMnn.se}} if multiple batches are present in \code{block}.
#' \item Running \code{\link{runAllNeighborSteps.se}} to obtain t-SNE and UMAP coordinates, and to perform graph-based clustering.
#' \item Running \code{\link{clusterKmeans.se}} to perform k-means clustering, depending on \code{kmeans.clusters}. 
#' \item Running \code{\link{scoreMarkers.se}} to compute markers for each modality based on one of the clusterings. 
#' }
#'
#' @author Aaron Lun
#' @examples
#' library(SingleCellExperiment)
#' sce <- getTestRnaData.se("start")
#' res <- analyze.se(
#'     sce, 
#'     rna.qc.subsets=list(mito=grep("^mt-", rownames(sce))),
#'     num.threads=2 # keep R CMD check happy
#' )
#' assayNames(res$x)
#' reducedDimNames(res$x)
#' colData(res$x)
#' previewMarkers(res$markers$rna[[1]], "cohens.d.mean")
#'
#' @export
#' @importFrom methods is
#' @importFrom BiocNeighbors AnnoyParam
#' @importFrom DelayedArray DelayedArray
analyze.se <- function(
    x,
    rna.altexp = NA,
    adt.altexp = NULL,
    crispr.altexp = NULL,
    rna.assay.type = "counts",
    adt.assay.type = "counts",
    crispr.assay.type = "counts",
    block = NULL,
    block.name = "block",
    rna.qc.subsets = list(),
    rna.qc.output.prefix = NULL,
    more.rna.qc.args = list(),
    adt.qc.subsets = list(),
    adt.qc.output.prefix = NULL,
    more.adt.qc.args = list(),
    crispr.qc.output.prefix = NULL,
    more.crispr.qc.args = list(),
    filter.cells = TRUE,
    rna.norm.output.name = "logcounts",
    more.rna.norm.args = list(),
    adt.norm.output.name = "logcounts",
    more.adt.norm.args = list(),
    crispr.norm.output.name = "logcounts",
    more.crispr.norm.args = list(),
    rna.hvg.output.prefix = NULL,
    more.rna.hvg.args = list(),
    rna.pca.output.name = "PCA",
    more.rna.pca.args = list(),
    adt.pca.output.name = "PCA",
    more.adt.pca.args = list(),
    use.rna.pcs = TRUE,
    use.adt.pcs = TRUE,
    scale.output.name = "combined",
    more.scale.args = list(),
    mnn.output.name = "MNN",
    more.mnn.args = list(),
    more.umap.args = list(),
    more.tsne.args = list(),
    cluster.graph.output.name = "graph.cluster",
    more.build.graph.args = list(),
    more.cluster.graph.args = list(),
    more.neighbor.args = list(),
    kmeans.clusters = NULL,
    kmeans.clusters.output.name = "kmeans.cluster",
    more.kmeans.args = list(),
    clusters.for.markers = c("graph", "kmeans"),
    more.rna.marker.args = list(),
    more.adt.marker.args = list(),
    more.crispr.marker.args = list(),
    BNPARAM = AnnoyParam(),
    num.threads = 3L
) {
    ############ Quality control o(*°▽°*)o #############

    collected.filters <- list()

    if (!is.null(rna.altexp)) {
        tmp <- .getModality(x, rna.altexp)
        tmp <- .call(
            quickRnaQc.se, 
            list(tmp, output.prefix=rna.qc.output.prefix),
            list(subsets=rna.qc.subsets, block=block, num.threads=num.threads, assay.type=rna.assay.type),
            more.rna.qc.args
        )
        collected.filters$rna <- .extractOrError(SummarizedExperiment::colData(tmp), paste0(rna.qc.output.prefix, "keep"))
        x <- .assignModality(x, rna.altexp, tmp)
    }

    if (!is.null(adt.altexp)) {
        tmp <- .getModality(x, adt.altexp)
        tmp <- .call(
            quickAdtQc.se,
            list(tmp, output.prefix=adt.qc.output.prefix),
            list(subsets=adt.qc.subsets, block=block, num.threads=num.threads, assay.type=adt.assay.type),
            more.adt.qc.args
        )
        collected.filters$adt <- .extractOrError(SummarizedExperiment::colData(tmp), paste0(adt.qc.output.prefix, "keep"))
        x <- .assignModality(x, adt.altexp, tmp)
    }

    if (!is.null(crispr.altexp)) {
        tmp <- .getModality(x, crispr.altexp)
        tmp <- .call(
            quickCrisprQc.se, 
            list(tmp, output.prefix=crispr.qc.output.prefix),
            list(num.threads=num.threads, block=block, assay.type=crispr.assay.type),
            more.crispr.qc.args
        )
        collected.filters$crispr <- .extractOrError(SummarizedExperiment::colData(tmp), paste0(crispr.qc.output.prefix, "keep"))
        x <- .assignModality(x, crispr.altexp, tmp)
    }

    # Combining all filters.
    combined.qc.filter <- !logical(ncol(x))
    for (keep in collected.filters) {
        combined.qc.filter <- combined.qc.filter & keep
    }
    if (length(collected.filters) > 1) {
        SummarizedExperiment::colData(x)[["combined.keep"]] <- combined.qc.filter
    }

    if (filter.cells) {
        x <- .delayifyAssays(x)
        x <- x[,combined.qc.filter]
        if (!is.null(block)) {
            block <- block[combined.qc.filter]
        }
    }

    ############ Normalization ( ꈍᴗꈍ) #############

    if (!is.null(rna.altexp)) {
        tmp <- .getModality(x, rna.altexp)
        lib.sizes <- .extractOrError(SummarizedExperiment::colData(tmp), paste0(rna.qc.output.prefix, "sum"))
        tmp <- .call(
            normalizeRnaCounts.se,
            list(tmp, assay.type=rna.assay.type, output.name=rna.norm.output.name),
            list(size.factors=lib.sizes, block=block),
            more.rna.norm.args
        )
        x <- .assignModality(x, rna.altexp, tmp)
    }

    if (!is.null(adt.altexp)) {
        tmp <- .getModality(x, adt.altexp)
        tmp <- .call(
            normalizeAdtCounts.se, 
            list(tmp, assay.type=adt.assay.type, output.name=adt.norm.output.name),
            list(block=block),
            more.adt.norm.args
        )
        x <- .assignModality(x, adt.altexp, tmp)
    }

    if (!is.null(crispr.altexp)) {
        tmp <- .getModality(x, crispr.altexp)
        lib.sizes <- .extractOrError(SummarizedExperiment::colData(tmp), paste0(crispr.qc.output.prefix, "sum"))
        tmp <- .call(
            normalizeCrisprCounts.se,
            list(tmp, assay.type=crispr.assay.type, output.name=crispr.norm.output.name),
            list(size.factors=lib.sizes, block=block),
            more.crispr.norm.args
        )
        x <- .assignModality(x, crispr.altexp, tmp)
    }

    ############ Variance modelling (～￣▽￣)～ #############

    if (!is.null(rna.altexp)) {
        tmp <- .getModality(x, rna.altexp)
        tmp <- .call(
            chooseRnaHvgs.se,
            list(tmp, output.prefix=rna.hvg.output.prefix),
            list(block=block),
            more.rna.hvg.args
        )
        x <- .assignModality(x, rna.altexp, tmp)
    }

    ############ Principal components analysis \(>⩊<)/ #############

    if (!is.null(rna.altexp)) {
        tmp <- .getModality(x, rna.altexp)
        is.hvg <- .extractOrError(SummarizedExperiment::rowData(tmp), paste0(rna.hvg.output.prefix, "hvg"))
        tmp <- .call(
            runPca.se,
            list(tmp, assay.type=rna.norm.output.name, output.name=rna.pca.output.name),
            list(features=is.hvg, block=block, delayed.transpose=TRUE),
            more.rna.pca.args
        )
        x <- .assignModality(x, rna.altexp, tmp)
    }

    if (!is.null(adt.altexp)) {
        tmp <- .getModality(x, adt.altexp)
        tmp <- .call(
            runPca.se,
            list(tmp, assay.type=adt.norm.output.name, output.name=adt.pca.output.name),
            list(features=NULL, block=block, delayed.transpose=TRUE),
            more.adt.pca.args
        )
        x <- .assignModality(x, adt.altexp, tmp)
    }

    ############ Combining modalities („• ᴗ •„) #############

    embeddings <- character()
    if (use.rna.pcs && !is.null(rna.altexp)) {
        embeddings <- c(embeddings, "rna")
    }
    if (use.adt.pcs && !is.null(adt.altexp)) {
        embeddings <- c(embeddings, "adt")
    }

    target.embedding <- NULL
    if (length(embeddings) == 0) {
        stop("at least one 'use.*' must be true")

    } else if (length(embeddings) == 1L) {
        if (embeddings == "rna") {
            target.embedding <- .defineSingleTargetEmbedding(x, rna.altexp, rna.pca.output.name)
        } else if (embeddings == "adt") {
            target.embedding <- .defineSingleTargetEmbedding(x, adt.altexp, adt.pca.output.name)
        }

    } else {
        all.reddims <- list(main=character(0), altexp=list())
        if ("rna" %in% embeddings) {
            all.reddims <- .addSourceEmbeddingToScale(x, rna.altexp, rna.pca.output.name, all.reddims)
        }
        if ("adt" %in% embeddings) {
            all.reddims <- .addSourceEmbeddingToScale(x, adt.altexp, adt.pca.output.name, all.reddims)
        }

        x <- .call(
            scaleByNeighbors.se,
            list(x, output.name=scale.output.name, main.reddims=all.reddims$main, altexp.reddims=all.reddims$altexp, delayed.transpose=TRUE),
            list(block=block, num.threads=num.threads, BNPARAM=BNPARAM),
            more.scale.args
        )
        target.embedding <- scale.output.name
    }

    ############ Performing batch correction ⸜(｡˃ ᵕ ˂ )⸝ #############

    if (!is.null(block)) {
        x <- .call(
            correctMnn.se,
            list(x, output.name=mnn.output.name, reddim.type=target.embedding, delayed.transpose=TRUE),
            list(block=block, num.threads=num.threads, BNPARAM=BNPARAM),
            more.mnn.args
        )
        target.embedding <- mnn.output.name

        if (!is.null(block.name)) {
            SummarizedExperiment::colData(x)[[block.name]] <- block
        }
    }

    ############ Assorted neighbor-related stuff ⸜(⸝⸝⸝´꒳`⸝⸝⸝)⸝ #############

    x <- do.call(
        runAllNeighborSteps.se,
        list(
            x,
            reddim.type=target.embedding,
            cluster.output.name=cluster.graph.output.name,
            more.umap.args=more.umap.args, 
            more.tsne.args=more.tsne.args, 
            more.build.graph.args=more.build.graph.args,
            more.cluster.graph.args=more.cluster.graph.args,
            more.neighbor.args=more.neighbor.args,
            BNPARAM=BNPARAM,
            num.threads=num.threads
        )
    )

    ############ Maybe some k-means clustering ⸜(⸝⸝⸝´꒳`⸝⸝⸝)⸝ #############

    if (!is.null(kmeans.clusters)) {
        x <- .call(
            clusterKmeans.se,
            list(x, reddim.type=target.embedding, output.name=kmeans.clusters.output.name),
            list(k=kmeans.clusters, num.threads=num.threads),
            more.kmeans.args
        )
    }

    clusters.for.markers <- match.arg(clusters.for.markers, several.ok=TRUE)
    chosen.clusters <- NULL
    for (c in clusters.for.markers) {
        if (c == "graph") {
            name <- cluster.graph.output.name
        } else if (c == "kmeans") {
            name <- kmeans.clusters.output.name
        }
        if (!is.null(name) && name %in% colnames(SummarizedExperiment::colData(x))) {
            chosen.clusters <- SummarizedExperiment::colData(x)[[name]]
            break
        }
    }

    ############ Finding markers (˶˃ ᵕ ˂˶) #############

    markers <- NULL

    if (!is.null(chosen.clusters)) {
        markers <- list()

        if (!is.null(rna.altexp)) {
            markers$rna <- .call(
                scoreMarkers.se,
                list(.getModality(x, rna.altexp), assay.type=rna.norm.output.name),
                list(groups=chosen.clusters, num.threads=num.threads, block=block),
                more.rna.marker.args
            )
        }

        if (!is.null(adt.altexp)) {
            markers$adt <- .call(
                scoreMarkers.se,
                list(.getModality(x, adt.altexp), assay.type=adt.norm.output.name),
                list(groups=chosen.clusters, num.threads=num.threads, block=block),
                more.adt.marker.args
            )
        }

        if (!is.null(crispr.altexp)) {
            markers$crispr <- .call(
                scoreMarkers.se,
                list(.getModality(x, crispr.altexp), assay.type=crispr.norm.output.name),
                list(groups=chosen.clusters, num.threads=num.threads, block=block),
                more.crispr.marker.args
            )
        }
    }

    list(x=x, markers=markers)
}

#' @importFrom methods is
.delayifyAssays <- function(x) {
    for (i in seq_along(SummarizedExperiment::assayNames(x))) {
        SummarizedExperiment::assay(x, i, withDimnames=FALSE) <- DelayedArray(SummarizedExperiment::assay(x, i, withDimnames=FALSE))
    }
    if (is(x, "SingleCellExperiment")) {
        for (i in seq_along(SingleCellExperiment::altExpNames(x))) {
            SingleCellExperiment::altExp(x, i, withDimnames=FALSE) <- .delayifyAssays(SingleCellExperiment::altExp(x, i, withDimnames=FALSE))
        }
    }
    x
}

.extractOrError <- function(df, name) {
    val <- df[[name]]
    if (is.null(val)) {
        stop("no DataFrame column named '", name, "'")
    }
    val
}

.getModality <- function(x, altexp) {
    if (is.na(altexp)) {
        x
    } else {
        SingleCellExperiment::altExp(x, altexp)
    }
}

.assignModality <- function(x, altexp, y) {
    if (is.na(altexp)) {
        y
    } else {
        SingleCellExperiment::altExp(x, altexp) <- y
        x
    }
}

.defineSingleTargetEmbedding <- function(x, altexp, output.name) {
    if (!is.na(altexp)) {
        if (is.numeric(altexp)) {
            nm <- SingleCellExperiment::altExpNames(x)[altexp]
        } else {
            nm <- altexp
        }
        names(output.name) <- nm
    }
    output.name
}

.addSourceEmbeddingToScale <- function(x, altexp, output.name, collected) {
    if (is.na(altexp)) {
        collected$main <- append(collected$main, output.name)
    } else {
        if (is.numeric(altexp)) {
            nm <- SingleCellExperiment::altExpNames(x)[altexp]
        } else {
            nm <- altexp
        }
        collected$altexp[[nm]] <- output.name
    }
    collected
}

# library(testthat); library(scrapper); source("test-se_analyze.R")

set.seed(12345)
library(SingleCellExperiment)
mean <- runif(100, 0, 5)
mat <- matrix(rpois(10000, mean * 10), ncol=100)
se <- SummarizedExperiment(list(counts=mat))
rownames(se) <- paste0("GENE_", seq_len(nrow(se)))

# Just fiddling with some of the parameters to cut down the runtime.
default <- analyze.se(se, more.tsne.args=list(max.iterations=10), more.umap.args=list(num.epochs=5), num.threads=1)

test_that("analyze.se works correctly with a default run", {
    res <- default

    expect_identical(rownames(res$x), rownames(se))
    expect_true(all(res$x$keep))
    expect_type(res$x$keep, "logical")
    expect_null(res$x$combined.keep)

    expect_s4_class(counts(res$x), "DelayedMatrix")
    expect_s4_class(logcounts(res$x), "DelayedMatrix")

    expect_type(rowData(res$x)$residuals, "double")

    expect_true("PCA" %in% reducedDimNames(res$x))
    expect_true("TSNE" %in% reducedDimNames(res$x))
    expect_true("UMAP" %in% reducedDimNames(res$x))
    expect_false("combined" %in% reducedDimNames(res$x))
    expect_false("MNN" %in% reducedDimNames(res$x))

    expect_true(is.factor(res$x$graph.cluster))
    expect_identical(names(res$markers$rna), levels(res$x$graph.cluster))
})

test_that("analyze.se works correctly with no filtering", {
    res <- analyze.se(se, more.tsne.args=list(max.iterations=10), filter.cells=FALSE, more.umap.args=list(num.epochs=5), num.threads=1)
    expect_identical(dim(res$x), dim(se))
    expect_true(is.matrix(counts(res$x)))
})

test_that("analyze.se works correctly with k-means clustering", {
    res <- analyze.se(se, more.tsne.args=list(max.iterations=10), more.umap.args=list(num.epochs=5), cluster.graph.output.name=NULL, kmeans.clusters=10, num.threads=1)
    expect_null(res$x$graph.cluster)
    expect_true(is.factor(res$x$kmeans.cluster))
    expect_identical(names(res$markers$rna), levels(res$x$kmeans.cluster))
})

test_that("analyze.se works correctly with no clustering", {
    res <- analyze.se(se, more.tsne.args=list(max.iterations=10), more.umap.args=list(num.epochs=5), cluster.graph.output.name=NULL, num.threads=1)
    expect_null(res$x$graph.cluster)
    expect_null(res$x$kmeans.cluster)
    expect_null(res$markers)
})

test_that("analyze.se works correctly with blocking", {
    block <- rep(LETTERS[1:4], length.out=ncol(mat))
    res <- analyze.se(se, block=block, more.tsne.args=list(max.iterations=10), more.umap.args=list(num.epochs=5), num.threads=1)
    expect_true("MNN" %in% reducedDimNames(res$x))
    expect_type(res$x$block, "character")
    expect_identical(names(metadata(res$x)$qc$thresholds$sum), LETTERS[1:4])

    # Confirm that blocked coordinates are actually used downstream.
    downstream <- runAllNeighborSteps(as.matrix(t(reducedDim(res$x, "MNN"))), runTsne.args=list(max.iterations=10), runUmap.args=list(num.epochs=5), num.threads=1)
    expect_identical(downstream$clusterGraph$membership, res$x$graph.cluster)
    expect_identical(downstream$runTsne, reducedDim(res$x, "TSNE"))
})

test_that("analyze.se works correctly with combined RNA+ADT", {
    sce <- as(se, "SingleCellExperiment")
    amat <- matrix(rpois(1000, 10), ncol=100)
    rownames(amat) <- paste0("TAG_", seq_len(nrow(amat)))
    altExp(sce, "ADT") <- SummarizedExperiment(list(counts=amat))

    # Works with RNA + ADT. We use a different number of PCs for the ADTs to get some variety.
    res <- analyze.se(sce, adt.altexp="ADT", more.adt.pca.args=list(number=10), more.tsne.args=list(max.iterations=10), more.umap.args=list(num.epochs=5), num.threads=1)

    expect_identical(rownames(altExp(res$x)), rownames(altExp(sce)))
    expect_type(altExp(res$x, "ADT")$keep, "logical")
    expect_type(res$x$combined.keep, "logical")
    expect_true(all(res$x$combined.keep))

    expect_s4_class(counts(altExp(res$x)), "DelayedMatrix")
    expect_s4_class(logcounts(altExp(res$x)), "DelayedMatrix")
    expect_null(rowData(altExp(res$x))$hvg)
    expect_identical(ncol(reducedDim(altExp(res$x), "PCA")), 10L)

    expect_type(rowData(res$x)$hvg, "logical")
    expect_identical(ncol(reducedDim(res$x, "PCA")), 25L)

    expect_identical(ncol(reducedDim(res$x, "combined")), 35L)
    expect_identical(names(metadata(res$x)$combined$main.scaling), "PCA")
    expect_identical(names(metadata(res$x)$combined$altexp.scaling$ADT), "PCA")

    expect_identical(names(res$markers$rna), levels(res$x$graph.cluster))
    expect_identical(names(res$markers$adt), levels(res$x$graph.cluster))

    # Also works with ADT + RNA where ADT is the main experiment.
    sce2 <- as(altExp(sce), "SingleCellExperiment")
    altExp(sce2, "RNA") <- se

    res2 <- analyze.se(sce2, rna.altexp="RNA", adt.altexp=NA, more.tsne.args=list(max.iterations=10), more.umap.args=list(num.epochs=5), num.threads=1)
    expect_null(rowData(res2$x)$hvg)
    expect_type(rowData(altExp(res2$x))$hvg, "logical")

    expect_identical(reducedDim(res2$x, "PCA"), reducedDim(altExp(res$x, "ADT"), "PCA"))
    expect_identical(reducedDim(altExp(res2$x, "RNA"), "PCA"), reducedDim(res$x, "PCA"))
    expect_identical(dim(reducedDim(res2$x, "combined")), dim(reducedDim(res$x, "combined")))

    expect_identical(res$markers, res2$markers)

    # Confirm that scaled coordinates are actually used downstream.
    ref <- runAllNeighborSteps(as.matrix(t(reducedDim(res$x, "combined"))), runTsne.args=list(max.iterations=10), runUmap.args=list(num.epochs=5), num.threads=1)
    expect_identical(ref$clusterGraph$membership, res$x$graph.cluster)
    expect_identical(ref$runTsne, reducedDim(res$x, "TSNE"))

    # Fails if neither ADT or RNA is supplied.
    expect_error(analyze.se(sce, use.rna.pcs=FALSE, use.adt.pcs=FALSE), "at least one")
})

test_that("analyze.se works correctly with ADT only", {
    amat <- matrix(rpois(1000, 10), ncol=100)
    rownames(amat) <- paste0("TAG_", seq_len(nrow(amat)))
    ase <- SummarizedExperiment(list(counts=amat))

    # Works with ADT in the main experiment.
    res <- analyze.se(ase, rna.altexp=NULL, adt.altexp=NA, more.tsne.args=list(max.iterations=10), more.umap.args=list(num.epochs=5), num.threads=1)
    expect_type(res$x$keep, "logical")
    expect_true("PCA" %in% reducedDimNames(res$x))
    expect_false("combined" %in% reducedDimNames(res$x))
    expect_null(res$markers$rna)
    expect_identical(names(res$markers), "adt")

    # Works after ignoring RNA in the main experiment.
    # This checks that the use of a named character vector for 'target.embedding' is respected within analyze.se().
    sce <- as(se, "SingleCellExperiment")
    altExp(sce, "ADT") <- ase

    res2 <- analyze.se(sce, rna.altexp=NULL, adt.altexp="ADT", more.tsne.args=list(max.iterations=10), more.umap.args=list(num.epochs=5), num.threads=1)
    expect_identical(reducedDim(altExp(res2$x, "ADT"), "PCA"), reducedDim(res$x, "PCA"))
    expect_identical(altExp(res2$x)$sizeFactor, res$x$sizeFactor)
    expect_identical(reducedDim(res2$x, "TSNE"), reducedDim(res$x, "TSNE")) # UMAP, TSNE, clusters still stored in the main experiment, though.
    expect_identical(reducedDim(res2$x, "UMAP"), reducedDim(res$x, "UMAP"))
    expect_identical(res2$x$graph.cluster, res$x$graph.cluster)
    expect_identical(res$markers, res2$markers)
})

test_that("analyze.se works correctly with the CRISPR modality", {
    sce <- as(se, "SingleCellExperiment")
    mat <- matrix(0, 20, 100)
    mat[cbind(sample(nrow(mat), ncol(mat), replace=TRUE), seq_len(ncol(mat)))] <- rpois(ncol(mat), 10)
    rownames(mat) <- paste0("GUIDE_", seq_len(nrow(mat)))
    altExp(sce, "CRISPR") <- SummarizedExperiment(list(counts=mat))

    res <- analyze.se(sce, crispr.altexp="CRISPR", more.tsne.args=list(max.iterations=10), more.umap.args=list(num.epochs=5), num.threads=1)
    expect_identical(rownames(altExp(res$x)), rownames(altExp(sce)))
    expect_type(altExp(res$x, "CRISPR")$keep, "logical")
    expect_type(res$x$combined.keep, "logical")
    expect_true(all(res$x$combined.keep))

    expect_s4_class(assay(altExp(res$x), "logcounts"), "DelayedMatrix")
    expect_identical(names(res$markers$crispr), levels(res$x$graph.cluster))
})

test_that(".extract_or_error works correctly", {
    se$foo <- runif(ncol(se))
    expect_identical(scrapper:::.extractOrError(colData(se), "foo"), se$foo)
    expect_error(scrapper:::.extractOrError(colData(se), "bar"), "no DataFrame")
})

test_that(".delayify_assays works correctly", {
    assay(se, "logcounts", withDimnames=FALSE) <- matrix(rpois(10000, mean * 10), ncol=100)
    sce <- as(se, "SingleCellExperiment")
    altExp(sce, "FOOBAR") <- se
    del <- scrapper:::.delayifyAssays(sce)

    expect_s4_class(assay(del, "counts"), "DelayedMatrix")
    expect_identical(as.matrix(assay(del, "counts")), assay(se, "counts"))

    expect_s4_class(assay(del, "logcounts"), "DelayedMatrix")
    expect_identical(as.matrix(assay(del, "logcounts")), assay(se, "logcounts"))

    expect_s4_class(assay(altExp(del), "counts"), "DelayedMatrix")
    expect_identical(as.matrix(assay(altExp(del), "counts")), assay(se, "counts"))
})

test_that(".define_single_target_embedding works as expected", {
    sce <- as(se, "SingleCellExperiment")
    altExp(sce, "FOOBAR") <- se
    expect_identical(scrapper:::.defineSingleTargetEmbedding(sce, NA, "PCA"), "PCA")
    expect_identical(scrapper:::.defineSingleTargetEmbedding(sce, 1, "PCA"), c(FOOBAR="PCA"))
    expect_identical(scrapper:::.defineSingleTargetEmbedding(sce, "STUFF", "PCA"), c(STUFF="PCA"))
})

test_that(".add_source_embedding_to_scale works as expected", {
    sce <- as(se, "SingleCellExperiment")
    altExp(sce, "FOOBAR") <- se

    everything <- list(main="A", altexp=list(YAY=2))
    expect_identical(scrapper:::.addSourceEmbeddingToScale(sce, NA, "PCA", everything), list(main=c("A", "PCA"), altexp=list(YAY=2)))
    expect_identical(scrapper:::.addSourceEmbeddingToScale(sce, 1, "PCA", everything), list(main="A", altexp=list(YAY=2, FOOBAR="PCA")))
    expect_identical(scrapper:::.addSourceEmbeddingToScale(sce, "STUFF", "PCA", everything), list(main="A", altexp=list(YAY=2, STUFF="PCA")))
})

#' Graph-based clustering of cells
#'
#' Identify clusters by applying community detection algorithms to a graph.
#' This assumes that that the nodes on the graph represent cells and weighted edges are formed between related cells.
#'
#' @param x List containing graph information or an external pointer to a graph, as returned by \code{\link{buildSnnGraph}}.
#' Alternatively, an \link[igraph]{igraph} object with edge weights.
#' @param method String specifying the algorithm to use.
#' \itemize{
#' \item \code{"multilevel"} uses multi-level modularity optimization, also known as the Louvain algorithm,
#' see \url{https://igraph.org/c/doc/igraph-Community.html#igraph_community_multilevel} for details.
#' \item \code{"walktrap"} uses the Walktrap community finding algorithm,
#' see \url{https://igraph.org/c/doc/igraph-Community.html#igraph_community_walktrap} for details.
#' \item \code{"leiden"} uses the Leiden algorithm,
#' see \url{https://igraph.org/c/doc/igraph-Community.html#igraph_community_leiden} for details.
#' }
#' @param multilevel.resolution Numeric scalar specifying the resolution when \code{method="multilevel"}.
#' Lower values favor fewer, larger communities; higher values favor more, smaller communities.
#' @param leiden.resolution Numeric scalar specifying the resolution when \code{method="leiden"}.
#' Lower values favor fewer, larger communities; higher values favor more, smaller communities.
#' @param leiden.objective String specifying the objective function when \code{method="leiden"}.
#' \code{"modularity"} uses the generalized modularity, \code{"cpm"} uses the Constant Potts Model, and \code{"er"} uses the Erd\"os-R\'enyi G(n, p) model.
#' The CPM typically yields more fine-grained clusters than the modularity at the same \code{leiden.resolution}.
#' @param walktrap.steps Integer scalar specifying the number of steps to use when \code{method="walktrap"}.
#' This determines the ability of the Walktrap algorithm to distinguish highly interconnected communities from the rest of the graph.
#' @param seed Integer scalar specifying the random seed to use for \code{method="multilevel"} or \code{"leiden"}.
#'
#' @return A list containing \code{membership}, a factor containing the cluster assignment for each cell.
#' Additional fields may be present depending on the \code{method}:
#' \itemize{
#' \item For \code{method="multilevel"}, the \code{levels} list contains the clustering result at each level of the algorithm.
#' A \code{modularity} numeric vector also contains the modularity at each level, the highest of which corresponds to the reported \code{membership}.
#' \item For \code{method="leiden"}, a \code{quality} numeric scalar containg the quality of the partitioning.
#' \item For \code{method="walktrap"}, a \code{merges} matrix specifies the pair of cells or clusters that were merged at each step of the algorithm.
#' A \code{modularity} numeric scalar also contains the modularity of the final partitioning.
#' }
#'
#' @author Aaron Lun
#'
#' @seealso
#' The various \code{cluster_*} functions in \url{https://libscran.github.io/scran_graph_cluster/}. 
#'
#' \code{\link{clusterGraph.se}}, which performs clustering on graph constructed from a \link[SingleCellExperiment]{SingleCellExperiment}.
#'
#' @examples
#' data <- matrix(rnorm(10000), ncol=1000)
#' gout <- buildSnnGraph(data)
#' str(gout)
#'
#' str(clusterGraph(gout))
#' str(clusterGraph(gout, method="leiden"))
#' str(clusterGraph(gout, method="walktrap"))
#'
#' @export
#' @importFrom methods is
clusterGraph <- function(
    x,
    method=c("multilevel", "leiden", "walktrap"),
    multilevel.resolution=1, 
    leiden.resolution=1, 
    leiden.objective=c("modularity", "cpm", "er"),
    walktrap.steps=4,
    seed=42
) {
    .checkSEX(x, "clusterGraph.se")
    if (is(x, "igraph")) {
        x <- list(
            vertices=igraph::vcount(x),
            edges=as.vector(t(igraph::as_edgelist(x, names=FALSE))),
            weights=igraph::E(x)$weight
        )
    }
    if (is.list(x)) {
        x <- list_to_graph(x)
    }

    method <- match.arg(method)
    if (method == "multilevel") {
        out <- cluster_multilevel(x, resolution=multilevel.resolution, seed=seed)
        out$levels <- lapply(out$levels, function(x) factor(x + 1L))
    } else if (method == "leiden") {
        out <- cluster_leiden(x, objective=match.arg(leiden.objective), resolution=leiden.resolution, seed=seed)
    } else if (method == "walktrap") {
        out <- cluster_walktrap(x, steps=walktrap.steps)
        out$merges <- out$merges + 1L
    }

    out$membership <- factor(out$membership + 1L)
    out
}

# library(testthat); library(scrapper); source("test-clusterGraph.R")

set.seed(98765)
data <- matrix(rnorm(10000), ncol=1000)
out <- buildSnnGraph(data)
ptr <- buildSnnGraph(data, as.pointer=TRUE)

unweighted <- out
unweighted[3] <- list(NULL)

test_that("clusterGraph works correctly for multilevel", {
    clust <- clusterGraph(out, method="multilevel")
    expect_identical(length(clust$membership), ncol(data))
    expect_lte(nlevels(clust$membership), ncol(data))
    expect_gte(nlevels(clust$membership), 1L)
    expect_false(anyNA(clust$membership))

    for (cl in clust$levels) {
        expect_identical(length(cl), ncol(data))
        expect_lte(nlevels(cl), ncol(data))
        expect_gte(nlevels(cl), 1L)
        expect_false(anyNA(cl))
    }

    # Something sensible happens with a pointer.
    clust2 <- clusterGraph(ptr, method="multilevel")
    expect_identical(clust, clust2)

    # Something sensible happens without weighting.
    clust.uw <- clusterGraph(unweighted, method="multilevel")
    expect_false(identical(clust, clust.uw))

    # Works with the graph.
    g <- igraph::make_undirected_graph(out$edges, n = out$vertices)
    igraph::E(g)$weight <- out$weights
    clusti <- clusterGraph(g, method="multilevel")
    expect_identical(clust, clusti)

    expect_error(clusterGraph(SummarizedExperiment::SummarizedExperiment(data)), "not supported")
})

test_that("clusterGraph works correctly for Leiden", {
    clust <- clusterGraph(out, method="leiden")
    expect_identical(length(clust$membership), ncol(data))
    expect_lte(nlevels(clust$membership), ncol(data))
    expect_gte(nlevels(clust$membership), 1L)
    expect_false(anyNA(clust$membership))

    clust.cpm <- clusterGraph(out, method="leiden", leiden.objective="cpm")
    expect_identical(length(clust.cpm$membership), ncol(data))
    expect_gte(nlevels(clust.cpm$membership), 1L)

    clust.er <- clusterGraph(out, method="leiden", leiden.objective="er")
    expect_identical(length(clust.er$membership), ncol(data))
    expect_gte(nlevels(clust.er$membership), 1L)

    # Something sensible happens without weighting.
    clust.uw <- clusterGraph(unweighted, method="leiden")
    expect_false(identical(clust, clust.uw))

    # Something sensible happens with a pointer.
    clust2 <- clusterGraph(ptr, method="leiden")
    expect_identical(clust, clust2)
})

test_that("clusterGraph works correctly for Walktrap", {
    clust <- clusterGraph(out, method="walktrap")
    expect_identical(length(clust$membership), ncol(data))
    expect_lte(nlevels(clust$membership), ncol(data))
    expect_gte(nlevels(clust$membership), 1L)
    expect_false(anyNA(clust$membership))

    # Something sensible happens without weighting.
    clust.uw <- clusterGraph(unweighted, method="walktrap")
    expect_false(identical(clust, clust.uw))

    # Something sensible happens with a pointer.
    clust2 <- clusterGraph(ptr, method="walktrap")
    expect_identical(clust, clust2)
})

#' Aggregate expression across cells
#'
#' Aggregate expression values across cells based on one or more grouping factors.
#' This is usually applied to a count matrix to create pseudo-bulk profiles for each cluster/sample combination.
#' These profiles can then be used as if they were counts from bulk data, e.g., for differential analyses with \pkg{edgeR}.
#'
#' @param x A matrix-like object where rows correspond to genes or genomic features and columns correspond to cells.
#' Values are typically expected to be counts.
#' @param factors A list or data frame (or their equivalents from \pkg{S4Vectors}) containing one or more grouping factors, see \code{\link{combineFactors}}.
#' @param compute.sum Boolean indicating whether to compute the sum in each group.
#' @param compute.detected Boolean indicating whether to compute the number of cells with detected expression in each group.
#' @param compute.median Boolean indicating whether to compute the median in each group.
#' @param num.threads Integer specifying the number of threads to be used for aggregation.
#'
#' @return A list containing:
#' \itemize{
#' \item \code{sums}, a numeric matrix where each row corresponds to a gene and each column corresponds to a unique combination of levels from \code{factors}.
#' Each entry contains the summed expression across all cells with that combination. 
#' Omitted if \code{compute.sum=FALSE}.
#' \item \code{detected}, an integer matrix where each row corresponds to a gene and each column corresponds to a unique combination of levels from \code{factors}.
#' Each entry contains the number of cells with detected (i.e., positive) expression in that combination.
#' Omitted if \code{compute.detected=FALSE}.
#' \item \code{medians}, a numeric matrix where each row corresponds to a gene and each column corresponds to a unique combination of levels from \code{factors}.
#' Each entry contains the median expression across all cells with that combination. 
#' Omitted if \code{compute.median=FALSE}.
#' \item \code{combinations}, a \link[S4Vectors]{DataFrame} containing the unique combination of levels from \code{factors}.
#' Rows correspond to columns of \code{sums} and \code{detected}, while columns correspond to the factors in \code{factors}.
#' \item \code{counts}, the number of cells associated with each combination.
#' Each entry corresponds to a row of \code{combinations}.
#' \item \code{index}, an integer vector of length equal to the number of cells in \code{x}.
#' This specifies the combination in \code{combinations} to which each cell was assigned.
#' }
#'
#' @seealso
#' The \code{aggregate_across_cells} function in \url{https://libscran.github.io/scran_aggregate/}.
#'
#' \code{\link{aggregateAcrossCells.se}}, to perform aggregation on a \link[SummarizedExperiment]{SummarizedExperiment}.
#'
#' \code{\link{aggregateAcrossGenes}}, to aggregate expression values across gene sets.
#'
#' @author Aaron Lun
#' @examples
#' # Mocking a matrix:
#' library(Matrix)
#' x <- round(abs(rsparsematrix(1000, 100, 0.1) * 100))
#'
#' # Simple aggregation:
#' clusters <- sample(LETTERS, 100, replace=TRUE)
#' agg <- aggregateAcrossCells(x, list(cluster=clusters))
#' str(agg)
#'
#' # Multi-factor aggregation
#' samples <- sample(1:5, 100, replace=TRUE)
#' agg2 <- aggregateAcrossCells(x, list(cluster=clusters, sample=samples))
#' str(agg2)
#' 
#' @export
#' @importFrom beachmat initializeCpp
#' @importFrom S4Vectors DataFrame
aggregateAcrossCells <- function(
    x,
    factors,
    compute.sum = TRUE,
    compute.detected = TRUE,
    compute.median = TRUE,
    num.threads = 1
) {
    .checkSEX(x, "aggregateAcrossCells.se")

    combined <- combineFactors(factors)

    ptr <- initializeCpp(x, .check.na=FALSE)
    output <- aggregate_across_cells(
        ptr,
        combined$index - 1L,
        compute_sum = compute.sum,
        compute_detected = compute.detected,
        compute_median = compute.median,
        num_threads = num.threads
    )

    output <- output[!vapply(output, is.null, TRUE)]
    for (can in names(output)) {
        rownames(output[[can]]) <- rownames(x)
    }

    output$combinations <- combined$levels
    output$counts <- tabulate(combined$index, nbins=nrow(output$combinations))
    output$index <- combined$index

    output
}

# library(testthat); library(scrapper); source("test-aggregateAcrossCells.R")

library(Matrix)

test_that("aggregateAcrossCells works for single factors", {
    x <- round(abs(rsparsematrix(1000, 100, 0.1) * 100))
    clusters <- sample(LETTERS, 100, replace=TRUE)
    agg <- aggregateAcrossCells(x, list(cluster=clusters))

    for (u in unique(clusters)) {
        chosen <- clusters == u
        m <- match(u, agg$combinations[,1])

        submat <- x[,chosen,drop=FALSE]
        sum.expected <- Matrix::rowSums(submat)
        expect_equal(sum.expected, agg$sums[,m])

        detected.expected <- Matrix::rowSums(submat > 0)
        expect_equal(detected.expected, agg$detected[,m])
    }

    expect_identical(agg$counts, as.integer(table(clusters)))

    expect_error(aggregateAcrossCells(SummarizedExperiment::SummarizedExperiment(x), list(cluster=clusters)), "not supported")
})

test_that("aggregateAcrossCells works for medians", {
    x <- matrix(runif(100000), ncol=100)
    clusters <- sample(LETTERS, 100, replace=TRUE)
    agg <- aggregateAcrossCells(x, list(cluster=clusters), compute.sum=FALSE, compute.detected=FALSE, compute.median=TRUE)

    for (u in unique(clusters)) {
        chosen <- clusters == u
        m <- match(u, agg$combinations[,1])
        submat <- x[,chosen,drop=FALSE]
        median.expected <- apply(submat, 1, median)
        expect_equal(median.expected, agg$medians[,m])
    }
})

test_that("aggregateAcrossCells works for multiple factors", {
    x <- round(abs(rsparsematrix(1000, 100, 0.1) * 100))
    clusters <- sample(LETTERS, 100, replace=TRUE)
    samples <- sample(1:3, 100, replace=TRUE)
    agg <- aggregateAcrossCells(x, list(sample=samples, cluster=clusters))

    full <- data.frame(samples, clusters)
    df <- unique(full)
    df <- df[order(df[,1], df[,2]),]

    expected.counts <- integer(nrow(df))
    for (u in seq_len(nrow(df))) {
        chosen <- clusters == df$clusters[u] & samples == df$samples[u]
        expected.counts[u] <- sum(chosen)
        m <- which(agg$combinations$cluster  == df$clusters[u] & agg$combinations$sample == df$samples[u])

        submat <- x[,chosen,drop=FALSE]
        sum.expected <- Matrix::rowSums(submat)
        expect_equal(sum.expected, agg$sums[,m])

        detected.expected <- Matrix::rowSums(submat > 0)
        expect_equal(detected.expected, agg$detected[,m])
    }

    expect_identical(agg$counts, expected.counts)
})

#' Build a shared nearest neighbor graph
#'
#' Build a shared nearest neighbor (SNN) graph where each node is a cell.
#' Edges are formed between cells that share one or more nearest neighbors, weighted by the number or ranking of those shared neighbors.
#' If two cells are close together but have distinct sets of neighbors, the corresponding edge is downweighted as the two cells are unlikely to be part of the same neighborhood.
#' In this manner, strongly weighted edges will only form within highly interconnected neighborhoods where many cells share the same neighbors.
#' This provides a more sophisticated definition of similarity between cells compared to a simpler (unweighted) nearest neighbor graph that just focuses on immediate proximity.
#'
#' @param x For \code{buildSnnGraph}, a numeric matrix where rows are dimensions and columns are cells,
#' typically containing a low-dimensional representation from, e.g., \code{\link{runPca}}.
#'
#' Alternatively, a named list of nearest-neighbor search results.
#' This should contain \code{index}, an integer matrix where rows are neighbors and columns are cells.
#' Each column contains 1-based indices for the nearest neighbors of the corresponding cell, ordered by increasing distance.
#' The number of neighbors for each cell should be equal to \code{num.neighbors}, otherwise a warning is raised.
#'
#' Alternatively, an index constructed by \code{\link[BiocNeighbors]{buildIndex}}.
#' @param num.neighbors Integer scalar specifying the number of neighbors to use to construct the graph.
#' Larger values increase the connectivity of the graph and reduce the granularity of subsequent community detection steps, at the cost of speed.
#' Ignored if \code{x} contains pre-computed neighbor search results. 
#' @param weight.scheme String specifying the weighting scheme to use for constructing the SNN graph.
#' This can be one of:
#' \itemize{
#' \item \code{"ranked"}, where the weight of the edge is defined by the smallest sum of ranks across all shared neighbors.
#' More shared neighbors, or shared neighbors that are close to both observations, will generally yield larger weights.
#' \item \code{"number"}, where the weight of the edge is the number of shared nearest neighbors between them. 
#' This is a simpler scheme that is also slightly faster but does not account for the ranking of neighbors within each set.
#' \item \code{"jaccard"}, where the weight of the edge is the Jaccard index of their neighbor sets,
#' This is a monotonic transformation of the weight used in \code{"number"}.
#' }
#' @param num.threads Integer scalar specifying the number of threads to use.
#' Only used if \code{x} is not a list of existing nearest-neighbor search results.
#' @param BNPARAM A \link[BiocNeighbors]{BiocNeighborParam} object specifying the algorithm to use.
#' Only used if \code{x} is not a list of existing nearest-neighbor search results.
#' @param as.pointer Logical scalar indicating whether to return an external pointer for direct use in \code{\link{clusterGraph}}.
#' This avoids the extra memory usage caused by conversion to/from an R list.
#'
#' @return If \code{as.pointer=FALSE}, a list is returned containing:
#' \itemize{
#' \item \code{vertices}, an integer scalar specifying the number of vertices in the graph (i.e., cells in \code{x}).
#' \item \code{edges}, an integer vector of 1-based indices for graph edges.
#' Pairs of values represent the endpoints of an (undirected) edge,
#' i.e., \code{edges[1:2]} form the first edge, \code{edges[3:4]} form the second edge and so on.
#' \item \code{weights}, a numeric vector of weights for each edge.
#' This has length equal to half the length of \code{edges}.
#' }
#'
#' If \code{as.pointer=TRUE}, an external pointer to the graph is returned that can be directly used in \code{\link{clusterGraph}}.
#'
#' @author Aaron Lun
#'
#' @seealso
#' The \code{build_snn_graph} function in \url{https://libscran.github.io/scran_graph_cluster/}.
#'
#' \code{\link{clusterGraph}}, to define clusters (i.e., communities) from the graph.
#'
#' \code{\link{clusterGraph.se}}, which builds an SNN graph from a \link[SingleCellExperiment]{SingleCellExperiment}.
#'
#' @examples
#' data <- matrix(rnorm(10000), ncol=1000)
#' out <- buildSnnGraph(data)
#' str(out)
#'
#' # We can use this to make an igraph::graph.
#' g <- igraph::make_undirected_graph(out$edges, n = out$vertices)
#' igraph::E(g)$weight <- out$weight
#'
#' @export 
#' @importFrom BiocNeighbors findKNN AnnoyParam
buildSnnGraph <- function(x, num.neighbors=10, weight.scheme="ranked", num.threads=1, BNPARAM=AnnoyParam(), as.pointer=FALSE) {
    .checkSEX(x, "clusterGraph.se")

    if (!is.list(x)) {
        x <- findKNN(x, k=num.neighbors, transposed=TRUE, get.index="transposed", get.distance=FALSE, num.threads=num.threads, BNPARAM=BNPARAM)
    } else {
        .checkNeighborResults(x$index, NULL)
    }

    out <- build_snn_graph(x$index, scheme=weight.scheme, num_threads=num.threads)
    if (!as.pointer) {
        out <- graph_to_list(out)
    }
    out
}

# library(testthat); library(scrapper); source("test-buildSnnGraph.R")

test_that("buildSnnGraph works correctly", {
    data <- matrix(rnorm(10000), ncol=1000)
    out <- buildSnnGraph(data)
    expect_identical(length(out$edges), 2L * length(out$weights))  
    expect_true(all(out$weights > 0))
    expect_true(all(out$edges >= 1 & out$edges <= ncol(data)))

    # Same results when given an index, and throwing in some parallelization.
    idx <- BiocNeighbors::buildIndex(data, transposed=TRUE, BNPARAM=BiocNeighbors::AnnoyParam())
    out2 <- buildSnnGraph(idx, num.threads=2)
    expect_identical(out, out2)

    # Same results when given a matrix of indices.
    res <- BiocNeighbors::findKNN(idx, k=10, get.distance=FALSE, get.index="transposed", BNPARAM=BiocNeighbors::AnnoyParam())
    out3 <- buildSnnGraph(res) 
    expect_identical(out, out3)

#    # Something sensible happens with a pointer.
#    out <- buildSnnGraph(data, as.pointer=TRUE)
#    expect_type(out, "externalptr")

    expect_error(buildSnnGraph(SummarizedExperiment::SummarizedExperiment(data)), "not supported")
})

#' Score gene set activity for each cell
#'
#' Compute per-cell scores for a gene set, defined as the column sums of a rank-1 approximation to the submatrix for the gene set.
#' This uses the same approach as the \pkg{GSDecon} package by Jason Hackney, adapted to use an approximate PCA (via IRLBA) and to support blocking.
#'
#' @param x A matrix-like object where rows correspond to genes or genomic features and columns correspond to cells.
#' Typically, the matrix is expected to contain log-expression values.
#' @param set Vector specifying the rows of \code{x} that belong to the gene set.
#' This may be an integer vector of row indices, a logical vector of length equal to the number of rows, or a character vector of row names.
#' For integer and character vectors, duplicate entries are ignored.
#' For a character vector, any string not present in \code{rownames(x)} is ignored.
#' @param rank Integer scalar specifying the rank of the approximation.
#' The default value of 1 assumes that each gene set only describes a single coordinated biological function.
#' @inheritParams runPca
#'
#' @return List containing:
#' \itemize{
#' \item \code{scores}, a numeric vector of per-cell scores for each column in \code{x}.
#' \item \code{weights}, a \link[S4Vectors]{DataFrame} containing \code{row}, an integer vector of ordered and unique row indices corresponding to the genes in \code{set};
#' and \code{weight}, a numeric vector of per-gene weights for each gene in \code{row}. 
#' }
#'
#' @author Aaron Lun
#'
#' @seealso
#' The \code{compute} and \code{compute_blocked} functions in \url{https://libscran.github.io/gsdecon/}.
#'
#' \code{\link{scoreGeneSet.se}}, to compute gene set scores from a \link[SummarizedExperiment]{SummarizedExperiment}.
#'
#' @author Aaron Lun
#' @examples
#' library(Matrix)
#' x <- round(abs(rsparsematrix(1000, 100, 0.1) * 100))
#' normed <- normalizeCounts(x, size.factors=centerSizeFactors(colSums(x)))
#' scoreGeneSet(normed, set=c(1,3,5,10,20,100))
#' 
#' @export
#' @importFrom beachmat initializeCpp tatami.dim tatami.subset
#' @importFrom S4Vectors DataFrame
scoreGeneSet <- function(
    x,
    set,
    rank=1, 
    scale=FALSE,
    block=NULL, 
    block.weight.policy=c("variable", "equal", "none"),
    variable.block.weight=c(0, 1000),
    extra.work=7,
    iterations=1000,
    seed=5489,
    realized=TRUE,
    num.threads=1) 
{
    .checkSEX(x, "scoreGeneSet.se")

    block <- .transformFactor(block)
    ptr <- initializeCpp(x, .check.na=FALSE)

    nr <- tatami.dim(ptr)[1]
    chosen <- .sanitizeGeneSet(set, n=nr, names=rownames(x))
    ptr <- tatami.subset(ptr, chosen, by.row=TRUE)

    out <- score_gene_set(
        ptr,
        block=block$index,
        rank=rank,
        scale=scale,
        block_weight_policy=match.arg(block.weight.policy),
        variable_block_weight=variable.block.weight,
        realized=realized,
        irlba_work=extra.work,
        irlba_iterations=iterations,
        irlba_seed=seed,
        num_threads=num.threads
    )

    out$weights <- DataFrame(row=chosen, weight=out$weights)
    out
}

.sanitizeGeneSet <- function(set, n, names, arg = "set") {
    # Make sure we obtain unique and sorted indices so that
    # the slicing operations in tatami.subset are efficient.
    if (is.numeric(set)) {
        set <- as.integer(set)
        if (anyDuplicated(set)) {
            set <- unique(set)
        }
        if (anyNA(set) || min(set) < 1 || max(set) > n) {
            stop("'", arg, "' contains out-of-range indices")
        }
        if (is.unsorted(set)) {
            set <- sort(set)
        }
        return(set)

    } else if (is.logical(set)) {
        if (length(set) != n) {
            stop("length of '", arg, "' should be equal to the number of rows")
        }
        return(which(set))

    } else {
        if (anyDuplicated(set)) {
            set <- unique(set)
        }
        set <- match(set, names)
        if (anyNA(set)) {
            set <- set[!is.na(set)]
        }
        if (is.unsorted(set)) {
            set <- sort(set)
        }
        return(set)
    }
}

# library(testthat); library(scrapper); source("test-scoreGeneSet.R")

library(Matrix)
x <- round(abs(rsparsematrix(1000, 100, 0.1) * 100))
normed <- normalizeCounts(x, size.factors=centerSizeFactors(Matrix::colSums(x)))

test_that("scoreGeneSet works more or less as expected", {
    pcs <- scoreGeneSet(normed, 1:50)
    expect_identical(length(pcs$scores), ncol(x))
    expect_identical(nrow(pcs$weights), 50L)

    # Now with blocking. 
    pcs <- scoreGeneSet(normed, block=sample(3, ncol(x), replace=TRUE), 1:50)
    expect_identical(length(pcs$scores), ncol(x))
    expect_identical(nrow(pcs$weights), 50L)

    expect_error(scoreGeneSet(SummarizedExperiment::SummarizedExperiment(normed), 1:50), "not supported")
})

test_that("sanitizeGeneSet works as expected", {
    expect_identical(scrapper:::.sanitizeGeneSet(5:50, n=100, names=NULL), 5:50)
    expect_identical(scrapper:::.sanitizeGeneSet(50:5, n=100, names=NULL), 5:50) # handles sorting
    expect_identical(scrapper:::.sanitizeGeneSet(c(5:50, 10:30), n=100, names=NULL), 5:50) # removes duplicates
    expect_error(scrapper:::.sanitizeGeneSet(5:50, n=10, names=NULL), "out-of-range")
    expect_error(scrapper:::.sanitizeGeneSet(NA_integer_, n=10, names=NULL), "out-of-range")

    x <- rbinom(100, 1, 0.5) == 1L
    expect_identical(scrapper:::.sanitizeGeneSet(x, n=100, names=NULL), which(x))
    expect_error(scrapper:::.sanitizeGeneSet(FALSE, n=10, names=NULL), "length of")

    expect_identical(scrapper:::.sanitizeGeneSet(LETTERS, n=100, names=LETTERS), 1:26)
    expect_identical(scrapper:::.sanitizeGeneSet(rev(LETTERS), n=100, names=LETTERS), 1:26) # handles sorting
    expect_identical(scrapper:::.sanitizeGeneSet(rep(LETTERS, each=2), n=100, names=LETTERS), 1:26)  # removes duplicates
    expect_identical(scrapper:::.sanitizeGeneSet(LETTERS[2:5], n=10, names=LETTERS[1:4]), c(2L, 3L, 4L)) # ignores missing values
})

#' Choose highly variable genes from a SummarizedExperiment 
#'
#' Model the mean-variance relationship across genes and choose highly variable genes (HVGs) based on the residuals of the fitted trend.
#' This calls \code{\link{modelGeneVariances}} on an assay of a \link[SummarizedExperiment]{SummarizedExperiment},
#' and then calls \code{\link{chooseHighlyVariableGenes}} on the residuals.
#'
#' @param x A \link[SummarizedExperiment]{SummarizedExperiment} object or one of its subclasses.
#' Rows correspond to genes and columns correspond to cells.
#' @param block Block assignment for each cell, to pass to \code{\link{modelGeneVariances}}.
#' @param num.threads Number of threads, to pass to \code{\link{modelGeneVariances}}.
#' @param more.var.args Named list of arguments to pass to \code{\link{modelGeneVariances}}.
#' @param top Number of HVGs to choose, to pass to \code{\link{chooseHighlyVariableGenes}}.
#' @param more.choose.args Named list of arguments to pass to \code{\link{chooseHighlyVariableGenes}}.
#' @param assay.type Integer or string specifying the assay of \code{x} containing the log-normalized expression matrix for the RNA data.
#' @param output.prefix String containing a prefix to add to the names of the \code{link[SummarizedExperiment]{rowData}} columns containing the output statistics.
#' @param include.per.block Logical scalar indicating whether the per-block statistics should be stored in the output \code{rowData}.
#' Only relevant if \code{block} is specified.
#' @param model.res List returned by \code{\link{modelGeneVariances}}.
#' @param choose.res Integer vector returned by \code{\link{chooseHighlyVariableGenes}}.
#' This may be \code{NULL}, in which case the identities of the HVGs will not be stored.
#'
#' @return
#' For \code{chooseRnaHvgs.se}, \code{x} is returned with the per-gene variance modelling statistics added to its \code{rowData}.
#' The \code{hvg} column in the \code{rowData} indicates whether a gene was chosen as a HVG.
#' If \code{include.per.block=TRUE} and \code{block} is specified, the per-block statistics are stored as a nested DataFrame in the \code{per.block} column.
#'
#' For \code{formatModelGeneVariancesResult}, a \link[S4Vectors]{DataFrame} is returned with the per-gene variance modelling statistics.
#' If \code{choose.res} is provided, a \code{hvg} column is also stored that indicates whether a gene was chosen as a HVG.
#'
#' @author Aaron Lun
#' @examples
#' library(SingleCellExperiment)
#' sce <- getTestRnaData.se("norm")
#' sce <- chooseRnaHvgs.se(sce, more.var.args=list(use.min.width=TRUE))
#' summary(rowData(sce)$hvg)
#'
#' plot(rowData(sce)$means, rowData(sce)$variances, col=factor(rowData(sce)$hvg))
#' curve(approxfun(rowData(sce)$means, rowData(sce)$fitted)(x), col="dodgerblue", add=TRUE)
#' 
#' @export
#' @importFrom S4Vectors cbind
chooseRnaHvgs.se <- function(
    x, 
    block = NULL,
    num.threads = 1,
    more.var.args = list(),
    top = 4000,
    more.choose.args = list(),
    assay.type = "logcounts",
    output.prefix = NULL,
    include.per.block = FALSE
) {
    info <- .call(
        modelGeneVariances,
        list(SummarizedExperiment::assay(x, assay.type)),
        list(block=block, num.threads=num.threads),
        more.var.args
    )

    hvg.index <- .call(
        chooseHighlyVariableGenes,
        list(info$statistics$residuals),
        list(top=top, larger=TRUE),
        more.choose.args
    )

    df <- formatModelGeneVariancesResult(info, choose.res=hvg.index, include.per.block=include.per.block)
    colnames(df) <- paste0(output.prefix, colnames(df))
    SummarizedExperiment::rowData(x) <- cbind(SummarizedExperiment::rowData(x), df)
    x
}

#' @export
#' @rdname chooseRnaHvgs.se
#' @importFrom S4Vectors make_zero_col_DFrame
formatModelGeneVariancesResult <- function(model.res, choose.res = NULL, include.per.block = FALSE) {
    df <- model.res$statistics

    if (include.per.block && !is.null(model.res$per.block)) {
        tmp <- make_zero_col_DFrame(nrow=nrow(df))
        for (n in names(model.res$per.block)) {
            tmp[[n]] <- model.res$per.block[[n]]
        }
        df[["per.block"]] <- tmp
    }

    if (!is.null(choose.res)) {
        is.hvg <- logical(nrow(df))
        is.hvg[choose.res] <- TRUE
        df[["hvg"]] <- is.hvg
    }

    df
}

# library(testthat); library(scrapper); source("test-se_chooseRnaHvgs.R")

set.seed(12341234)
library(SummarizedExperiment)
mean <- runif(100, 0, 5)
mat <- matrix(rpois(1000, mean * 10), ncol=10)
se <- SummarizedExperiment(list(counts=mat))
se <- normalizeRnaCounts.se(se)

test_that("chooseRnaHvgs.se works as expected", {
    out <- chooseRnaHvgs.se(se, top=20)
    expect_false(is.null(rowData(out)$means))
    expect_true(any(rowData(out)$hvg))
    expect_false(all(rowData(out)$hvg))

    # Works with a prefix.
    out <- chooseRnaHvgs.se(se, top=2000, output.prefix="VAR.")
    expect_type(rowData(out)$VAR.means, "double")
    expect_type(rowData(out)$VAR.variances, "double")
    expect_true(all(rowData(out)$VAR.hvgs))

    # Works with blocking.
    block <- rep(LETTERS[1:2], each=5)
    out <- chooseRnaHvgs.se(se, top=20, block=block)
    expect_null(rowData(out)$per.block)

    out <- chooseRnaHvgs.se(se, top=20, block=block, include.per.block=TRUE)
    expect_s4_class(rowData(out)$per.block, "DFrame")
    expect_type(rowData(out)$per.block$A$means, "double")
    expect_type(rowData(out)$per.block$B$residuals, "double")
})

#' Center size factors
#'
#' Scale the size factors so they are centered at unity.
#' This ensures that the original scale of the counts is preserved in the normalized values from \code{\link{normalizeCounts}},
#' which simplifies interpretation and ensures that any pseudo-count added prior to log-transformation has a predictable shrinkage effect.
#'
#' @param size.factors Numeric vector of size factors across cells.
#' Invalid size factors (e.g., non-positive, non-finite) will be ignored.
#' @param block Vector or factor of length equal to \code{size.factors}, specifying the block of origin for each cell.
#' Alternatively \code{NULL}, in which case all cells are assumed to be in the same block.
#' @param mode String specifying how to scale size factors across blocks.
#' This can be either \code{"lowest"} or \code{"per-block"}, see Details.
#' Only used if \code{block} is provided.
#' 
#' @return Numeric vector of length equal to \code{size.factors}, containing the centered size factors.
#'
#' @details
#' \code{"lowest"} will compute the average size factor in each block, identify the lowest average across all blocks, and then scale all size factors by that value.
#' Here, our normalization strategy involves downscaling all blocks to match the coverage of the lowest-coverage block.
#' This is useful for datasets with big differences in coverage between blocks as it avoids egregious upscaling of low-coverage blocks.
#' Specifically, strong upscaling allows the log-transformation to ignore any shrinkage from the pseudo-count.
#' This is problematic as it inflates differences between cells at log-values derived from low counts, increasing noise and overstating log-fold changes. 
#' Downscaling is safer as it allows the pseudo-count to shrink the log-differences between cells towards zero at low counts,
#' effectively sacrificing some information in the higher-coverage batches so that they can be compared to the low-coverage batches
#' (which is preferable to exaggerating the informativeness of the latter for comparison to the former).
#'
#' \code{"per-block"} will compute the average size factor in each block, and then scale each size factor by the average of block to which it belongs.
#' The scaled size factors are identical to those obtained by separate invocations of `center_size_factors()` on the size factors for each block.
#' This can be desirable to ensure consistency with independent analyses of each block - otherwise, the centering would depend on the size factors in other blocks.
#' However, any systematic differences in the size factors between blocks are lost, i.e., systematic changes in coverage between blocks will not be normalized.
#'
#' @author Aaron Lun
#'
#' @seealso
#' The \code{center_size_factors} and \code{center_size_factors_blocked} functions in \url{https://libscran.github.io/scran_norm/}.
#'
#' \code{\link{normalizeRnaCounts.se}} and related functions, which center the size factors prior to normalization of a \link[SummarizedExperiment]{SummarizedExperiment}.
#'
#' @examples
#' centerSizeFactors(runif(100))
#'
#' centerSizeFactors(runif(100), block=sample(3, 100, replace=TRUE))
#'
#' @export
centerSizeFactors <- function(size.factors, block=NULL, mode=c("lowest", "per-block")) {
    block <- .transformFactor(block)
    center_size_factors(size.factors, block$index, match.arg(mode) == "lowest")
}

# library(testthat); library(scrapper); source("test-centerSizeFactors.R")

set.seed(99999)

test_that("centerSizeFactors works as expected", {
    f <- runif(100)
    cf <- centerSizeFactors(f)
    expect_equal(mean(cf), 1)
    expect_true(abs(diff(range(cf/f))) < 1e-8)

    # Ignores weirdos.
    f[1] <- 0
    f[2] <- NA
    f[3] <- Inf
    f[4] <- -1
    cf <- centerSizeFactors(f)
    tmp <- centerSizeFactors(tail(f, -4))
    expect_equal(tail(cf, -4), tmp)
})

test_that("centerSizeFactors works with blocking", {
    f <- runif(100)
    b <- sample(3, 100, replace=TRUE)
    scaling <- vapply(split(f, b), mean, 0)

    cf <- centerSizeFactors(f, b)
    expect_equal(f / min(scaling), cf)

    cf <- centerSizeFactors(f, b, mode="per-block")
    expect_equal(f / unname(scaling[b]), cf)
})

# library(testthat); library(scrapper); source("test-runPca.R")

set.seed(991)
library(Matrix)
x <- round(abs(rsparsematrix(1000, 100, 0.1) * 100))
normed <- normalizeCounts(x, size.factors=centerSizeFactors(Matrix::colSums(x)))

test_that("runPCA works more or less as expected", {
    pcs <- runPca(normed)

    rm <- rowMeans(pcs$components)
    expect_true(all(abs(rm) < 1e-8))
    expect_identical(length(rm), 25L)

    expect_equal(pcs$center, Matrix::rowMeans(normed))
    expect_null(pcs$scale)

    expect_false(is.unsorted(-pcs$variance.explained))
    rv <- apply(pcs$components, 1, var)
    expect_equal(rv, pcs$variance.explained)

    expect_identical(nrow(pcs$rotation), 1000L)
    expect_identical(ncol(pcs$rotation), 25L)

    # Works with scaling.
    spcs <- runPca(normed, scale=TRUE)
    expect_identical(length(spcs$scale), 1000L)

    expect_error(runPca(SummarizedExperiment::SummarizedExperiment(x)), "not supported")
})

test_that("runPCA works with blocking", {
    block <- sample(3, ncol(x), replace=TRUE)
    pcs <- runPca(normed, block=block)
    expect_identical(pcs$block.ids, 1:3)

    rm <- rowMeans(pcs$components)
    expect_true(all(abs(rm) < 1e-8))
    expect_identical(length(rm), 25L)

    expect_equal(pcs$center[1,], Matrix::rowMeans(normed[,block==1]))
    expect_equal(pcs$center[2,], Matrix::rowMeans(normed[,block==2]))
    expect_equal(pcs$center[3,], Matrix::rowMeans(normed[,block==3]))

    # Variance isn't so easily computed this time, so we just check it's sorted.
    expect_false(is.unsorted(-pcs$variance.explained))

    expect_identical(nrow(pcs$rotation), 1000L)
    expect_identical(ncol(pcs$rotation), 25L)

    # Check that the PCs are different from the unblocked case.
    ref <- runPca(normed)
    expect_false(isTRUE(all.equal(ref$components, pcs$components)))
})

test_that("runPCA works with residual components", {
    block <- sample(3, ncol(x), replace=TRUE)
    pcs <- runPca(normed, block=block, components.from.residuals=TRUE)

    rm <- rowMeans(pcs$components)
    expect_true(all(abs(rm) < 1e-8))
    expect_identical(length(rm), 25L)

    expect_false(is.unsorted(-pcs$variance.explained))
    rv <- apply(pcs$components, 1, var)
    ratio <- rv / pcs$variance.explained
    expect_lt(diff(range(ratio)), 1e-8)

    expect_identical(nrow(pcs$rotation), 1000L)
    expect_identical(ncol(pcs$rotation), 25L)

    # Make sure it's different from the other options.
    ref1 <- runPca(normed)
    expect_false(isTRUE(all.equal(ref1$components, pcs$components)))
    ref2 <- runPca(normed, block=block)
    expect_false(isTRUE(all.equal(ref2$components, pcs$components)))
})

test_that("runPCA works when more PCs are requested than available", {
    pcs <- runPca(normed, number=150)
    expect_identical(nrow(pcs$components), min(dim(normed)))
    expect_identical(ncol(pcs$rotation), min(dim(normed)))
})

test_that("runPCA works with subsets", {
    ref <- runPca(normed)
    doubled <- rbind(normed, normed)
    out <- runPca(doubled, subset=seq_len(nrow(normed)) + nrow(normed))
    expect_identical(ref$components, out$components)
    expect_equal(ref$rotation, out$rotation[seq_len(nrow(normed)),], tolerance=1e-5)
    expect_equal(ref$rotation, out$rotation[nrow(normed) + seq_len(nrow(normed)),])

    # Throwing in some blocks.
    block <- sample(3, ncol(x), replace=TRUE)
    ref <- runPca(normed, block=block)
    doubled <- rbind(normed, normed)
    out <- runPca(doubled, block=block, subset=seq_len(nrow(normed)) + nrow(normed))
    expect_identical(ref$components, out$components)
    expect_equal(ref$rotation, out$rotation[seq_len(nrow(normed)),], tolerance=1e-5)
    expect_equal(ref$rotation, out$rotation[nrow(normed) + seq_len(nrow(normed)),])
})

# library(testthat); library(scrapper); source("test-normalizeCounts.R")

set.seed(8888)
library(Matrix)
x <- round(abs(rsparsematrix(1000, 100, 0.1) * 100))
sf <- centerSizeFactors(Matrix::colSums(x))

test_that("normalizeCounts gives the same results in all modes", {
    y <- normalizeCounts(x, sf)
    expect_s4_class(y, "LogNormalizedMatrix")
    z <- normalizeCounts(beachmat::initializeCpp(x), sf)
    expect_equal(y[1,], beachmat::tatami.get(z, 1, row=TRUE))
    expect_equal(y[,1], beachmat::tatami.get(z, 1, row=FALSE))

    y <- normalizeCounts(x, sf, delayed=FALSE)
    expect_s4_class(y, "dgCMatrix")
    expect_equal(y[1,], beachmat::tatami.get(z, 1, row=TRUE))
    expect_equal(y[,1], beachmat::tatami.get(z, 1, row=FALSE))

    y <- normalizeCounts(x, sf, log=FALSE)
    expect_s4_class(y, "DelayedMatrix")
    z <- normalizeCounts(beachmat::initializeCpp(x), sf, log=FALSE)
    expect_equal(y[1,], beachmat::tatami.get(z, 1, row=TRUE))
    expect_equal(y[,1], beachmat::tatami.get(z, 1, row=FALSE))

    y <- normalizeCounts(x, sf, log=FALSE, delayed=FALSE)
    expect_s4_class(y, "dgCMatrix")
    expect_equal(y[1,], beachmat::tatami.get(z, 1, row=TRUE))
    expect_equal(y[,1], beachmat::tatami.get(z, 1, row=FALSE))

    y <- normalizeCounts(x, sf, log.base=10)
    z <- normalizeCounts(beachmat::initializeCpp(x), sf, log.base=10)
    expect_equal(y[1,], beachmat::tatami.get(z, 1, row=TRUE))
    expect_equal(y[,1], beachmat::tatami.get(z, 1, row=FALSE))

    y <- normalizeCounts(x, sf, pseudo.count=3)
    z <- normalizeCounts(beachmat::initializeCpp(x), sf, pseudo.count=3)
    expect_equal(y[1,], beachmat::tatami.get(z, 1, row=TRUE))
    expect_equal(y[,1], beachmat::tatami.get(z, 1, row=FALSE))

    y <- normalizeCounts(x, sf, pseudo.count=3, preserve.sparsity=TRUE)
    z <- normalizeCounts(beachmat::initializeCpp(x), sf, pseudo.count=3, preserve.sparsity=TRUE)
    expect_equal(y[1,], beachmat::tatami.get(z, 1, row=TRUE))
    expect_equal(y[,1], beachmat::tatami.get(z, 1, row=FALSE))

    expect_error(normalizeCounts(SummarizedExperiment::SummarizedExperiment(x)), "not supported")
})

# library(testthat); library(scrapper); source("test-LogNormalizedMatrix.R")

test_that("LogNormalizedMatrix works for dense matrices", {
    mat <- matrix(rpois(1000, lambda=2), ncol=10)
    sf <- centerSizeFactors(colSums(mat))

    norm <- LogNormalizedMatrix(mat, sf)
    expect_s4_class(norm, "LogNormalizedMatrix")
    expect_identical(dim(norm), dim(mat))
    expect_false(is_sparse(norm))
    expect_output(show(norm), "LogNormalizedMatrix")

    ref <- log1p(t(t(mat) / sf)) / log(2)
    expect_equal(as.matrix(norm), ref)
    expect_equal(extract_array(norm, list(1:10, 1:10)), ref[1:10,1:10])

    norm <- LogNormalizedMatrix(mat, sf, pseudo.count=2)
    ref <- log2(t(t(mat) / sf) + 2)
    expect_equal(as.matrix(norm), ref)

    norm <- LogNormalizedMatrix(mat, sf, log.base=10)
    ref <- log10(t(t(mat) / sf) + 1)
    expect_equal(as.matrix(norm), ref)

    dimnames(mat) <- list(paste0("GENE-", 1:100), LETTERS[1:10])
    norm <- LogNormalizedMatrix(mat, sf)
    expect_identical(dimnames(mat), dimnames(norm))

    expect_error(LogNormalizedMatrix(mat, 1), "equal number of columns")
})

library(Matrix)
test_that("LogNormalizedMatrix works for sparse matrices", {
    smat <- abs(rsparsematrix(50, 20, density=0.21))
    sf <- centerSizeFactors(Matrix::colSums(smat))

    norm <- LogNormalizedMatrix(smat, sf)
    expect_true(is_sparse(norm))

    ref <- as.matrix(log1p(t(t(smat) / sf)) / log(2))
    expect_equal(extract_array(norm, list(1:10, 1:10)), ref[1:10,1:10])

    expect_equal({
        tmp <- as.matrix(norm)
        dimnames(tmp) <- NULL
        tmp
    }, as.matrix(ref))

    extracted <- extract_sparse_array(norm, list(NULL, NULL))
    expect_s4_class(extracted, "SVT_SparseMatrix")
    expect_equal(as.matrix(extracted), ref)

    extracted <- extract_sparse_array(norm, list(20:50, 10:20))
    expect_s4_class(extracted, "SVT_SparseMatrix")
    expect_equal(as.matrix(extracted), ref[20:50,10:20])
})

test_that("LogNormalizedMatrix works with initializeCpp", {
    mat <- matrix(rpois(1000, lambda=1), ncol=10)
    sf <- centerSizeFactors(colSums(mat))

    norm <- LogNormalizedMatrix(mat, sf)
    ref <- as.matrix(norm)
    ptr <- beachmat::initializeCpp(norm)
    expect_equal(ref, beachmat::tatami.realize(ptr, num.threads=1))

    # Also works with sparse matrices.
    smat <- as(mat, "dgCMatrix")
    snorm <- LogNormalizedMatrix(smat, sf)
    sptr <- beachmat::initializeCpp(snorm)
    expect_equal(as(snorm, "dgCMatrix"), beachmat::tatami.realize(sptr, num.threads=1))
})

#' Count cells in groups and blocks
#'
#' Tabulate the frequency of cells in each combination of groups and blocks.
#' This is typically used to examine the distribution of cells across batches for each cluster -
#' the presence of a batch-specific cluster may be indicative of a batch effect.
#'
#' @param groups Factor specifying the group to which each cell was assigned.
#' This is typically used for clusters.
#' @param block Factor specifying the block to which each cell was assigned.
#' This is typically used for batches or samples.
#' @param normalize.block Boolean indicating whether to normalize the number of cells across blocks.
#' If \code{TRUE}, frequencies are divided by the column sums.
#' @param normalize.groups Boolean indicating whether to normalize the number of cells across groups.
#' If \code{TRUE}, frequencies are divided by the row sums.
#' This is performed after normalization of the block counts if \code{normalize.block=TRUE}.
#' 
#' @return Matrix of (normalized) frequencies.
#' Each row corresponds to a group and each column corresponds to a block.
#'
#' @author Aaron Lun
#' @examples
#' groups <- sample(10, 100, replace=TRUE)
#' block <- sample(LETTERS[1:6], 100, replace=TRUE)
#'
#' countGroupsByBlock(groups, block)
#' countGroupsByBlock(groups, block, normalize.block=TRUE)
#' countGroupsByBlock(groups, block, normalize.groups=TRUE)
#' countGroupsByBlock(groups, block, normalize.block=TRUE, normalize.groups=TRUE)
#'
#' @seealso
#' \code{\link{table}}, which is used internally by this function.
#'
#' @export
countGroupsByBlock <- function(groups, block, normalize.block=FALSE, normalize.groups=FALSE) {
    tab <- table(groups, block)
    if (normalize.block) {
        tab <- t(t(tab) / colSums(tab))
    }
    if (normalize.groups) {
        tab <- tab / rowSums(tab)
    }
    tab
}

#' Quality control for CRISPR count data
#'
#' Compute per-cell QC metrics from an initialized matrix of CRISPR counts,
#' and use the metrics to suggest filter thresholds to retain high-quality cells.
#' 
#' @param x A matrix-like object where rows are CRISPRs and columns are cells.
#' Values are expected to be counts.
#' @param num.threads Integer scalar specifying the number of threads to use.
#' @param metrics \link[S4Vectors]{DataFrame} of per-cell QC metrics.
#' This should have the same structure as the return value of \code{computeCrisprQcMetrics}.
#' @param block Factor specifying the block of origin (e.g., batch, sample) for each cell in \code{metrics}.
#' Alternatively \code{NULL} if all cells are from the same block.
#'
#' For \code{filterCrisprQcMetrics}, a blocking factor should be provided if \code{block} was used to construct \code{thresholds}. 
#' @param num.mads Number of median from the median, to define the threshold for outliers in each metric.
#' @param thresholds List with the same structure as produced by \code{suggestCrisprQcThresholds}.
#'
#' @return For \code{computeCrisprQcMetrics}, a \link[S4Vectors]{DataFrame} is returned with one row per cell in \code{x}.
#' This contains the following columns:
#' \itemize{
#' \item \code{sum}, a numeric vector containing the total CRISPR count for each cell.
#' Low counts indicate that the cell was not successfully transfected with a construct or that library preparation and sequencing failed.
#' \item \code{detected}, an integer vector containing the number of detected guides per cell.
#' In theory, this should be 1, as each cell should express no more than one guide construct.
#' However, ambient contamination may introduce non-zero counts for multiple guides, without necessarily interfering with downstream analyses.
#' As such, this metric is less useful for guide data, though we compute it anyway.
#' \item \code{max.value}, a numeric vector containing the count for the most abundant guide in cell.
#' Low values indicate that the cell was not successfully transfected or that library preparation and sequencing failed.
#' \item \code{max.index}, an integer vector containing the row index of the most abundant guide in cell.
#' }
#' Each vector is of length equal to the number of cells.
#'
#' For \code{suggestCrisprQcThresholds}, a named list is returned.
#' \itemize{
#' \item If \code{block=NULL}, the list contains:
#' \itemize{
#' \item \code{max.value}, a numeric scalar containing the lower bound on the maximum count. 
#' This is defined as \code{num.mads} MADs below the median of the log-transformed metrics across cells with high maximum proportions (see Details).
#' }
#' \item Otherwise, if \code{block} is supplied, the list contains:
#' \itemize{
#' \item \code{max.value}, a numeric vector containing the lower bound on the maximum counts for each blocking level.
#' Here, the threshold is computed independently for each block, using the same method as the unblocked case.
#' \item \code{block.ids}, a vector containing the identities of the unique blocks.
#' }
#' Each vector is of length equal to the number of levels in \code{block} and is named accordingly.
#' }
#'
#' For \code{filterCrisprQcMetrics}, a logical vector of length \code{ncol(x)} is returned indicating which cells are of high quality. 
#' High-quality cells are defined as those with maximum counts above the \code{max.value} threshold.
#'
#' @details
#' In CRISPR data, a cell is considered to be of low quality if it has a low count for its most abundant guide.
#' However, directly defining a MAD-based outlier threshold on the maximum count is somewhat tricky as unsuccessful transfection can be common.
#' This often results in a large subpopulation with low maximum counts, inflating the MAD and compromising the threshold calculation.
#' Instead, we use the following approach:
#' \itemize{
#' \item Compute the proportion of counts in the most abundant guide (i.e., the maximum proportion) in each cell.
#' Cells that were successfully transfected should have high maximum proportions.
#' In contrast, unsuccessfully transfected cells will be dominated by ambient contamination and have low proportions.
#' \item Subset the dataset to only retain those cells with maximum proportions above the median.
#' This assumes that at least 50% of cells were successfully transfected.
#' Thus, we remove all of the unsucessful transfections and enrich for mostly-high-quality cells.
#' \item Define a MAD-based threshold for low outliers on the log-transformed maximum count within the subset (see `choose_filter_thresholds()` for details).
#' This is now possible as we can assume that most of the remaining cells are of high quality.
#' }
#' Note that the maximum proportion is only used to define the subset for threshold calculation.
#' Once the maximum count threshold is computed, it is applied to all cells regardless of their maximum proportions.
#' This ensures that we correctly remove cells with low coverage, even if the proportion is high.
#' It also allows us to retain cells transfected with multiple guides, as long as the maximum is high enough -
#' such cells are not necessarily uninteresting, e.g., for examining interaction effects, so we will err on the side of caution and leave them in.
#'
#' @seealso
#' The \code{compute_crispr_qc_metrics}, \code{compute_crispr_qc_filters} and \code{compute_crispr_qc_filters_blocked} functions in \url{https://libscran.github.io/scran_qc/}.
#'
#' \code{\link{quickCrisprQc.se}}, to run all of the CRISPR-related QC functions on a \link[SummarizedExperiment]{SummarizedExperiment}.
#'
#' @author Aaron Lun
#' @examples
#' # Mocking a matrix:
#' library(Matrix)
#' x <- round(abs(rsparsematrix(100, 100, 0.1) * 100))
#'
#' qc <- computeCrisprQcMetrics(x)
#' qc
#'
#' filt <- suggestCrisprQcThresholds(qc)
#' str(filt)
#'
#' keep <- filterCrisprQcMetrics(filt, qc)
#' summary(keep)
#'
#' @export
#' @name crispr_quality_control
#' @importFrom S4Vectors DataFrame
computeCrisprQcMetrics <- function(x, num.threads = 1) {
    .checkSEX(x, "quickCrisprQc.se")
    y <- initializeCpp(x, .check.na=FALSE)
    output <- compute_crispr_qc_metrics(y, num_threads=num.threads)
    output$max.index <- output$max.index + 1L
    DataFrame(output, row.names=colnames(x))
}

#' @export
#' @rdname crispr_quality_control
suggestCrisprQcThresholds <- function(metrics, block=NULL, num.mads=3) {
    block <- .transformFactor(block)
    metrics <- as.list(metrics)
    metrics$max.index <- metrics$max.index - 1L # restore 0-based indexing.
    thresholds <- suggest_crispr_qc_thresholds(metrics, block=block$index, num_mads=num.mads)

    if (!is.null(block$names)) {
        names(thresholds$max.value) <- block$names
        thresholds$block.ids <- block$names # store it separately to preserve any non-character type.
    }

    thresholds
}

#' @export
#' @rdname crispr_quality_control
filterCrisprQcMetrics <- function(thresholds, metrics, block=NULL) {
    block <- .matchBlockThresholds(block, thresholds$block.ids)
    metrics <- as.list(metrics)
    metrics$max.index <- metrics$max.index - 1L # restore 0-based indexing.
    .checkThresholdNames(thresholds, c("max.value"), NULL, NULL)
    filter_crispr_qc_metrics(thresholds, metrics, block=block)
}

# library(testthat); library(scrapper); source("test-crispr_quality_control.R")

set.seed(66667777)
library(Matrix)
x <- round(abs(rsparsematrix(25, 1000, 0.1) * 100))
z <- as.matrix(x)

test_that("computeCrisprQcMetrics works as expected", { 
    qc <- computeCrisprQcMetrics(x)
    expect_identical(qc$sum, Matrix::colSums(x))
    expect_identical(qc$detected, Matrix::colSums(x > 0))

    top <- max.col(t(z))
    expect_equal(qc$max.value, z[cbind(top, seq_along(top))])

    # Can't compare top and top2 directly in case of ties.
    top2 <- qc$max.index
    expect_equal(qc$max.value, z[cbind(top2, seq_along(top2))])

    expect_error(computeCrisprQcMetrics(SummarizedExperiment::SummarizedExperiment(x)), "not supported")
})

test_that("suggestCrisprQcFilters works as expected", { 
    qc <- computeCrisprQcMetrics(x)
    num.mads <- 1.5
    thresholds <- suggestCrisprQcThresholds(qc, num.mads=num.mads)

    expected <- qc$max.value >= thresholds$max.value
    expected[is.na(expected)] <- TRUE
    observed <- filterCrisprQcMetrics(thresholds, qc)
    expect_identical(expected, observed)
})

test_that("suggestCrisprQcFilters works as expected with blocking", { 
    qc <- computeCrisprQcMetrics(x)
    block <- sample(3, ncol(x), replace=TRUE)
    num.mads <- 1.5
    thresholds <- suggestCrisprQcThresholds(qc, block=block, num.mads=num.mads)

    expect_identical(thresholds$block.ids, 1:3)

    expected <- qc$max.value >= unname(thresholds$max.value[block])
    expected[is.na(expected)] <- TRUE
    observed <- filterCrisprQcMetrics(thresholds, qc, block=block)
    expect_identical(expected, observed)

    # Same filtering with just the last block.
    last <- block == 3
    last_observed <- filterCrisprQcMetrics(
        thresholds,
        lapply(qc, function(x) x[last]),
        block=block[last]
    )
    expect_identical(observed[last], last_observed)
})

#' Uniform manifold approximation and projection
#'
#' Compute UMAP coordinates to visualize similarities between cells.
#'
#' @inheritParams runTsne
#' @param num.dim Integer scalar specifying the number of dimensions of the output embedding.
#' @param local.connectivity Numeric scalar specifying the number of nearest neighbors that are assumed to be always connected, with maximum membership confidence.
#' Larger values increase the connectivity of the embedding and reduce the focus on local structure.
#' This may be a fractional number of neighbors, in which case interpolation is performed when computing the membership confidence.
#' @param bandwidth Numeric scalar specifying the effective bandwidth of the kernel when converting the distance to a neighbor into a fuzzy set membership confidence.
#' Larger values reduce the decay in confidence with respect to distance, increasing connectivity and favoring global structure. 
#' @param mix.ratio Numeric scalar between 0 and 1 specifying the mixing ratio when combining fuzzy sets.
#' A mixing ratio of 1 will take the union of confidences, a ratio of 0 will take the intersection, and intermediate values will interpolate between them.
#' Larger values favor connectivity and more global structure.
#' @param spread Numeric scalar specifying the scale of the coordinates of the final low-dimensional embedding.
#' Ignored if \code{a} and \code{b} are provided.
#' @param min.dist Numeric scalar specifying the minimum distance between observations in the final low-dimensional embedding.
#' Smaller values will increase local clustering while larger values favor a more even distribution of observations throughout the low-dimensional space.
#' This is interpreted relative to \code{spread}.
#' Ignored if \code{a} and \code{b} are provided.
#' @param a Numeric scalar specifying the \eqn{a} parameter for the fuzzy set membership strength calculations.
#' Larger values yield a sharper decay in membership strength with increasing distance between observations.
#' If this or \code{b} are \code{NULL}, a suitable value for this parameter is automatically determined from \code{spread} and \code{min.dist}.
#' @param b Numeric scalar specifying the \eqn{b} parameter for the fuzzy set membership strength calculations.
#' Larger values yield an earlier decay in membership strength with increasing distance between observations.
#' If this or \code{a} are \code{NULL}, a suitable value for this parameter is automatically determined from \code{spread} and \code{min.dist}.
#' @param repulsion.strength Numeric scalar specifying the modifier for the repulsive force.
#' Larger values increase repulsion and favor local structure.
#' @param initialize.method String specifying how to initialize the embedding.
#' This should be one of:
#' \itemize{
#' \item \code{spectral}: spectral decomposition of the normalized graph Laplacian.
#' Specifically, the initial coordinates are defined from the eigenvectors corresponding to the smallest non-zero eigenvalues.
#' This fails in the presence of multiple graph components or if the approximate SVD fails to converge.
#' \item \code{random}: fills the embedding with random draws from a normal distribution.
#' \item \code{none}: uses initial values from \code{initial.coordinates}.
#' }
#' @param initialize.random.on.spectral.fail Logical scalar indicating whether to fall back to random sampling (i.e., same as \code{random})
#' if spectral initialization fails due to the presence of multiple components in the graph.
#' If \code{FALSE}, the values in \code{initial.coordinates} will be used instead, i.e., same as \code{none}.
#' Only relevant if \code{initialize.method = "spectral"} and spectral initialization fails.
#' @param initialize.spectral.scale Numeric scalar specifying the maximum absolute magnitude of the coordinates after spectral initialization.
#' All initial coordinates are scaled such that the maximum of the absolute values is equal to \code{initialize.spectral.scale}.
#' This ensures that outlier observations will not have large absolute distances that may interfere with optimization.
#' Only relevant if \code{initialize.method = "spectral"} and spectral initialization does not fail.
#' @param initialize.spectral.jitter Logical scalar indicating whether to jitter coordinates after spectral initialization to separate duplicate observations (e.g., to avoid overplotting).
#' This is done using normally-distributed noise of mean zero and standard deviation of \code{initialize.spectral.jitter.sd}.
#' Only relevant if \code{initialize.method = "spectral"} and spectral initialization does not fail.
#' @param initialize.spectral.jitter.sd Numeric scalar specifying the standard deviation of the jitter to apply after spectral initialization.
#' Only relevant if \code{initialize.method = "spectral"} and spectral initialization does not fail and \code{initialize.spectral.jitter = TRUE}.
#' @param initialize.random.scale Numeric scalar specifying the scale of the randomly generated initial coordinates.
#' Coordinates are sampled from a uniform distribution from \eqn{[-x, x)} where \eqn{x} is \code{initialize.random.scale}.
#' Only relevant if \code{initialize.method = "random"},
#' or \code{initialize.method = "spectral"} and spectral initialization fails and \code{initialize.random.on.spectral.fail = TRUE}.
#' @param initialize.seed Numeric scalar specifying the seed for the random number generation during initialization.
#' Only relevant if \code{initialize.method = "random"},
#' or \code{initialize.method = "spectral"} and \code{initialize.spectral.jitter = TRUE};
#' or \code{initialize.method = "spectral"} and spectral initialization fails and \code{initialize.random.on.spectral.fail = TRUE}.
#' @param initial.coordinates Numeric matrix of initial coordinates, with number of rows equal to the number of observations and number of columns equal to \code{num.dim}.
#' Only relevant if \code{initialize.method = "none"};
#' or \code{initialize.method = "spectral"} and spectral initialization fails and \code{initialize.random.on.spectral.fail = FALSE}.
#' @param num.epochs Integer scalar specifying the number of epochs for the gradient descent, i.e., optimization iterations. 
#' Larger values improve accuracy at the cost of increased compute time.
#' If \code{NULL}, a value is automatically chosen based on the size of the dataset:
#' \itemize{
#' \item For datasets with no more than 10000 observations, the default number of epochs is set to 500.
#' \item For larger datasets, the number of epochs is inversely proportional to the number of cells, starting from 500 and decreasing asymptotically to a lower limit of 200.
#' This choice aims to reduce computational work for very large datasets. 
#' }
#' @param learning.rate Numeric scalar specifying the initial learning rate used in the gradient descent.
#' Larger values can accelerate convergence but at the risk of skipping over suitable local optima.
#' @param negative.sample.rate Numeric scalar specifying the rate of sampling negative observations to compute repulsive forces.
#' Greater values will improve accuracy but increase compute time. 
#' @param num.neighbors Integer scalar specifying the number of neighbors to use to define the fuzzy sets.
#' Larger values improve connectivity and favor preservation of global structure, at the cost of increased compute time.
#' Ignored if \code{x} contains pre-computed neighbor search results. 
#' @param optimize.seed Numeric scalar specifying the seed to use for the optimization epochs. 
#' @param num.threads Integer scalar specifying the number of threads to use.
#' @param parallel.optimization Logical scalar specifying whether to parallelize the optimization step.
#' 
#' @return 
#' A numeric matrix where rows are cells and columns are the two dimensions of the embedding.
#' 
#' @author Aaron Lun
#'
#' @seealso
#' \url{https://libscran.github.io/umappp/}, for details on the underlying implementation.
#'
#' \code{\link{runUmap.se}}, to run UMAP on a \link[SingleCellExperiment]{SingleCellExperiment}.
#'
#' @references
#' McInnes L, Healy J, Melville J (2020).
#' UMAP: Uniform Manifold Approximation and Projection for Dimension Reduction.
#' \emph{arXiv}, \url{https://arxiv.org/abs/1802.03426}
#'
#' @examples
#' x <- t(as.matrix(iris[,1:4]))
#' embedding <- runUmap(x)
#' plot(embedding[,1], embedding[,2], col=iris[,5])
#' 
#' @export
#' @importFrom BiocNeighbors findKNN AnnoyParam
runUmap <- function(
    x, 
    num.dim=2,
    local.connectivity=1,
    bandwidth=1,
    mix.ratio=1,
    spread=1,
    min.dist=0.1, 
    a=NULL,
    b=NULL,
    repulsion.strength=1,
    initialize.method=c("spectral", "random", "none"),
    initial.coordinates=NULL,
    initialize.random.on.spectral.fail=TRUE,
    initialize.spectral.scale=10,
    initialize.spectral.jitter=FALSE,
    initialize.spectral.jitter.sd=0.0001,
    initialize.random.scale=10,
    initialize.seed=9876543210,
    num.epochs=NULL, 
    learning.rate=1,
    negative.sample.rate=5,
    num.neighbors=15, 
    optimize.seed=1234567890, 
    num.threads=1,
    parallel.optimization=FALSE,
    BNPARAM=AnnoyParam())
{
    .checkSEX(x, "runUmap.se")

    if (!is.list(x)) {
        x <- findKNN(x, k=num.neighbors, transposed=TRUE, get.index="transposed", get.distance="transposed", num.threads=num.threads, BNPARAM=BNPARAM)
    } else {
        .checkNeighborResults(x$index, x$distance)
    }

    if (!is.null(initial.coordinates)) {
        num.obs <- as.integer(ncol(x$index))
        stopifnot(identical(nrow(initial.coordinates), num.obs))
        stopifnot(identical(ncol(initial.coordinates), as.integer(num.dim)))
        initial.coordinates <- t(initial.coordinates)
    }

    output <- run_umap(
        nnidx=x$index, 
        nndist=x$distance, 
        num_dim=num.dim,
        local_connectivity=local.connectivity,
        bandwidth=bandwidth,
        mix_ratio=mix.ratio,
        spread=spread,
        min_dist=min.dist,
        a=a,
        b=b,
        repulsion_strength=repulsion.strength,
        initialize_method=match.arg(initialize.method),
        initial_coordinates=initial.coordinates,
        initialize_random_on_spectral_fail=initialize.random.on.spectral.fail,
        initialize_spectral_scale=initialize.spectral.scale,
        initialize_spectral_jitter=initialize.spectral.jitter,
        initialize_spectral_jitter_sd=initialize.spectral.jitter.sd,
        initialize_random_scale=initialize.random.scale,
        initialize_seed=initialize.seed,
        num_epochs=num.epochs,
        learning_rate=learning.rate,
        negative_sample_rate=negative.sample.rate,
        optimize_seed=optimize.seed,
        num_threads=num.threads,
        parallel_optimization=parallel.optimization
    )

    t(output)
}

# library(testthat); library(scrapper); source("test-runUmap.R")

library(BiocNeighbors)
x <- t(as.matrix(iris[,1:4]))
ref <- runUmap(x)

test_that("runUmap works in basic mode", {
    expect_identical(nrow(ref), ncol(x))
    expect_identical(ncol(ref), 2L)

    again <- runUmap(x)
    expect_identical(ref, again) # check it's reproducible

    alt <- runUmap(x, num.neighbors=20)
    expect_identical(dim(alt), dim(ref))
    expect_false(identical(alt, ref)) # check that it has an effect

    res <- findKNN(x, transposed=TRUE, k=15, get.distance="transposed", get.index="transposed", BNPARAM=AnnoyParam())
    nnin <- runUmap(res)
    expect_identical(nnin, ref)
})

test_that("runUmap works with initial coordinates", {
    expect_error(runUmap(x, initialize.method="none"), "expected initial coordinates")
    expect_error(runUmap(x, initialize.method="none", initial.coordinates=cbind(1,2)), "nrow")
    expect_error(runUmap(x, initialize.method="none", initial.coordinates=cbind(seq_len(ncol(x)))), "ncol")

    stuff <- matrix(rnorm(ncol(x) * 2), ncol=2)
    init.out <- runUmap(x, initialize.method="none", initial.coordinates=stuff)
    expect_false(identical(init.out, ref))
})

test_that("runUmap works with pre-specified a/b", {
    custom <- runUmap(x, a=2, b=1)
    expect_false(identical(custom, ref))

    expect_error(runUmap(x, a=1:2, b=1), "should be a numeric scalar")
    expect_error(runUmap(x, a=2, b=1:2), "should be a numeric scalar")
})

test_that("runUmap works with an explicit number of epochs", {
    custom <- runUmap(x, num.epochs=600)
    expect_false(identical(custom, ref))
    expect_error(runUmap(x, num.epochs=1:1000), "should be an integer scalar")
})

#' Combine multiple factors 
#'
#' Combine multiple factors into a single factor where each level of the latter is a unique combination of levels from the former.
#'
#' @param factors An ordinary list or \link[S4Vectors]{List} of vectors or factors of the same length.
#' Corresponding elements across all vectors/factors represent the combination of levels for a single observation.
#' For factors, any existing levels are respected.
#' For other vectors, the sorted and unique values are used as levels.
#'
#' Alternatively, a data frame or \link[S4Vectors]{DataFrame} where each column is a vector or factor and each row corresponds to an observation.
#' @param keep.unused Logical scalar indicating whether to report unused combinations of levels.
#'
#' @return List containing:
#' \itemize{
#' \item \code{levels}, a \link[S4Vectors]{DataFrame} containing the sorted and unique combinations of levels from \code{factors}.
#' Each column corresponds to a factor in \code{factors} while each row corresponds to a unqiue combination.
#' \item \code{index}, an integer vector specifying the index into \code{levels} for each observation.
#' }
#' For observation \code{i} and factor \code{j}, \code{levels[[[j]][index[i]]} will recover \code{factors[[j]][i]}.
#'
#' @author Aaron Lun
#'
#' @seealso
#' The \code{combine_to_factor} function in \url{https://libscran.github.io/factorize/}. 
#'
#' @examples
#' combineFactors(list(
#'     sample(LETTERS[1:5], 100, replace=TRUE),
#'     sample(3, 100, replace=TRUE)
#' ))
#'
#' combineFactors(list(
#'     factor(sample(LETTERS[1:5], 10, replace=TRUE), LETTERS[1:5]),
#'     factor(sample(5, 10, replace=TRUE), 1:5)
#' ), keep.unused=TRUE)
#'
#' @export
combineFactors <- function(factors, keep.unused=FALSE) {
    out <- .combineFactors(factors, keep.unused=keep.unused)
    out$index <- out$index + 1L
    out
}

#' @importFrom S4Vectors DataFrame
.combineFactors <- function(factors, keep.unused=FALSE) {
    factors <- as.list(factors)
    nfac <- length(factors)
    f0 <- vector("list", nfac)
    levels0 <- vector("list", nfac)

    for (f in seq_len(nfac)) {
        current <- factors[[f]]
        if (is.factor(current)) {
            f0[[f]] <- as.integer(current)
            levels0[[f]] <- levels(current)
        } else {
            curlevels <- sort(unique(current))
            levels0[[f]] <- curlevels
            f0[[f]] <- match(current, curlevels)
        }

        f0[[f]] <- f0[[f]] - 1L # get to 0-based indices.
    }

    combined <- combine_factors(f0, keep.unused, lengths(levels0))

    for (f in seq_len(nfac)) {
        combined$levels[[f]] <- levels0[[f]][combined$levels[[f]] + 1L]
    }
    names(combined$levels) <- names(factors)

    if (is.null(names(combined$levels))) {
        names(combined$levels) <- seq_len(nfac)
    }
    combined$levels <- DataFrame(combined$levels, check.names=FALSE)

    combined 
}

# library(testthat); library(scrapper); source("test-combineFactors.R")

test_that("combineFactors works for multiple factors", {
    set.seed(10000)
    x <- sample(LETTERS[1:5], 100, replace=TRUE)
    y <- sample(3, 100, replace=TRUE)
    z <- sample(letters[1:4], 100, replace=TRUE)

    out <- combineFactors(list(X=x, Y=y, Z=z))
    expect_identical(names(out$levels), c("X", "Y", "Z"))
    expect_identical(out$levels[[1]][out$index], x)
    expect_identical(out$levels[[2]][out$index], y)
    expect_identical(out$levels[[3]][out$index], z)
})

test_that("combineFactors works for multiple factors with unused combinations", {
    set.seed(10002)
    x <- factor(sample(LETTERS[1:5], 10, replace=TRUE), LETTERS[1:5])
    y <- factor(sample(3, 10, replace=TRUE), 1:3)
    z <- factor(sample(letters[1:4], 10, replace=TRUE), letters[1:4])

    # Sanity check.
    out <- combineFactors(list(x, y, z))
    expect_identical(names(out$levels), as.character(1:3))
    expect_lt(length(out$levels[[1]]), 60L)
    expect_identical(out$levels[[1]][out$index], as.character(x))
    expect_identical(out$levels[[2]][out$index], as.character(y))
    expect_identical(out$levels[[3]][out$index], as.character(z))

    out <- combineFactors(list(x, y, z), keep.unused=TRUE)
    expect_equal(length(out$levels[[1]]), 60L)
    expect_identical(out$levels[[1]][out$index], as.character(x))
    expect_identical(out$levels[[2]][out$index], as.character(y))
    expect_identical(out$levels[[3]][out$index], as.character(z))
})

test_that("combineFactors works for a single factor", {
    set.seed(10001)
    x <- sample(LETTERS[1:5], 100, replace=TRUE)
    out <- combineFactors(list(x))
    expect_identical(out$levels[[1]][out$index], x)

    # Drops unused elements.
    x <- factor(sample(LETTERS, 10, replace=TRUE), LETTERS)
    out <- combineFactors(list(x))
    expect_identical(out$levels[[1]][out$index], as.character(x))
    expect_lt(length(out$levels[[1]]), length(LETTERS))

    out <- combineFactors(list(x), keep.unused=TRUE)
    expect_identical(out$levels[[1]][out$index], as.character(x))
    expect_identical(out$levels[[1]], LETTERS)
})

#' Aggregate expression across genes
#'
#' Aggregate expression values across genes, potentially with weights.
#' This is typically used to summarize expression values for gene sets into a single per-cell score.
#'
#' @param x A matrix-like object where rows correspond to genes or genomic features and columns correspond to cells.
#' Values are usually normalized expression values, possibly log-transformed depending on the application.
#' @param sets List of vectors where each entry corresponds to a gene set.
#' Each entry may be an integer vector of row indices, a logical vector of length equal to the number of rows, or a character vector of row names.
#' For integer and character vectors, duplicate elements are ignored.
#' For character vectors, any strings not present in \code{rownames(x)} are ignored.
#'
#' Alternatively, each entry may be a list of two vectors.
#' The first vector should be either integer (row indices) or character (row names), specifying the genes in the set. 
#' The second vector should be numeric and of the same length as the first vector, specifying the weight associated with each gene.
#' If duplicate genes are present, only the first occurrence is used.
#' If the first vector contains gene names not present in \code{x}, those genes are ignored.
#' @param average Logical scalar indicating whether to compute the average rather than the sum.
#' @param convert Logical scalar indicating whether to convert gene identities to non-duplicate row indices in each entry of \code{sets}.
#' Can be set to \code{FALSE} for greater efficiency if the \code{sets} already contains non-duplicated integer vectors. 
#' @param num.threads Integer specifying the number of threads to be used for aggregation.
#'
#' @return A list of length equal to that of \code{sets}.
#' Each entry is a numeric vector of length equal to the number of columns in \code{x}, 
#' containing the (weighted) sum/mean of expression values for the corresponding set across all cells.
#'
#' @author Aaron Lun
#' @seealso
#' The \code{aggregate_across_genes} function in \url{https://libscran.github.io/scran_aggregate/}. 
#'
#' \code{\link{aggregateAcrossGenes.se}}, to perform aggregation on a \link[SummarizedExperiment]{SummarizedExperiment}.
#'
#' \code{\link{aggregateAcrossCells}}, to aggregate expression values across groups of cells.
#'
#' @examples
#' # Mocking a matrix:
#' library(Matrix)
#' x <- round(abs(rsparsematrix(1000, 100, 0.1) * 100))
#'
#' # Unweighted aggregation:
#' sets <- list(
#'    foo = sample(nrow(x), 20),
#'    bar = sample(nrow(x), 10)
#' )
#' agg <- aggregateAcrossGenes(x, sets)
#' str(agg)
#'
#' # Weighted aggregation:
#' sets <- list(
#'    foo = list(sample(nrow(x), 20), runif(20)),
#'    bar = list(sample(nrow(x), 10), runif(10))
#' )
#' agg2 <- aggregateAcrossGenes(x, sets, average = TRUE)
#' str(agg2)
#' 
#' @export
#' @importFrom beachmat initializeCpp
aggregateAcrossGenes <- function(x, sets, average = FALSE, convert = TRUE, num.threads = 1) {
    .checkSEX(x, "aggregateAcrossGenes.se")

    if (convert) {
        nr <- nrow(x)
        rn <- rownames(x)
        for (i in seq_along(sets)) {
            curset <- sets[[i]]
            if (!is.list(curset)) {
                sets[[i]] <- .sanitizeGeneSet(curset, n=nr, names=rn, arg=sprintf("sets[[%i]]", i))
            } else {
                if (length(curset) != 2L) {
                    stop("expected 'sets[[", i, "]]' to contain two vectors")
                }
                sets[[i]] <- .sanitizeGeneSetWithWeights(curset[[1]], curset[[2]], n=nr, names=rn, index=i)
            }
        }
    }

    ptr <- initializeCpp(x, .check.na=FALSE)
    output <- aggregate_across_genes(ptr, sets, average, num.threads)
    names(output) <- names(sets)
    output
}

.sanitizeGeneSetWithWeights <- function(genes, weights, n, names, index) {
    if (length(genes) != length(weights)) {
        stop("expected all vectors of 'sets[[", index, "]]' to have the same length")
    }

    if (is.numeric(genes)) {
        genes <- as.integer(genes)
        if (anyDuplicated(genes)) {
            keep <- !duplicated(genes)
            genes <- genes[keep]
            weights <- weights[keep]
        }
        if (anyNA(genes) || min(genes) < 1 || max(genes) > n) {
            stop("'sets[[", index, "]]' contains out-of-range indices")
        }

    } else {
        if (anyDuplicated(genes)) {
            keep <- !duplicated(genes)
            genes <- genes[keep]
            weights <- weights[keep]
        }
        genes <- match(genes, names)
        if (anyNA(genes)) {
            keep <- !is.na(genes)
            genes <- genes[keep]
            weights <- weights[keep]
        }
    }

    if (is.unsorted(genes)) {
        o <- order(genes)
        genes <- genes[o]
        weights <- weights[o]
    }
    list(genes, weights)
}

# library(testthat); library(scrapper); source("test-aggregateAcrossGenes.R")

library(Matrix)

set.seed(9999)

test_that("aggregateAcrossGenes works for unweighted sets", {
    x <- round(abs(rsparsematrix(1000, 100, 0.1) * 100))

    sets <- list(
       foo = sample(nrow(x), 20),
       bar = sample(nrow(x), 10),
       whee = sample(nrow(x), 500)
    )

    agg <- aggregateAcrossGenes(x, sets)
    expect_identical(names(agg), names(sets))
    for (s in names(sets)) {
        expect_equal(Matrix::colSums(x[sets[[s]],,drop=FALSE]), agg[[s]])
    }

    agg <- aggregateAcrossGenes(x, sets, average=TRUE)
    expect_identical(names(agg), names(sets))
    for (s in names(sets)) {
        expect_equal(Matrix::colMeans(x[sets[[s]],,drop=FALSE]), agg[[s]])
    }

    expect_error(aggregateAcrossGenes(x, list(-1L)), "out-of-range")
    expect_error(aggregateAcrossGenes(SummarizedExperiment::SummarizedExperiment(x), sets), "not supported")
})

test_that("aggregateAcrossGenes works for weighted sets", {
    x <- round(abs(rsparsematrix(1000, 100, 0.1) * 100))

    sets <- list(
       foo = list(sample(nrow(x), 20), runif(20)),
       bar = list(sample(nrow(x), 10), runif(10)),
       whee = list(sample(nrow(x), 500), runif(500))
    )

    agg <- aggregateAcrossGenes(x, sets)
    expect_identical(names(agg), names(sets))
    for (s in names(sets)) {
        current <- sets[[s]]
        expect_equal(Matrix::colSums(x[current[[1]],,drop=FALSE] * current[[2]]), agg[[s]])
    }

    avagg <- aggregateAcrossGenes(x, sets, average=TRUE)
    expect_identical(names(avagg), names(sets))
    for (s in names(sets)) {
        current <- sets[[s]]
        expect_equal(Matrix::colSums(x[current[[1]],,drop=FALSE] * current[[2]]) / sum(current[[2]]), avagg[[s]])
    }

    expect_error(aggregateAcrossGenes(x, list(list(1L, 2, 3))), "contain two vectors")
    expect_error(aggregateAcrossGenes(x, list(list(1L, numeric(0)))), "same length")
    expect_error(aggregateAcrossGenes(x, list(list(-1L, 2))), "out-of-range")

    # Duplicated indices are just ignored.
    sets2 <- lapply(sets,
        function(x) {
            first <- x[[1]]
            list(
                rep(first, 2),
                c(x[[2]], runif(length(first)))
            )
        }
    )
    expect_identical(aggregateAcrossGenes(x, sets2), agg)
})

test_that("aggregateAcrossGenes works for other vector types", {
    x <- round(abs(rsparsematrix(1000, 100, 0.1) * 100))
    rownames(x) <- sprintf("GENE_%s", seq_len(nrow(x)))

    sets <- list(
       foo = rbinom(nrow(x), 1, 0.2) == 1,
       bar = sample(rownames(x), 20),
       stuff = list(sample(rownames(x), 50), runif(50))
    )
    agg <- aggregateAcrossGenes(x, sets)

    refsets <- sets
    refsets$foo <- which(sets$foo)
    refsets$bar <- match(sets$bar, rownames(x))
    refsets$stuff[[1]] <- match(sets$stuff[[1]], rownames(x))
    refagg <- aggregateAcrossGenes(x, refsets)

    expect_identical(agg, refagg)

    # Duplicated and missing names are just ignored, with or without weights.
    sets2 <- sets
    sets2$bar <- c("foo", sets$bar, "whee", sets$bar)
    sets2$stuff <- local({
        idx <- sets$stuff[[1]]
        list(
            c(rep(idx, 2), "absent"),
            c(sets$stuff[[2]], runif(length(idx)), pi)
        )
    })
    expect_identical(aggregateAcrossGenes(x, sets2), agg)
})

#' Delayed log-normalization of a matrix
#'
#' Delayed calculation of log-normalized expression values, typically returned by \code{\link{normalizeCounts}}.
#'
#' @param x Count matrix to be normalized. 
#' @param size.factors Numeric vector of size factors, of length equal to the number of columns of \code{x}.
#' @param pseudo.count Number specifying the pseudo-count to add prior to log-transformation.
#' @param log.base Number specifying the base of the log-transformation. 
#' 
#' @return An instance of a LogNormalizedMatrix(Seed).
#'
#' @details
#' This is based on the \pkg{DelayedArray} framework and 
#'
#' @author Aaron Lun
#' @examples
#' mat <- matrix(rpois(1000, lambda=2), ncol=10)
#' sf <- centerSizeFactors(colSums(mat))
#' norm <- LogNormalizedMatrix(mat, sf)
#' norm
#'
#' # Also works with sparse matrices.
#' library(Matrix)
#' smat <- abs(rsparsematrix(50, 20, density=0.21))
#' ssf <- centerSizeFactors(colSums(smat))
#' snorm <- LogNormalizedMatrix(smat, ssf)
#' snorm
#'
#' @aliases
#' LogNormalizedMatrix-class
#' LogNormalizedMatrixSeed-class
#' dim,LogNormalizedMatrixSeed-method
#' dimnames,LogNormalizedMatrixSeed-method
#' type,LogNormalizedMatrixSeed-method
#' is_sparse,LogNormalizedMatrixSeed-method
#' extract_array,LogNormalizedMatrixSeed-method
#' extract_sparse_array,LogNormalizedMatrixSeed-method
#' matrixClass,LogNormalizedMatrixSeed-method
#' DelayedArray,LogNormalizedMatrixSeed-method
#' initializeCpp,LogNormalizedMatrixSeed-method
#'
#' @export
LogNormalizedMatrix <- function(x, size.factors, pseudo.count=1, log.base=2) {
    DelayedArray(LogNormalizedMatrixSeed(x, size.factors=size.factors, pseudo.count=pseudo.count, log.base=log.base))
}

#' @export
#' @rdname LogNormalizedMatrix
#' @importFrom methods new
LogNormalizedMatrixSeed <- function(x, size.factors, pseudo.count=1, log.base=2) {
    new("LogNormalizedMatrixSeed",
        seed=x,
        size.factors=as.double(size.factors), 
        pseudo.count=as.double(pseudo.count),
        log.base=as.double(log.base)
    )
}

#' @export
#' @import methods
setClass("LogNormalizedMatrixSeed", slots=c(seed="ANY", size.factors="numeric", pseudo.count="numeric", log.base="numeric"))

#' @importFrom S4Vectors setValidity2
setValidity2("LogNormalizedMatrixSeed", function(object) {
    if (ncol(object@seed) != length(object@size.factors)) {
        return("length of 'size.factors' should equal number of columns")
    }
})

#' @export
setMethod("dim", "LogNormalizedMatrixSeed", function(x) dim(x@seed))

#' @export
setMethod("dimnames", "LogNormalizedMatrixSeed", function(x) dimnames(x@seed))

#' @export
#' @importFrom DelayedArray type
setMethod("type", "LogNormalizedMatrixSeed", function(x) "double")

#' @export
#' @importFrom SparseArray is_sparse
setMethod("is_sparse", "LogNormalizedMatrixSeed", function(x) is_sparse(x@seed) && x@pseudo.count == 1)

.slice_size_factors <- function(x, index) { 
    sf <- x@size.factors
    if (!is.null(index[[2]])) {
        sf <- sf[index[[2]]]
    }
    sf
}

#' @importFrom DelayedArray t
.log_normalize <- function(mat, sf, pseudo.count, log.base) {
    norm <- t(t(mat)/sf)
    if (pseudo.count == 1) {
        log1p(norm) / log(log.base)
    } else {
        log(norm + pseudo.count, log.base)
    }
}

#' @export
#' @importFrom DelayedArray extract_array
setMethod("extract_array", "LogNormalizedMatrixSeed", function(x, index) {
    core <- extract_array(x@seed, index)
    .log_normalize(core, .slice_size_factors(x, index), pseudo.count=x@pseudo.count, log.base=x@log.base)
})

#' @export
#' @importFrom SparseArray extract_sparse_array
setMethod("extract_sparse_array", "LogNormalizedMatrixSeed", function(x, index) {
    core <- extract_sparse_array(x@seed, index)
    .log_normalize(core, .slice_size_factors(x, index), pseudo.count=x@pseudo.count, log.base=x@log.base)
})

#' @export
setClass("LogNormalizedMatrix", contains="DelayedMatrix", slots=c(seed="LogNormalizedMatrixSeed"))

#' @export
#' @importFrom DelayedArray DelayedArray new_DelayedArray
setMethod("DelayedArray", "LogNormalizedMatrixSeed", function(seed) new_DelayedArray(seed, Class = "LogNormalizedMatrix"))

#' @export
#' @importFrom beachmat initializeCpp
setMethod("initializeCpp", "LogNormalizedMatrixSeed", function(x, ...) {
    seed <- initializeCpp(x@seed)
    initialize_LogNormalizedMatrix(seed, x@size.factors, x@pseudo.count, x@log.base)
})

#' Sanitize size factors
#'
#' Replace invalid size factors, i.e., zero, negative, infinite or NaN values.
#' Such size factors can occasionally arise if, e.g., insufficient quality control was performed upstream.
#' Removing them ensures that the normalized values from \code{\link{normalizeCounts}} remain finite for sensible downstream processing.
#'
#' @param size.factors Numeric vector of size factors across cells.
#' @param replace.zero Logical scalar indicating whether to replace size factors of zero with the lowest positive factor in \code{size.factors}.
#' This ensures that the normalized values will be large to reflect the extremity of the scaling, but still finite for sensible downstream processing. 
#' If \code{FALSE}, zeros are retained.
#' @param replace.negative Logical scalar indicating whether to replace negative size factors with the lowest positive factor in \code{size.factors}.
#' This ensures that the normalized values will be large to reflect the extremity of the scaling, but still finite for sensible downstream processing. 
#' If \code{FALSE}, negative values are retained.
#' @param replace.infinite Logical scalar indicating whether to replace infinite size factors with the largest positive factor in \code{size.factors}.
#' This ensures that any normalized values will be, at least, finite; the choice of a relatively large replacement value reflects the extremity of the scaling.
#' If \code{FALSE}, infinite values are retained.
#' @param replace.nan Logical scalar indicating whether to replace NaN size factors with unity, e.g., scaling normalization is a no-op.
#' If \code{FALSE}, NaN values are retained.
#' 
#' @return Numeric vector of length equal to \code{size.factors}, containing the sanitized size factors.
#'
#' @author Aaron Lun
#'
#' @examples
#' sf <- 2^rnorm(100)
#' sf[1] <- 0
#' sf[2] <- -1
#' sf[3] <- Inf
#' sf[4] <- NaN
#' sanitizeSizeFactors(sf)
#'
#' @seealso
#' The \code{sanitize_size_factors} function in \url{https://libscran.github.io/scran_norm/}. 
#'
#' @export
sanitizeSizeFactors <- function(size.factors, replace.zero=TRUE, replace.negative=TRUE, replace.infinite=TRUE, replace.nan=TRUE) {
    sanitize_size_factors(size.factors, handle_zero=replace.zero, handle_negative=replace.negative, handle_infinite=replace.infinite, handle_nan=replace.nan)
}

# library(testthat); library(scrapper); source("test-sanitizeSizeFactors.R")

set.seed(99999)

test_that("sanitizeSizeFactors works as expected", {
    f <- 2^rnorm(100)
    f[1] <- 0
    f[2] <- NA
    f[3] <- Inf
    f[4] <- -1

    cf <- sanitizeSizeFactors(f)
    expect_true(all(is.finite(cf) & cf > 0))
    expect_equal(tail(cf, -4), tail(f, -4))
})

#' Quality control for ADT count data
#'
#' Compute per-cell QC metrics from an initialized matrix of ADT counts,
#' and use the metrics to suggest filter thresholds to retain high-quality cells.
#' 
#' @param x A matrix-like object where rows are ADTs and columns are cells.
#' Values are expected to be counts.
#' @param subsets Named list of vectors specifying tag subsets of interest, typically control tags like IgGs.
#' Each vector may be logical (whether to keep each row), integer (row indices) or character (row names).
#' For character vectors, strings not present in \code{rownames(x)} are ignored.
#' @param num.threads Integer scalar specifying the number of threads to use.
#' @param metrics \link[S4Vectors]{DataFrame} of per-cell QC metrics.
#' This should have the same structure as the return value of \code{computeAdtQcMetrics}.
#' @param block Factor specifying the block of origin (e.g., batch, sample) for each cell in \code{metrics}.
#' Alternatively \code{NULL} if all cells are from the same block.
#'
#' For \code{filterAdtQcMetrics}, a blocking factor should be provided if \code{block} was used to construct \code{thresholds}. 
#' @param min.detected.drop Minimum drop in the number of detected features from the median, in order to consider a cell to be of low quality.
#' @param num.mads Number of median from the median, to define the threshold for outliers in each metric.
#' @param thresholds List with the same structure as produced by \code{suggestAdtQcThresholds}.
#'
#' @return For \code{computeAdtQcMetrics}, a \link[S4Vectors]{DataFrame} is returned with one row per cell in \code{x}.
#' This contains the following columns:
#' \itemize{
#' \item \code{sum}, a numeric vector containing the total ADT count for each cell.
#' In theory, this represents the efficiency of library preparation and sequencing.
#' Compared to RNA, the sum is less useful as a QC metric for ADT data as it is strongly influenced by biological variation in the abundance of the targeted features.
#' Nonetheless, we compute it for diagnostic purposes.
#' \item \code{detected}, an integer vector containing the number of detected tags per cell.
#' Even though ADTs are typically used in situations where few features are highly abundant (e.g., cell type-specific markers), 
#' we still expect detectable coverage of most features due to ambient contamination, non-specific binding or some background expression.
#' Low numbers of detected tags indicates that library preparation or sequencing depth was suboptimal.
#' \item \code{subsets}, a nested DataFrame where each column corresponds to a control subset and is a numeric vector containing the total count in that subset. 
#' The exact interpretation depends on the nature of the feature subset but the most common use case involves isotype control (IgG) features.
#' IgG antibodies should not bind to anything so a high subset sum suggests that non-specific binding is a problem, e.g., due to antibody conjugates.
#' (Unlike RNA quality control, we do not use proportions here as it is entirely possible for a cell to have low counts for other tags due to the absence of their targeted features;
#' this would result in a high proportion even if the cell has a "normal" level of non-specific binding.)
#' }
#' Each vector is of length equal to the number of cells.
#'
#' For \code{suggestAdtQcThresholds}, a named list is returned:
#' \itemize{
#' \item If \code{block=NULL}, the list contains:
#' \itemize{
#' \item \code{detected}, a numeric scalar containing the lower bound on the number of detected tags. 
#' This is defined as the lower of (i) \code{num.mads} MADs below the median for the log-transformed values across all cells,
#' and (ii) the product of \code{1 - min.detected.drop} and the median across all cells.
#' The latter avoids overly aggressive filtering when the MAD is zero.
#' \item \code{subsets}, a numeric vector containing the upper bound on the sum of counts in each control subset. 
#' This is defined as \code{num.mads} MADs above the median of the log-transformed metrics across all cells.
#' }
#' \item Otherwise, if \code{block} is supplied, the list contains:
#' \itemize{
#' \item \code{detected}, a numeric vector containing the lower bound on the number of detected tags for each blocking level.
#' Here, the threshold is computed independently for each block, using the same method as the unblocked case.
#' \item \code{subsets}, a list of numeric vectors containing the upper bound on the sum of counts in each control subset for each blocking level.
#' Here, the threshold is computed independently for each block, using the same method as the unblocked case.
#' \item \code{block.ids}, a vector containing the identities of the unique blocks.
#' }
#' Each vector is of length equal to the number of levels in \code{block} and is named accordingly.
#' }
#'
#' For \code{filterAdtQcMetrics}, a logical vector of length \code{ncol(x)} is returned indicating which cells are of high quality. 
#' High-quality cells are defined as those with numbers of detected tags above the \code{detected} threshold and control subset sums below the \code{subsets} threshold.
#' 
#' @seealso
#' The \code{compute_adt_qc_metrics}, \code{compute_adt_qc_filters} and \code{compute_adt_qc_filters_blocked} functions in \url{https://libscran.github.io/scran_qc/}.
#'
#' \code{\link{quickAdtQc.se}}, to run all of the ADT-related QC functions on a \link[SummarizedExperiment]{SummarizedExperiment}.
#'
#' @author Aaron Lun
#' @examples
#' # Mocking a matrix:
#' library(Matrix)
#' x <- round(abs(rsparsematrix(1000, 100, 0.1) * 100))
#'
#' # Mocking up a control set.
#' sub <- list(IgG=rbinom(nrow(x), 1, 0.1) > 0)
#'
#' qc <- computeAdtQcMetrics(x, sub)
#' qc 
#'
#' filt <- suggestAdtQcThresholds(qc)
#' str(filt)
#'
#' keep <- filterAdtQcMetrics(filt, qc)
#' summary(keep)
#'
#' @export
#' @name adt_quality_control
#' @importFrom beachmat initializeCpp tatami.dim
#' @importFrom S4Vectors DataFrame I
computeAdtQcMetrics <- function(x, subsets, num.threads = 1) {
    if (length(subsets) > 0 && is.null(names(subsets))) {
        stop("subsets should be a named list")
    }
    .checkSEX(x, "quickAdtQc.se")

    ptr <- initializeCpp(x, .check.na=FALSE)

    subsets <- as.list(subsets)
    subsets <- lapply(subsets, .subsetToLogical, n=tatami.dim(ptr)[1], names=rownames(x))

    output <- compute_adt_qc_metrics(ptr, subsets, num_threads=num.threads)
    .reformatQcMetrics(output, "subsets", subsets, x)
}

#' @export
#' @rdname adt_quality_control
suggestAdtQcThresholds <- function(metrics, block=NULL, min.detected.drop=0.1, num.mads=3) {
    block <- .transformFactor(block) 
    metrics <- .simplifyQcMetrics(metrics)
    thresholds <- suggest_adt_qc_thresholds(metrics, block=block$index, min_detected_drop=min.detected.drop, num_mads=num.mads)

    names(thresholds$subsets) <- names(metrics$subsets)

    if (!is.null(block$names)) {
        names(thresholds$detected) <- block$names
        for (i in seq_along(metrics$subsets)) {
            names(thresholds$subsets[[i]]) <- block$names
        }
        thresholds$block.ids <- block$names # store it separately to preserve any non-character type.
    }

    thresholds
}

#' @export
#' @rdname adt_quality_control
filterAdtQcMetrics <- function(thresholds, metrics, block=NULL) {
    block <- .matchBlockThresholds(block, thresholds$block.ids)
    metrics <- .simplifyQcMetrics(metrics)
    .checkThresholdNames(thresholds, c("detected"), "subsets", names(metrics$subsets))
    filter_adt_qc_metrics(thresholds, metrics, block=block)
}

# library(testthat); library(scrapper); source("test-adt_quality_control.R")

set.seed(112233)
library(Matrix)
x <- round(abs(rsparsematrix(200, 1000, 0.1) * 100))

test_that("computeAdtQcMetrics works as expected", { 
    sub <- list(IgG=rbinom(nrow(x), 1, 0.1) > 0)
    qc <- computeAdtQcMetrics(x, sub)
    expect_identical(qc$sum, Matrix::colSums(x))
    expect_identical(qc$detected, Matrix::colSums(x > 0))
    expect_identical(qc$subsets$IgG, Matrix::colSums(x[sub$IgG,]))

    nosub <- computeAdtQcMetrics(x, list())
    expect_identical(qc$sum, nosub$sum)

    expect_error(computeAdtQcMetrics(x, list(1)), "should be a named list")
    expect_error(computeAdtQcMetrics(SummarizedExperiment::SummarizedExperiment(x), NULL), "not supported")
})

test_that("suggestAdtQcThresholds works as expected", { 
    sub <- list(IgG=rbinom(nrow(x), 1, 0.1) > 0)
    qc <- computeAdtQcMetrics(x, sub)
    num.mads <- 1.5
    thresholds <- suggestAdtQcThresholds(qc, num.mads=num.mads)

    # Check the thresholds.
    expect_lt(thresholds$detected, median(qc$detected))
    expect_gt(thresholds$subsets[[1]], median(qc$subsets[[1]]))
    expect_equal(names(thresholds$subsets), "IgG")

    # Check the filter.
    expected <- qc$detected >= thresholds$detected & qc$subsets[[1]] <= thresholds$subsets[[1]]
    observed <- filterAdtQcMetrics(thresholds, qc)
    expect_identical(expected, observed) 
})

test_that("suggestAdtQcThresholds works as expected with blocking", { 
    sub <- list(IgG=rbinom(nrow(x), 1, 0.1) > 0)
    qc <- computeAdtQcMetrics(x, sub)
    block <- sample(3, ncol(x), replace=TRUE)
    num.mads <- 1.5
    thresholds <- suggestAdtQcThresholds(qc, block=block, num.mads=num.mads)

    expect_identical(thresholds$block.ids, 1:3)

    # Check the thresholds.
    for (b in 1:3) {
        keep <- block == b
        expect_lt(thresholds$detected[b], median(qc$detected[keep]))
        expect_gt(thresholds$subsets[[1]][b], median(qc$subsets[[1]][keep]))
    }

    # Check the filter.
    expected <- qc$detected >= thresholds$detected[block] & qc$subsets[[1]] <= thresholds$subsets[[1]][block]
    observed <- filterAdtQcMetrics(thresholds, qc, block=block)
    expect_identical(unname(expected), observed)

    # Same filtering with just the last block.
    last <- block == 3
    last_observed <- filterAdtQcMetrics(
        thresholds,
        list(sum=qc$sum[last], detected=qc$detected[last], subsets=list(IgG=qc$subsets[[1]][last])),
        block=block[last]
    )
    expect_identical(observed[last], last_observed)
})

#' Principal components analysis
#'
#' Run a PCA on the gene-by-cell log-expression matrix and extract the top principal components (PCs).
#' This yields a low-dimensional representation that reduces noise and compute time in downstream analyses.
#' For efficiency, the PCA itself is approximated using IRLBA.
#'
#' @param x A matrix-like object where rows correspond to genes or genomic features and columns correspond to cells.
#' Typically, the matrix is expected to contain log-expression values (see \code{\link{normalizeCounts}}) for \dQuote{interesting} genes (see \code{\link{chooseHighlyVariableGenes}}).
#' @param number Integer scalar specifying the number of top PCs to retain.
#' More PCs will capture more biological signal at the cost of increasing noise and compute time.
#' If this is greater than the maximum number of PCs (i.e., the smaller dimension of \code{x}), only the maximum number of PCs will be reported in the results.
#' @param scale Logical scalar indicating whether to scale all genes to have the same variance.
#' This ensures that each gene contributes equally to the PCA, favoring consistent variation across many genes rather than large variation in a few genes.
#' If \code{block} is specified, each gene's variance is calculated as a weighted sum of the variances from each block. 
#' Genes with zero variance are ignored.
#' @param block Factor specifying the block of origin (e.g., batch, sample) for each cell in \code{x}.
#' The PCA will be performed on the residuals after regressing out the block effect, ensuring that differences between block do not dominate the variation in the dataset.
#' Alternatively \code{NULL} if all cells are from the same block.
#' @param block.weight.policy String specifying the policy to use for weighting the contribution of different blocks to the PCA.
#' See the argument of the same name in \code{\link{computeBlockWeights}} for more detail.
#' Only used if \code{block} is not \code{NULL}.
#' @param variable.block.weight Numeric vector of length 2, specifying the parameters for variable block weighting.
#' See the argument of the same name in \code{\link{computeBlockWeights}} for more detail.
#' Only used if \code{block} is not \code{NULL} and \code{block.weight.policy = "variable"}.
#' @param components.from.residuals Logical scalar indicating whether to compute the PC scores from the residuals in the presence of a blocking factor.
#' By default, the residuals are only used to compute the rotation matrix, and the original expression values of the cells are projected onto this new space (see Details).
#' Only used if \code{block} is not \code{NULL}.
#' @param subset Integer, logical or character vector specifying the rows of \code{x} to use for the PCA.
#' This yields the same results as \code{runPca} on \code{x[subset,]}, except that entries of the rotation matrix will also be computed for rows outside of the subset.
#' If \code{NULL}, all rows of \code{x} are used.
#' @param extra.work Integer scalar specifying the extra dimensions for the IRLBA workspace.
#' Larger values improve accuracy at the cost of compute time.
#' @param iterations Integer scalar specifying the maximum number of restart iterations for IRLBA.
#' Larger values improve accuracy at the cost of compute time.
#' @param seed Integer scalar specifying the seed for the initial random vector in IRLBA.
#' @param realized Logical scalar indicating whether to realize \code{x} into an optimal memory layout for IRLBA.
#' This speeds up computation at the cost of increased memory usage.
#' @param warn Boolean specifying whether a warning should be emitted if IRLBA failed to converge.
#' @param num.threads Number of threads to use.
#'
#' @return List containing:
#' \itemize{
#' \item \code{components}, a matrix of PC scores.
#' Rows are dimensions (i.e., PCs) and columns are cells.
#' \item \code{rotation}, the rotation matrix.
#' Rows are genes and columns are dimensions.
#' \item \code{variance.explained}, the vector of variances explained by each PC.
#' \item \code{total.variance}, the total variance in the dataset.
#' This can be used to divide \code{variance.explained} to obtain the proportion of variance explained by each PC.
#' \item \code{center}, a numeric vector containing the mean for each gene.
#' If \code{block} is provided, this is instead a matrix containing the mean for each gene (column) in each block (row).
#' \item \code{block.ids}, a vector containing the identities of the unique blocks in the same order as the rows of \code{center}.
#' Only reported if \code{block} is provided. 
#' \item \code{scale}, a numeric vector containing the scaling for each gene.
#' Only reported if \code{scale=TRUE}.
#' \item \code{metrics}, a named list containing IRLBA metrics.
#' This includes \code{converged}, whether the algorithm converged successfully;
#' \code{iterations}, the number of IRLBA restart iterations;
#' and \code{multiplications}, the number of matrix multiplications.
#' }
#'
#' @details
#' When \code{block} is specified, the nature of the reported PC scores depends on the choice of \code{components.from.residuals}:
#' \itemize{
#' \item If \code{TRUE}, the PC scores are computed from the matrix of residuals.
#' This yields a low-dimensional space where inter-block differences have been removed,
#' assuming that all blocks have the same subpopulation composition and the inter-block differences are consistent for all cell subpopulations.
#' Under these assumptions, we could use these components for downstream analysis without any concern for block-wise effects.
#' \item If \code{FALSE}, the rotation vectors are first computed from the matrix of residuals.
#' To obtain PC scores, each cell is then projected onto the associated subspace using its original expression values.
#' This approach ensures that inter-block differences do not contribute to the PCA but does not attempt to explicitly remove them.
#' }
#' In complex datasets, the assumptions mentioned for \code{TRUE} not hold and more sophisticated batch correction methods like MNN correction are required.
#' Functions like \code{\link{correctMnn}} will accept a low-dimensional embedding of cells that can be created as described above with \code{FALSE}.
#'
#' @author Aaron Lun
#'
#' @seealso
#' The \code{simple_pca} and \code{blocked_pca} functions for \url{https://libscran.github.io/scran_pca/}. 
#'
#' \code{\link{runPca.se}}, to run a PCA on a \link[SummarizedExperiment]{SummarizedExperiment}.
#'
#' @examples
#' library(Matrix)
#' x <- abs(rsparsematrix(1000, 100, 0.1) * 10)
#' y <- normalizeCounts(x, size.factors=centerSizeFactors(colSums(x)))
#'
#' # A simple PCA:
#' out <- runPca(y)
#' str(out)
#'
#' # Blocking on uninteresting factors:
#' block <- sample(LETTERS[1:3], ncol(y), replace=TRUE)
#' bout <- runPca(y, block=block)
#' str(bout)
#' 
#' @export
#' @importFrom beachmat initializeCpp
runPca <- function(
    x,
    number=25,
    scale=FALSE,
    block=NULL, 
    block.weight.policy=c("variable", "equal", "none"),
    variable.block.weight=c(0, 1000),
    components.from.residuals=FALSE,
    subset=NULL,
    extra.work=7,
    iterations=1000,
    seed=5489,
    realized=TRUE,
    warn=TRUE,
    num.threads=1
) {
    .checkSEX(x, "runPca.se")
    block <- .transformFactor(block)

    if (!is.null(subset)) {
        subset <- which(.subsetToLogical(subset, nrow(x), rownames(x))) - 1L
    }

    out <- run_pca(
        initializeCpp(x, .check.na=FALSE),
        number=number,
        scale=scale,
        block=block$index,
        block_weight_policy=match.arg(block.weight.policy),
        variable_block_weight=variable.block.weight,
        components_from_residuals=components.from.residuals,
        subset=subset,
        realized=realized,
        irlba_work=extra.work,
        irlba_iterations=iterations,
        irlba_seed=seed,
        num_threads=num.threads
    )

    out$converged <- out$metrics$converged # For back-compatibility.
    if (!out$metrics$converged && warn) {
        warning("convergence failure for the approximate PCA")
    }

    if (!is.null(block$index)) {
        rownames(out$center) <- block$names
        colnames(out$center) <- rownames(x)
        out$block.ids <- block$names
    } else {
        names(out$center) <- rownames(x)
    }

    if (!scale) {
        out <- out[setdiff(names(out), "scale")]
    } else {
        names(out$scale) <- rownames(x)
    }

    rownames(out$rotation) <- rownames(x)

    out
}

#' Score a gene set in a SummarizedExperiment
#'
#' Compute a gene set activity score for each cell based on the expression values of the genes in the set,
#' by calling \code{\link{scoreGeneSet}} on an assay of a \link[SummarizedExperiment]{SummarizedExperiment}.
#'
#' @param x A \link[SummarizedExperiment]{SummarizedExperiment} object or one of its subclasses.
#' Rows correspond to genes and columns correspond to cells.
#' @param set Vector containing the gene set, see \code{?\link{scoreGeneSet}} for details.
#' @param block Block assignment for each cell, passed to \code{\link{scoreGeneSet}}.
#' @param num.threads Number of threads for \code{\link{scoreGeneSet}}.
#' @param more.score.args Named list of further arguments to pass to \code{\link[scrapper]{scoreGeneSet}}.
#' @param assay.type Integer or string specifying the relevant assay in \code{x}, usually containing log-normalized expression values.
#'
#' @return List containing \code{scores}, a numeric vector of the gene set scores across all cells in \code{x};
#' and \code{weights}, a numeric vector of weights for all genes in \code{set}.
#'
#' @author Aaron Lun
#' @examples
#' # Defining a gene set of oligodendrocyte genes.
#' library(org.Mm.eg.db)
#' oligo.set <- select(org.Mm.eg.db, keytype="GO", keys="GO:0048709", columns="SYMBOL")
#' oligo.set <- unique(oligo.set$SYMBOL)
#'
#' sce <- getTestRnaData.se("norm")
#' oligo.scores <- scoreGeneSet.se(sce, oligo.set)
#' summary(oligo.scores$scores)
#'
#' @export
scoreGeneSet.se <- function(
    x,
    set,
    block = NULL,
    num.threads = 1,
    more.score.args = list(),
    assay.type = "logcounts"
) {
    out <- .call(
        scoreGeneSet,
        list(SummarizedExperiment::assay(x, assay.type)),
        list(set=set, block=block, num.threads=num.threads),
        more.score.args
    )

    names(out$scores) <- colnames(x)
    w <- out$weights$weight
    names(w) <- rownames(x)[out$weights$row]
    out$weights <- w

    out
}

# library(testthat); library(scrapper); source("test-se_scoreGeneSet.R")

library(SummarizedExperiment)
set.seed(1111)
sce <- SummarizedExperiment(list(logcounts = matrix(runif(2000), 200, 10)))

test_that("scoreGeneSet.se works as expected", {
    out <- scoreGeneSet.se(sce, set=1:10)
    expect_identical(length(out$scores), ncol(sce))
    expect_identical(length(out$weights), 10L)

    # Respects names in the output object.
    rownames(sce) <- paste0("GENE_", seq_len(nrow(sce)))
    out2 <- scoreGeneSet.se(sce, set=10:1)
    expect_equal(out$scores, out2$scores)
    expect_identical(names(out2$weights), rownames(sce)[1:10])
    expect_identical(unname(out2$weights), out$weights)
})

.transformFactor <- function(f) {
    if (is.null(f)) {
        list(index=NULL, names=NULL)
    } else {
        # Don't use 'factor', to preserve the type of 'f' in the 'names'.
        lev <- sort(unique(f))
        list(index=match(f, lev) - 1L, names=lev)
    }
}

#' @importFrom methods is
.checkSEX <- function(x, alt) {
    if (is(x, "SummarizedExperiment")) {
        stop("SummarizedExperiment inputs are not supported, use '", alt, "()' or extract the relevant 'assay()' instead")
    }
}

.checkNeighborResults <- function(index, distance) {
    stopifnot(length(dim(index)) == 2L)
    r <- range(index)
    if (!is.finite(r[1]) || r[1] < 1L) {
        stop("'index' should contain finite positive integers")
    }
    if (!is.finite(r[2]) || r[2] > ncol(index)) {
        stop("'index' should contain finite integers no greater than 'ncol(index)'")
    }
    if (!is.null(distance) && !identical(dim(index), dim(distance))) {
        stop("'index' and 'distance' should have the same dimensions")
    }
}

# We keep arguments that are explicitly named in 'more.args' for back-compatibility,
# in case some of these are promoted into top-level arguments at a later date.
.call <- function(FUN, essential.args, named.args, more.args) {
    all.args <- c(
        essential.args,
        named.args[!(names(named.args) %in% names(more.args))],
        more.args
    )
    do.call(FUN, all.args)
}

.sanitizeAltexpAssays <- function(altexps, all.altexps, default.assay.type) {
    if (!is.null(names(altexps))) {
        altexps[!duplicated(names(altexps))]
    } else {
        altexps <- unique(altexps)
        output <- rep(default.assay.type, length(altexps))
        if (is.numeric(altexps)) {
            names(output) <- all.altexps[altexps]
        } else {
            names(output) <- altexps
        }
        output
    }
}

#' @importFrom DelayedArray t
#' @importFrom DelayedArray DelayedArray
.addTransposedReddim <- function(x, name, mat, delayed) {
    if (delayed) {
        mat <- DelayedArray(mat)
    }
    SingleCellExperiment::reducedDim(x, name) <- t(mat)
    x
}

#' @importFrom DelayedArray t
#' @importFrom methods is
#' @importClassesFrom DelayedArray DelayedArray
.getTransposedReddim <- function(x, name) {
    if (is.null(names(name))) {
        mat <- SingleCellExperiment::reducedDim(x, name, withDimnames=FALSE)
    } else {
        mat <- SingleCellExperiment::reducedDim(SingleCellExperiment::altExp(x, names(name), withDimnames=FALSE), name, withDimnames=FALSE)
    }

    mat <- t(mat)

    if (is.matrix(mat)) {
        colnames(mat) <- colnames(x)
        return(mat)
    } else if (is(mat, "DelayedArray")) {
        # Possibly a no-op if .add_transposed_reddim was set with delayed=TRUE.
        if (is.matrix(mat@seed)) {
            out <- mat@seed
            colnames(out) <- colnames(x)
            return(out)
        }
    }

    colnames(mat) <- colnames(x)
    return(as.matrix(mat))
}

#' Batch correction with mutual nearest neighbors
#'
#' Apply mutual nearest neighbor (MNN) correction to remove batch effects from a low-dimensional embedding.
#'
#' @param x Numeric matrix where rows are dimensions and columns are cells,
#' typically containing coordinates in a low-dimensional embedding (e.g., from \code{\link{runPca}}).
#' @param block Factor specifying the block of origin (e.g., batch, sample) for each cell in \code{x}.
#' @param num.neighbors Integer scalar specifying the number of neighbors in the various search steps.
#' Larger values improve the stability of the correction by increasing the number of MNN pairs and including more observations in each center of mass.
#' However, this comes at the cost of reduced resolution when matching subpopulations across batches.
#' @param num.steps Integer scalar specifying the number of steps for the recursive neighbor search to compute the center of mass.
#' Larger values mitigate the kissing effect but increase the risk of including inappropriately distant subpopulations into the center of mass.
#' @param num.mads Deprecated and ignored.
#' @param robust.iterations Deprecated and ignored.
#' @param robust.trim Deprecated and ignored.
#' @param mass.cap Deprecated and ignored.
#' @param num.threads Integer scalar specifying the number of threads to use.
#' @param order Deprecated and ignored, the merge order is now always automatically determined.
#' @param reference.policy Deprecated, use \code{merge.policy} instead. 
#' @param merge.policy String specifying the policy to use to choose the order of batches to merge.
#' \itemize{
#' \item \code{"input"} will use the input order of the batches.
#' Observations in the last batch are corrected first, and then the second-last batch, and so on.
#' This allows users to control the merge order by simply changing the inputs.
#' \item \code{"size"} will merge batches in order of increasing size (i.e., the number of observations).
#' So, the smallest batch is corrected first while the largest batch is unchanged.
#' The aim is to lower compute time by reducing the number of observations that need to be reprocessed in later merge steps.
#' \item \code{"variance"} will merge batches in order of increasing variance between observations. 
#' So, the batch with the lowest variance is corrected first while the batch with the highest variance is unchanged.
#' The aim is to lower compute time by encouraging more observations to be corrected to the most variable batch, thus avoid reprocessing in later merge steps.
#' \item \code{"rss"} will merge batches in order of increasing residual sum of squares (RSS).
#' This is effectively a compromise between \code{"variance"} and \code{"size"}.
#' }
#' @param BNPARAM A \link[BiocNeighbors]{BiocNeighborParam} object specifying the nearest-neighbor algorithm to use.
#'
#' @return List containing \code{corrected}, a numeric matrix of the same dimensions as \code{x}, containing the corrected values.
#'
#' @seealso
#' The \code{compute} function in \url{https://libscran.github.io/mnncorrect/}.
#'
#' \code{\link{correctMnn.se}}, to perform MNN correction on a \link[SingleCellExperiment]{SingleCellExperiment}.
#'
#' @references 
#' Haghverdi L, Lun ATL, Morgan MD, Marioni JC (2018).
#' Batch effects in single-cell RNA-sequencing data are corrected by matching mutual nearest neighbors.
#' \emph{Nat. Biotechnol.} 36(5):421-427
#'
#' @author Aaron Lun
#'
#' @examples
#' # Mocking up a dataset with multiple batches.
#' x <- matrix(rnorm(10000), nrow=10)
#' b <- sample(3, ncol(x), replace=TRUE)
#' x[,b==2] <- x[,b==2] + 3
#' x[,b==3] <- x[,b==3] + 5
#' lapply(split(colMeans(x), b), mean) # indeed the means differ...
#'
#' corrected <- correctMnn(x, b)
#' str(corrected)
#' lapply(split(colMeans(corrected$corrected), b), mean) # now merged.
#'
#' @export
#' @importFrom BiocNeighbors defineBuilder
correctMnn <- function(
    x,
    block,
    num.neighbors=15,
    num.steps=1,
    merge.policy=c("rss", "size", "variance", "input"),
    num.mads=NULL,
    robust.iterations=NULL,
    robust.trim=NULL,
    mass.cap=NULL,
    order=NULL,
    reference.policy=NULL,
    BNPARAM=AnnoyParam(),
    num.threads=1)
{
    .checkSEX(x, "correctMnn.se")
    block <- .transformFactor(block)

    if (!is.null(num.mads)) {
        .Deprecated(old="num.mads=")
    }
    if (!is.null(robust.iterations)) {
        .Deprecated(old="robust.iterations=")
    }
    if (!is.null(robust.trim)) {
        .Deprecated(old="robust.trim=")
    }
    if (!is.null(mass.cap)) {
        .Deprecated(old="mass.cap=")
    }
    if (!is.null(order)) {
        .Deprecated(old="order=")
    }

    if (!is.null(reference.policy)) {
        merge.policy <- sub("^max-", "", reference.policy)
        .Deprecated(old=sprintf("reference.policy=%s", deparse(reference.policy)), new=sprintf("merge.policy=%s", deparse(merge.policy)))
    }
    merge.policy <- match.arg(merge.policy)

    output <- correct_mnn(
        x, 
        block$index, 
        num_neighbors=num.neighbors,
        num_steps=num.steps,
        num_threads=num.threads,
        merge_policy=merge.policy,
        builder=defineBuilder(BNPARAM)$builder
    )

    output
}

# library(testthat); library(scrapper); source("test-correctMnn.R")

set.seed(99991)
test_that("correctMnn works correctly in simple cases", {
    x <- matrix(rnorm(10000), nrow=10)
    b <- rep(1:3, c(500, 300, 200))
    x[,b==2] <- x[,b==2] + 3
    x[,b==3] <- x[,b==3] + 5

    means <- vapply(split(colMeans(x), b), mean, 0)
    expect_gt(means[2] - means[1], 2)
    expect_gt(means[3] - means[1], 4)

    corrected <- correctMnn(x, b)
    expect_identical(dim(corrected$corrected), dim(x))

    means <- vapply(split(colMeans(corrected$corrected), b), mean, 0)
    expect_lt(abs(means[2] - means[1]), 1) # means are within 1 SD of the truth.
    expect_lt(abs(means[3] - means[1]), 1)

    expect_error(correctMnn(SummarizedExperiment::SummarizedExperiment(x)), "not supported")
})

#' Test for gene set enrichment 
#'
#' Perform a hypergeometric test for enrichment of gene sets in a list of interesting genes (e.g., markers).
#'
#' @param x Vector of identifiers for some interesting genes, e.g., symbols or Ensembl IDs.
#' This is usually derived from a selection of top markers, e.g., from \code{\link{scoreMarkers}}.
#' @param sets List of vectors of identifiers for the pre-defined gene sets.
#' Each inner vector corresponds to a gene set and should contain the same type of identifiers as \code{x}.
#' @param universe Vector of identifiers for the universe of genes in the dataset.
#' \code{x} and each vector in \code{sets} will be subsetted to only include those genes in \code{universe}.
#' If \code{NULL}, the universe is defined as the union of all genes in \code{x} and \code{sets}. 
#'
#' Alternatively, an integer scalar specifying the number of genes in the universe.
#' This is assumed to be greater than or equal to the number of unique genes in \code{x} and \code{sets}. 
#' @param log Logical scalar indicating whether to report log-transformed p-values.
#' This may be desirable to avoid underflow at near-zero p-values.
#' @param num.threads Integer scalar specifying the number of threads to use.
#'
#' @return \link[S4Vectors]{DataFrame} with one row per gene set and the following columns:
#' \itemize{
#' \item \code{overlap}, the overlap between \code{x} and each entry of \code{sets}, i.e., the number of genes in the intersection.
#' \item \code{size}, the set of each entry of \code{sets}.
#' \item \code{p.value}, the (possibly log-transformed) p-value for overrepresentation of the gene set in \code{x}.
#' }
#' 
#' @author Aaron Lun
#'
#' @seealso
#' \code{\link{phyper}} and \url{https://libscran.github.io/phyper/},
#' which is the basis for the underlying calculation.
#'
#' @examples
#' testEnrichment(
#'     x=LETTERS[1:5], 
#'     sets=list(
#'         first=LETTERS[1:10],
#'         second=LETTERS[1:5 * 2],
#'         third=LETTERS[10:20]
#'     ),
#'     universe=LETTERS
#' )
#' 
#' @export
#' @importFrom S4Vectors DataFrame
testEnrichment <- function(x, sets, universe=NULL, log=FALSE, num.threads=1) {
    num.sets <- length(sets)
    all.genes <- unlist(sets)
    set.sizes <- lengths(sets)
    set.ids <- rep(seq_along(sets), set.sizes)

    if (is.null(universe)) {
        universe <- length(unique(c(x, all.genes)))
    } else if (!is.numeric(universe) || length(universe) != 1) {
        x <- intersect(x, universe)
        keep <- all.genes %in% universe
        if (!all(keep)) {
            all.genes <- all.genes[keep]
            set.ids <- set.ids[keep]
            set.sizes <- tabulate(set.ids, nbins=num.sets)
        }
        universe <- length(universe)
    }

    overlap <- tabulate(set.ids[all.genes %in% x], nbins=num.sets)
    out <- test_enrichment(overlap, length(x), set.sizes, universe, log=log, num_threads=num.threads)

    DataFrame(
        overlap=overlap,
        size=set.sizes,
        p.value=out,
        row.names=names(sets)
    )
}

# library(scrapper); library(testthat); source("test-testEnrichment.R")

test_that("testEnrichment works as expected", {
    sets <- list(
        first=LETTERS[1:10],
        second=LETTERS[1:5 * 2],
        third=LETTERS[10:20]
    )

    out <- testEnrichment(x=LETTERS[1:5], sets, universe=LETTERS)
    expected <- phyper(c(5, 2, 0) - 1L, lengths(sets), 26 - lengths(sets), 5L, lower.tail=FALSE)
    expect_identical(rownames(out), c("first", "second", "third"))
    expect_equal(out$p.value, expected)

    # Works with log-transformation.
    lout <- testEnrichment(x=LETTERS[1:5], sets, universe=LETTERS, log=TRUE)
    expected <- phyper(c(5, 2, 0) - 1L, lengths(sets), 26 - lengths(sets), 5L, log=TRUE, lower.tail=FALSE)
    expect_equal(lout$p.value, expected)

    # Works with parallelization.
    pout <- testEnrichment(x=LETTERS[1:5], sets, universe=LETTERS, num.threads=2)
    expect_equal(pout, out)
})

test_that("testEnrichment works with a truncated universe", {
    sets <- list(LETTERS[1:10])
    out <- testEnrichment(x=LETTERS[1:5], sets, universe=setdiff(LETTERS, "G"))
    expected <- phyper(5 - 1L, 9, 25 - 9, 5, lower.tail=FALSE)
    expect_equal(out$p.value, expected)

    sets <- list(LETTERS[1:10])
    out <- testEnrichment(x=LETTERS[1:5], sets, universe=setdiff(LETTERS, "A"))
    expected <- phyper(4 - 1L, 9, 25 - 9, 4, lower.tail=FALSE)
    expect_equal(out$p.value, expected)
})

test_that("testEnrichment works with a NULL universe", {
    sets <- list(
        first=LETTERS[1:10],
        second=LETTERS[1:5 * 2],
        third=LETTERS[10:20]
    )

    ref <- testEnrichment(x=LETTERS[1:5], sets, universe=LETTERS[1:20])
    out <- testEnrichment(x=LETTERS[1:5], sets, universe=NULL)
    expect_identical(ref, out)

    ref <- testEnrichment(x=LETTERS[c(1:4, 22:26)], sets, universe=LETTERS[-21])
    out <- testEnrichment(x=LETTERS[c(1:4, 22:26)], sets, universe=NULL)
    expect_identical(ref, out)
})

test_that("testEnrichment works with an integer universe", {
    sets <- list(LETTERS[1:10])
    out <- testEnrichment(x=LETTERS[1:5], sets, universe=26)
    expected <- phyper(5 - 1L, 10, 26 - 10, 5, lower.tail=FALSE)
    expect_equal(out$p.value, expected)
})

#' Analyze single-cell data
#'
#' Execute a simple single-cell analysis pipeline, starting from a count matrix and ending with clusters, visualizations and markers.
#' This also supports integration of multiple modalities and correction of batch effects.
#'
#' @param rna.x Matrix-like object containing RNA counts.
#' This should have the same number of columns as the other \code{*.x} arguments.
#'
#' Alternatively, a \link[SummarizedExperiment]{SummarizedExperiment} instance containing such a matrix in its \code{rna.assay}.
#'
#' Alternatively \code{NULL}, if no RNA counts are available.
#' @param adt.x Matrix-like object containing ADT counts.
#' This should have the same number of columns as the other \code{*.x} arguments.
#'
#' Alternatively, a \link[SummarizedExperiment]{SummarizedExperiment} instance containing such a matrix in its \code{adt.assay}.
#'
#' Alternatively \code{NULL}, if no ADT counts are available.
#' @param crispr.x Matrix-like object containing ADT counts.
#' This should have the same number of columns as the other \code{*.x} arguments.
#'
#' Alternatively, a \link[SummarizedExperiment]{SummarizedExperiment} instance containing such a matrix in its \code{crispr.assay}.
#'
#' Alternatively \code{NULL}, if no ADT counts are available.
#' @param block Factor specifying the block of origin (e.g., batch, sample) for each cell in the \code{*_x} matrices.
#' Alternatively \code{NULL}, if all cells are from the same block.
#' @param rna.subsets Gene subsets for quality control, typically used for mitochondrial genes.
#' See the \code{subsets} arguments in \code{\link{computeRnaQcMetrics}} for details.
#' @param adt.subsets ADT subsets for quality control, typically used for IgG controls.
#' See the \code{subsets} arguments in \code{\link{computeAdtQcMetrics}} for details.
#' @param suggestRnaQcThresholds.args Named list of arguments to pass to \code{\link{suggestRnaQcThresholds}}.
#' @param suggestAdtQcThresholds.args Named list of arguments to pass to \code{\link{suggestAdtQcThresholds}}.
#' @param suggestCrisprQcThresholds.args Named list of arguments to pass to \code{\link{suggestCrisprQcThresholds}}.
#' @param filter.cells Logical scalar indicating whether to filter the count matrices to only retain high-quality cells in all modalities.
#' If \code{FALSE}, QC metrics and thresholds are still computed but are not used to filter the count matrices.
#' @param centerSizeFactors.args Named list of arguments to pass to \code{\link{centerSizeFactors}}.
#' @param computeClrm1Factors.args Named list of arguments to pass to \code{\link{computeClrm1Factors}}.
#' Only used if \code{adt.x} is provided.
#' @param normalizeCounts.args Named list of arguments to pass to \code{\link{normalizeCounts}}.
#' @param modelGeneVariances.args Named list of arguments to pass to \code{\link{modelGeneVariances}}.
#' Only used if \code{rna.x} is provided.
#' @param chooseHighlyVariableGenes.args Named list of arguments to pass to \code{\link{chooseHighlyVariableGenes}}.
#' Only used if \code{rna.x} is provided.
#' @param runPca.args Named list of arguments to pass to \code{\link{runPca}}.
#' @param use.rna.pcs Logical scalar indicating whether to use the RNA-derived PCs for downstream steps (i.e., clustering, visualization).
#' Only used if \code{rna.x} is provided.
#' @param use.adt.pcs Logical scalar indicating whether to use the ADT-derived PCs for downstream steps (i.e., clustering, visualization).
#' Only used if \code{adt.x} is provided.
#' @param use.crispr.pcs Logical scalar indicating whether to use the CRISPR-derived PCs for downstream steps (i.e., clustering, visualization).
#' Only used if \code{crispr.x} is provided.
#' @param scaleByNeighbors.args Named list of arguments to pass to \code{\link{scaleByNeighbors}}.
#' Only used if multiple modalities are available and their corresponding \code{use.*.pcs} arguments are \code{TRUE}.
#' @param correctMnn.args Named list of arguments to pass to \code{\link{correctMnn}}.
#' Only used if \code{block} is supplied.
#' @param runTsne.args Named list of arguments to pass to \code{\link{runTsne}}.
#' If \code{NULL}, t-SNE is not performed.
#' @param runUmap.args Named list of arguments to pass to \code{\link{runUmap}}.
#' If \code{NULL}, UMAP is not performed.
#' @param buildSnnGraph.args Named list of arguments to pass to \code{\link{buildSnnGraph}}.
#' Ignored if \code{clusterGraph.args = NULL}.
#' @param clusterGraph.args Named list of arguments to pass to \code{\link{clusterGraph}}.
#' If \code{NULL}, graph-based clustering is not performed.
#' @param runAllNeighborSteps.args Named list of arguments to pass to \code{\link{runAllNeighborSteps}}.
#' @param kmeans.clusters Integer scalar specifying the number of clusters to use in k-means clustering.
#' If \code{NULL}, k-means clustering is not performed.
#' @param clusterKmeans.args Named list of arguments to pass to \code{\link{clusterKmeans}}.
#' Ignored if \code{kmeans.clusters = NULL}.
#' @param clusters.for.markers Character vector of clustering algorithms (either \code{"graph"} or \code{"kmeans"}, specifying the clustering to be used for marker detection.
#' The first available clustering will be chosen.
#' @param scoreMarkers.args Named list of arguments to pass to \code{\link{scoreMarkers}}.
#' Ignored if no suitable clusterings are available.
#' @param BNPARAM A \link[BiocNeighbors]{BiocNeighborParam} instance specifying the nearest-neighbor search algorithm to use.
#' @param rna.assay Integer scalar or string specifying the assay to use if \code{rna.x} is a \link[SummarizedExperiment]{SummarizedExperiment}.
#' @param adt.assay Integer scalar or string specifying the assay to use if \code{adt.x} is a \link[SummarizedExperiment]{SummarizedExperiment}.
#' @param crispr.assay Integer scalar or string specifying the assay to use if \code{crispr.x} is a \link[SummarizedExperiment]{SummarizedExperiment}.
#' @param num.threads Integer scalar specifying the number of threads to use in each step.
#'
#' @return List containing the results of the entire analysis:
#' \describe{
#' \item{\code{rna.qc.metrics}:}{Results of \code{\link{computeRnaQcMetrics}}.
#' If RNA data is not available, this is set to \code{NULL} instead.}
#' \item{\code{rna.qc.thresholds}:}{Results of \code{\link{suggestRnaQcThresholds}}.
#' If RNA data is not available, this is set to \code{NULL} instead.}
#' \item{\code{rna.qc.filter}:}{Results of \code{\link{filterRnaQcMetrics}}.
#' If RNA data is not available, this is set to \code{NULL} instead.}
#' \item{\code{adt.qc.metrics}:}{Results of \code{\link{computeAdtQcMetrics}}.
#' If ADT data is not available, this is set to \code{NULL} instead.}
#' \item{\code{adt.qc.thresholds}:}{Results of \code{\link{suggestAdtQcThresholds}}.
#' If ADT data is not available, this is set to \code{NULL} instead.}
#' \item{\code{adt.qc.filter}:}{Results of \code{\link{filterAdtQcMetrics}}.
#' If ADT data is not available, this is set to \code{NULL} instead.}
#' \item{\code{crispr.qc.metrics}:}{Results of \code{\link{computeCrisprQcMetrics}}.
#' If CRISPR data is not available, this is set to \code{NULL} instead.}
#' \item{\code{crispr.qc.thresholds}:}{Results of \code{\link{suggestCrisprQcThresholds}}.
#' If CRISPR data is not available, this is set to \code{NULL} instead.}
#' \item{\code{crispr.qc.filter}:}{Results of \code{\link{filterCrisprQcMetrics}}.
#' If CRISPR data is not available, this is set to \code{NULL} instead.}
#' \item{\code{combined.qc.filter}:}{Logical vector indicating which cells are of high quality and should be retained for downstream analyses.}
#' \item{\code{rna.filtered}:}{Matrix of RNA counts that has been filtered to only contain the high-quality cells in \code{combined.qc.filter}.
#' If RNA data is not available, this is set to \code{NULL} instead.}
#' \item{\code{adt.filtered}:}{Matrix of ADT counts that has been filtered to only contain the high-quality cells in \code{combined.qc.filter}.
#' If ADT data is not available, this is set to \code{NULL} instead.}
#' \item{\code{crispr.filtered}:}{Matrix of CRISPR counts that has been filtered to only contain the high-quality cells in \code{combined.qc.filter}.
#' If CRISPR data is not available, this is set to \code{NULL} instead.}
#' \item{\code{rna.size.factors}:}{Size factors for the RNA count matrix, derived from the sum of counts for each cell and centered with \code{\link{centerSizeFactors}}.
#' If RNA data is not available, this is set to \code{NULL} instead.}
#' \item{\code{rna.normalized}:}{Matrix of (log-)normalized expression values derived from RNA counts, as computed by \code{\link{normalizeCounts}} using \code{rna.size.factors}.
#' If RNA data is not available, this is set to \code{NULL} instead.}
#' \item{\code{adt.size.factors}:}{Size factors for the ADT count matrix, computed by \code{\link{computeClrm1Factors}} and centered with \code{\link{centerSizeFactors}}.
#' If ADT data is not available, this is set to \code{NULL} instead.}
#' \item{\code{adt.normalized}:}{Matrix of (log-)normalized expression values derived from ADT counts, as computed by \code{\link{normalizeCounts}} using \code{adt.size.factors}.
#' If ADT data is not available, this is set to \code{NULL} instead.}
#' \item{\code{crispr.size.factors}:}{Size factors for the CRISPR count matrix, derived from the sum of counts for each cell and centered with \code{\link{centerSizeFactors}}.
#' If CRISPR data is not available, this is set to \code{NULL} instead.}
#' \item{\code{crispr.normalized}:}{Matrix of (log-)normalized expression values derived from CRISPR counts, as computed by \code{\link{normalizeCounts}} using \code{crispr.size.factors}.
#' If CRISPR data is not available, this is set to \code{NULL} instead.}
#' \item{\code{rna.gene.variances}:}{Results of \code{\link{modelGeneVariances}}.
#' If RNA data is not available, this is set to \code{NULL} instead.}
#' \item{\code{rna.highly.variable.genes}:}{Results of \code{\link{chooseHighlyVariableGenes}}.
#' If RNA data is not available, this is set to \code{NULL} instead.}
#' \item{\code{rna.pca}:}{Results of calling \code{\link{runPca}} on \code{rna.normalized} with the \code{rna.highly.variable.genes} subset.
#' If RNA data is not available, this is set to \code{NULL} instead.}
#' \item{\code{adt.pca}:}{Results of calling \code{\link{runPca}} on \code{adt.normalized}.
#' If ADT data is not available, this is set to \code{NULL} instead.}
#' \item{\code{crispr.pca}:}{Results of calling \code{\link{runPca}} on \code{crispr.normalized}.
#' If CRISPR data is not available, this is set to \code{NULL} instead.}
#' \item{\code{combined.pca}:}{If only one modality is used for the downstream analysis, this is a string specifying the list element containing the components to be used, e.g., \code{"rna.pca"}.
#' If multiple modalities are to be combined for downstream analysis, this contains the results of \code{\link{scaleByNeighbors}} on the PCs of those modalities.}
#' \item{\code{block}:}{Vector or factor containing the blocking factor for all cells (after filtering, if \code{filter.cells = TRUE}).
#' This is set to \code{NULL} if no blocking factor was supplied.}
#' \item{\code{mnn.corrected}:}{Results of \code{\link{correctMnn}} on the PCs in or referenced by \code{combined.pca}.
#' If no blocking factor is supplied, this is set to \code{NULL} instead.}
#' \item{\code{tsne}:}{Results of \code{\link{runTsne}}.
#' This is \code{NULL} if t-SNE was not performed.}
#' \item{\code{umap}:}{Results of \code{\link{runUmap}}.
#' This is \code{NULL} if UMAP was not performed.}
#' \item{\code{snn.graph}:}{Results of \code{\link{buildSnnGraph}}.
#' This is \code{NULL} if graph-based clustering was not performed, or if \code{return.graph=FALSE} in \code{\link{runAllNeighborSteps}}.}
#' \item{\code{graph.clusters}:}{Results of \code{\link{clusterGraph}}.
#' This is \code{NULL} if graph-based clustering was not performed.}
#' \item{\code{kmeans.clusters}:}{Results of \code{\link{clusterKmeans}}.
#' This is \code{NULL} if k-means clustering was not performed.}
#' \item{\code{clusters}:}{Integer vector containing the cluster assignment for each cell (after filtering, if \code{filter.cells = TRUE}).
#' This may be derived from \code{graph.clusters} or \code{kmeans.clusters} depending on the choice of \code{clusters.for.markers}.
#' If no suitable clusterings are available, this is set to NULL.}
#' \item{\code{rna.markers}:}{Results of calling \code{\link{scoreMarkers}} on \code{rna.normalized}.
#' This is \code{NULL} if RNA data is not available or no suitable clusterings are available.}
#' \item{\code{adt.markers}:}{Results of calling \code{\link{scoreMarkers}} on \code{adt.normalized}.
#' This is \code{NULL} if ADT data is not available or no suitable clusterings are available.}
#' \item{\code{crispr.markers}:}{Results of calling \code{\link{scoreMarkers}} on \code{crispr.normalized}.
#' This is \code{NULL} if CRISPR data is not available or no suitable clusterings are available.}
#' }
#'
#' @seealso
#' \code{\link{convertAnalyzeResults}}, to convert the results into a \link[SingleCellExperiment]{SingleCellExperiment}.
#'
#' @author Aaron Lun
#' @examples
#' sce <- getTestRnaData.se()
#' sce <- sce[,1:500] # smaller dataset for a faster runtime for R CMD check. 
#' res <- analyze(
#'     sce, 
#'     rna.subsets=list(mito=grep("^mt-", rownames(sce))),
#'     num.threads=2 # keep R CMD check happy
#' )
#' str(res)
#' convertAnalyzeResults(res)
#'
#' @export
#' @importFrom methods is
#' @importFrom BiocNeighbors AnnoyParam
#' @importFrom DelayedArray DelayedArray
analyze <- function(
    rna.x,
    adt.x = NULL,
    crispr.x = NULL,
    block = NULL,
    rna.subsets = list(),
    adt.subsets = list(),
    suggestRnaQcThresholds.args = list(),
    suggestAdtQcThresholds.args = list(),
    suggestCrisprQcThresholds.args = list(),
    filter.cells = TRUE,
    centerSizeFactors.args = list(),
    computeClrm1Factors.args = list(),
    normalizeCounts.args = list(),
    modelGeneVariances.args = list(),
    chooseHighlyVariableGenes.args = list(),
    runPca.args = list(),
    use.rna.pcs = TRUE,
    use.adt.pcs = TRUE,
    use.crispr.pcs = TRUE,
    scaleByNeighbors.args = list(),
    correctMnn.args = list(),
    runUmap.args = list(),
    runTsne.args = list(),
    buildSnnGraph.args = list(),
    clusterGraph.args = list(),
    runAllNeighborSteps.args = list(),
    kmeans.clusters = NULL,
    clusterKmeans.args = list(),
    clusters.for.markers = c("graph", "kmeans"),
    scoreMarkers.args = list(),
    BNPARAM = AnnoyParam(),
    rna.assay = 1L,
    adt.assay = 1L,
    crispr.assay = 1L,
    num.threads = 3L)
{
    store <- list()
    .Deprecated(new="analyze.se")

    try.se.extract <- function(x, assay) {
        if (is(x, "SummarizedExperiment")) {
            SummarizedExperiment::assay(x, assay)
        } else {
            x
        }
    }

    all.ncols <- integer()
    if (!is.null(rna.x)) {
        all.ncols <- c(all.ncols, ncol(rna.x))
        rna.x <- try.se.extract(rna.x, rna.assay)
    }
    if (!is.null(adt.x)) {
        all.ncols <- c(all.ncols, ncol(adt.x))
        adt.x <- try.se.extract(adt.x, adt.assay)
    }
    if (!is.null(crispr.x)) {
        all.ncols <- c(all.ncols, ncol(crispr.x))
        crispr.x <- try.se.extract(crispr.x, crispr.assay)
    }

    ncells <- unique(all.ncols)
    if (length(ncells) > 1) {
        stop("'*.x' have differing numbers of columns")
    }
    if (length(ncells) == 0) {
        stop("at least one '*.x' must be non-NULL")
    }

    ############ Quality control o(*°▽°*)o #############

    if (!is.null(rna.x)) {
        store[["rna.qc.metrics"]] <- computeRnaQcMetrics(rna.x, subsets=rna.subsets, num.threads=num.threads)
        store[["rna.qc.thresholds"]] <- do.call(suggestRnaQcThresholds, c(list(store[["rna.qc.metrics"]], block=block), suggestRnaQcThresholds.args))
        store[["rna.qc.filter"]] <- filterRnaQcMetrics(store[["rna.qc.thresholds"]], store[["rna.qc.metrics"]], block=block)
    } else {
        store["rna.qc.metrics"] <- list(NULL)
        store["rna.qc.thresholds"] <- list(NULL)
        store["rna.qc.filter"] <- list(NULL)
    }

    if (!is.null(adt.x)) {
        store[["adt.qc.metrics"]] <- computeAdtQcMetrics(adt.x, subsets=adt.subsets, num.threads=num.threads)
        store[["adt.qc.thresholds"]] <- do.call(suggestAdtQcThresholds, c(list(store[["adt.qc.metrics"]], block=block), suggestAdtQcThresholds.args))
        store[["adt.qc.filter"]] <- filterAdtQcMetrics(store[["adt.qc.thresholds"]], store[["adt.qc.metrics"]], block=block)
    } else {
        store["adt.qc.metrics"] <- list(NULL)
        store["adt.qc.thresholds"] <- list(NULL)
        store["adt.qc.filter"] <- list(NULL)
    }

    if (!is.null(crispr.x)) {
        store[["crispr.qc.metrics"]] <- computeCrisprQcMetrics(crispr.x, num.threads=num.threads)
        store[["crispr.qc.thresholds"]] <- do.call(suggestCrisprQcThresholds, c(list(store[["crispr.qc.metrics"]], block=block), suggestCrisprQcThresholds.args))
        store[["crispr.qc.filter"]] <- filterCrisprQcMetrics(store[["crispr.qc.thresholds"]], store[["crispr.qc.metrics"]], block=block)
    } else {
        store["crispr.qc.metrics"] <- list(NULL)
        store["crispr.qc.thresholds"] <- list(NULL)
        store["crispr.qc.filter"] <- list(NULL)
    }

    # Combining all filters.
    combined.qc.filter <- !logical(ncells)
    for (mod in c("rna", "adt", "crispr")) {
        keep <- store[[paste0(mod, ".qc.filter")]]
        if (!is.null(keep)) {
            combined.qc.filter <- combined.qc.filter & keep
        }
    }
    store[["combined.qc.filter"]] <- combined.qc.filter

    if (!is.null(rna.x)) {
        store[["rna.filtered"]] <- DelayedArray(rna.x)
    } else {
        store["rna.filtered"] <- list(NULL)
    }
    if (!is.null(adt.x)) {
        store[["adt.filtered"]] <- DelayedArray(adt.x)
    } else {
        store["adt.filtered"] <- list(NULL)
    }
    if (!is.null(crispr.x)) {
        store[["crispr.filtered"]] <- DelayedArray(crispr.x)
    } else {
        store["crispr.filtered"] <- list(NULL)
    }

    if (filter.cells) {
        if (!is.null(rna.x)) {
            store[["rna.filtered"]] <- store[["rna.filtered"]][,combined.qc.filter,drop=FALSE]
        }
        if (!is.null(adt.x)) {
            store[["adt.filtered"]] <- store[["adt.filtered"]][,combined.qc.filter,drop=FALSE]
        }
        if (!is.null(crispr.x)) {
            store[["crispr.filtered"]] <- store[["crispr.filtered"]][,combined.qc.filter,drop=FALSE]
        }
        if (!is.null(block)) {
            block <- block[combined.qc.filter]
        }
    }

    ############ Normalization ( ꈍᴗꈍ) #############

    maybe.filter <- function(x) {
        if (filter.cells && !all(combined.qc.filter)) {
            x[combined.qc.filter]
        } else {
            x
        }
    }

    if (!is.null(rna.x)) {
        store[["rna.size.factors"]] <- do.call(centerSizeFactors, c(list(maybe.filter(store[["rna.qc.metrics"]]$sum), block=block), centerSizeFactors.args))
        store[["rna.normalized"]] <- do.call(normalizeCounts, c(list(store[["rna.filtered"]], store[["rna.size.factors"]]), normalizeCounts.args))
    } else {
        store["rna.size.factors"] <- list(NULL)
        store["rna.normalized"] <- list(NULL)
    }

    if (!is.null(adt.x)) {
        raw.adt.sf <- do.call(computeClrm1Factors, c(list(store[["adt.filtered"]], num.threads=num.threads), computeClrm1Factors.args))
        store[["adt.size.factors"]] <- do.call(centerSizeFactors, c(list(raw.adt.sf, block=block), centerSizeFactors.args))
        store[["adt.normalized"]] <- do.call(normalizeCounts, c(list(store[["adt.filtered"]], store[["adt.size.factors"]]), normalizeCounts.args))
    } else {
        store["adt.size.factors"] <- list(NULL)
        store["adt.normalized"] <- list(NULL)
    }

    if (!is.null(crispr.x)) {
        store[["crispr.size.factors"]] <- do.call(centerSizeFactors, c(list(maybe.filter(store[["crispr.qc.metrics"]]$sum), block=block), centerSizeFactors.args))
        store[["crispr.normalized"]] <- do.call(normalizeCounts, c(list(store[["crispr.filtered"]], store[["crispr.size.factors"]]), normalizeCounts.args))
    } else {
        store["crispr.size.factors"] <- list(NULL)
        store["crispr.normalized"] <- list(NULL)
    }

    ############ Variance modelling (～￣▽￣)～ #############

    if (!is.null(rna.x)) {
        store[["rna.gene.variances"]] <- do.call(modelGeneVariances, c(list(store[["rna.normalized"]], block=block, num.threads=num.threads), modelGeneVariances.args))
        store[["rna.highly.variable.genes"]] <- do.call(chooseHighlyVariableGenes, c(list(store[["rna.gene.variances"]]$statistics$residuals), chooseHighlyVariableGenes.args))
    } else {
        store["rna.gene.variances"] <- list(NULL)
        store["rna.highly.variable.genes"] <- list(NULL)
    }

    ############ Principal components analysis \(>⩊<)/ #############

    if (!is.null(rna.x)) {
        hvg.mat <- store[["rna.normalized"]][store[["rna.highly.variable.genes"]],,drop=FALSE]
        store[["rna.pca"]] <- do.call(runPca, c(list(hvg.mat, block=block, num.threads=num.threads), runPca.args))
    } else {
        store["rna.pca"] <- list(NULL)
    }

    if (!is.null(adt.x)) {
        store[["adt.pca"]] <- do.call(runPca, c(list(store[["adt.normalized"]], block=block, num.threads=num.threads), runPca.args))
    } else {
        store["adt.pca"] <- list(NULL)
    }

    if (!is.null(crispr.x)) {
        store[["crispr.pca"]] <- do.call(runPca, c(list(store[["crispr.normalized"]], block=block, num.threads=num.threads), runPca.args))
    } else {
        store["crispr.pca"] <- list(NULL)
    }

    ############ Combining modalities and batches („• ᴗ •„) #############

    embeddings <- character()
    if (use.rna.pcs && !is.null(rna.x)) {
        embeddings <- c(embeddings, "rna.pca")
    }
    if (use.adt.pcs && !is.null(adt.x)) {
        embeddings <- c(embeddings, "adt.pca")
    }
    if (use.crispr.pcs && !is.null(crispr.x)) {
        embeddings <- c(embeddings, "crispr.pca")
    }

    if (length(embeddings) == 0) {
        stop("at least one 'use.*' must be True")
    }
    if (length(embeddings) == 1) {
        store[["combined.pca"]] <- embeddings[1]
        chosen.pcs <- store[[embeddings[1]]]$components
    } else {
        embeddings <- lapply(embeddings, function(e) store[[e]]$components)
        store[["combined.pca"]] <- do.call(scaleByNeighbors, c(list(embeddings, BNPARAM=BNPARAM, num.threads=num.threads), scaleByNeighbors.args))
        chosen.pcs <- store[["combined.pca"]]$combined
    }

    store[["block"]] <- block
    if (is.null(block)) {
        store["mnn.corrected"] <- list(NULL)
    } else {
        store[["mnn.corrected"]] <- do.call(correctMnn, c(list(chosen.pcs, block=block, BNPARAM=BNPARAM, num.threads=num.threads), correctMnn.args))
        chosen.pcs <- store[["mnn.corrected"]]$corrected
    }

    ############ Assorted neighbor-related stuff ⸜(⸝⸝⸝´꒳`⸝⸝⸝)⸝ #############

    nn.out <- do.call(runAllNeighborSteps, c(
        list(
            chosen.pcs,
            runUmap.args=runUmap.args, 
            runTsne.args=runTsne.args, 
            buildSnnGraph.args=buildSnnGraph.args,
            clusterGraph.args=clusterGraph.args,
            BNPARAM=BNPARAM,
            num.threads=num.threads
        ),
        runAllNeighborSteps.args
    ))

    store["tsne"] <- nn.out["runTsne"]
    store["umap"] <- nn.out["runUmap"]
    store["snn.graph"] <- nn.out["buildSnnGraph"]
    store["graph.clusters"] <- nn.out["clusterGraph"]

    ############ Finding markers (˶˃ ᵕ ˂˶) #############

    if (!is.null(kmeans.clusters)) {
        store[["kmeans.clusters"]] <- do.call(clusterKmeans, c(list(chosen.pcs, k=kmeans.clusters, num.threads=num.threads), clusterKmeans.args))
    } else {
        store["kmeans.clusters"] <- list(NULL)
    }

    store["clusters"] <- list(NULL)
    for (c in clusters.for.markers) {
        if (c == "graph") {
            candidate <- store[["graph.clusters"]]
            if (!is.null(candidate)) {
                store[["clusters"]] <- candidate$membership
                break
            }
        } else if (c == "kmeans") {
            candidate <- store[["kmeans.clusters"]]
            if (!is.null(candidate)) {
                store[["clusters"]] <- candidate$clusters
                break
            }
        }
    }

    store["rna.markers"] <- list(NULL)
    store["adt.markers"] <- list(NULL)
    store["crispr.markers"] <- list(NULL)

    clusters <- store[["clusters"]]
    if (!is.null(clusters)) {
        if (!is.null(rna.x)) {
            store[["rna.markers"]] <- do.call(scoreMarkers, c(list(store[["rna.normalized"]], groups=clusters, num.threads=num.threads, block=block), scoreMarkers.args))
        }
        if (!is.null(adt.x)) {
            store[["adt.markers"]] <- do.call(scoreMarkers, c(list(store[["adt.normalized"]], groups=clusters, num.threads=num.threads, block=block), scoreMarkers.args))
        }
        if (!is.null(crispr.x)) {
            store[["crispr.markers"]] <- do.call(scoreMarkers, c(list(store[["crispr.normalized"]], groups=clusters, num.threads=num.threads, block=block), scoreMarkers.args))
        }
    }

    store
}

convertQcMetricsToDataFrame <- function(metrics, flatten = TRUE) {
    nrow <- length(metrics$sum) # all QC functions produce 'sum'.
    output <- S4Vectors::make_zero_col_DFrame(nrow=nrow)

    for (n in names(metrics)) {
        current <- metrics[[n]]
        if (!is.list(current)) {
            output[[n]] <- current
        } else {
            if (is.null(names(current))) {
                names(current) <- seq_along(current)
            }
            if (flatten) {
                for (x in names(current)) {
                    output[[paste0(n, ".", x)]] <- current[[x]]
                }
            } else {
                tmp <- S4Vectors::make_zero_col_DFrame(nrow=nrow)
                for (x in names(current)) {
                    tmp[[x]] <- current[[x]]
                }
                output[[n]] <- tmp
            }
        }
    }

    output
}

#' Convert analysis results into a SingleCellExperiment
#'
#' Convert results from \code{\link{analyze}} into a \link[SingleCellExperiment]{SingleCellExperiment} for further analysis with Bioconductor packages.
#'
#' @param results List of results produced by \code{\link{analyze}}.
#' @param main.modality String specifying the modality to use as the main experiment of a \link[SingleCellExperiment]{SingleCellExperiment}.
#' @param flatten.qc.subsets Logical scalar indicating whether QC metrics for subsets should be flattened in the column data.
#' If \code{FALSE}, subset metrics are reported as a nested \link[S4Vectors]{DataFrame}.
#' @param include.per.block.variances Logical scalar indicating whether the per-block variances should be reported as a nested \link[S4Vectors]{DataFrame} in the row data.
#'
#' @return A \link[SingleCellExperiment]{SingleCellExperiment} containing most of the analysis results.
#' Filtered and normalized matrices are stored in the assays.
#' QC metrics, size factors and clusterings are stored in the column data.
#' Gene variances are stored in the row data.
#' PCA, t-SNE and UMAP results are stored in the reduced dimensions.
#' Further modalities are stored as alternative experiments.
#'
#' @author Aaron Lun
#' @seealso
#' \code{\link{analyze}}, to generate \code{results}.
#'
#' @export
convertAnalyzeResults <- function(
    results,
    main.modality = NULL,
    flatten.qc.subsets = TRUE,
    include.per.block.variances = FALSE)
{
    objects <- list()
    .Deprecated(new="analyze.se")

    maybe.filter <- function(x, ncells) {
        if (nrow(x) > ncells) {
            x[results$combined.qc.filter,,drop=FALSE]
        } else {
            x
        }
    }

    if (!is.null(results$rna.filtered)) {
        rna.sce <- SingleCellExperiment::SingleCellExperiment(
            assays=list(filtered=results$rna.filtered, normalized=results$rna.normalized),
            reducedDims=list(pca=t(results$rna.pca$components))
        )

        cd <- SummarizedExperiment::colData(rna.sce)
        qcdf <- convertQcMetricsToDataFrame(results$rna.qc.metrics, flatten=flatten.qc.subsets)
        cd <- cbind(cd, maybe.filter(qcdf, nrow(cd)))
        SummarizedExperiment::colData(rna.sce) <- cd
        SingleCellExperiment::sizeFactors(rna.sce) <- results$rna.size.factors

        rd <- SummarizedExperiment::rowData(rna.sce)
        rd <- cbind(rd, results$rna.gene.variances$statistics)
        if (include.per.block.variances) {
            per.block <- results$rna.gene.variances$per.block
            if (!is.null(per.block)) {
                for (i in seq_along(per.block)) {
                    per.block[[i]] <- S4Vectors::I(S4Vectors::DataFrame(per.block[[i]]))
                }
                rd$per.block <- S4Vectors::DataFrame(per.block)
            }
        }
        hvgs <- logical(nrow(rd))
        hvgs[results$rna.highly.variable.genes] <- TRUE
        rd$is.highly.variable <- hvgs
        SummarizedExperiment::rowData(rna.sce) <- rd

        objects$rna <- rna.sce
    }

    if (!is.null(results$adt.filtered)) {
        adt.sce <- SingleCellExperiment::SingleCellExperiment(
            assays=list(filtered=results$adt.filtered, normalized=results$adt.normalized),
            reducedDims=list(pca=t(results$adt.pca$components))
        )

        cd <- SummarizedExperiment::colData(adt.sce)
        qcdf <- convertQcMetricsToDataFrame(results$adt.qc.metrics, flatten=flatten.qc.subsets)
        cd <- cbind(cd, maybe.filter(qcdf, nrow(cd)))
        SummarizedExperiment::colData(adt.sce) <- cd
        SingleCellExperiment::sizeFactors(adt.sce) <- results$adt.size.factors

        objects$adt <- adt.sce
    }

    if (!is.null(results$crispr.filtered)) {
        crispr.sce <- SingleCellExperiment::SingleCellExperiment(
            assays=list(filtered=results$crispr.filtered, normalized=results$crispr.normalized),
            reducedDims=list(pca=t(results$crispr.pca$components))
        )

        cd <- SummarizedExperiment::colData(crispr.sce)
        qcdf <- convertQcMetricsToDataFrame(results$crispr.qc.metrics)
        cd <- cbind(cd, maybe.filter(qcdf, nrow(cd)))
        SummarizedExperiment::colData(crispr.sce) <- cd
        SingleCellExperiment::sizeFactors(crispr.sce) <- results$crispr.size.factors

        objects$crispr <- crispr.sce
    }

    if (is.null(main.modality)) {
        main.modality <- names(objects)[1]
    }

    sce <- objects[[main.modality]]
    SingleCellExperiment::mainExpName(sce) <- main.modality
    objects[[main.modality]] <- NULL
    SingleCellExperiment::altExps(sce) <- objects

    if (!is.character(results$combined.pca)) {
        SingleCellExperiment::reducedDim(sce, "combined.pca") <- t(results$combined.pca$combined)
    }

    if (!is.null(results$mnn.corrected)) {
        sce$block <- results$block
        SingleCellExperiment::reducedDim(sce, "mnn.corrected") <- t(results$mnn.corrected$corrected)
    }

    if (!is.null(results$tsne)) {
        SingleCellExperiment::reducedDim(sce, "tsne") <- results$tsne
    }

    if (!is.null(results$umap)) {
        SingleCellExperiment::reducedDim(sce, "umap") <- results$umap
    }

    if (!is.null(results$clusters)) {
        sce$clusters <- results$clusters
    }

    sce
}

#' Aggregate expression across cells in a SummarizedExperiment
#' 
#' Aggregate expression values across groups of cells for each gene,
#' by calling \code{\link{aggregateAcrossCells}} on an assay in a \link[SummarizedExperiment]{SummarizedExperiment}.
#'
#' @param x A \link[SummarizedExperiment]{SummarizedExperiment} object or one of its subclasses.
#' Rows correspond to genes and columns correspond to cells.
#' @param factors List or data frame (or their equivalents from \pkg{S4Vectors}) containing grouping factors, see \code{\link{aggregateAcrossCells}} for more details.
#'
#' Alternatively, an atomic vector or factor representing a single variable.
#' @param num.threads Number of threads, passed to \code{\link{aggregateAcrossCells}}.
#' @param more.aggr.args Named list of additional arguments to pass to \code{\link{aggregateAcrossCells}}.
#' @param assay.type Integer or string specifying the assay of \code{x} to be aggregated.
#' @param output.prefix String specifying a prefix to add to the names of the \code{link[SummarizedExperiment]{colData}} columns storing the factor combinations. 
#' If \code{NULL}, no prefix is added.
#' @param counts.name String specifying the name of the \code{\link[SummarizedExperiment]{colData}} column in which to store the cell count for each factor combination. 
#' If \code{NULL}, the cell counts are not reported.
#' @param meta.name String specifying the name of the \code{\link[S4Vectors]{metadata}} entry in which to store additional outputs like the combination indices.
#' If \code{NULL}, additional outputs are not reported.
#' @param include.coldata Logical scalar indicating whether to add the aggregated \code{colData} from \code{x} to the output.
#' @param more.coldata.args Named list of additional arguments to pass to \code{aggregateColData}.
#' Only relevant if \code{include.coldata=TRUE}.
#' @param altexps Unnamed integer or character vector specifying the indices/names of alternative experiments of \code{x} to aggregate.
#' The aggregated assay from each alternative experiment is determined by \code{assay.type}.
#'
#' Alternatively, this may be a named integer or character vector.
#' Each name specifies an alternative experiment while each value is the index/name of the assay to aggregate from that experiment.
#'
#' Only relevant if \code{x} is a \link[SingleCellExperiment]{SingleCellExperiment}.
#' @param copy.altexps Logical scalar indicating whether to copy the \code{colData} and \code{metadata} of the output SingleCellExperiment into each of its alternative experiments.
#' @param coldata \link[S4Vectors]{DataFrame} of column data, containing one row for each cell.
#' @param index Integer vector containing the index of the factor combination to which each cell in \code{coldata} was assigned.
#' @param number Integer specifying the total number of unique factor combinations.
#' All elements of \code{index} should be less than or equal to \code{number}.
#' @param only.atomic Logical scalar specifying whether to skip non-atomic, non-factor columns.
#' @param placeholder Placeholder value to store in the output column when a factor combination does not have a single unique value. 
#'
#' @return
#' For \code{aggregateAcrossCells.se}, a SummarizedExperiment is returned where each column corresponds to a factor combination.
#' Each row corresponds to a gene in \code{x} and the \code{\link[SummarizedExperiment]{rowData}} is taken from \code{x}.
#' The assays contain the sum of counts (\code{"sums"}) and the number of detected cells (\code{"detected"}) in each combination for each gene.
#' The \code{colData} contains:
#' \itemize{
#' \item The factor combinations, with column names prefixed by \code{output.prefix}.
#' \item The cell count for each combination, named by \code{counts.name}.
#' \item Additional \code{colData} from \code{x} if \code{include.coldata=TRUE}.
#' This is aggregated with \code{aggregateColData} on the combination indices.
#' }
#' The metadata contains a list named as \code{meta.name}, containing a \code{index} integer vector of length equal to the number of cells in \code{x}.
#' Each entry of this vector is an index of the factor combination (i.e., column of the output object) to which the corresponding cell was assigned.
#'
#' If \code{altexps} is specified, a SingleCellExperiment is returned instead.
#' The same aggregation for the main experiment is applied to each alternative experiment.
#' If \code{copy.altexps=TRUE}, the \code{colData} for each alternative experiment will contain a copy of the factor combinations and cell counts,
#' and the \code{metadata} will contain a copy of the index vector.
#'
#' For \code{aggregateColData}, a \link[S4Vectors]{DataFrame} is returned with number of rows equal to \code{number}.
#' Each atomic or factor column in \code{coldata} is represented by a column in the output DataFrame.
#' In each column, the \code{j}-th entry is equal to the unique value of all rows where \code{index == j},
#' or \code{placeholder} if there is not exactly one unique value.
#' If \code{only.atomic=FALSE}, any non-atomic/non-factor columns of \code{coldata} are represented in the output DataFrame by a vector of \code{placeholder} values.
#' If \code{only.atomic=TRUE}, any non-atomic/non-factor columns of \code{coldata} are skipped.
#' 
#' @author Aaron Lun
#' @examples
#' library(SingleCellExperiment)
#' sce <- getTestRnaData.se("start")
#' aggr <- aggregateAcrossCells.se(sce, sce$level1class)
#' head(assay(aggr))
#' colData(aggr)
#'
#' # We can also aggregate within alternative experiments.
#' aggr2 <- aggregateAcrossCells.se(sce, sce$level1class, altexps="ERCC")
#' head(assay(altExp(aggr2, "ERCC")))
#'
#' @export
#' @importFrom methods is
#' @importFrom S4Vectors cbind metadata metadata<-
#' @importClassesFrom S4Vectors List DataFrame
aggregateAcrossCells.se <- function(
    x,
    factors,
    num.threads = 1,
    more.aggr.args = list(),
    assay.type = "counts",
    output.prefix = "factor.",
    counts.name = "counts",
    meta.name = "aggregated",
    include.coldata = TRUE,
    more.coldata.args = list(),
    altexps = NULL,
    copy.altexps = FALSE
) {
    if (is.list(factors) || is.data.frame(factors) || is(factors, "List") || is(factors, "DataFrame")) {
        # this is fine.
    } else {
        factors <- list(factors)
        names(factors) <- "1" # mimic what happens in combineFactors.
    }

    out <- .call(
        aggregateAcrossCells,
        list(SummarizedExperiment::assay(x, assay.type)),
        list(factors=factors, num.threads=num.threads),
        more.aggr.args
    )

    CON <- SummarizedExperiment::SummarizedExperiment
    if (length(altexps)) {
        CON <- SingleCellExperiment::SingleCellExperiment
    }
    se <- CON(out[intersect(names(out), c("sums", "detected", "medians"))], rowData=SummarizedExperiment::rowData(x))

    common.cd <- out$combinations
    colnames(common.cd) <- paste0(output.prefix, colnames(common.cd))
    if (!is.null(counts.name)) {
        common.cd[[counts.name]] <- out$counts
    }
    SummarizedExperiment::colData(se) <- common.cd
    if (include.coldata) {
        aggr.cd <- do.call(aggregateColData, c(list(SummarizedExperiment::colData(x), out$index, number=nrow(out$combinations)), more.coldata.args))
        SummarizedExperiment::colData(se) <- cbind(SummarizedExperiment::colData(se), aggr.cd)
    }

    if (!is.null(meta.name)) {
        metadata(se)[[meta.name]] <- list(index=out$index)
    }

    if (length(altexps)) {
        SingleCellExperiment::mainExpName(se) <- SingleCellExperiment::mainExpName(x)
        altexps <- .sanitizeAltexpAssays(altexps, all.altexps=SingleCellExperiment::altExpNames(x), default.assay.type=assay.type)

        for (ae in names(altexps)) {
            ae.se <- Recall(
                SingleCellExperiment::altExp(x, ae),
                out$index,
                num.threads=num.threads,
                more.aggr.args=more.aggr.args,
                assay.type=altexps[[ae]],
                altexps=NULL,
                output.prefix=NULL,
                counts.name=NULL,
                meta.name=NULL,
                include.coldata=include.coldata
            )

            ae.cd <- SummarizedExperiment::colData(ae.se)[,-1,drop=FALSE] # remove uninteresting factor combination
            if (copy.altexps) {
                ae.cd <- cbind(common.cd, ae.cd)
            }
            SummarizedExperiment::colData(ae.se) <- ae.cd

            if (copy.altexps) {
                metadata(ae.se) <- metadata(se)
            }
            SingleCellExperiment::altExp(se, ae) <- ae.se
        }
    }

    se
}

#' @export
#' @rdname aggregateAcrossCells.se
#' @importFrom S4Vectors make_zero_col_DFrame
aggregateColData <- function(coldata, index, number, only.atomic = TRUE, placeholder = NA) {
    collected <- make_zero_col_DFrame(nrow=number)
    index <- factor(index, seq_len(number))

    for (cn in colnames(coldata)) {
        curcol <- coldata[[cn]]
        if (!is.atomic(curcol) && !is.factor(curcol)) {
            if (!only.atomic) {
                collected[[cn]] <- rep(placeholder, number)
            }
            next
        }

        grouped <- split(curcol, index)
        alloc <- rep(curcol[1], number)
        for (i in seq_along(grouped)) {
            u <- unique(grouped[[i]])
            if (length(u) != 1L) {
                alloc[i] <- placeholder
            } else {
                alloc[i] <- u
            }
        }

        collected[[cn]] <- alloc
    }

    collected
}

# library(testthat); library(scrapper); source("test-se_aggregateAcrossCells.R")

library(SingleCellExperiment)
mat <- matrix(rpois(1000, 10), ncol=100)
se <- SummarizedExperiment(list(counts=mat))
se$stuff <- sample(LETTERS[1:5], ncol(se), replace=TRUE)
se$whee <- sample(c(TRUE, FALSE), ncol(se), replace=TRUE)
rowData(se)$foo <- runif(nrow(se))

test_that("aggregateAcrossCells.se works as expected", {
    out <- aggregateAcrossCells.se(se, colData(se)[,"stuff",drop=FALSE])
    expect_identical(ncol(out), length(unique(se$stuff)))
    expect_identical(rowData(out), rowData(se))
    expect_identical(out$factor.stuff, out$stuff)
    expect_identical(sort(out$factor.stuff), sort(unique(se$stuff)))
    expect_identical(out$counts, as.integer(table(se$stuff)[out$stuff]))
    expect_identical(metadata(out)$aggregated$index, match(se$stuff, out$stuff))

    # Works with direct specification of the factor.
    out2 <- aggregateAcrossCells.se(se, se$stuff)
    expect_identical(out2$factor.1, out2$stuff)
    expect_identical(assay(out2), assay(out))

    # Works if we provide a list.
    out3 <- aggregateAcrossCells.se(se, list(foo=se$stuff), output.prefix=NULL)
    expect_identical(out3$foo, out$stuff)
    expect_identical(assay(out3), assay(out))

    # Works with multiple factors.
    mult <- aggregateAcrossCells.se(se, colData(se)[,c("stuff", "whee"),drop=FALSE])
    expect_identical(mult$factor.stuff, mult$stuff)
    expect_identical(mult$factor.whee, mult$whee)
    counters <- S4Vectors::match(colData(se)[,c("stuff", "whee"),drop=FALSE], colData(mult)[,c("stuff", "whee"),drop=FALSE])
    expect_identical(as.integer(table(counters)), mult$counts)
    expect_identical(metadata(mult)$aggregated$index, counters)

    # Works if we disable the outputs.
    out4 <- aggregateAcrossCells.se(se, se$stuff, meta.name=NULL, counts.name=NULL, include.coldata=FALSE)
    expect_null(out4$counts)
    expect_null(metadata(out4)$aggregated)
    expect_null(out4$stuff)
})

test_that("aggregateAcrossCells.se works with alternative experiments", {
    sce <- as(se, "SingleCellExperiment")
    is.even <- 1:10 %% 2 == 0
    altExp(sce, "GuP") <- se[is.even,]
    sce$random <- paste0("FOO_", sce$stuff)

    out <- aggregateAcrossCells.se(sce, colData(se)[,c("stuff", "whee"),drop=FALSE])
    expect_false(is(out, "SingleCellExperiment"))

    out <- aggregateAcrossCells.se(sce, colData(se)[,c("stuff", "whee"),drop=FALSE], altexps="GuP")
    expect_s4_class(out, "SingleCellExperiment")
    expect_identical(assay(out)[is.even,], assay(altExp(out)))
    expect_identical(rowData(out)[is.even,,drop=FALSE], rowData(altExp(out)))
    expect_identical(out$stuff, altExp(out)$stuff)
    expect_identical(out$random, paste0("FOO_", out$stuff))
    expect_null(altExp(out)$factor.stuff)
    expect_null(altExp(out)$factor.whee)
    expect_null(altExp(out)$counts)
    expect_null(metadata(altExp(out))$aggregated)
    expect_null(altExp(out)$random)

    # Copying restores the extra outputs.
    out <- aggregateAcrossCells.se(sce, colData(se)[,c("stuff", "whee"),drop=FALSE], altexps="GuP", copy.altexps=TRUE)
    expect_s4_class(out, "SingleCellExperiment")
    expect_identical(out$stuff, altExp(out)$factor.stuff)
    expect_identical(out$whee, altExp(out)$factor.whee)
    expect_identical(out$counts, altExp(out)$counts)
    expect_identical(metadata(out)$aggregated, metadata(altExp(out))$aggregated)
})

test_that("aggregateColData works as expected", {
    df <- DataFrame(stuff=rep(LETTERS[1:5], each=4), whee=rep(c(TRUE, FALSE), 10))

    adf <- aggregateColData(df, as.integer(factor(df$stuff)), 5)
    expect_identical(adf$stuff, LETTERS[1:5])
    expect_identical(adf$whee, rep(NA, 5))

    adf <- aggregateColData(df, as.integer(factor(df$whee)), 2)
    expect_identical(adf$stuff, rep(NA_character_, 2))
    expect_identical(adf$whee, c(FALSE, TRUE))

    # Still works as a factor.
    df$stuff <- factor(df$stuff)
    adf <- aggregateColData(df, as.integer(factor(df$stuff)), 5)
    expect_identical(adf$stuff, factor(LETTERS[1:5]))

    # Handles non-atomic, non-factor columns.
    df$foo <- DataFrame(bar=runif(nrow(df)))
    adf <- aggregateColData(df, as.integer(factor(df$whee)), 2)
    expect_null(adf$foo)
    adf <- aggregateColData(df, as.integer(factor(df$whee)), 2, only.atomic=FALSE)
    expect_identical(adf$foo, rep(NA, 2))

    zeroed <- aggregateColData(df[0,], integer(0), 0)
    expect_true(is.factor(zeroed$stuff))
    expect_type(zeroed$whee, "logical")
})

#' Run all nearest neighbor steps on a SummarizedExperiment
#'
#' Concurrently run all steps involving a nearest-neighbor search (t-SNE, UMAP and graph-based clustering) using the same nearest-neighbor index,
#' by calling \code{\link{runAllNeighborSteps}} on a reduced dimension entry of a \link[SingleCellExperiment]{SingleCellExperiment}.
#'
#' @param x A \link[SingleCellExperiment]{SingleCellExperiment} object or one of its subclasses.
#' Rows correspond to genomic features and columns correspond to cells.
#' @param umap.output.name String containing the name of the \code{\link[SingleCellExperiment]{reducedDim}} entry to store the UMAP coordinates.
#' If \code{NULL}, the UMAP is not computed.
#' @param more.umap.args Named list of additional arguments to pass to \code{\link{runAllNeighborSteps}} as \code{runUmap.args}.
#' @param tsne.output.name String containing the name of the \code{\link[SingleCellExperiment]{reducedDim}} entry to store the t-SNE coordinates.
#' If \code{NULL}, the t-SNE is not computed.
#' @param more.tsne.args Named list of additional arguments to pass to \code{\link{runAllNeighborSteps}} as \code{runTsne.args}.
#' @param build.graph.name String containing the name of the \code{\link[S4Vectors]{metadata}} entry in which to store the nearest neighbor graph.
#' If \code{NULL}, the graph is not stored.
#' @param more.build.graph.args Named list of additional arguments to pass to \code{\link{runAllNeighborSteps}} as \code{buildSnnGraph.args}.
#' @param cluster.output.name String containing the name of the \code{\link[SummarizedExperiment]{colData}} column in which to store the cluster assignments.
#' If \code{NULL}, graph-based clustering is not performed.
#' @param cluster.meta.name String containing the name of the \code{\link[S4Vectors]{metadata}} entry in which to store additional clustering outputs.
#' If \code{NULL}, these additional outputs are not stored.
#' @param more.cluster.graph.args Named list of additional arguments to pass to \code{\link{runAllNeighborSteps}} as \code{clusterGraph.args}.
#' @param BNPARAM,num.threads Arguments to pass to \code{\link{runAllNeighborSteps}}.
#' @param more.neighbor.args Named list of additional arguments to pass to \code{\link{runAllNeighborSteps}}.
#' @param reddim.type String or integer specifying the \code{\link[SingleCellExperiment]{reducedDim}} entry on which to perform a nearest neighbor search.
#' Alternatively, a named integer or character vector of length 1,
#' where the name specifies an alternative experiment of \code{x} and the value is the name/index of a \code{reducedDim} entry in that alternative experiment.
#'
#' @return \code{x} is returned with additional coordinates stored in its \code{\link[SingleCellExperiment]{reducedDims}}
#' and clustering output in its \code{\link[SummarizedExperiment]{colData}}.
#' Additional information may also be stored in its \code{\link[S4Vectors]{metadata}}.
#' 
#' @author Aaron Lun
#'
#' @examples
#' library(SingleCellExperiment)
#' sce <- getTestRnaData.se("pca")
#' sce <- runAllNeighborSteps.se(
#'    sce,
#'    more.tsne.args=list(max.iterations=50),
#'    more.umap.args=list(num.epochs=50),
#'    num.threads=2 # to keep R CMD check happy
#' )
#' reducedDimNames(sce)
#' table(sce$clusters)
#' 
#' @export 
#' @importFrom BiocNeighbors AnnoyParam
runAllNeighborSteps.se <- function(
    x,
    umap.output.name = "UMAP",
    more.umap.args = list(),
    tsne.output.name = "TSNE",
    more.tsne.args = list(),
    build.graph.name = NULL,
    more.build.graph.args = list(),
    cluster.output.name = "clusters",
    cluster.meta.name = NULL,
    more.cluster.graph.args = list(),
    BNPARAM = AnnoyParam(),
    num.threads = 3,
    more.neighbor.args = list(),
    reddim.type = "PCA"
) {
    if (is.null(umap.output.name)) {
        more.umap.args <- NULL 
    }
    if (is.null(tsne.output.name)) {
        more.tsne.args <- NULL 
    }
    if (is.null(cluster.output.name)) {
        more.cluster.graph.args <- NULL 
    }
    more.neighbor.args$return.graph <- !is.null(build.graph.name)

    outputs <- .call(
        runAllNeighborSteps,
        list(.getTransposedReddim(x, reddim.type)),
        list(
            runUmap.args=more.umap.args,
            runTsne.args=more.tsne.args,
            buildSnnGraph.args=more.build.graph.args,
            clusterGraph.args=more.cluster.graph.args,
            BNPARAM = BNPARAM,
            num.threads=num.threads
        ),
        more.neighbor.args
    )

    if (!is.null(outputs$runTsne)) {
        x <- .addTsneResults(x, outputs$runTsne, output.name=tsne.output.name)
    }
    if (!is.null(outputs$runUmap)) {
        x <- .addUmapResults(x, outputs$runUmap, output.name=umap.output.name)
    }
    if (!is.null(outputs$clusterGraph)) {
        x <- .addBuildGraphResults(x, outputs$buildSnnGraph, graph.name=build.graph.name)
        x <- .addClusterGraphResults(x, outputs$clusterGraph, output.name=cluster.output.name, meta.name=cluster.meta.name)
    }

    x
}

# library(testthat); library(scrapper); source("test-se_runAllNeighborSteps.R")

library(SingleCellExperiment)
set.seed(6666)
sce <- SingleCellExperiment(list(counts = matrix(0, 0, 200)))
reducedDim(sce, "PCA") <- matrix(runif(1000), nrow=200)

test_that("runAllNeighborSteps.se works as expected", {
    out <- runAllNeighborSteps.se(sce, num.threads=2)
    expect_true("TSNE" %in% reducedDimNames(out))
    expect_true("UMAP" %in% reducedDimNames(out))
    expect_true(is.factor(out$clusters))

    # Nullifying everything.
    null <- runAllNeighborSteps.se(sce, umap.output.name=NULL, tsne.output.name=NULL, cluster.output.name=NULL, num.threads=2)
    expect_false("TSNE" %in% reducedDimNames(null))
    expect_false("UMAP" %in% reducedDimNames(null))
    expect_null(null$clusters)

    # Checking that we can get the graph.
    wt.graph <- runAllNeighborSteps.se(sce, umap.output.name=NULL, tsne.output.name=NULL, build.graph.name="graph", num.threads=2)
    expect_false("TSNE" %in% reducedDimNames(wt.graph))
    expect_false("UMAP" %in% reducedDimNames(wt.graph))
    expect_true(is.factor(wt.graph$clusters))
    expect_type(metadata(wt.graph)$graph, "list")
})

#' Quality control for RNA count data
#'
#' Compute per-cell QC metrics from an initialized matrix of RNA counts,
#' and use the metrics to suggest filter thresholds to retain high-quality cells.
#' 
#' @param x A matrix-like object where rows are genes and columns are cells.
#' Values are expected to be counts.
#' @param subsets Named list of vectors specifying gene subsets of interest, typically for control-like features like mitochondrial genes or spike-in transcripts.
#' Each vector may be logical (whether to keep each row), integer (row indices) or character (row names).
#' For character vectors, strings not present in \code{rownames(x)} are ignored.
#' @param num.threads Integer scalar specifying the number of threads to use.
#' @param metrics \link[S4Vectors]{DataFrame} of per-cell QC metrics.
#' This should have the same structure as the return value of \code{computeRnaQcMetrics}.
#' @param block Factor specifying the block of origin (e.g., batch, sample) for each cell in \code{metrics}.
#' Alternatively \code{NULL} if all cells are from the same block.
#'
#' For \code{filterRnaQcMetrics}, a blocking factor should be provided if \code{block} was used to construct \code{thresholds}. 
#' @param num.mads Number of median from the median, to define the threshold for outliers in each metric.
#' @param thresholds List with the same structure as produced by \code{suggestRnaQcThresholds}.
#'
#' @return For \code{computeRnaQcMetrics}, a \link[S4Vectors]{DataFrame} is returned with one row per cell in \code{x}.
#' This contains the following columns:
#' \itemize{
#' \item \code{sum}, a numeric vector containing the total RNA count for each cell.
#' This represents the efficiency of library preparation and sequencing.
#' Low totals indicate that the library was not successfully captured.
#' \item \code{detected}, an integer vector containing the number of detected genes per cell.
#' This also quantifies library preparation efficiency but with greater focus on capturing transcriptional complexity.
#' \item \code{subsets}, a nested DataFrame where each column corresponds to a feature subset and is a numeric vector containing the proportion of counts in that subset.
#' The exact interpretation of which depends on the nature of the subset.
#' For example, if one subset contains all genes on the mitochondrial chromosome, higher proportions are representative of cell damage;
#' the assumption is that cytoplasmic transcripts leak through tears in the cell membrane while the mitochondria are still trapped inside.
#' The proportion of spike-in transcripts can be interpreted in a similar manner, where the loss of endogenous transcripts results in higher spike-in proportions.
#' }
#' Each vector is of length equal to the number of cells.
#'
#' For \code{suggestRnaQcThresholds}, a named list is returned.
#' \itemize{
#' \item If \code{block=NULL}, the list contains:
#' \itemize{
#' \item \code{sum}, a numeric scalar containing the lower bound on the sum.
#' This is defined as \code{num.mads} MADs below the median of the log-transformed metrics across all cells.
#' \item \code{detected}, a numeric scalar containing the lower bound on the number of detected genes. 
#' This is defined as \code{num.mads} MADs below the median of the log-transformed metrics across all cells.
#' \item \code{subsets}, a numeric vector containing the upper bound on the sum of counts in each feature subset. 
#' This is defined as \code{num.mads} MADs above the median across all cells.
#' }
#' \item Otherwise, if \code{block} is supplied, the list contains:
#' \itemize{
#' \item \code{sum}, a numeric vector containing the lower bound on the sum for each blocking level.
#' Here, the threshold is computed independently for each block, using the same method as the unblocked case.
#' \item \code{detected}, a numeric vector containing the lower bound on the number of detected genes for each blocking level.
#' Here, the threshold is computed independently for each block, using the same method as the unblocked case.
#' \item \code{subsets}, a list of numeric vectors containing the upper bound on the sum of counts in each feature subset for each blocking level.
#' Here, the threshold is computed independently for each block, using the same method as the unblocked case.
#' \item \code{block.ids}, a vector containing the identities of the unique blocks.
#' }
#' Each vector is of length equal to the number of levels in \code{block} and is named accordingly.
#' }
#'
#' For \code{filterRnaQcMetrics}, a logical vector of length \code{ncol(x)} is returned indicating which cells are of high quality. 
#' High-quality cells are defined as those with sums and detected genes above their respective thresholds and subset proportions below the \code{subsets} threshold.
#'
#' @seealso
#' The \code{compute_rna_qc_metrics}, \code{compute_rna_qc_filters} and \code{compute_rna_qc_filters_blocked} functions in \url{https://libscran.github.io/scran_qc/}.
#'
#' \code{\link{quickRnaQc.se}}, to run all of the RNA-related QC functions on a \link[SummarizedExperiment]{SummarizedExperiment}.
#'
#' @author Aaron Lun
#' @examples
#' # Mocking a matrix:
#' library(Matrix)
#' x <- round(abs(rsparsematrix(1000, 100, 0.1) * 100))
#'
#' # Mocking up a control set.
#' sub <- list(mito=rbinom(nrow(x), 1, 0.1) > 0)
#'
#' qc <- computeRnaQcMetrics(x, sub)
#' qc
#'
#' filt <- suggestRnaQcThresholds(qc)
#' str(filt)
#'
#' keep <- filterRnaQcMetrics(filt, qc)
#' summary(keep)
#'
#' @export
#' @name rna_quality_control
#' @importFrom beachmat initializeCpp tatami.dim
computeRnaQcMetrics <- function(x, subsets, num.threads = 1) {
    if (length(subsets) > 0 && is.null(names(subsets))) {
        stop("subsets should be a named list")
    }
    .checkSEX(x, "quickRnaQc.se")

    ptr <- initializeCpp(x, .check.na=FALSE)

    subsets <- as.list(subsets)
    subsets <- lapply(subsets, .subsetToLogical, n=tatami.dim(ptr)[1], names=rownames(x))

    output <- compute_rna_qc_metrics(ptr, subsets, num_threads=num.threads)
    .reformatQcMetrics(output, "subsets", subsets, x)
}

#' @export
#' @rdname rna_quality_control
suggestRnaQcThresholds <- function(metrics, block=NULL, num.mads=3) {
    block <- .transformFactor(block)
    metrics <- .simplifyQcMetrics(metrics)
    thresholds <- suggest_rna_qc_thresholds(metrics, block=block$index, num_mads=num.mads)

    names(thresholds$subsets) <- names(metrics$subsets)

    if (!is.null(block$names)) {
        names(thresholds$sum) <- block$names
        names(thresholds$detected) <- block$names
        for (i in seq_along(metrics$subsets)) {
            names(thresholds$subsets[[i]]) <- block$names
        }
        thresholds$block.ids <- block$names # store it separately to preserve any non-character type.
    }

    thresholds
}

#' @export
#' @rdname rna_quality_control
filterRnaQcMetrics <- function(thresholds, metrics, block=NULL) {
    block <- .matchBlockThresholds(block, thresholds$block.ids)
    metrics <- .simplifyQcMetrics(metrics)
    .checkThresholdNames(thresholds, c("sum", "detected"), "subsets", names(metrics$subsets))
    filter_rna_qc_metrics(thresholds, metrics, block=block)
}

#' @importFrom S4Vectors DataFrame I
.reformatQcMetrics <- function(res, subfield, subsets, mat) {
    if (length(res$subsets) > 0) {
        names(res$subsets) <- names(subsets)
        subdf <- DataFrame(res$subsets, row.names=colnames(mat))
    } else {
        subdf <- make_zero_col_DFrame(ncol(mat))
        rownames(subdf) <- colnames(mat)
    }
    res$subsets <- I(subdf)
    DataFrame(res, row.names=colnames(mat))
}

.simplifyQcMetrics <- function(df) {
    metrics <- as.list(df)
    metrics$subsets <- as.list(metrics$subsets)
    metrics
}

.checkThresholdNames <- function(thresholds, regular.fields, subset.field, subset.names) {
    if (!is.null(subset.field)) {
        if (!identical(names(thresholds[[subset.field]]), subset.names)) {
            stop("expected identical names for 'thresholds$", subset.field, "' and 'metrics$", subset.field, "'")
        }
    }

    if (!is.null(thresholds$block.ids)) {
        block.names <- as.character(thresholds$block.ids)
        for (r in regular.fields) {
            if (!identical(names(thresholds[[r]]), block.names)) {
                stop("expected names for 'thresholds$", r, "' to be identical to 'thresholds$block.ids'")
            }
        }

        if (!is.null(subset.field)) {
            for (s in subset.names) {
                if (!identical(names(thresholds[[subset.field]][[s]]), block.names)) {
                    stop("expected names for each entry of 'thresholds$", subset.field, "' to be identical to 'thresholds$block.ids'")
                }
            }
        }
    }
}

.subsetToLogical <- function(x, n, names) {
    if (is.logical(x)) {
        if (length(x) != n) {
            stop("length of a 'subsets' entry must be equal to the number of rows")
        }
    } else if (is.numeric(x)) {
        if (length(x)) {
            if (anyNA(x) || min(x) < 1 || max(x) > n) {
                stop("'subsets' entry contains out-of-range indices")
            }
        }
        tmp <- logical(n)
        tmp[x] <- TRUE
        x <- tmp
    } else if (is.character(x)) {
        if (is.null(names)) {
            stop("no row names available for matching a 'subsets' entry")
        }
        x <- names %in% x
    }
    x
}

.matchBlockThresholds <- function(block, levels) {
    if (is.null(block) && is.null(levels)) {
        return(NULL)
    } else if (is.null(block)) {
        stop("expected 'block=' to be supplied for blocked 'thresholds'")
    } else if (is.null(levels)) {
        stop("'block=' should be set to NULL for unblocked 'thresholds'")
    }

    m <- match(block, levels)
    if (anyNA(m)) {
        stop("entries of 'block' are not present in 'thresholds'")
    }

    m - 1L
}

# library(testthat); library(scrapper); source("test-rna_quality_control.R")

set.seed(444555666)
library(Matrix)
x <- round(abs(rsparsematrix(5000, 1000, 0.1) * 100))

test_that("computeRnaQcMetrics works as expected", { 
    sub <- list(Mito=rbinom(nrow(x), 1, 0.1) > 0)
    qc <- computeRnaQcMetrics(x, sub)
    expect_identical(qc$sum, Matrix::colSums(x))
    expect_identical(qc$detected, Matrix::colSums(x > 0))
    expect_identical(qc$subsets$Mito, Matrix::colSums(x[sub$Mito,]) / qc$sum)

    expect_error(computeRnaQcMetrics(x, list(1)), "should be a named list")
    expect_error(computeRnaQcMetrics(x, list(foo=TRUE)), "number of rows")
    expect_error(computeRnaQcMetrics(x, list(foo=-1)), "out-of-range")
    expect_error(computeRnaQcMetrics(x, list(foo="foo")), "no row names")

    # Works with strings, where missing values are ignored..
    rownames(x) <- sprintf("GENE-%s", seq_len(nrow(x)))
    qc2 <- computeRnaQcMetrics(x, list(Mito=c("FOO", rownames(x)[sub$Mito], "BAR")))
    expect_identical(qc, qc2)

    expect_error(computeRnaQcMetrics(SummarizedExperiment::SummarizedExperiment(x), NULL), "not supported")
})

test_that("suggestRnaQcThresholds works as expected", { 
    sub <- list(Mito=rbinom(nrow(x), 1, 0.1) > 0)
    qc <- computeRnaQcMetrics(x, sub)
    num.mads <- 1.5
    thresholds <- suggestRnaQcThresholds(qc, num.mads=num.mads)

    # Check the thresholds.
    expect_lt(thresholds$sum, median(qc$sum))
    expect_lt(thresholds$detected, median(qc$detected))

    sub <- qc$subsets[[1]]
    expect_gt(thresholds$subsets[[1]], median(sub))
    expect_equal(names(thresholds$subsets), "Mito")

    # Check the filter.
    expected <- qc$sum >= thresholds$sum & qc$detected >= thresholds$detected & qc$subsets[[1]] <= thresholds$subsets[[1]]
    observed <- filterRnaQcMetrics(thresholds, qc)
    expect_identical(expected, observed)

    expect_error(filterRnaQcMetrics(thresholds, qc, block=1:10), "should be set to NULL")
})

test_that("suggestRnaQcThresholds works as expected with blocking", { 
    sub <- list(Mito=rbinom(nrow(x), 1, 0.1) > 0)
    qc <- computeRnaQcMetrics(x, sub)
    block <- sample(3, ncol(x), replace=TRUE)
    num.mads <- 1.5
    thresholds <- suggestRnaQcThresholds(qc, block=block, num.mads=num.mads)

    expect_identical(thresholds$block.ids, 1:3)

    # Check the thresholds.
    for (b in 1:3) {
        keep <- block == b
        expect_lt(thresholds$sum[[b]], median(qc$sum[keep]))
        expect_lt(thresholds$detected[[b]], median(qc$detected[keep]))
        expect_gt(thresholds$subsets[[1]][[b]], median(qc$subsets[[1]][keep]))
    }

    # Check the filter.
    expected <- qc$sum >= thresholds$sum[block] & qc$detected >= thresholds$detected[block] & qc$subsets[[1]] <= thresholds$subsets[[1]][block]
    observed <- filterRnaQcMetrics(thresholds, qc, block=block)
    expect_identical(unname(expected), observed)

    # Same filtering with just the last block.
    last <- block == 3
    last_observed <- filterRnaQcMetrics(
        thresholds,
        list(sum=qc$sum[last], detected=qc$detected[last], subsets=list(Mito=qc$subsets[[1]][last])),
        block=block[last]
    )
    expect_identical(observed[last], last_observed)

    expect_error(filterRnaQcMetrics(thresholds, qc), "expected 'block='")
    expect_error(filterRnaQcMetrics(thresholds, qc, block=1:10), "not present in 'thresholds'")
})

#' k-means clustering of cells in a SingleCellExperiment
#'
#' Perform k-means clustering on an existing low-dimensional embedding
#' by calling \code{\link{clusterKmeans}} on a reduced dimension entry in a \link[SingleCellExperiment]{SingleCellExperiment}.
#'
#' @param x A \link[SingleCellExperiment]{SingleCellExperiment} object or one of its subclasses.
#' Rows correspond to genomic features and columns correspond to cells.
#' @param k Number of clusters, passed to \code{\link{clusterKmeans}}.
#' @param num.threads Number of threads, passed to \code{\link{clusterKmeans}}.
#' @param more.kmeans.args Named list of further arguments to be passed to \code{\link{clusterKmeans}}.
#' @param reddim.type Integer or string specifying the existing embedding in the \code{\link[SingleCellExperiment]{reducedDim}} of \code{x}.
#' Alternatively, a named integer or character vector of length 1,
#' where the name specifies an alternative experiment of \code{x} and the value is the name/index of a \code{reducedDim} entry in that alternative experiment.
#' @param output.name String containing the name of the \code{\link[SummarizedExperiment]{colData}} column in which to store the cluster assignments.
#' @param meta.name String containing the name of the \code{\link[S4Vectors]{metadata}} entry in which to store extra clustering output.
#' If \code{NULL}, no extra clustering output is stored. 
#'
#' @return \code{x} is returned with the cluster assignment for each cell stored in the \code{colData}.
#' Additional clustering output is stored in the \code{metadata}.
#'
#' @author Aaron Lun
#' @examples
#' sce <- getTestRnaData.se("pca")
#' sce <- clusterKmeans.se(sce, k=10)
#' table(sce$clusters)
#' 
#' @export
#' @importFrom S4Vectors metadata metadata<-
clusterKmeans.se <- function(x, k, num.threads=1, more.kmeans.args = list(), reddim.type = "PCA", output.name = "clusters", meta.name = NULL) {
    clout <- .call(
        clusterKmeans,
        list(.getTransposedReddim(x, reddim.type)),
        list(k=k, num.threads=num.threads),
        more.kmeans.args
    )

    SummarizedExperiment::colData(x)[[output.name]] <- clout$clusters
    if (!is.null(meta.name)) {
        clout$clusters <- NULL
        metadata(x)[[meta.name]] <- clout
    }

    x
}

# library(testthat); library(scrapper); source("test-se_clusterKmeans.R")

library(SingleCellExperiment)
set.seed(8888)
sce <- SingleCellExperiment(list(counts = matrix(0, 0, 200)))
reducedDim(sce, "PCA") <- matrix(runif(1000), nrow=200)

test_that("clusterKmeans.se works as expected", {
    out <- clusterKmeans.se(sce, k=2)
    expect_identical(nlevels(out$clusters), 2L)

    out <- clusterKmeans.se(sce, k=10, meta.name="kmeans")
    expect_identical(nlevels(out$clusters), 10L)
    expect_identical(ncol(metadata(out)$kmeans$centers), 10L)
})

#' Normalize the count matrix
#'
#' Apply scaling normalization and log-transformation to a count matrix.
#' This yields normalized expression values that can be used in downstream procedures like PCA. 
#'
#' @param x A matrix-like object where rows correspond to genes or genomic features and columns correspond to cells.
#' Values are expected to be non-negative counts.
#' Alternatively, an external pointer created by \code{\link[beachmat]{initializeCpp}}.
#' @param size.factors A numeric vector of length equal to the number of cells in \code{x}, containing positive size factors for all cells.
#' Any invalid values should be replaced with \code{\link{sanitizeSizeFactors}}.
#' For most applications, these size factors should also be centered with \code{\link{centerSizeFactors}}.
#' @param delayed Logical scalar indicating whether operations on a matrix-like \code{x} should be delayed.
#' This improves memory efficiency at the cost of some speed in downstream operations.
#' @param log Logical scalar indicating whether log-transformation should be performed.
#' This ensures that downstream analyses like t-tests and distance calculations focus on relative fold-changes rather than absolute differences.
#' The log-transformation also provides some measure of variance stabilization so that the downstream analyses are not dominated by sampling noise at large counts.
#' @param pseudo.count Numeric scalar specifying the positive pseudo-count to add before any log-transformation.
#' Larger values shrink the differences between cells towards zero, reducing spurious differences (but also signal) at low counts - see \code{\link{choosePseudoCount}} for comments.
#' Ignored if \code{log=FALSE}.
#' @param log.base Numeric scalar specifying the base of the log-transformation.
#' Ignored if \code{log=FALSE}.
#' @param preserve.sparsity Logical scalar indicating whether to preserve sparsity for \code{pseudo.count != 1}.
#' If \code{TRUE}, users should manually add \code{log(pseudo.count, log.base)} to the returned matrix to obtain the desired log-transformed expression values.
#' Ignored if \code{log = FALSE} or \code{pseudo.count = 1}.
#'
#' @return If \code{x} is a matrix-like object and \code{delayed=TRUE}, a \link[DelayedArray]{DelayedArray} is returned containing the (log-transformed) normalized expression matrix.
#' If \code{delayed=FALSE}, the type of the (log-)normalized matrix will depend on the operations applied to \code{x}.
#'
#' If \code{x} is an external pointer produced by \code{\link[beachmat]{initializeCpp}}, a new external pointer is returned containing the normalized expression matrix.
#'
#' @seealso
#' The \code{normalize_counts} function in \url{https://libscran.github.io/scran_norm/}.
#'
#' \code{\link{normalizeRnaCounts.se}} and related functions, which compute normalized expression values from a \link[SummarizedExperiment]{SummarizedExperiment}.
#'
#' @author Aaron Lun
#' @examples
#' # Mocking a matrix:
#' library(Matrix)
#' x <- round(abs(rsparsematrix(1000, 100, 0.1) * 100))
#' sf <- centerSizeFactors(colSums(x))
#' normed <- normalizeCounts(x, size.factors=sf)
#' normed
#'
#' # Passing a pointer.
#' ptr <- beachmat::initializeCpp(x)
#' optr <- normalizeCounts(ptr, sf)
#' optr
#'
#' @export
#' @importFrom methods is
#' @importFrom DelayedArray DelayedArray t
normalizeCounts <- function(x, size.factors, log=TRUE, pseudo.count=1, log.base=2, preserve.sparsity=FALSE, delayed=TRUE) {
    .checkSEX(x, "normalizeRnaCounts.se")
    if (is(x, "externalptr")) {
        return(normalize_counts(x, size.factors, log=log, log_base=log.base, pseudo_count=pseudo.count, preserve_sparsity=preserve.sparsity))
    }

    if (!log) {
        if (delayed) {
            x <- DelayedArray(x)
        }
        return(t(t(x)/size.factors))
    }

    if (pseudo.count != 1 && preserve.sparsity) {
        size.factors <- size.factors * pseudo.count
        pseudo.count <- 1
    }

    if (!delayed) {
        .log_normalize(x, size.factors, pseudo.count=pseudo.count, log.base=log.base)
    } else {
        LogNormalizedMatrix(x, size.factors, pseudo.count=pseudo.count, log.base=log.base)
    }
}

#' Model per-gene variances in expression
#'
#' Model the per-gene variances as a function of the mean in single-cell expression data.
#' Highly variable genes can then be selected for downstream analyses. 
#'
#' @param x A matrix-like object where rows correspond to genes or genomic features and columns correspond to cells.
#' It is typically expected to contain log-expression values, e.g., from \code{\link{normalizeCounts}}.
#' @param block Factor specifying the block of origin (e.g., batch, sample) for each cell in \code{x}.
#' If provided, calculation of means/variances and trend fitting are performed within each block to ensure that block effects do not confound the estimates.
#' The weighted average of each statistic across all blocks is reported for each gene.
#' Alternatively \code{NULL}, if all cells are from the same block.
#' @param block.average.policy String specifying the policy to use for average statistics across blocks.
#' This can either be a (weighted) \code{"mean"} or a \code{"quantile"}.
#' Only used if \code{block} is not \code{NULL}. 
#' @param block.weight.policy String specifying the policy to use for weighting different blocks when computing the average for each statistic.
#' See the argument of the same name in \code{\link{computeBlockWeights}} for more detail.
#' Only used if \code{block} is not \code{NULL} and \code{block.average.policy="mean"}.
#' @param variable.block.weight Numeric vector of length 2, specifying the parameters for variable block weighting.
#' See the argument of the same name in \code{\link{computeBlockWeights}} for more detail.
#' Only used if \code{block} is not \code{NULL}, \code{block.average.policy="mean"} and \code{block.weight.policy = "variable"}.
#' @param block.quantile Number specifying the probability of the quantile of statistics across blocks. 
#' Defaults to 0.5, i.e., the median of per-block statistics.
#' Only used if \code{block} is not \code{NULL} and \code{block.average.policy="quantile"}.
#' @inheritParams fitVarianceTrend
#' @param num.threads Integer scalar specifying the number of threads to use.
#'
#' @details
#' We compute the mean and variance for each gene and fit a trend to the variances with respect to the means using \code{\link{fitVarianceTrend}}.
#' We assume that most genes at any given abundance are not highly variable, such that the fitted value of the trend is interpreted as the \dQuote{uninteresting} variance - 
#' this is mostly attributed to technical variation like sequencing noise, but can also represent constitutive biological noise like transcriptional bursting.
#' Under this assumption, the residual can be treated as a measure of biologically interesting variation.
#' Genes with large residuals can then be selected for downstream analyses, e.g., with \code{\link{chooseHighlyVariableGenes}}.
#'
#' @return A list containing \code{statistics}, a \link[S4Vectors]{DataFrame} with number of rows equal to the number of genes.
#' This contains the columns \code{means}, \code{variances}, \code{fitted} and \code{residuals},
#' each of which is a numeric vector containing the statistic of the same name across all genes.
#'
#' If \code{block} is supplied, each of the column vectors described above contains the average across all blocks.
#' The list will also contain \code{per.block}, a list of DataFrames containing the equivalent statistics for each block;
#' and \code{block.ids}, a vector containing the identities of the unique blocks in the same order as \code{per.block}.
#'
#' @author Aaron Lun
#'
#' @seealso
#' The \code{model_gene_variances} function in \url{https://libscran.github.io/scran_variances/}.
#'
#' \code{\link{chooseRnaHvgs.se}}, which computes the variances and trend from a \link[SummarizedExperiment]{SummarizedExperiment}.
#'
#' @examples
#' library(Matrix)
#' x <- abs(rsparsematrix(1000, 100, 0.1) * 10)
#' out <- modelGeneVariances(x)
#' out
#'
#' # Throwing in some blocking.
#' block <- sample(letters[1:4], ncol(x), replace=TRUE)
#' out <- modelGeneVariances(x, block=block)
#' out
#'
#' @export
#' @importFrom beachmat initializeCpp
#' @importFrom S4Vectors DataFrame
modelGeneVariances <- function(
    x,
    block=NULL,
    block.average.policy=c("mean", "quantile"),
    block.weight.policy=c("variable", "equal", "none"),
    variable.block.weight=c(0, 1000),
    block.quantile=0.5,
    mean.filter=TRUE,
    min.mean=0.1, 
    transform=TRUE, 
    span=0.3,
    use.min.width=FALSE,
    min.width=1,
    min.window.count=200,
    num.threads=1
) {
    .checkSEX(x, "chooseRnaHvgs.se")
    block <- .transformFactor(block)

    stats <- model_gene_variances(
        initializeCpp(x, .check.na=FALSE),
        block=block$index,
        nblocks=length(block$names),
        block_average_policy=match.arg(block.average.policy),
        block_weight_policy=match.arg(block.weight.policy),
        variable_block_weight=variable.block.weight,
        block_quantile=block.quantile,
        mean_filter=mean.filter,
        min_mean=min.mean,
        transform=transform,
        span=span,
        use_min_width=use.min.width,
        min_width=min.width,
        min_window_count=min.window.count,
        num_threads=num.threads
    )

    output <- list(
        statistics = DataFrame(
            means = stats$means,
            variances = stats$variances,
            fitted = stats$fitted,
            residuals = stats$residuals,
            row.names = rownames(x)
        )
    )

    if (!is.null(block$index)) {
        pb <- stats$per.block
        for (i in seq_along(pb)) {
            curstats <- pb[[i]]
            pb[[i]] <- DataFrame( 
                means = curstats$means,
                variances = curstats$variances,
                fitted = curstats$fitted,
                residuals = curstats$residuals,
                row.names = rownames(x)
            )
        }
        names(pb) <- block$names
        output$per.block <- pb
        output$block.ids <- block$names
    }

    output
}

#' Quick quality control for RNA data in a SummarizedExperiment
#'
#' Quickly compute quality control (QC) metrics, thresholds and filters from RNA data in a \link[SummarizedExperiment]{SummarizedExperiment}.
#' This calls \code{\link{computeRnaQcMetrics}} on an assay in a SummarizedExperiment,
#' followed by \code{\link{suggestRnaQcThresholds}} and \code{\link{filterRnaQcMetrics}} to identify high-quality cells.
#' 
#' @param x A \link[SummarizedExperiment]{SummarizedExperiment} object or one of its subclasses.
#' Rows correspond to genes and columns correspond to cells.
#' @param subsets List of subsets of control genes, see \code{?\link{computeRnaQcMetrics}} for more details.
#' @param num.threads Number of threads, to pass to \code{\link{computeRnaQcMetrics}}.
#' @param block Block assignment for each cell, to pass to \code{\link{suggestRnaQcThresholds}} and \code{\link{filterRnaQcMetrics}}.
#' @param thresholds List containing pre-defined thresholds for each QC metric,
#' see the return value of \code{\link{suggestRnaQcThresholds}} for the expected format.
#' @param more.suggest.args Named list of additional arguments to pass to \code{\link{suggestRnaQcThresholds}}.
#' @param altexp.proportions Alternative experiments for which to compute QC metrics.
#' This is typically used to refer to alternative experiments holding spike-in data.
#' For each alternative experiment, the proportion is defined as \eqn{X/(X+Y)} where \eqn{X} is the alternative experiment's total and \eqn{Y} is the RNA total.
#' These proportions will be used for filtering in the same manner as the proportions computed from \code{subsets}.
#'
#' More specifically, \code{altexp.proportions} should be an unnamed integer or character vector containing the names/indices of the alternative experiments of interest.
#' The assay to use from each alternative experiment is determined by \code{assay.type}.
#'
#' Alternatively, \code{altexp.proportions} may be a named integer or character vector.
#' Each name specifies an alternative experiment while each value is the index/name of the assay to use from that experiment.
#' 
#' Only relevant if \code{x} is a \link[SingleCellExperiment]{SingleCellExperiment}.
#' @param assay.type Integer or string specifying the assay of \code{x} containing the RNA count matrix.
#' @param output.prefix String containing a prefix to add to the names of the \code{link[SummarizedExperiment]{colData}} columns containing the output statistics.
#' @param meta.name String containing the name of the \code{\link[S4Vectors]{metadata}} entry containing the additional outputs such as the filtering thresholds.
#' If \code{NULL}, additional outputs are not reported. 
#' @param flatten Logical scalar indicating whether to flatten the subset proportions into separate columns of the \code{link[SummarizedExperiment]{colData}}.
#' If \code{FALSE}, the subset proportions are stored in a nested \link[S4Vectors]{DataFrame}.
#' @param compute.res \link[S4Vectors]{DataFrame} returned by \code{\link{computeRnaQcMetrics}}.
#' 
#' @return
#' For \code{quickRnaQc.se}, \code{x} is returned with additional columns added to its \code{\link[SummarizedExperiment]{colData}}.
#' Each column contains per-cell values for one of the QC metrics, see \code{\link{computeRnaQcMetrics}} for details.
#' The suggested thresholds are stored as a list in \code{\link[S4Vectors]{metadata}}.
#' The \code{colData} also contains a \code{keep} column, specifying which cells are to be retained.
#' If \code{altexp.proportions} is provided, QC metrics are added to the \code{colData} of the specified alternative experiments in the output object.
#'
#' For \code{computeRnaQcMetricsWithAltExps}, a list is returned containing:
#' \itemize{
#' \item \code{main}, the result of calling \code{\link{computeRnaQcMetrics}} on the RNA count matrix in \code{x}.
#' The proportion of counts in each alternative experiment is added to the \code{subsets}.
#' \item \code{altexp}, a named list of length equal to \code{altexp.proportions}.
#' Each inner list is the result of calling \code{computeRnaQcMetrics} on the RNA count matrix of the corresponding alternative experiment of \code{x}.
#' }
#'
#' For \code{formatComputeRnaQcMetricsResult}, a \link[S4Vectors]{DataFrame} is returned containing the per-cell QC metrics.
#'
#' @author Aaron Lun
#'
#' @examples
#' library(SingleCellExperiment)
#' sce <- getTestRnaData.se()
#' sce <- quickRnaQc.se(sce, subsets=list(mito=grepl("^mt", rownames(sce))))
#' colData(sce)[,c("sum", "detected", "subset.proportion.mito")]
#' metadata(sce)$qc$thresholds
#' summary(sce$keep)
#'
#' # Computing spike-in proportions, if available.
#' sce <- getTestRnaData.se()
#' sce <- quickRnaQc.se(
#'    sce,
#'    subsets=list(mito=grepl("^mt", rownames(sce))),
#'    altexp.proportions="ERCC"
#' )
#' colData(sce)[,c("sum", "detected", "subset.proportion.mito", "subset.proportion.ERCC")]
#' colData(altExp(sce, "ERCC"))[,c("sum", "detected")]
#'
#' @export
#' @importFrom S4Vectors cbind metadata metadata<-
quickRnaQc.se <- function( 
    x,
    subsets,
    num.threads = 1,
    thresholds = NULL,
    block = NULL,
    more.suggest.args = list(),
    altexp.proportions = NULL,
    assay.type = "counts",
    output.prefix = NULL, 
    meta.name = "qc",
    flatten = TRUE
) {
    metrics <- computeRnaQcMetricsWithAltExps(
        x,
        subsets,
        altexp.proportions=altexp.proportions,
        assay.type=assay.type,
        num.threads=num.threads
    )

    if (is.null(thresholds)) {
        thresholds <- .call(
            suggestRnaQcThresholds,
            list(metrics$main),
            list(block=block),
            more.suggest.args
        )
    } else {
        names(thresholds)[names(thresholds) == "subset.proportion"] <- "subsets"
        thresholds <- .populateSubsetThresholds(thresholds, "subsets", !is.null(block))
    }

    keep <- filterRnaQcMetrics(thresholds, metrics$main, block=block)

    df <- formatComputeRnaQcMetricsResult(metrics$main, flatten=flatten)
    df$keep <- keep
    colnames(df) <- paste0(output.prefix, colnames(df))
    SummarizedExperiment::colData(x) <- cbind(SummarizedExperiment::colData(x), df)

    if (!is.null(altexp.proportions)) {
        for (ae.name in names(metrics$altexp)) {
            ae.df <- formatComputeRnaQcMetricsResult(metrics$altexp[[ae.name]], flatten=flatten)
            colnames(ae.df) <- paste0(output.prefix, colnames(ae.df))

            ae.se <- SingleCellExperiment::altExp(x, ae.name)
            ae.cd <- SummarizedExperiment::colData(ae.se) <- cbind(SummarizedExperiment::colData(ae.se), ae.df)
            SingleCellExperiment::altExp(x, ae.name) <- ae.se
        }
    }

    if (!is.null(meta.name)) {
        names(thresholds)[names(thresholds) == "subsets"] <- "subset.proportion"
        metadata(x)[[meta.name]] <- list(thresholds=thresholds)
    }

    x
}

#' @export
#' @rdname quickRnaQc.se
computeRnaQcMetricsWithAltExps <- function(x, subsets, altexp.proportions, num.threads = 1, assay.type = "counts") {
    metrics <- computeRnaQcMetrics(SummarizedExperiment::assay(x, assay.type, withDimnames=FALSE), subsets, num.threads=num.threads)

    # Adding more proportions from the alternative experiments, mostly for spike-ins.
    altexp.collected <- list()
    if (!is.null(altexp.proportions)) {
        altexp.proportions <- .sanitizeAltexpAssays(
            altexp.proportions,
            all.altexps=SingleCellExperiment::altExpNames(x),
            default.assay.type=assay.type
        )

        for (ae.name in names(altexp.proportions)) {
            alt.assay <- SummarizedExperiment::assay(SingleCellExperiment::altExp(x, ae.name), altexp.proportions[[ae.name]])
            alt.metrics <- computeRnaQcMetrics(alt.assay, subsets=list(), num.threads=num.threads)
            altexp.collected[[ae.name]] <- alt.metrics
            metrics$subsets[[ae.name]] <- alt.metrics$sum/(metrics$sum + alt.metrics$sum)
        }
    }

    list(main=metrics, altexp=altexp.collected)
}

#' @export
#' @rdname quickRnaQc.se
#' @importFrom S4Vectors DataFrame cbind
formatComputeRnaQcMetricsResult <- function(compute.res, flatten = TRUE) {
    if (flatten) {
        extra <- compute.res$subsets
        colnames(extra) <- sprintf("subset.proportion.%s", colnames(extra))
        cbind(DataFrame(sum=compute.res$sum, detected=compute.res$detected), extra)
    } else {
        colnames(compute.res)[colnames(compute.res) == "subsets"] <- "subset.proportion"
        compute.res
    }
}

# library(testthat); library(scrapper); source("test-se_quickRnaQc.R")

library(SingleCellExperiment)
mat <- matrix(rpois(1000, 1), ncol=10)
se <- SummarizedExperiment(list(counts=mat))

test_that("quickRnaQc.se works as expected", {
    out <- quickRnaQc.se(se, subsets=list())
    expect_type(out$sum, "double")
    expect_type(out$detected, "integer")
    expect_type(out$keep, "logical")
    expect_type(metadata(out)$qc$thresholds, "list")

    out <- quickRnaQc.se(se, subsets=list(mt=1:5))
    expect_type(out$subset.proportion.mt, "double")

    out2 <- quickRnaQc.se(se, subsets=list(mt=1:5), output.prefix="WHEE.")
    expect_identical(out2$WHEE.sum, out$sum)
    expect_identical(out2$WHEE.detected, out$detected)
    expect_identical(out2$WHEE.subset.proportion.mt, out$subset.proportion.mt)
    expect_identical(out2$WHEE.keep, out$keep)

    out3 <- quickRnaQc.se(se, subsets=list(mt=1:5, bar=6:10), flatten=FALSE)
    expect_identical(out$subset.proportion.mt, out3$subset.proportion$mt)
    expect_type(out3$subset.proportion$bar, "double")

    out4 <- quickRnaQc.se(se, subsets=list(), meta.name=NULL)
    expect_null(metadata(out4)$qc)
})

test_that("quickRnaQc.se works with alternative experiments", {
    sce <- as(se, "SingleCellExperiment")
    altExp(sce, "ERCC") <- se[1:20,]

    out <- quickRnaQc.se(sce, subsets=list(), altexp.proportions="ERCC")
    expect_type(out$subset.proportion.ERCC, "double")
    expect_equal(1 / (1 + out$sum / altExp(out)$sum), out$subset.proportion.ERCC)

    out <- quickRnaQc.se(sce, subsets=list(), altexp.proportions="ERCC", output.prefix="WHEE.")
    expect_type(out$WHEE.subset.proportion.ERCC, "double")
    expect_equal(1 / (1 + out$WHEE.sum / altExp(out, "ERCC")$WHEE.sum), out$WHEE.subset.proportion.ERCC)
})

test_that("quickRnaQc.se works with custom thresholds", {
    fixed <- list(sum=1, detected=1)
    out <- quickRnaQc.se(se, subsets=list(), thresholds=fixed)

    refixed <- fixed
    refixed$subset.proportion <- setNames(numeric(0), character(0))
    expect_identical(metadata(out)$qc$thresholds, refixed)
})

#' Normalize CRISPR counts in a SummarizedExperiment
#'
#' Compute (log-)normalized expression values after performing scaling normalization of an CRISPR count matrix.
#' This calls \code{\link{normalizeCounts}} on an assay of a \link[SummarizedExperiment]{SummarizedExperiment},
#' after centering the size factors with \code{\link{centerSizeFactors}}.
#'
#' @param x A \link[SummarizedExperiment]{SummarizedExperiment} object or one of its subclasses.
#' Rows correspond to CRISPR guides and columns correspond to cells.
#' @param size.factors Numeric vector of length equal to the number of columns of \code{x},
#' containing the size factor for each cell in \code{x}.
#' If \code{NULL}, this defaults to the column sums of the count matrix in \code{x}.
#' @param center Logical scalar indicating whether to center the \code{size.factors},
#' see \code{?\link{centerSizeFactors}} for more details.
#' @param block Block assignments for each cell, passed to \code{\link{centerSizeFactors}}.
#' @param mode How to center size factors in different blocks, see \code{?\link{centerSizeFactors}} for more details.
#' @param log Whether to log-transform the normalized expression values, see \code{?\link{normalizeCounts}} for more details.
#' @param pseudo.count The pseudo-count for log-transformation, see \code{?\link{normalizeCounts}} for more details.
#' @param more.norm.args Named list of additional arguments to pass to \code{\link{normalizeCounts}}.
#' @param assay.type Integer or string specifying the assay of \code{x} with the count matrix.
#' @param output.name String containing the name of the assay to store the normalized matrix.
#' @param factor.name String containing the name of the \code{\link[SummarizedExperiment]{colData}} column in which to store the size factors in the output object.
#' If \code{NULL}, the size factors are not stored. 
#' @param num.threads Integer specifying the number of threads for computing column sums.
#' Only used if \code{size.factors=NULL}.
#'
#' @return \code{x} is returned with a new assay containing the (log-)normalized matrix.
#' Size factors are also stored in the \code{\link[SummarizedExperiment]{colData}}.
#'
#' @author Aaron Lun
#'
#' @examples
#' library(SingleCellExperiment)
#' sce <- altExp(getTestCrisprData.se("qc"), "CRISPR Guide Capture")
#' sce <- normalizeCrisprCounts.se(sce, size.factors=sce$sum)
#' assayNames(sce)
#' summary(sizeFactors(sce))
#'
#' @export
#' @importFrom beachmat initializeCpp tatami.column.sums
normalizeCrisprCounts.se <- function(
    x,
    size.factors = NULL,
    center = TRUE,
    block = NULL,
    mode = "lowest",
    log = TRUE,
    pseudo.count = 1,
    more.norm.args = list(),
    assay.type = "counts",
    output.name = "logcounts",
    factor.name = "sizeFactor",
    num.threads = 1
) {
    y <- SummarizedExperiment::assay(x, assay.type)

    if (is.null(size.factors)) {
        size.factors <- tatami.column.sums(initializeCpp(y), num.threads=num.threads)
    }
    if (center) {
        size.factors <- centerSizeFactors(size.factors, block=block, mode=mode)
    }

    SummarizedExperiment::assay(x, output.name) <- .call(
        normalizeCounts,
        list(y, size.factors=size.factors),
        list(log=log, pseudo.count=pseudo.count),
        more.norm.args
    )

    if (!is.null(factor.name)) {
        SummarizedExperiment::colData(x)[[factor.name]] <- size.factors
    }

    x
}

#' MNN correction on a SingleCellExperiment
#'
#' Correct batch effects from an existing embedding with mutual nearest neighbors (MNNs),
#' by calling \code{\link{correctMnn}} on a reduced dimension entry of a \link[SingleCellExperiment]{SingleCellExperiment}.
#'
#' @param x A \link[SingleCellExperiment]{SingleCellExperiment} object or one of its subclasses.
#' Rows correspond to genomic features and columns correspond to cells.
#' @param block Block assignment for each cell, passed to \code{\link{correctMnn}}.
#' @param BNPARAM Algorithm for the nearest neighbor search, passed to \code{\link{correctMnn}}.
#' @param num.threads Number of threads, passed to \code{\link{correctMnn}}.
#' @param more.mnn.args Named list of additional arguments to pass to \code{\link{correctMnn}}.
#' @param reddim.type String or integer specifying the \code{\link[SingleCellExperiment]{reducedDim}} entry on which to perform MNN correction.
#' Alternatively, a named integer or character vector of length 1,
#' where the name specifies an alternative experiment of \code{x} and the value is the name/index of a \code{reducedDim} entry in that alternative experiment.
#' @param output.name String containing the name of the \code{\link[SingleCellExperiment]{reducedDim}} entry in which to store the corrected embedding.
#' @param delayed.transpose Logical scalar indicating whether to delay the transposition when storing coordinates in the \code{\link[SingleCellExperiment]{reducedDims}}.
#'
#' @return \code{x} is returned with the corrected embedding stored as a \code{reducedDim} entry.
#' @author Aaron Lun
#'
#' @examples
#' library(SingleCellExperiment)
#' sce <- getTestRnaData.se("pca")
#' # Treating the tissue of origin as the batch.
#' sce <- correctMnn.se(sce, sce$tissue)
#' reducedDimNames(sce)
#'
#' @export
#' @importFrom BiocNeighbors AnnoyParam
correctMnn.se <- function(
    x,
    block,
    BNPARAM = AnnoyParam(),
    num.threads = 1,
    more.mnn.args = list(),
    reddim.type = "PCA", 
    output.name = "MNN",
    delayed.transpose = FALSE
) {
    out <- .call(
        correctMnn,
        list(.getTransposedReddim(x, reddim.type)),
        list(block=block, num.threads=num.threads, BNPARAM=BNPARAM),
        more.mnn.args
    )

    .addTransposedReddim(x, output.name, out$corrected, delayed.transpose)
}

# library(testthat); library(scrapper); source("test-se_correctMnn.R")

library(SingleCellExperiment)
set.seed(8888)
sce <- SingleCellExperiment(list(counts = matrix(0, 0, 200)))
reducedDim(sce, "PCA") <- matrix(runif(1000), nrow=200)

test_that("correctMnn.se works as expected", {
    block <- rep(LETTERS[1:2], each=100)
    out <- correctMnn.se(sce, block=block)
    expect_true(is.matrix(reducedDim(out, "MNN")))

    out <- correctMnn.se(sce, block=block, delayed.transpose=TRUE)
    expect_s4_class(reducedDim(out, "MNN"), "DelayedMatrix")
})

cache <- new.env()
cache$rna <- list()
cache$adt <- list()
cache$crispr <- list()

#' Get datasets for testing
#'
#' Get single-cell datasets from the \pkg{scRNAseq} package with varying levels of processing.
#' This is primarily intended for testing other \pkg{scrapper} functions, e.g., in their Examples section.
#'
#' @param at String specifying the level of processing.
#' For \code{"start"}, no processing was performed.
#' Otherwise, the dataset is returned after quality control (\code{"qc"}),
#' normalization (\code{"norm"}),
#' feature selection (\code{"hvg"}),
#' PCA (\code{"PCA"})
#' or graph-based clustering (\code{"cluster"}).
#'
#' @return A \link[SingleCellExperiment]{SingleCellExperiment} containing a dataset at the specified level of processing.
#' 
#' @details
#' For \code{getTestRnaData}, this is a scRNA-seq dataset of the mouse brain,
#' where the main experiment contains RNA counts and the alternative experiments contain ERCC and repeat element counts.
#' This is obtained with \code{\link[scRNAseq]{fetchDataset}("zeisel-brain-2015", "2023-12-14")}.
#'
#' For \code{getTestAdtData}, this is a CITE-seq dataset of human PBMCs,
#' where the main experiment contains RNA counts and the alternative experiment contains ADT counts.
#' This is obtained with \code{\link[scRNAseq]{fetchDataset}("kotliarov-pbmc-2020", "2024-04-18")}.
#' Only the first 5000 cells are loaded for speed.
#'
#' For \code{getTestCrisprData}, this is a Perturb-seq dataset of a pancreatic beta cell line,
#' where the main experiment contains RNA counts and the alternative experiment contains CRISPR guide counts.
#' This is obtained with \code{\link[scRNAseq]{fetchDataset}("cao-pancreas-2025", "2025-10-10", "rqc")}.
#' Only the first 5000 cells are loaded for speed.
#'
#' @author Aaron Lun
#' @examples
#' getTestRnaData.se()
#' getTestAdtData.se()
#' getTestCrisprData.se()
#'
#' @seealso
#' \code{\link[scRNAseq]{fetchDataset}}, used to obtain each dataset.
#' 
#' @export
#' @name getTestData.se
getTestRnaData.se <- function(at = c("start", "qc", "norm", "hvg", "pca", "cluster")) {
    at <- match.arg(at)

    if (!("start" %in% names(cache$rna))) {
        cache$rna$start <- scRNAseq::fetchDataset("zeisel-brain-2015", "2023-12-14", realize.assays=TRUE)
    }
    sce <- cache$rna$start
    if (at == "start") {
        return(sce)
    }

    if (!("qc" %in% names(cache$rna))) {
        sce <- quickRnaQc.se(sce, subsets=list(mito=startsWith(rownames(sce), "mt-")), altexp.proportions="ERCC")
        sce <- sce[,sce$keep]
        cache$rna$qc <- sce
    }
    sce <- cache$rna$qc
    if (at == "qc") {
        return(sce)
    }

    if (!("norm" %in% names(cache$rna))) {
        sce <- normalizeRnaCounts.se(sce)
        cache$rna$norm <- sce
    }
    sce <- cache$rna$norm
    if (at == "norm") {
        return(sce)
    }

    if (!("hvg" %in% names(cache$rna))) {
        sce <- chooseRnaHvgs.se(sce, more.var.args=list(use.min.width=TRUE))
        cache$rna$hvg <- sce
    }
    sce <- cache$rna$hvg
    if (at == "hvg") {
        return(sce)
    }

    if (!("pca" %in% names(cache$rna))) {
        sce <- runPca.se(sce, features=SummarizedExperiment::rowData(sce)$hvg)
        cache$rna$pca <- sce
    }
    sce <- cache$rna$pca
    if (at == "pca") {
        return(sce)
    }

    if (!("cluster" %in% names(cache$rna))) {
        sce <- clusterGraph.se(sce)
        cache$rna$cluster <- sce
    }
    sce <- cache$rna$cluster
    if (at == "cluster") {
        return(sce)
    }
}

#' @export
#' @rdname getTestData.se
#' @importFrom methods as
getTestAdtData.se <- function(at = c("start", "qc", "norm", "hvg", "pca")) {
    at <- match.arg(at)

    if (!("start" %in% names(cache$adt))) {
        loadNamespace("Matrix")
        raw.sce <- scRNAseq::fetchDataset("kotliarov-pbmc-2020", "2024-04-18")
        raw.sce <- raw.sce[,1:5000] # Cutting it down a bit for speed.
        SummarizedExperiment::assay(raw.sce) <- as(SummarizedExperiment::assay(raw.sce), "dgCMatrix")

        raw.ae <- SingleCellExperiment::altExp(raw.sce, "ADT")
        SummarizedExperiment::assay(raw.ae) <- as(SummarizedExperiment::assay(raw.ae), "dgCMatrix")
        SingleCellExperiment::altExp(raw.sce, "ADT") <- raw.ae

        cache$adt$start <- raw.sce
    }
    sce <- cache$adt$start
    if (at == "start") {
        return(sce)
    }

    if (!("qc" %in% names(cache$adt))) {
        sce <- quickRnaQc.se(sce, subsets=list(mito=startsWith(rownames(sce), "MT-")))
        alt.se <- SingleCellExperiment::altExp(sce, "ADT")
        alt.se <- quickAdtQc.se(alt.se, subsets=list(igg=SummarizedExperiment::rowData(alt.se, "ADT")$isotype))
        SingleCellExperiment::altExp(sce, "ADT") <- alt.se
        sce <- sce[,sce$keep & alt.se$keep]
        cache$adt$qc <- sce
    }
    sce <- cache$adt$qc
    if (at == "qc") {
        return(sce)
    }

    if (!("norm" %in% names(cache$adt))) {
        sce <- normalizeRnaCounts.se(sce)
        SingleCellExperiment::altExp(sce, "ADT") <- normalizeAdtCounts.se(SingleCellExperiment::altExp(sce, "ADT"))
        cache$adt$norm <- sce
    }
    sce <- cache$adt$norm
    if (at == "norm") {
        return(sce)
    }

    if (!("hvg" %in% names(cache$adt))) {
        sce <- chooseRnaHvgs.se(sce)
        cache$adt$hvg <- sce
    }
    sce <- cache$adt$hvg
    if (at == "hvg") {
        return(sce)
    }

    if (!("pca" %in% names(cache$adt))) {
        sce <- runPca.se(sce, features=SummarizedExperiment::rowData(sce)$hvg)
        SingleCellExperiment::altExp(sce, "ADT") <- runPca.se(SingleCellExperiment::altExp(sce, "ADT"), features=NULL)
        cache$adt$pca <- sce
    }
    sce <- cache$adt$pca
    if (at == "pca") {
        return(sce)
    }
}

#' @export
#' @rdname getTestData.se
#' @importFrom methods as
getTestCrisprData.se <- function(at = c("start", "qc")) {
    at <- match.arg(at)

    if (!("start" %in% names(cache$crispr))) {
        loadNamespace("Matrix")
        raw.sce <- scRNAseq::fetchDataset("cao-pancreas-2025", "2025-10-10", "rqc")
        raw.sce <- raw.sce[,1:5000] # Cutting it down a bit for speed.
        SummarizedExperiment::assay(raw.sce) <- as(SummarizedExperiment::assay(raw.sce), "dgCMatrix")

        raw.ae <- SingleCellExperiment::altExp(raw.sce, "CRISPR Guide Capture")
        SummarizedExperiment::assay(raw.ae) <- as(SummarizedExperiment::assay(raw.ae), "dgCMatrix")
        SingleCellExperiment::altExp(raw.sce, "CRISPR Guide Capture") <- raw.ae

        cache$crispr$start <- raw.sce
    }
    sce <- cache$crispr$start
    if (at == "start") {
        return(sce)
    }

    if (!("qc" %in% names(cache$crispr))) {
        sce <- quickRnaQc.se(sce, subsets=list(mito=startsWith(rownames(sce), "MT-")))
        SingleCellExperiment::altExp(sce, "CRISPR Guide Capture") <- quickCrisprQc.se(SingleCellExperiment::altExp(sce, "CRISPR Guide Capture"))
        sce <- sce[,sce$keep & SingleCellExperiment::altExp(sce, "CRISPR Guide Capture")$keep]
        cache$crispr$qc <- sce
    }
    sce <- cache$crispr$qc
    if (at == "qc") {
        return(sce)
    }
}

#' Subsample cells based on their neighbors
#'
#' Subsample a dataset by selecting cells to represent all of their nearest neighbors.
#' The aim is to preserve the relative density of the original dataset while guaranteeing representation of low-frequency subpopulations. 
#'
#' @param x A numeric matrix where rows are dimensions and columns are cells,
#' typically containing a low-dimensional representation from, e.g., \code{\link{runPca}}.
#'
#' Alternatively, an index constructed by \code{\link[BiocNeighbors]{buildIndex}}.
#'
#' Alternatively, a list containing existing nearest-neighbor search results.
#' This should contain:
#' \itemize{
#' \item \code{index}, an integer matrix where rows are neighbors and columns are cells.
#' Each column contains 1-based indices for the nearest neighbors of the corresponding cell, ordered by increasing distance.
#' \item \code{distance}, a numeric matrix of the same dimensions as \code{index},
#' containing the distances to each of the nearest neighbors.
#' }
#' The number of neighbors should be equal to \code{num.neighbors}, otherwise a warning is raised.
#' @param num.neighbors Integer scalar specifying the number of neighbors to use.
#' Larger values result in stronger downsampling. 
#' Ignored if \code{x} contains pre-computed neighbor search results.
#' @param BNPARAM A \link[BiocNeighbors]{BiocNeighborParam} object specifying the algorithm to use.
#' Only used if \code{x} does not contain existing nearest-neighbor results.
#' @param min.remaining Integer scalar specifying the minimum number of remaining neighbors that a cell must have in order to be considered for selection.
#' This should be less than or equal to \code{num.neighbors}.
#' Larger values result in stronger downsampling. 
#' @param num.threads Integer scalar specifying the number of threads to use for the nearest-neighbor search.
#' Only used if \code{x} does not contain existing nearest-neighbor results.
#'
#' @details
#' Starting from the densest region in the high-dimensional space, we select an observation for inclusion into the subsampled dataset.
#' Every time we select an observation, we remove it and all of its nearest neighbors from the dataset.
#' We then select the next observation with the most remaining neighbors, with ties broken by density; this is repeated until there are no more observations.
#'
#' The premise is that each selected observation serves as a representative for its nearest neighbors.
#' This ensures that the subsampled points are well-distributed across the original dataset.
#' Low-frequency subpopulations will always have at least a few representatives if they are sufficiently distant from other subpopulations.
#' We also preserve the relative density of the original dataset as more representatives will be generated from high-density regions. 
#' 
#' @return Integer vector with the indices of the selected cells in the subsample.
#'
#' @author Aaron Lun
#'
#' @seealso
#' \url{https://libscran.github.io/nenesub/}, for more details on the underlying algorithm.
#'
#' @examples
#' x <- matrix(rnorm(10000), nrow=2)
#' keep <- subsampleByNeighbors(x, 10)
#' plot(x[1,], x[2,])
#' points(x[1,keep], x[2,keep], col="red")
#' legend('topright', col=c('black', 'red'), legend=c('all', 'subsample'), pch=1)
#'
#' @export
#' @importFrom BiocNeighbors findKNN AnnoyParam
subsampleByNeighbors <- function(x, num.neighbors=20, min.remaining=10, num.threads=1, BNPARAM=AnnoyParam()) {
    if (!is.list(x)) {
        x <- findKNN(x, k=num.neighbors, transposed=TRUE, get.index="transposed", get.distance="transposed", num.threads=num.threads, BNPARAM=BNPARAM)
    } else {
        .checkNeighborResults(x$index, x$distance)
    }
    subsample_by_neighbors(x$index, x$distance, min_remaining=min.remaining) 
}

# library(testthat); library(scrapper); source("test-subsampleByNeighbors.R")

set.seed(10000)
x <- matrix(runif(10000), nrow=10)

test_that("subsampleByNeighbors works as expected", {
    keep <- subsampleByNeighbors(x, num.neighbors=5, min.remaining=2)
    expect_lt(length(keep), ncol(x))
    expect_gt(min(keep), 0)
    expect_lte(max(keep), ncol(x))

    keep2 <- subsampleByNeighbors(x, num.neighbors=10, min.remaining=2)
    expect_lt(length(keep2), length(keep))
    expect_gt(min(keep2), 0)
    expect_lte(max(keep2), ncol(x))

    # Same results when given some nearest neighbor results.
    res <- BiocNeighbors::findKNN(x, k=10, transposed=TRUE, BNPARAM=BiocNeighbors::AnnoyParam(), get.index="transposed", get.distance="transposed")
    keep.alt <- subsampleByNeighbors(res, num.neighbors=10, min.remaining=2)
    expect_identical(keep2, keep.alt)

    expect_error(subsampleByNeighbors(x, num.neighbors=10, min.remaining=1000), "should not be greater")

    # Make sure that we're actually using the latest distance.
    copyres <- res
    copyres$distance[seq_len(9),] <- NA_real_ 
    keep.copy <- subsampleByNeighbors(copyres, num.neighbors=10, min.remaining=2)
    expect_identical(keep.copy, keep.alt)
})

#' Score marker genes
#'
#' Score marker genes for each group using a variety of effect sizes from pairwise comparisons between groups.
#' This includes Cohen's d, the area under the curve (AUC), the difference in the means (delta-mean) and the difference in the proportion of detected cells (delta-detected).
#' For each group, the strongest markers are those genes with the largest effect sizes (i.e., upregulated) when compared to all other groups.
#'
#' @param x A matrix-like object where rows correspond to genes or genomic features and columns correspond to cells.
#' It is typically expected to contain log-expression values, e.g., from \code{\link{normalizeCounts}}.
#' @param groups A vector specifying the group assignment for each cell in \code{x}.
#' @param block Factor specifying the block of origin (e.g., batch, sample) for each cell in \code{x}.
#' If provided, comparisons are performed within each block to ensure that block effects do not confound the estimates.
#' The weighted average of the effect sizes across all blocks is reported for each gene.
#' Alternatively \code{NULL}, if all cells are from the same block.
#' @param block.average.policy String specifying the policy to use for average statistics across blocks.
#' This can either be a (weighted) \code{"mean"} or a \code{"quantile"}.
#' Only used if \code{block} is not \code{NULL}. 
#' @param block.weight.policy String specifying the policy to use for weighting different blocks when computing the average for each statistic.
#' See the argument of the same name in \code{\link{computeBlockWeights}} for more detail.
#' Only used if \code{block} is not \code{NULL}.
#' @param variable.block.weight Numeric vector of length 2, specifying the parameters for variable block weighting.
#' See the argument of the same name in \code{\link{computeBlockWeights}} for more detail.
#' Only used if \code{block} is not \code{NULL} and \code{block.weight.policy = "variable"}.
#' @param block.quantile Number specifying the probability of the quantile of statistics across blocks. 
#' Defaults to 0.5, i.e., the median of per-block statistics.
#' Only used if \code{block} is not \code{NULL} and \code{block.average.policy="quantile"}.
#' @param threshold Non-negative numeric scalar specifying the minimum threshold on the differences in means (i.e., the log-fold change, if \code{x} contains log-expression values). 
#' This is incorporated into the effect sizes for Cohen's d and the AUC.
#' Larger thresholds will favor genes with large differences at the expense of genes with low variance that would otherwise have comparable effect sizes.
#' @param compute.group.mean Logical scalar indicating whether to compute the group-wise mean expression for each gene.
#' @param compute.group.detected Logical scalar indicating whether to compute the group-wise proportion of detected cells for each gene.
#' @param compute.delta.mean Logical scalar indicating whether to compute the delta-means, i.e., the log-fold change when \code{x} contains log-expression values.
#' @param compute.delta.detected Logical scalar indicating whether to compute the delta-detected, i.e., differences in the proportion of cells with detected expression.
#' @param compute.cohens.d Logical scalar indicating whether to compute Cohen's d.
#' @param compute.auc Logical scalar indicating whether to compute the AUC.
#' Setting this to \code{FALSE} can improve speed and memory efficiency.
#' @param compute.summary.min Boolean specifying whether to compute the minimum as a summary statistic for each effect size.
#' Only used if \code{all.pairwise=FALSE}.
#' @param compute.summary.mean Boolean specifying whether to compute the mean as a summary statistic for each effect size.
#' Only used if \code{all.pairwise=FALSE}.
#' @param compute.summary.median Boolean specifying whether to compute the median as a summary statistic for each effect size.
#' Only used if \code{all.pairwise=FALSE}.
#' @param compute.summary.max Boolean specifying whether to compute the maximum as a summary statistic for each effect size.
#' Only used if \code{all.pairwise=FALSE}.
#' @param compute.summary.quantiles Numeric scalars containing the probabilities of quantiles to compute as summary statistics for each effect size.
#' If \code{NULL}, no quantiles are computed.
#' Only used if \code{all.pairwise=FALSE}.
#' @param compute.summary.min.rank Boolean specifying whether to compute the mininum rank as a summary statistic for each effect size.
#' If \code{NULL}, no quantiles are computed.
#' Only used if \code{all.pairwise=FALSE}.
#' @param all.pairwise Logical scalar indicating whether to report the effect sizes for every pairwise comparison between groups.
#' Alternatively, an integer scalar indicating the number of top markers to report from each pairwise comparison between groups.
#' If \code{FALSE}, only the summary statistics are reported.
#' @param top.index.only Boolean indicating whether to only report the indices of the top genes when \code{all.pairwise} is an integer.
#' This is more efficient when the effect sizes of the top genes are not required.
#' Ignored for all other values of \code{all.pairwise}.
#' @param min.rank.limit Integer scalar specifying the maximum value of the min-rank to report.
#' Lower values improve memory efficiency at the cost of discarding information about lower-ranked genes.
#' Only used if \code{all.pairwise=FALSE}.
#' @param num.threads Integer scalar specifying the number of threads to use.
#'
#' @return
#' A named list containing:
#' \itemize{
#' \item \code{nrow}, integer specifying the number of rows in \code{x}.
#' \item \code{row.names}, character vector or \code{NULL} containing the row names of \code{x}.
#' \item \code{group.ids}, vector contaning the identities of the unique groups.
#' \item \code{mean}, a numeric matrix containing the mean expression for each group.
#' Each row is a gene and each column is a group in \code{group.ids}.
#' Omitted if \code{compute.group.mean=FALSE}.
#' \item \code{detected}, a numeric matrix containing the proportion of detected cells in each group.
#' Each row is a gene and each column is a group in \code{group.ids}.
#' Omitted if \code{compute.group.detected=FALSE}.
#' }
#' 
#' If \code{all.pairwise=FALSE}, the list also contains:
#' \itemize{
#' \item \code{cohens.d}, a list of \link[S4Vectors]{DataFrame}s where each DataFrame corresponds to a group in \code{group.ids}.
#' Each row of a DataFrame represents a gene, while each column contains a summary of Cohen's d from pairwise comparisons to all other groups.
#' This includes \code{min}, \code{mean}, \code{median}, \code{max}, \code{quantile.*} and \code{min.rank} - check out \code{?\link{summarizeEffects}} for details.
#' Omitted if \code{compute.cohens.d=FALSE}.
#' \item \code{auc}, a list like \code{cohens.d} but containing the summaries of the AUCs from each pairwise comparison.
#' Omitted if \code{compute.auc=FALSE}.
#' \item \code{delta.mean}, a list like \code{cohens.d} but containing the summaries of the delta-mean from each pairwise comparison.
#' Omitted if \code{compute.delta.mean=FALSE}.
#' \item \code{delta.detected}, a list like \code{cohens.d} but containing the summaries of the delta-detected from each pairwise comparison.
#' Omitted if \code{compute.delta.detected=FALSE}.
#' }
#'
#' If \code{all.pairwise=TRUE}, the list also contains:
#' \itemize{
#' \item \code{cohens.d}, a 3-dimensional numeric array containing the Cohen's d from each pairwise comparison between groups.
#' The extents of the first two dimensions are equal to the number of groups in \code{group.ids}, while the extent of the final dimension is equal to the number of genes.
#' The entry \code{cohens.d[i, j, k]} represents Cohen's d from the comparison of group \code{group.ids[j]} over group \code{group.ids[i]} for gene \code{k}.
#' Omitted if \code{compute.cohens.d=FALSE}.
#' \item \code{auc}, an array like \code{cohens.d} but containing the AUCs from each pairwise comparison.
#' Omitted if \code{compute.auc=FALSE}.
#' \item \code{delta.mean}, an array like \code{cohens.d} but containing the delta-mean from each pairwise comparison.
#' Omitted if \code{compute.delta.mean=FALSE}.
#' \item \code{delta.detected}, an array like \code{cohens.d} but containing the delta-detected from each pairwise comparison.
#' Omitted if \code{compute.delta.detected=FALSE}.
#' }
#'
#' If \code{all.pairwise} is an integer and \code{top.index.only=FALSE}, the list also contains:
#' \itemize{
#' \item \code{cohens.d}, a list of list of \link[S4Vectors]{DataFrame}s containing the top genes with the largest Cohen's d for each pairwise comparison.
#' Specifically, \code{cohens.d[[i]][[j]]} is a DataFrame that contains the top \code{all.pairwise} genes from the comparison of group \code{group.ids[i]} over group \code{group.ids[j]}. 
#' Each DataFrame contains an \code{index} column, the row index of the gene; and an \code{effect} column, the Cohen's d for that gene.
#' Omitted if \code{compute.cohens.d=FALSE}.
#' \item \code{auc}, a list of list of DataFrames like \code{cohens.d} but containing the AUCs from each pairwise comparison.
#' Omitted if \code{compute.auc=FALSE}.
#' \item \code{delta.mean}, a list of list of DataFrames like \code{cohens.d} but containing the delta-mean from each pairwise comparison.
#' Omitted if \code{compute.delta.mean=FALSE}.
#' \item \code{delta.detected}, a list of list of DataFrames like \code{cohens.d} but containing the delta-detected from each pairwise comparison.
#' Omitted if \code{compute.delta.detected=FALSE}.
#' }
#'
#' If \code{all.pairwise} is an integer and \code{top.index.only=TRUE}, the list also contains:
#' \itemize{
#' \item \code{cohens.d}, a list of list of integer vectors containing the row indices of the top genes with the largest Cohen's d for each pairwise comparison.
#' Specifically, \code{cohens.d[[i]][[j]]} is a vector that contains the top \code{all.pairwise} genes from the comparison of group \code{group.ids[i]} over group \code{group.ids[j]}. 
#' Omitted if \code{compute.cohens.d=FALSE}.
#' \item \code{auc}, a list of list of DataFrames like \code{cohens.d} but containing the AUCs from each pairwise comparison.
#' Omitted if \code{compute.auc=FALSE}.
#' \item \code{delta.mean}, a list of list of DataFrames like \code{cohens.d} but containing the delta-mean from each pairwise comparison.
#' Omitted if \code{compute.delta.mean=FALSE}.
#' \item \code{delta.detected}, a list of list of DataFrames like \code{cohens.d} but containing the delta-detected from each pairwise comparison.
#' Omitted if \code{compute.delta.detected=FALSE}.
#' }
#'
#' @section Choice of effect size:
#' The delta-mean is the difference in the mean expression between groups.
#' This is fairly straightforward to interpret - a positive delta-mean corresponds to increased expression in the first group compared to the second. 
#' The delta-mean can also be treated as the log-fold change if the input matrix contains log-transformed normalized expression values.
#' 
#' The delta-detected is the difference in the proportion of cells with detected expression between groups.
#' This lies between 1 and -1, with the extremes occurring when a gene is silent in one group and detected in all cells of the other group.
#' For this interpretation, we assume that the input matrix contains non-negative expression values, where a value of zero corresponds to lack of detectable expression.
#' 
#' Cohen's d is the standardized difference between two groups.
#' This is defined as the difference in the mean for each group scaled by the average standard deviation across the two groups.
#' (Technically, we should use the pooled variance; however, this introduces some unintuitive asymmetry depending on the variance of the larger group, so we take a simple average instead.)
#' A positive value indicates that the gene has increased expression in the first group compared to the second.
#' Cohen's d is analogous to the t-statistic in a two-sample t-test and avoids spuriously large effect sizes from comparisons between highly variable groups.
#' We can also interpret Cohen's d as the number of standard deviations between the two group means.
#' 
#' The area under the curve (AUC) is the probability that a randomly chosen observation in one group is greater than a randomly chosen observation in the other group. 
#' Values greater than 0.5 indicate that a gene is upregulated in the first group.
#' The AUC is closely related to the U-statistic used in the Wilcoxon rank sum test. 
#' The key difference between the AUC and Cohen's d is that the former is less sensitive to the variance within each group, e.g.,
#' if two distributions exhibit no overlap, the AUC is the same regardless of the variance of each distribution. 
#' This may or may not be desirable as it improves robustness to outliers but reduces the information available to obtain a fine-grained ranking. 
#' 
#' @section With a minimum change threshold:
#' Setting a minimum change threshold (i.e., \code{threshold}) prioritizes genes with large shifts in expression instead of those with low variances.
#' Currently, only positive thresholds are supported, which focuses on genes that are upregulated in the first group compared to the second.
#' The effect size definitions are generalized when testing against a non-zero threshold:
#' \itemize{
#' \item Cohen's d is redefined as the standardized difference between the difference in means and the specified threshold,
#' analogous to the TREAT method from the \pkg{limma} package.
#' Large positive values are only obtained when the observed difference in means is significantly greater than the threshold.
#' For example, if we had a threshold of 2 and we obtained a Cohen's d of 3, this means that the observed difference in means was 3 standard deviations greater than 2.
#' Note that a negative Cohen's d cannot be intepreted as downregulation, as the difference in means may still be positive but less than the threshold.
#' \item The AUC is generalized to the probability of obtaining a random observation in one group that is greater than a random observation plus the threshold in the other group.
#' For example, if we had a threshold of 2 and we obtained an AUC of 0.8, this means that, 80% of the time,
#' the random observation from the first group would be greater than a random observation from the second group by 2 or more.
#' Again, AUCs below 0.5 cannot be interpreted as downregulation, as it may be caused by a positive shift that is less than the threshold.
#' }
#' 
#' @examples
#' # Mocking a matrix:
#' library(Matrix)
#' x <- round(abs(rsparsematrix(1000, 100, 0.1) * 100))
#' normed <- normalizeCounts(x, size.factors=centerSizeFactors(colSums(x)))
#'
#' # Compute marker summaries for each group:
#' g <- sample(letters[1:4], ncol(x), replace=TRUE)
#' scores <- scoreMarkers(normed, g)
#' names(scores)
#' head(scores$mean)
#' head(scores$cohens.d[["a"]])
#'
#' # Report marker statistics for a single group:
#' reportGroupMarkerStatistics(scores, "b")
#'
#' @seealso
#' The \code{score_markers_summary}, \code{score_markers_pairwise} and \code{score_markers_best} functions in \url{https://libscran.github.io/scran_markers/}.
#' See their blocked equivalents (e.g., \code{score_markers_summary_blocked}) when \code{block} is specified.
#'
#' \code{\link{summarizeEffects}}, to summarize the pairwise effects returned when \code{all.pairwise=TRUE}.
#'
#' \code{\link{reportGroupMarkerStatistics}}, to consolidate the statistics for a single group into its own data frame.
#'
#' \code{\link{scoreMarkers.se}}, to score markers from a \link[SummarizedExperiment]{SummarizedExperiment}.
#'
#' @export
#' @importFrom beachmat initializeCpp
#' @importFrom S4Vectors DataFrame
scoreMarkers <- function(
    x, 
    groups, 
    block=NULL, 
    block.average.policy=c("mean", "quantile"),
    block.weight.policy=c("variable", "equal", "none"),
    variable.block.weight=c(0, 1000),
    block.quantile=0.5,
    compute.group.mean=TRUE,
    compute.group.detected=TRUE,
    compute.delta.mean=TRUE,
    compute.delta.detected=TRUE,
    compute.cohens.d=TRUE,
    compute.auc=TRUE,
    compute.summary.min=TRUE,
    compute.summary.mean=TRUE,
    compute.summary.median=TRUE,
    compute.summary.max=TRUE,
    compute.summary.quantiles=NULL,
    compute.summary.min.rank=TRUE,
    threshold=0, 
    all.pairwise=FALSE, 
    top.index.only=FALSE,
    min.rank.limit=500,
    num.threads=1
) {
    .checkSEX(x, "scoreMarkers.se")

    rn <- rownames(x)
    ngenes <- nrow(x)
    x <- initializeCpp(x, .check.na=FALSE)

    groups <- .transformFactor(groups)
    if (is.null(groups$index)) {
        stop("'groups' should be non-NULL")
    }
    block <- .transformFactor(block)

    args <- list(
        groups=groups$index,
        num_groups=length(groups$names),
        block=block$index,
        block_average_policy=match.arg(block.average.policy),
        block_weight_policy=match.arg(block.weight.policy),
        variable_block_weight=variable.block.weight,
        block_quantile=block.quantile,
        threshold=threshold,
        compute_group_mean=compute.group.mean,
        compute_group_detected=compute.group.detected,
        compute_cohens_d=compute.cohens.d,
        compute_delta_mean=compute.delta.mean,
        compute_delta_detected=compute.delta.detected,
        compute_auc=compute.auc,
        num_threads=num.threads
    )

    if (isTRUE(all.pairwise)) {
        output <- do.call(score_markers_pairwise, c(list(x), args))
        for (nm in setdiff(names(output), c("mean", "detected"))) {
            current <- output[[nm]]
            dimnames(current) <- list(groups$names, groups$names, rn)
            output[[nm]] <- current
        }

    } else if (isFALSE(all.pairwise)) {
        if (!is.null(compute.summary.quantiles)) {
            compute.summary.quantiles <- sort(unique(compute.summary.quantiles))
        }

        output <- do.call(
            score_markers_summary,
            c(
                list(
                    x,
                    min_rank_limit=min.rank.limit,
                    compute_summary_mean=compute.summary.mean,
                    compute_summary_min=compute.summary.min,
                    compute_summary_median=compute.summary.median,
                    compute_summary_max=compute.summary.max,
                    compute_summary_quantiles=compute.summary.quantiles,
                    compute_summary_min_rank=compute.summary.min.rank
                ),
                args
            )
        )

        for (nm in setdiff(names(output), c("mean", "detected"))) {
            current <- output[[nm]]
            names(current) <- groups$names
            output[[nm]] <- format_summary_output(current, ngenes, rn, compute.summary.quantiles)
        }

    } else {
        output <- do.call(score_markers_best, c(list(x, top=all.pairwise, index_only=top.index.only), args))
        for (nm in setdiff(names(output), c("mean", "detected"))) {
            current <- output[[nm]]
            names(current) <- groups$names
            for (i in seq_along(current)) {
                names(current[[i]]) <- groups$names
            }
            output[[nm]] <- current
        }
    }

    output$nrow <- ngenes
    output$row.names <- rn
    output$group.ids <- groups$names

    if (compute.group.mean) {
        dimnames(output$mean) <- list(rn, groups$names)
    }
    if (compute.group.detected) {
        dimnames(output$detected) <- list(rn, groups$names)
    }

    output
}

#' Report marker statistics for a single group
#'
#' Combine all marker statistics for a single group into a data frame for easy inspection.
#' Users can pick one of the columns for sorting potential marker genes. 
#'
#' @param results Named list of marker statistics, typically generated by \code{\link{scoreMarkers}} with \code{all.pairwise=FALSE}.
#' @param group String or integer scalar specifying the group of interest.
#' @param effect.sizes Character vector specifying the effect sizes of interest.
#' If \code{NULL}, all effect sizes are reported in the returned data frame.
#' @param summaries Character vector specifying the summary statistics of interest.
#' If \code{NULL}, all summaries are reported in the returned data frame.
#' @param include.mean Logical scalar indicating whether the mean expression should be reported in the returned data frame.
#' @param include.detected Logical scalar indicating whether the proportion of detected cells should be reported in the returned data frame.
#'
#' @return Data frame where each row corresponds to a gene.
#' Each column contains the requested statistics for \code{group}.
#' Effect size summary columns are named as \code{<EFFECT>.<SUMMARY>}.
#'
#' @author Aaron Lun
#'
#' @seealso
#' \code{\link{scoreMarkers}}, to generate \code{results}.
#'
#' \code{\link{summarizeEffects}}, for the trade-offs between effect size summaries.
#'
#' @export
reportGroupMarkerStatistics <- function(
    results,
    group,
    effect.sizes = NULL,
    summaries = NULL,
    include.mean = TRUE,
    include.detected = TRUE)
{
    if (is.null(effect.sizes)) {
        effect.sizes <- c("cohens.d", "auc", "delta.mean", "delta.detected")
    }
    if (is.null(summaries)) {
        summaries <- c("min", "mean", "median", "max", "min.rank")
    }

    current <- data.frame(row.names=rownames(results$mean), matrix(0L, nrow(results$mean), 0L))

    if (include.mean) {
        current$mean <- results$mean[,group]
    }
    if (include.detected) {
        current$detected <- results$detected[,group]
    }

    for (eff in effect.sizes) {
        eff.all <- results[[eff]]
        if (is.null(eff.all)) {
            next
        }
        eff.df <- eff.all[[group]]
        for (summ in summaries) {
            current[[paste0(eff, ".", summ)]] <- eff.df[[summ]]
        }
    }

    current
}

# library(testthat); library(scrapper); source("test-scoreMarkers.R")

set.seed(999)
library(Matrix)
x <- abs(rsparsematrix(1000, 100, 0.1))
x@x <- jitter(x@x)

check_summaries <- function(summary) {
    for (x in names(summary)) {
        current <- summary[[x]]
        expect_true(all(current$min <= current$max))
        expect_true(all(current$min <= current$mean + 1e-8)) # add some tolerance for numerical imprecision when averaging identical effects.
        expect_true(all(current$min <= current$median))
        expect_true(all(current$mean <= current$max + 1e-8))
        expect_true(all(current$median <= current$max))
        expect_true(all(current$min_rank <= 1000))
        expect_true(all(current$min_rank > 0))
    }
}

test_that("scoreMarkers works as expected for simple cases", {
    g <- sample(letters[1:4], ncol(x), replace=TRUE)
    out <- scoreMarkers(x, g)

    expect_equal(out$mean[,"a"], Matrix::rowMeans(x[,g=="a"]))
    expect_equal(out$detected[,"d"], Matrix::rowMeans(x[,g=="d"] > 0))

    check_summaries(out$cohens.d)
    check_summaries(out$auc)
    check_summaries(out$delta.mean)
    check_summaries(out$delta.detected)

    for (n in names(out$auc)) {
        collected <- unlist(out$auc[[n]][c("min", "mean", "median", "max")])
        expect_true(all(collected >= 0))
        expect_true(all(collected <= 1))
    }

    # Same results for multiple threads.
    pout <- scoreMarkers(x, g, num.threads=2)
    expect_identical(out, pout)

    # Works without the AUC and the groupwise stats.
    aout <- scoreMarkers(x, g, compute.auc=FALSE, compute.group.mean=FALSE, compute.group.detected=FALSE)
    expect_false("auc" %in% names(aout))
    expect_false("mean" %in% names(aout))
    expect_false("detected" %in% names(aout))
    expect_equal(out$delta.detected, aout$delta.detected)

    # Works without anything.
    empty <- scoreMarkers(x, g, compute.auc=FALSE, compute.cohens.d=FALSE, compute.delta.detected=FALSE, compute.delta.mean=FALSE)
    expect_identical(names(empty), c("mean", "detected", "nrow", "group.ids"))

    expect_error(scoreMarkers(SummarizedExperiment::SummarizedExperiment(x), g), "not supported")
    expect_error(scoreMarkers(x, NULL), "non-NULL")
    expect_error(scoreMarkers(x, 1), "length equal")
})

test_that("scoreMarkers works with blocking", {
    g <- sample(letters[1:4], ncol(x), replace=TRUE)
    b <- rep(1:3, length.out=ncol(x))
    out <- scoreMarkers(x, g, block=b, block.weight.policy="equal")

    bkeep <- g == "b"
    expect_equal(out$mean[,"b"], (Matrix::rowMeans(x[,bkeep & b == 1]) + Matrix::rowMeans(x[,bkeep & b == 2]) + Matrix::rowMeans(x[,bkeep & b == 3]))/3)
    ckeep <- g == "c"
    expect_equal(out$mean[,"c"], (Matrix::rowMeans(x[,ckeep & b == 1]) + Matrix::rowMeans(x[,ckeep & b == 2]) + Matrix::rowMeans(x[,ckeep & b == 3]))/3)

    check_summaries(out$cohens.d)
    check_summaries(out$auc)
    check_summaries(out$delta.mean)
    check_summaries(out$delta.detected)

    for (x in names(out$auc)) {
        collected <- unlist(out$auc[[x]][c("min", "mean", "median", "max")])
        expect_true(all(collected >= 0))
        expect_true(all(collected <= 1))
    }
})

test_that("scoreMarkers works with quantile summaries", {
    g <- sample(4, ncol(x), replace=TRUE)
    res <- scoreMarkers(x, g, compute.summary.quantiles=c(0, 0.5, 1))
    for (i in seq_len(max(g))) {
        for (e in c("cohens.d", "auc", "delta.mean", "delta.detected")) {
            current <- res[[e]][[i]]
            expect_identical(current$min, current$quantile[["0"]])
            expect_identical(current$median, current$quantile[["0.5"]])
            expect_identical(current$max, current$quantile[["1"]])
        }
    }
})

test_that("scoreMarkers works with all summaries disabled", {
    g <- sample(4, ncol(x), replace=TRUE)
    res <- scoreMarkers(
        x, 
        g,
        compute.group.mean=FALSE,
        compute.group.detected=FALSE,
        compute.summary.min=FALSE,
        compute.summary.mean=FALSE,
        compute.summary.median=FALSE,
        compute.summary.max=FALSE,
        compute.summary.min.rank=FALSE
    )

    for (i in seq_len(max(g))) {
        for (e in c("cohens.d", "auc", "delta.mean", "delta.detected")) {
            expect_identical(ncol(res[[e]][[i]]), 0L)
            expect_identical(nrow(res[[e]][[i]]), nrow(x))
        }
    }
})

test_that("scoreMarkers works as expected for the full pairwise statistics", {
    g <- sample(4, ncol(x), replace=TRUE)
    full <- scoreMarkers(x, g, all.pairwise=TRUE)

    # Checking that we set the dimensions correctly.
    for (g1 in seq_len(4)) {
        for (g2 in seq_len(g1)) {
            if (g1 == g2) {
                expect_true(all(full$delta.mean[g1, g2, ] == 0))
                expect_true(all(full$auc[g1, g2, ] == 0))
            } else {
                expect_equal(full$delta.mean[g1, g2, ], -full$delta.mean[g2, g1, ])
                expect_equal(full$auc[g1, g2, ], 1 - full$auc[g2, g1, ])
            }
        }
    }

    expect_true(all(full$auc >= 0))
    expect_true(all(full$auc <= 1))

    # Works without AUCs and the groupwise stats.
    aout <- scoreMarkers(x, g, all.pairwise=TRUE, compute.auc=FALSE, compute.group.mean=FALSE, compute.group.detected=FALSE)
    expect_false("auc" %in% names(aout))
    expect_false("mean" %in% names(aout))
    expect_false("detected" %in% names(aout))
    expect_equal(full$cohens.d, aout$cohens.d)

    # Works without anything.
    empty <- scoreMarkers(x, g, compute.auc=FALSE, compute.cohens.d=FALSE, compute.delta.detected=FALSE, compute.delta.mean=FALSE)
    expect_identical(names(empty), c("mean", "detected", "nrow", "group.ids"))

    # Works with blocking.
    b <- rep(1:3, length.out=ncol(x))
    bout <- scoreMarkers(x, g, block=b, block.weight.policy="equal", all.pairwise=TRUE)
    sbout <- scoreMarkers(x, g, block=b, block.weight.policy="equal")
    expect_identical(bout$mean, sbout$mean)
})

test_that("scoreMarkers works as expected for the best pairwise statistics", {
    convert_to_best <- function(pairwise, n, bound) {
        all.labels <- dimnames(pairwise)[[1]]
        all.genes <- dimnames(pairwise)[[3]]
        output <- vector("list", length(all.labels))
        names(output) <- all.labels
        for (g1 in all.labels) {
            current <- vector("list", length(all.labels))
            names(current) <- all.labels
            for (g2 in all.labels) {
                if (g1 == g2) {
                    next
                }
                stats <- pairwise[g2, g1,] # remember, second dimension is the first group in the comparison.
                keep <- which(stats > bound)
                o <- order(stats[keep], decreasing=TRUE)
                indices <- keep[head(o, n)]
                current[[g2]] <- S4Vectors::DataFrame(index=indices, effect=stats[indices])
            }
            output[[g1]] <- current
        }
        output
    }

    g <- sample(4, ncol(x), replace=TRUE)
    full <- scoreMarkers(x, g, all.pairwise=TRUE)
    best <- scoreMarkers(x, g, all.pairwise=10)

    expect_equal(full$mean, best$mean)
    expect_equal(full$detected, best$detected)
    expect_identical(convert_to_best(full$cohens.d, 10, 0), best$cohens.d)
    expect_identical(convert_to_best(full$auc, 10, 0.5), best$auc)
    expect_identical(convert_to_best(full$delta.mean, 10, 0), best$delta.mean)
    expect_identical(convert_to_best(full$delta.detected, 10, 0), best$delta.detected)

    # Works without AUCs and the groupwise means.
    aout <- scoreMarkers(x, g, all.pairwise=10, compute.auc=FALSE, compute.group.mean=FALSE, compute.group.detected=FALSE)
    expect_false("auc" %in% names(aout))
    expect_false("mean" %in% names(aout))
    expect_false("detected" %in% names(aout))
    expect_equal(best$cohens.d, aout$cohens.d)

    # Same results if we only want the top indices.
    best.index <- scoreMarkers(x, g, all.pairwise=10, compute.auc=FALSE, compute.group=FALSE, compute.group.detected=FALSE, top.index.only=TRUE)
    expect_identical(names(best.index$cohens.d), names(best$cohens.d))
    for (n in names(best$cohens.d)) {
        expect_identical(names(best.index$cohens.d[[n]]), names(best$cohens.d[[n]]))
        for (n2 in names(best$cohens.d[[n]])) {
            expect_identical(best.index$cohens.d[[n]][[n2]], best$cohens.d[[n]][[n2]]$index)
        }
    }

    # Works with blocking.
    b <- rep(1:3, length.out=ncol(x))
    bout <- scoreMarkers(x, g, block=b, block.weight.policy="equal", all.pairwise=10)
    sbout <- scoreMarkers(x, g, block=b, block.weight.policy="equal")
    expect_identical(bout$mean, sbout$mean)
})

test_that("reportGroupMarkerStatistics works as expected", {
    g <- sample(letters[1:4], ncol(x), replace=TRUE)
    out <- scoreMarkers(x, g)

    df <- reportGroupMarkerStatistics(out, "b")
    expect_identical(length(df$mean), nrow(x))
    expect_identical(length(df$detected), nrow(x))
    expect_identical(length(df$cohens.d.min), nrow(x))
    expect_identical(length(df$auc.median), nrow(x))
    expect_identical(length(df$delta.mean.min.rank), nrow(x))
    expect_identical(length(df$delta.detected.max), nrow(x))

    df <- reportGroupMarkerStatistics(out, 3, include.mean=FALSE, include.detected=FALSE, effect.sizes=character(0), summaries=character(0))
    expect_identical(nrow(df), nrow(x))
    expect_identical(ncol(df), 0L)

    # Works with row names and missing statistics.
    rownames(x) <- sprintf("GENE_%i", seq_len(nrow(x)))
    out <- scoreMarkers(x, g, compute.auc=FALSE)
    df <- reportGroupMarkerStatistics(out, 1)
    expect_identical(rownames(df), rownames(x))
    expect_null(df$auc.median)
})

#' Quick quality control for ADT data in a SummarizedExperiment
#'
#' Quickly compute quality control (QC) metrics, thresholds and filters from ADT data in a \link[SummarizedExperiment]{SummarizedExperiment}.
#' This calls \code{\link{computeAdtQcMetrics}} on an assay in a SummarizedExperiment,
#' followed by \code{\link{suggestAdtQcThresholds}} and \code{\link{filterAdtQcMetrics}} to identify high-quality cells.
#' 
#' @param x A \link[SummarizedExperiment]{SummarizedExperiment} object or one of its subclasses.
#' Rows correspond to antibody-derived tags (ADTs) and columns correspond to cells.
#' @param subsets List of subsets of control tags, see \code{?\link{computeAdtQcMetrics}} for more details.
#' @param num.threads Number of threads, to pass to \code{\link{computeAdtQcMetrics}}.
#' @param block Block assignment for each cell, to pass to \code{\link{suggestAdtQcThresholds}} and \code{\link{filterAdtQcMetrics}}.
#' @param thresholds List containing pre-defined thresholds for each QC metric,
#' see the return value of \code{\link{suggestAdtQcThresholds}} for the expected format.
#' @param more.suggest.args Named list of additional arguments to pass to \code{\link{suggestAdtQcThresholds}}.
#' @param assay.type Integer or string specifying the assay of \code{x} containing the ADT count matrix.
#' @param output.prefix String containing a prefix to add to the names of the \code{link[SummarizedExperiment]{colData}} columns containing the output statistics.
#' @param meta.name String containing the name of the \code{\link[S4Vectors]{metadata}} entry containing additional outputs like the filtering thresholds.
#' If \code{NULL}, additional outputs are not reported.
#' @param flatten Logical scalar indicating whether to flatten the subset proportions into separate columns of the \code{link[SummarizedExperiment]{colData}}.
#' If \code{FALSE}, the subset proportions are stored in a nested \link[S4Vectors]{DataFrame}.
#' @param compute.res \link[S4Vectors]{DataFrame} returned by \code{\link{computeAdtQcMetrics}}.
#' 
#' @return
#' For \code{quickAdtQc.se}, \code{x} is returned with additional columns added to its \code{\link[SummarizedExperiment]{colData}}.
#' Each column contains per-cell values for one of the QC metrics, see \code{\link{computeAdtQcMetrics}} for details.
#' The suggested thresholds are stored as a list in \code{\link[S4Vectors]{metadata}}.
#' The \code{colData} also contains a \code{keep} column, specifying which cells are to be retained.
#'
#' For \code{formatComputeAdtQcMetricsResult}, a \link[S4Vectors]{DataFrame} is returned with the per-cell QC metrics.
#'
#' @author Aaron Lun
#'
#' @examples
#' library(SingleCellExperiment)
#' sce <- altExp(getTestAdtData.se(), "ADT")
#' sce <- quickAdtQc.se(sce, subsets=list(igg=grepl("IgG", rownames(sce))))
#' colData(sce)[,c("sum", "detected", "subset.sum.igg")]
#' metadata(sce)$qc$thresholds
#' summary(sce$keep)
#' 
#' @export
#' @importFrom S4Vectors cbind metadata metadata<-
quickAdtQc.se <- function( 
    x,
    subsets,
    num.threads = 1,
    thresholds = NULL,
    block = NULL,
    more.suggest.args = list(),
    assay.type = "counts",
    output.prefix = NULL, 
    meta.name = "qc",
    flatten = TRUE
) {
    metrics <- computeAdtQcMetrics(SummarizedExperiment::assay(x, assay.type, withDimnames=FALSE), subsets, num.threads=num.threads)

    if (is.null(thresholds)) {
        thresholds <- .call(
            suggestAdtQcThresholds,
            list(metrics=metrics),
            list(block=block),
            more.suggest.args
        )
    } else {
        names(thresholds)[names(thresholds) == "subset.sum"] <- "subsets"
        thresholds <- .populateSubsetThresholds(thresholds, "subsets", !is.null(block))
    }

    keep <- filterAdtQcMetrics(thresholds, metrics, block=block)

    df <- formatComputeAdtQcMetricsResult(metrics, flatten=flatten)
    df$keep <- keep
    colnames(df) <- paste0(output.prefix, colnames(df))
    SummarizedExperiment::colData(x) <- cbind(SummarizedExperiment::colData(x), df)

    if (!is.null(meta.name)) {
        names(thresholds)[names(thresholds) == "subsets"] <- "subset.sum"
        metadata(x)[[meta.name]] <- list(thresholds=thresholds)
    }

    x
}

#' @export
#' @rdname quickAdtQc.se
#' @importFrom S4Vectors DataFrame cbind
formatComputeAdtQcMetricsResult <- function(compute.res, flatten = TRUE) {
    if (flatten) {
        extra <- compute.res$subsets
        colnames(extra) <- sprintf("subset.sum.%s", colnames(extra))
        cbind(DataFrame(sum=compute.res$sum, detected=compute.res$detected), extra)
    } else {
        colnames(compute.res)[colnames(compute.res) == "subsets"] <- "subset.sum"
        compute.res
    }
}

.populateSubsetThresholds <- function(thresholds, subset.field, has.block) {
    if (is.null(thresholds[[subset.field]])) {
        if (has.block) {
            stub <- list()
        } else {
            stub <- numeric(0)
        }
        names(stub) <- character(0)
        thresholds[[subset.field]] <- stub
    }
    thresholds
}

# library(testthat); library(scrapper); source("test-se_quickAdtQc.R")

library(SummarizedExperiment)
mat <- matrix(rpois(1000, 1), ncol=10)
se <- SummarizedExperiment(list(counts=mat))

test_that("quickAdtQc.se works as expected", {
    out <- quickAdtQc.se(se, subsets=list())
    expect_type(out$sum, "double")
    expect_type(out$detected, "integer")
    expect_type(out$keep, "logical")
    expect_type(metadata(out)$qc$thresholds, "list")

    out <- quickAdtQc.se(se, subsets=list(igg=1:5))
    expect_type(out$subset.sum.igg, "double")

    out2 <- quickAdtQc.se(se, subsets=list(igg=1:5), output.prefix="WHEE.")
    expect_identical(out2$WHEE.sum, out$sum)
    expect_identical(out2$WHEE.detected, out$detected)
    expect_identical(out2$WHEE.subset.sum.igg, out$subset.sum.igg)
    expect_identical(out2$WHEE.keep, out$keep)

    out3 <- quickAdtQc.se(se, subsets=list(igg=1:5, bar=6:10), flatten=FALSE)
    expect_identical(out$subset.sum.igg, out3$subset.sum$igg)
    expect_type(out3$subset.sum$bar, "double")

    out4 <- quickAdtQc.se(se, subsets=list(), meta.name=NULL)
    expect_null(metadata(out4)$qc)
})

test_that("quickRnaQc.se works with custom thresholds", {
    fixed <- list(sum=1, detected=1)
    out <- quickAdtQc.se(se, subsets=list(), thresholds=fixed)

    refixed <- fixed
    refixed$subset.sum <- setNames(numeric(0), character(0))
    expect_identical(metadata(out)$qc$thresholds, refixed)
})

#' Choose highly variable genes
#'
#' Choose highly variable genes (HVGs) based on a variance-related statistic.
#' This is typically used to subset the gene-cell matrix prior to calling \code{\link{runPca}}.
#'
#' @param stats Numeric vector of variances (or a related statistic) across all genes.
#' Typically, the residuals from \code{\link{modelGeneVariances}} are used here.
#' @param top Integer specifying the number of top genes to retain.
#' Note that the actual number of retained genes may not be equal to \code{top}, depending on the other options.
#' @param larger Logical scalar indicating whether larger values of \code{stats} correspond to more variable genes.
#' If \code{TRUE}, HVGs are defined as those with the largest values of \code{stats}.
#' This is typically the case for variances or related statistics, e.g., residuals.
#' @param keep.ties Logical scalar indicating whether to keep tied values of \code{stats}, even if \code{top} may be exceeded.
#' @param bound Numeric scalar specifying the lower bound (if \code{larger=TRUE}) or upper bound (otherwise) to be applied to \code{stats}.
#' Genes are not considered to be HVGs if they do not satisfy this bound, even if they are within the \code{top} genes.
#' For example, residuals from the fitted trend should be positive, which can be enforced by setting \code{bound} to zero.
#' Ignored if \code{NULL}.
#'
#' @return Integer vector containing the indices of genes in \code{stats} that are considered to be highly variable.
#' 
#' @examples
#' resids <- rexp(10000)
#' str(chooseHighlyVariableGenes(resids))
#'
#' @seealso
#' The \code{choose_highly_variable_genes} function in \url{https://libscran.github.io/scran_variances/}. 
#'
#' \code{\link{chooseRnaHvgs.se}}, which choose the HVGs from the residuals computed from a \link[SummarizedExperiment]{SummarizedExperiment}.
#'
#' @author Aaron Lun
#' @export
chooseHighlyVariableGenes <- function(stats, top=4000, larger=TRUE, keep.ties=TRUE, bound=0) {
    top <- min(length(stats), top) # protect against top=Inf, which Rcpp can't cast to an int.
    out <- choose_highly_variable_genes(stats, top=top, larger=larger, keep_ties=keep.ties, bound=bound)
    out + 1L
}

# library(testthat); library(scrapper); source("test-chooseHighlyVariableGenes.R")

test_that("chooseHighlyVariableGenes works", {
    set.seed(20000)

    stats <- runif(10000)
    out <- chooseHighlyVariableGenes(stats, top=2000)
    expect_equal(length(out), 2000)
    expect_gt(min(stats[out]), max(stats[-out]))

    out <- chooseHighlyVariableGenes(stats, top=Inf)
    expect_identical(out, seq_along(stats))

    out <- chooseHighlyVariableGenes(stats, larger=FALSE, bound=NULL)
    expect_equal(length(out), 4000)
    expect_lt(max(stats[out]), min(stats[-out]))

    out <- chooseHighlyVariableGenes(stats, bound=0.9)
    expect_lt(length(out), 2000)
    expect_gt(min(stats[out]), 0.9)

    out <- chooseHighlyVariableGenes(numeric(10000), keep.ties=TRUE, bound=NULL)
    expect_equal(out, seq_len(10000))
})

#' Principal components analysis of a Summarizedexperiment
#'
#' Compact and denoise the dataset by performing PCA on the (log-)normalized expression matrix,
#' by calling \code{\link{runPca}} on an assay of a \link[SummarizedExperiment]{SummarizedExperiment}.
#'
#' @param x A \link[SummarizedExperiment]{SummarizedExperiment} object or one of its subclasses.
#' Rows correspond to genomic features and columns correspond to cells.
#' @param features Integer, logical or character vector containing the features of interest to use in the PCA.
#' For RNA data, this is typically the \code{hvg} vector added by \code{\link{chooseRnaHvgs.se}}.
#' If \code{NULL}, all available features are used.
#' @param number Number of PCs to retain, passed to \code{\link{runPca}}.
#' @param block Block assignment for each cell, passed to \code{\link{runPca}}.
#' @param num.threads Number of threads for the PCA, passed to \code{\link{runPca}}.
#' @param more.pca.args Named list of additional arguments to pass to \code{\link{runPca}}.
#' @param assay.type Integer or string specifying the assay of \code{x} to be used for PCA.
#' This is typically the log-normalized expression matrix created by \code{\link{normalizeRnaCounts.se}}.
#' @param output.name String containing the name of the \code{\link[SingleCellExperiment]{reducedDim}} entry in which to store the PC scores.
#' @param meta.name String containing the name of the \code{link[S4Vectors]{metadata}} entry in which to store other PCA statistics.
#' @param delayed.transpose Logical scalar indicating whether to delay the transposition when storing coordinates in the \code{\link[SingleCellExperiment]{reducedDims}}.
#'
#' @return \code{x} is returned with the principal component scores in the \code{reducedDim}.
#' (This is converted to a \link[SingleCellExperiment]{SingleCellExperiment} if it wasn't one already.)
#' Additional outputs (e.g., rotation matrix, variance explained) are stored in the \code{\link[S4Vectors]{metadata}}.
#'
#' @author Aaron Lun
#' @examples
#' library(SingleCellExperiment)
#' sce <- getTestRnaData.se("hvg")
#' sce <- runPca.se(sce, rowData(sce)$hvg)
#' dim(reducedDim(sce, "PCA"))
#' plot(metadata(sce)$PCA$variance.explained / metadata(sce)$PCA$total.variance)
#'
#' @export
#' @importFrom methods is as
#' @importFrom DelayedArray DelayedArray
#' @importFrom S4Vectors metadata metadata<-
runPca.se <- function(
    x,
    features,
    number = 25,
    block = NULL,
    num.threads = 1,
    more.pca.args = list(),
    assay.type = "logcounts",
    output.name = "PCA",
    meta.name = "PCA",
    delayed.transpose = FALSE
) {
    y <- SummarizedExperiment::assay(x, assay.type)
    if (!is.null(features)) {
        y <- DelayedArray(y)[features,,drop=FALSE] # ensure that no copy is made.
    }

    out <- .call(
        runPca,
        list(y),
        list(number=number, block=block, num.threads=num.threads),
        more.pca.args
    )

    if (!is(x, "SingleCellExperiment")) {
        loadNamespace("SingleCellExperiment")
        x <- as(x, "SingleCellExperiment")
    }
    x <- .addTransposedReddim(x, output.name, out$components, delayed.transpose)

    if (!is.null(meta.name)) {
        out$components <- NULL
        metadata(x)[[meta.name]] <- out
    }

    x
}

# library(testthat); library(scrapper); source("test-se_runPca.R")

library(SingleCellExperiment)
set.seed(8888)
sce <- SingleCellExperiment(list(logcounts = matrix(rnorm(100000), 500, 200)))

test_that("runPca.se works as expected", {
    out <- runPca.se(sce, features=1:50) 
    expect_true("PCA" %in% reducedDimNames(out))
    expect_identical(nrow(metadata(out)$PCA$rotation), 50L)
    expect_identical(length(metadata(out)$PCA$variance.explained), 25L)

    # Works with features set to NULL.
    null <- runPca.se(sce[1:50,], features=NULL) 
    expect_identical(reducedDim(null), reducedDim(out))
    expect_identical(metadata(null)$PCA$rotation, metadata(out)$PCA$rotation)

    no.meta <- runPca.se(sce, features=1:10, meta.name=NULL)
    expect_null(metadata(no.meta)$PCA)
})

test_that("runPca.se works with a basic SE", {
    se <- SummarizedExperiment(list(logcounts=assay(sce)))
    out <- runPca.se(se, features=5:100) 
    expect_s4_class(out, "SingleCellExperiment")
    expect_identical(reducedDimNames(out), "PCA")
})

#' Choose a suitable pseudo-count
#'
#' Choose a suitable pseudo-count to control the bias introduced by log-transformation of normalized counts from \code{\link{normalizeCounts}}.
#' Larger pseudo-counts shrink log-expression values towards the zero-expression baseline,
#' reducing the impact of the transformation bias at the cost of some sensitivity.
#'
#' @param size.factors Numeric vector of size factors for all cells.
#' It is expected that these have already been centered, e.g., with \code{\link{centerSizeFactors}}.
#' Invalid size factors (e.g., non-positive, non-finite) will be ignored.
#' @param quantile Numeric scalar specifying the quantile to use for finding the smallest/largest size factors.
#' Setting this to zero will use the observed minimum and maximum, though in practice, this is usually too sensitive to outliers.
#' The default is to take the 5th and 95th percentile to obtain a range that captures most of the distribution.
#' @param max.bias Numeric scalar specifying the maximum allowed bias.
#' This is the maximum absolute value of any spurious log2-fold change between the cells with the smallest and largest size factors.
#' @param min.value Numeric scalar specifying the minimum value for the pseudo-count.
#' Defaults to 1 to stabilize near-zero normalized expression values, otherwise these manifest as avoid large negative values.
#'
#' @return A choice of pseudo-count for \code{\link{normalizeCounts}}.
#'
#' @author Aaron Lun
#'
#' @seealso
#' \code{choose_pseudo_count} in \url{https://libscran.github.io/scran_norm/}.
#'
#' @references
#' Lun ATL (2018).
#' Overcoming systematic errors caused by log-transformation of normalized single-cell RNA sequencing data.
#' \emph{biorXiv} doi:10.1101/404962
#'
#' @examples
#' sf <- centerSizeFactors(runif(100))
#' choosePseudoCount(sf)
#' choosePseudoCount(sf, quantile=0.01)
#' choosePseudoCount(sf, max.bias=0.5)
#'
#' @export
choosePseudoCount <- function(size.factors, quantile = 0.05, max.bias = 1, min.value = 1) {
    choose_pseudo_count(size.factors, quantile, max.bias, min.value)
}

# library(testthat); library(scrapper); source("test-choosePseudoCount.R")

set.seed(9999)
test_that("choosePseudoCount works as expected", {
    sf <- 2^rnorm(100, sd=2)
    out <- choosePseudoCount(sf)
    expect_gt(out, 0)
    expect_lt(out, choosePseudoCount(sf, quantile=0.01))
    expect_lt(out, choosePseudoCount(sf, max.bias=0.5))
})

#' Scale and combine multiple embeddings 
#'
#' Scale multiple embeddings (usually derived from different modalities for the same cells) so that their within-population variances are comparable,
#' and then combine them into a single embedding matrix for further analyses like clustering, t-SNE, etc.
#' The aim is to equalize uninteresting variance across modalities so that high technical variance in one modality does not drown out interesting biology in another modality.
#'
#' @param x List of numeric matrices of principal components or other embeddings, one for each modality.
#' For each entry, rows are dimensions and columns are cells.
#' All entries should have the same number of columns but may have different numbers of rows.
#' @param block Factor specifying the block of origin (e.g., batch, sample) for each cell in \code{x}.
#' If provided, the scaling factor is computed as a weighted average across blocks to ensure that block effects do not inflate the within-population variance.
#' Alternatively \code{NULL}, if all cells are from the same block.
#' @param block.weight.policy String specifying the policy to use for weighting different blocks when computing the average scaling factor.
#' See the argument of the same name in \code{\link{computeBlockWeights}} for more detail.
#' Only used if \code{block} is not \code{NULL}.
#' @param variable.block.weight Numeric vector of length 2, specifying the parameters for variable block weighting.
#' See the argument of the same name in \code{\link{computeBlockWeights}} for more detail.
#' Only used if \code{block} is not \code{NULL} and \code{block.weight.policy = "variable"}.
#' @param num.neighbors Integer scalar specifying the number of neighbors to use to define the scaling factor.
#' @param BNPARAM A \link[BiocNeighbors]{BiocNeighborParam} object specifying how to perform the neighbor search.
#' @param num.threads Integer scalar specifying the number of threads to use.
#' @param weights Numeric vector of length equal to that of \code{x}, specifying the weights to apply to each modality.
#' Each value represents a multiplier of the within-population variance of its modality, i.e., larger values increase the contribution of that modality in the combined output matrix.
#' \code{NULL} is equivalent to an all-1 vector, i.e., all modalities are scaled to have the same within-population variance.
#'
#' @return List containing \code{scaling}, a vector of scaling factors to be aplied to each embedding;
#' and \code{combined}, a numeric matrix creating by scaling each entry of \code{x} by \code{scaling} and then \code{rbind}ing them together.
#'
#' @seealso
#' \url{https://libscran.github.io/mumosa/}, for the basis and caveats of this approach.
#'
#' \code{\link{scaleByNeighbors.se}}, to combine embeddings in a \link[SingleCellExperiment]{SingleCellExperiment}.
#'
#' @examples
#' pcs <- list(
#'     gene = matrix(rnorm(10000), ncol=200),
#'     protein = matrix(rnorm(1000, sd=3), ncol=200),
#'     guide = matrix(rnorm(2000, sd=5), ncol=200)
#' )
#'
#' out <- scaleByNeighbors(pcs)
#' out$scaling
#' dim(out$combined)
#' @author Aaron Lun
#' @export
#' @importFrom BiocNeighbors defineBuilder AnnoyParam
scaleByNeighbors <- function(
    x,
    num.neighbors=20, 
    block=NULL,
    block.weight.policy=c("variable", "equal", "none"),
    variable.block.weight=c(0, 1000),
    num.threads=1,
    weights=NULL,
    BNPARAM=AnnoyParam()
) {
    .checkSEX(x, "scaleByNeighbors.se")

    x <- as.list(x)
    nmod <- length(x)
    ref.ncol <- ncol(x[[1]])
    for (i in seq_along(x)) {
        x[[i]] <- as.matrix(x[[i]])
        if (ncol(x[[i]]) != ref.ncol) {
            stop("all entries of 'x' should have the same number of columns")
        }
    }

    block <- .transformFactor(block)

    scaling <- scale_by_neighbors(
        num_cells=ref.ncol,
        embedding=x,
        num_neighbors=num.neighbors,
        block=block$index,
        block_weight_policy=match.arg(block.weight.policy),
        variable_block_weight=variable.block.weight,
        nn_builder=defineBuilder(BNPARAM)$builder,
        num_threads=num.threads
    )

    if (!is.null(weights)) {
        scaling <- scaling * sqrt(weights)
    }

    names(scaling) <- names(x)
    for (i in seq_along(x)) {
        x[[i]] <- x[[i]] * scaling[i]
    }

    list(
         scaling = scaling,
         combined = do.call(rbind, x)
    )
}

# library(testthat); library(scrapper); source("test-scaleByNeighbors.R")

set.seed(2000)
pcs <- list(
    gene = matrix(rnorm(10000), ncol=200),
    protein = matrix(rnorm(1000, sd=3), ncol=200),
    guide = matrix(rnorm(2000, sd=5), ncol=200)
)

manual_knn_distance <- function(x, k = 20) {
    value <- numeric(ncol(x))
    for (i in seq_len(ncol(x))) {
        chosen <- x[,i]
        distances <- sqrt(colSums((x[,-i] - chosen)^2))
        value[i] <- sort(distances)[k]
    } 
    median(value)
}

test_that("scaleByNeighbors works as expected", {
    out <- scaleByNeighbors(pcs, BNPARAM=BiocNeighbors::VptreeParam())
    expect_identical(names(out$scaling), names(pcs))

    references <- lapply(pcs, manual_knn_distance)
    expect_equal(out$scaling, references[[1]]/unlist(references))

    expect_identical(ncol(out$combined), ncol(pcs[[1]]))
    expect_identical(nrow(out$combined), sum(vapply(pcs, nrow, 0L)))

    expect_error(scaleByNeighbors(SummarizedExperiment::SummarizedExperiment()), "not supported")
})

test_that("scaleByNeighbors works as expected with blocking", {
    # Same results with duplicated blocks.
    {
        shuffle <- sample(200)
        block <- rep(LETTERS[1:3], each=200)[shuffle]
        pcs3 <- lapply(pcs, function(x) {
            cbind(x, x, x)[,shuffle,drop=FALSE]
        })

        ref <- scaleByNeighbors(pcs, BNPARAM=BiocNeighbors::VptreeParam())
        out <- scaleByNeighbors(pcs3, block=block, BNPARAM=BiocNeighbors::VptreeParam())
        expect_identical(out$scaling, ref$scaling)
    }

    # Manual comparison using a randomly sampled block.
    {
        set.seed(20001)
        block <- sample(LETTERS[1:3], 200, replace=TRUE)
        out <- scaleByNeighbors(pcs, block=block, block.weight.policy="equal", BNPARAM=BiocNeighbors::VptreeParam())

        references <- lapply(pcs, function(x) {
            cur.d <- numeric()
            for (b in LETTERS[1:3]) {
                cur.d <- append(cur.d, manual_knn_distance(x[,b == block,drop=FALSE]))
            }
            mean(cur.d)
        })

        references <- 1/unlist(references)
        normalized <- references / (references[1] / out$scaling[1])
        expect_equal(normalized, out$scaling)
    }
})

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

compute_adt_qc_metrics <- function(x, subsets, num_threads) {
    .Call('_scrapper_compute_adt_qc_metrics', PACKAGE = 'scrapper', x, subsets, num_threads)
}

suggest_adt_qc_thresholds <- function(metrics, block, min_detected_drop, num_mads) {
    .Call('_scrapper_suggest_adt_qc_thresholds', PACKAGE = 'scrapper', metrics, block, min_detected_drop, num_mads)
}

filter_adt_qc_metrics <- function(filters, metrics, block) {
    .Call('_scrapper_filter_adt_qc_metrics', PACKAGE = 'scrapper', filters, metrics, block)
}

aggregate_across_cells <- function(x, groups, compute_sum, compute_detected, compute_median, num_threads) {
    .Call('_scrapper_aggregate_across_cells', PACKAGE = 'scrapper', x, groups, compute_sum, compute_detected, compute_median, num_threads)
}

aggregate_across_genes <- function(x, sets, average, nthreads) {
    .Call('_scrapper_aggregate_across_genes', PACKAGE = 'scrapper', x, sets, average, nthreads)
}

build_snn_graph <- function(neighbors, scheme, num_threads) {
    .Call('_scrapper_build_snn_graph', PACKAGE = 'scrapper', neighbors, scheme, num_threads)
}

graph_to_list <- function(ptr0) {
    .Call('_scrapper_graph_to_list', PACKAGE = 'scrapper', ptr0)
}

list_to_graph <- function(contents) {
    .Call('_scrapper_list_to_graph', PACKAGE = 'scrapper', contents)
}

center_size_factors <- function(size_factors, block, lowest) {
    .Call('_scrapper_center_size_factors', PACKAGE = 'scrapper', size_factors, block, lowest)
}

choose_highly_variable_genes <- function(stats, top, larger, keep_ties, bound) {
    .Call('_scrapper_choose_highly_variable_genes', PACKAGE = 'scrapper', stats, top, larger, keep_ties, bound)
}

choose_pseudo_count <- function(size_factors, quantile, max_bias, min_value) {
    .Call('_scrapper_choose_pseudo_count', PACKAGE = 'scrapper', size_factors, quantile, max_bias, min_value)
}

cluster_multilevel <- function(ptr0, resolution, seed) {
    .Call('_scrapper_cluster_multilevel', PACKAGE = 'scrapper', ptr0, resolution, seed)
}

cluster_leiden <- function(ptr0, resolution, objective, seed) {
    .Call('_scrapper_cluster_leiden', PACKAGE = 'scrapper', ptr0, resolution, objective, seed)
}

cluster_walktrap <- function(ptr0, steps) {
    .Call('_scrapper_cluster_walktrap', PACKAGE = 'scrapper', ptr0, steps)
}

cluster_kmeans <- function(x, nclusters, tatami, init_method, refine_method, var_part_optimize_partition, var_part_size_adjustment, lloyd_iterations, hartigan_wong_iterations, hartigan_wong_quick_transfer_iterations, hartigan_wong_quit_quick_transfer_failure, seed, nthreads) {
    .Call('_scrapper_cluster_kmeans', PACKAGE = 'scrapper', x, nclusters, tatami, init_method, refine_method, var_part_optimize_partition, var_part_size_adjustment, lloyd_iterations, hartigan_wong_iterations, hartigan_wong_quick_transfer_iterations, hartigan_wong_quit_quick_transfer_failure, seed, nthreads)
}

combine_factors <- function(factors, keep_unused, nlevels) {
    .Call('_scrapper_combine_factors', PACKAGE = 'scrapper', factors, keep_unused, nlevels)
}

compute_block_weights <- function(sizes, policy, variable_block_weight) {
    .Call('_scrapper_compute_block_weights', PACKAGE = 'scrapper', sizes, policy, variable_block_weight)
}

compute_clrm1_factors <- function(x, num_threads) {
    .Call('_scrapper_compute_clrm1_factors', PACKAGE = 'scrapper', x, num_threads)
}

correct_mnn <- function(x, block, num_neighbors, num_steps, num_threads, merge_policy, builder) {
    .Call('_scrapper_correct_mnn', PACKAGE = 'scrapper', x, block, num_neighbors, num_steps, num_threads, merge_policy, builder)
}

compute_crispr_qc_metrics <- function(x, num_threads) {
    .Call('_scrapper_compute_crispr_qc_metrics', PACKAGE = 'scrapper', x, num_threads)
}

suggest_crispr_qc_thresholds <- function(metrics, block, num_mads) {
    .Call('_scrapper_suggest_crispr_qc_thresholds', PACKAGE = 'scrapper', metrics, block, num_mads)
}

filter_crispr_qc_metrics <- function(filters, metrics, block) {
    .Call('_scrapper_filter_crispr_qc_metrics', PACKAGE = 'scrapper', filters, metrics, block)
}

fit_variance_trend <- function(means, variances, mean_filter, min_mean, transform, span, use_min_width, min_width, min_window_count, num_threads) {
    .Call('_scrapper_fit_variance_trend', PACKAGE = 'scrapper', means, variances, mean_filter, min_mean, transform, span, use_min_width, min_width, min_window_count, num_threads)
}

model_gene_variances <- function(x, block, nblocks, block_average_policy, block_weight_policy, variable_block_weight, block_quantile, mean_filter, min_mean, transform, span, use_min_width, min_width, min_window_count, num_threads) {
    .Call('_scrapper_model_gene_variances', PACKAGE = 'scrapper', x, block, nblocks, block_average_policy, block_weight_policy, variable_block_weight, block_quantile, mean_filter, min_mean, transform, span, use_min_width, min_width, min_window_count, num_threads)
}

normalize_counts <- function(x, size_factors, log, pseudo_count, log_base, preserve_sparsity) {
    .Call('_scrapper_normalize_counts', PACKAGE = 'scrapper', x, size_factors, log, pseudo_count, log_base, preserve_sparsity)
}

initialize_LogNormalizedMatrix <- function(seed, size_factors, pseudo_count, log_base) {
    .Call('_scrapper_initialize_LogNormalizedMatrix', PACKAGE = 'scrapper', seed, size_factors, pseudo_count, log_base)
}

compute_rna_qc_metrics <- function(x, subsets, num_threads) {
    .Call('_scrapper_compute_rna_qc_metrics', PACKAGE = 'scrapper', x, subsets, num_threads)
}

suggest_rna_qc_thresholds <- function(metrics, block, num_mads) {
    .Call('_scrapper_suggest_rna_qc_thresholds', PACKAGE = 'scrapper', metrics, block, num_mads)
}

filter_rna_qc_metrics <- function(filters, metrics, block) {
    .Call('_scrapper_filter_rna_qc_metrics', PACKAGE = 'scrapper', filters, metrics, block)
}

run_pca <- function(x, number, block, block_weight_policy, variable_block_weight, components_from_residuals, scale, subset, realized, irlba_work, irlba_iterations, irlba_seed, num_threads) {
    .Call('_scrapper_run_pca', PACKAGE = 'scrapper', x, number, block, block_weight_policy, variable_block_weight, components_from_residuals, scale, subset, realized, irlba_work, irlba_iterations, irlba_seed, num_threads)
}

run_tsne <- function(nnidx, nndist, perplexity, theta, early_exaggeration_iterations, exaggeration_factor, momentum_switch_iterations, start_momentum, final_momentum, eta, max_depth, leaf_approx, max_iter, seed, num_threads) {
    .Call('_scrapper_run_tsne', PACKAGE = 'scrapper', nnidx, nndist, perplexity, theta, early_exaggeration_iterations, exaggeration_factor, momentum_switch_iterations, start_momentum, final_momentum, eta, max_depth, leaf_approx, max_iter, seed, num_threads)
}

perplexity_to_neighbors <- function(p) {
    .Call('_scrapper_perplexity_to_neighbors', PACKAGE = 'scrapper', p)
}

run_umap <- function(nnidx, nndist, num_dim, local_connectivity, bandwidth, mix_ratio, spread, min_dist, a, b, repulsion_strength, initialize_method, initial_coordinates, initialize_random_on_spectral_fail, initialize_spectral_scale, initialize_spectral_jitter, initialize_spectral_jitter_sd, initialize_random_scale, initialize_seed, num_epochs, learning_rate, negative_sample_rate, optimize_seed, num_threads, parallel_optimization) {
    .Call('_scrapper_run_umap', PACKAGE = 'scrapper', nnidx, nndist, num_dim, local_connectivity, bandwidth, mix_ratio, spread, min_dist, a, b, repulsion_strength, initialize_method, initial_coordinates, initialize_random_on_spectral_fail, initialize_spectral_scale, initialize_spectral_jitter, initialize_spectral_jitter_sd, initialize_random_scale, initialize_seed, num_epochs, learning_rate, negative_sample_rate, optimize_seed, num_threads, parallel_optimization)
}

sanitize_size_factors <- function(size_factors, handle_zero, handle_negative, handle_nan, handle_infinite) {
    .Call('_scrapper_sanitize_size_factors', PACKAGE = 'scrapper', size_factors, handle_zero, handle_negative, handle_nan, handle_infinite)
}

scale_by_neighbors <- function(num_cells, embedding, num_neighbors, block, block_weight_policy, variable_block_weight, num_threads, nn_builder) {
    .Call('_scrapper_scale_by_neighbors', PACKAGE = 'scrapper', num_cells, embedding, num_neighbors, block, block_weight_policy, variable_block_weight, num_threads, nn_builder)
}

score_gene_set <- function(x, rank, block, block_weight_policy, variable_block_weight, scale, realized, irlba_work, irlba_iterations, irlba_seed, num_threads) {
    .Call('_scrapper_score_gene_set', PACKAGE = 'scrapper', x, rank, block, block_weight_policy, variable_block_weight, scale, realized, irlba_work, irlba_iterations, irlba_seed, num_threads)
}

score_markers_summary <- function(x, groups, num_groups, block, block_average_policy, block_weight_policy, variable_block_weight, block_quantile, threshold, num_threads, compute_group_mean, compute_group_detected, compute_delta_mean, compute_delta_detected, compute_cohens_d, compute_auc, compute_summary_min, compute_summary_mean, compute_summary_median, compute_summary_max, compute_summary_quantiles, compute_summary_min_rank, min_rank_limit) {
    .Call('_scrapper_score_markers_summary', PACKAGE = 'scrapper', x, groups, num_groups, block, block_average_policy, block_weight_policy, variable_block_weight, block_quantile, threshold, num_threads, compute_group_mean, compute_group_detected, compute_delta_mean, compute_delta_detected, compute_cohens_d, compute_auc, compute_summary_min, compute_summary_mean, compute_summary_median, compute_summary_max, compute_summary_quantiles, compute_summary_min_rank, min_rank_limit)
}

score_markers_pairwise <- function(x, groups, num_groups, block, block_average_policy, block_weight_policy, variable_block_weight, block_quantile, threshold, num_threads, compute_group_mean, compute_group_detected, compute_delta_mean, compute_delta_detected, compute_cohens_d, compute_auc) {
    .Call('_scrapper_score_markers_pairwise', PACKAGE = 'scrapper', x, groups, num_groups, block, block_average_policy, block_weight_policy, variable_block_weight, block_quantile, threshold, num_threads, compute_group_mean, compute_group_detected, compute_delta_mean, compute_delta_detected, compute_cohens_d, compute_auc)
}

score_markers_best <- function(x, top, groups, num_groups, block, block_average_policy, block_weight_policy, variable_block_weight, block_quantile, threshold, num_threads, compute_group_mean, compute_group_detected, compute_delta_mean, compute_delta_detected, compute_cohens_d, compute_auc, index_only) {
    .Call('_scrapper_score_markers_best', PACKAGE = 'scrapper', x, top, groups, num_groups, block, block_average_policy, block_weight_policy, variable_block_weight, block_quantile, threshold, num_threads, compute_group_mean, compute_group_detected, compute_delta_mean, compute_delta_detected, compute_cohens_d, compute_auc, index_only)
}

set_executor <- function(ptr) {
    .Call('_scrapper_set_executor', PACKAGE = 'scrapper', ptr)
}

subsample_by_neighbors <- function(indices, distances, min_remaining) {
    .Call('_scrapper_subsample_by_neighbors', PACKAGE = 'scrapper', indices, distances, min_remaining)
}

summarize_effects <- function(num_genes, num_groups, effects, compute_summary_min, compute_summary_mean, compute_summary_median, compute_summary_max, compute_summary_quantiles, compute_summary_min_rank, num_threads) {
    .Call('_scrapper_summarize_effects', PACKAGE = 'scrapper', num_genes, num_groups, effects, compute_summary_min, compute_summary_mean, compute_summary_median, compute_summary_max, compute_summary_quantiles, compute_summary_min_rank, num_threads)
}

test_enrichment <- function(overlap, num_interest, set_sizes, universe, log, num_threads) {
    .Call('_scrapper_test_enrichment', PACKAGE = 'scrapper', overlap, num_interest, set_sizes, universe, log, num_threads)
}

#' Run all neighbor-related steps
#'
#' Run all steps that require a nearest-neighbor search.
#' This includs \code{\link{runUmap}}, \code{\link{runTsne}} and \code{\link{buildSnnGraph}} with \code{\link{clusterGraph}}.
#' The idea is to build the index once, perform the neighbor search, and run each task in parallel to save time.
#'
#' @param x Numeric matrix where rows are dimensions and columns are cells,
#' typically containing a low-dimensional representation from, e.g., \code{\link{runPca}}.
#'
#' Alternatively, an index constructed by \code{\link[BiocNeighbors]{buildIndex}}.
#' @param runUmap.args Named list of further arguments to pass to \code{\link{runUmap}}.
#' This can be set to \code{NULL} to omit the UMAP.
#' @param runTsne.args Named list of further arguments to pass to \code{\link{runTsne}}.
#' This can be set to \code{NULL} to omit the t-SNE.
#' @param buildSnnGraph.args Named list of further arguments to pass to \code{\link{buildSnnGraph}}.
#' Ignored if \code{clusterGraph.args=NULL}.
#' @param clusterGraph.args Named list of further arguments to pass to \code{\link{clusterGraph}}.
#' This can be set to \code{NULL} to omit the clustering.
#' @param BNPARAM A \link[BiocNeighbors]{BiocNeighborParam} instance specifying the nearest-neighbor search algorithm to use.
#' @param collapse.search Logical scalar indicating whether to collapse the nearest-neighbor search for each step into a single search.
#' Steps that need fewer neighbors will take a subset of the neighbors from the collapsed search.
#' Setting this to \code{TRUE} is faster but may not give the same results as separate searches for some nearest-neighbor algorithms (e.g., approximate methods).
#' @param return.graph Logical scalar indicating whether to return the output of \code{\link{buildSnnGraph}}.
#' By default, only the output of \code{\link{clusterGraph}} is returned.
#' @param num.threads Integer scalar specifying the number of threads to use.
#' At least one thread should be available for each step. 
#'
#' @return A named list containing the results of each step.
#' See each individual function for the format of the results.
#'
#' @seealso
#' \code{\link{runAllNeighborSteps.se}}, to run each neighbor-related step on a \link[SingleCellExperiment]{SingleCellExperiment}.
#' 
#' @author Aaron Lun
#' @examples
#' x <- t(as.matrix(iris[,1:4]))
#' # (Turning down the number of threads so that R CMD check is happy.)
#' res <- runAllNeighborSteps(x, num.threads=2)
#' str(res)
#'
#' @export
#' @importFrom BiocNeighbors buildIndex AnnoyParam findKNN
#' @importFrom parallel makeCluster stopCluster parLapply
runAllNeighborSteps <- function(
    x,
    runUmap.args=list(),
    runTsne.args=list(),
    buildSnnGraph.args=list(),
    clusterGraph.args=list(),
    BNPARAM=AnnoyParam(),
    return.graph=FALSE,
    collapse.search=TRUE,
    num.threads=3
) {
    .checkSEX(x, "runAllNeighborSteps.se")
    k.choices <- list()

    if (!is.null(runUmap.args)) {
        umap.k <- runUmap.args$num.neighbors
        if (is.null(umap.k)) {
            umap.k <- formals(runUmap)$num.neighbors
        }
        k.choices$runUmap <- umap.k
    }

    if (!is.null(runTsne.args)) {
        tsne.perplexity <- runTsne.args$perplexity
        if (is.null(tsne.perplexity)) {
            tsne.perplexity <- formals(runTsne)$perplexity 
        }
        k.choices$runTsne <- tsnePerplexityToNeighbors(tsne.perplexity)
    }

    if (!is.null(clusterGraph.args)) {
        snn.k <- buildSnnGraph.args$num.neighbors
        if (is.null(snn.k)) {
            snn.k <- formals(buildSnnGraph)$num.neighbors
        }
        k.choices$clusterGraph <- snn.k
    }

    if (length(k.choices) == 0L) {
        output <- list()
        names(output) <- character(0)
        return(output)
    }

    if (is.matrix(x)) {
        index <- buildIndex(x, BNPARAM=BNPARAM, transposed=TRUE)
    } else {
        index <- x
    }
    nn.res <- list()

    if (collapse.search) {
        all.res <- findKNN(
            index,
            k=max(unlist(k.choices)),
            num.threads=num.threads,
            get.index="transposed",
            get.distance="transposed"
        )
        for (n in names(k.choices)) {
            curk <- k.choices[[n]]
            curi <- all.res$index
            curd <- all.res$distance
            if (curk < nrow(curi)) {
                curi <- curi[seq_len(curk),,drop=FALSE]
                curd <- curd[seq_len(curk),,drop=FALSE]
            }
            nn.res[[n]] <- list(index=curi, distance=curd)
        }

    } else {
        precomputed <- list()
        for (n in names(k.choices)) {
            curk <- k.choices[[n]]
            key <- as.character(curk) 
            if (key %in% names(precomputed)) {
                curres <- precomputed[[key]]
            } else {
                curres <- findKNN(
                    index,
                    k=curk,
                    num.threads=num.threads,
                    get.index="transposed",
                    get.distance="transposed"
                )
                precomputed[[key]] <- curres
            }
            nn.res[[n]] <- curres
        }
    }

    jobs <- list()
    if ("runUmap" %in% names(nn.res)) {
        jobs$runUmap <- list(runUmap=c(list(x=nn.res$runUmap), runUmap.args))
    }
    if ("runTsne" %in% names(nn.res)) {
        if (!("num.threads" %in% names(runTsne.args))) {
            runTsne.args$num.threads <- max(1L, num.threads - length(nn.res) + 1L)
        }
        jobs$runTsne <- list(runTsne=c(list(x=nn.res$runTsne), runTsne.args))
    }
    if ("clusterGraph" %in% names(nn.res)) {
        jobs$clusterGraph <- list(
            buildSnnGraph=c(list(x=nn.res$clusterGraph), buildSnnGraph.args),
            clusterGraph=clusterGraph.args,
            return.graph=return.graph
        )
    }

    # Don't attempt to be smart and use a FORK cluster on Unix-likes to squeeze
    # out more efficiency, as forking does not play nice with OpenMP; see
    # https://stackoverflow.com/questions/49049388/understanding-openmp-shortcomings-regarding-fork.
    # Just keep the PSOCK default.
    clust <- makeCluster(min(length(jobs), num.threads))
    on.exit(stopCluster(clust), add=TRUE, after=FALSE)
    output <- parLapply(clust, jobs, fun=.internal_executor)

    if (return.graph && "clusterGraph" %in% names(output)) {
        tmp <- output$clusterGraph
        output$clusterGraph <- NULL
        c(output, tmp)
    } else {
        output
    }
}

.internal_executor <- function(job) {
    if ("runUmap" %in% names(job)) {
        do.call(runUmap, job$runUmap)
    } else if ("runTsne" %in% names(job)) {
        do.call(runTsne, job$runTsne)
    } else {
        snn.graph <- do.call(buildSnnGraph, job$buildSnnGraph)
        clust <- do.call(clusterGraph, c(list(snn.graph), job$clusterGraph))

        if (job$return.graph) {
            list(buildSnnGraph=snn.graph, clusterGraph=clust)
        } else {
            return(clust)
        }
    }
}

#' Compute block weights
#'
#' Compute a weight for each block based on the number of cells in each block.
#' This is typically used to aggregate statistics across blocks, e.g., with weighted sums/averages.
#' 
#' @param sizes Numeric vector containing the size of (i.e., number of cells in) each block.
#' @param block.weight.policy String specifying the policy to use for weighting different blocks. 
#' This should be one of:
#' \itemize{
#' \item \code{"size"}: the contribution of each block is proportional to its size.
#' \code{"none"} is also a deprecated alias for \code{"size"}. 
#' \item \code{"equal"}: blocks are equally weighted regardless of their size.
#' The exception is that of empty blocks with no cells, which receive zero weight.
#' \item \code{"variable"}: blocks are equally weighted past a certain threshold size.
#' Below that size, the contribution of each block is proportional to its size.
#' This avoids outsized contributions from very large blocks.
#' }
#' @param variable.block.weight Numeric vector of length 2, specifying the parameters for variable block weighting.
#' The first and second values are used as the lower and upper bounds, respectively, for the variable weight calculation.
#' Only used if \code{block.weight.policy = "variable"}.
#'
#' @return Numeric vector containing the relative block weights.
#'
#' @author Aaron Lun
#' @examples
#' computeBlockWeights(c(1, 10, 100, 1000, 10000))
#' computeBlockWeights(c(1, 10, 100, 1000, 10000), block.weight.policy="equal")
#' computeBlockWeights(c(1, 10, 100, 1000, 10000), variable.block.weight=c(50, 5000))
#'
#' @seealso
#' The \code{compute_weights} function from \url{https://libscran.github.io/scran_blocks/}.
#'
#' @export
computeBlockWeights <- function(
    sizes,
    block.weight.policy=c("variable", "equal", "size", "none"),
    variable.block.weight=c(0, 1000))
{
    compute_block_weights(sizes, match.arg(block.weight.policy), variable.block.weight)
}

# library(testthat); library(scrapper); source("test-computeBlockWeights.R")

test_that("computeBlockWeights works as expected", {
    sizes <- c(1, 10, 100, 1000, 10000)
    expect_equal(computeBlockWeights(sizes), pmin(1, sizes/1000))
    expect_equal(computeBlockWeights(sizes, variable.block.weight=c(50, 500)), c(0, 0, 50/450, 1, 1))
    expect_equal(computeBlockWeights(sizes, block.weight.policy="equal"), rep(1, length(sizes)))
    expect_equal(computeBlockWeights(sizes, block.weight.policy="size"), sizes)
    expect_equal(computeBlockWeights(sizes, block.weight.policy="none"), sizes)
})

#' Score marker genes in a SummarizedExperiment
#'
#' Identify candidate marker genes based on effect sizes from pairwise comparisons between groups of cells,
#' by calling \code{\link{scoreMarkers}} on an assay of a \link[SummarizedExperiment]{SummarizedExperiment}.
#'
#' @param x A \link[SummarizedExperiment]{SummarizedExperiment} object or one of its subclasses.
#' Rows correspond to genes and columns correspond to cells.
#' @param groups Group assignment for each cell, passed to \code{\link{scoreMarkers}}.
#' @param block Block assignment for each cell, passed to \code{\link{scoreMarkers}}.
#' @param num.threads Number of threads for marker scoring, passed to \code{\link{scoreMarkers}}.
#' @param more.marker.args Named list of additional arguments to pass to \code{\link{scoreMarkers}}.
#' @param assay.type Integer or string specifying the assay to use for differential comparisons, usually containing log-normalized expression values.
#' @param extra.columns \link[S4Vectors]{DataFrame} containing extra columns to add each DataFrame.
#' This should have the same number of rows as \code{x}.
#' For \code{scoreMarkers.se}, this may also be a character vector specifying the columns of \code{\link[SummarizedExperiment]{rowData}} to be added.
#' @param order.by String specifying the column to order each DataFrame by.
#' Alternatively \code{TRUE}, a column is automatically chosen from the effect size summaries.
#' If \code{NULL} or \code{FALSE}, no ordering is performed.
#' @param marker.res List containing the result of \code{\link{scoreMarkers}}.
#' @param marker.df DataFrame containing the marker statistics for a single group.
#' @param columns Character vector of the names of columns to retain in the preview.
#' This may be named, in which the names are used as the column names.
#' @param pre.columns,post.columns Character vector of the names of additional columns to retain in the preview.
#' These are added before or after the columns in \code{columns}, for \code{pre.columns} and \code{post.columns} respectively.
#' @param include.order.by Boolean indicating whether the column specified by \code{order.by} should be included in the output DataFrame.
#' A string may also be supplied and will be treated as \code{TRUE}; the value of the string will be used as the column name in the output DataFrame.
#' @param rows Integer specifying the number of rows to show.
#' If \code{NULL}, all rows are returned.
#' 
#' @return 
#' For \code{scoreMarkers.se} and \code{formatScoreMarkersResult}, a \link[S4Vectors]{List} of \link[S4Vectors]{DataFrame}s is returned.
#' Each DataFrame corresponds to a unique group in \code{groups}.
#' Each row contains statistics for a gene in \code{x}, with the following columns:
#' \itemize{
#' \item \code{mean}, the mean expression in the current group.
#' \item \code{detected}, the proportion of cells with detected expression in the current group.
#' \item \code{<effect>.<summary>}, a summary statistic for an effect size,
#' e.g., \code{cohens.d.mean} contains the mean Cohen's d across comparisons involving the current group.
#' }
#'
#' For \code{previewMarkers}, a DataFrame is returned containing the specified columns and rows.
#'
#' @author Aaron Lun
#'
#' @examples
#' sce <- getTestRnaData.se("cluster")
#' markers <- scoreMarkers.se(sce, sce$clusters)
#' previewMarkers(markers[["1"]])
#'
#' @export
scoreMarkers.se <- function(
    x,
    groups,
    block = NULL,
    num.threads = 1,
    more.marker.args = list(),
    assay.type = "logcounts",
    extra.columns = NULL,
    order.by = TRUE
) {
    res <- .call(
        scoreMarkers,
        list(SummarizedExperiment::assay(x, assay.type), groups=groups),
        list(block=block, num.threads=num.threads),
        more.marker.args
    )

    if (is.character(extra.columns)) {
        extra.columns <- SummarizedExperiment::rowData(x)[,extra.columns,drop=FALSE]
    }
    formatScoreMarkersResult(res, extra.columns=extra.columns, order.by=order.by)
}

.findOrderBy <- function(marker.df, order.by) {
    if (isTRUE(order.by)) {
        # Find something decent to use for ordering.
        for (summ in c("mean", "median", "min.rank", "min", "max")) {
            for (eff in c("cohens.d", "auc", "delta.mean", "delta.detected")) {
                candidate <- paste0(eff, ".", summ)
                if (candidate %in% names(marker.df)) {
                    return(candidate)
                }
            }
        }
        return(NULL)
    } else if (isFALSE(order.by)) {
        return(NULL)
    } else {
        # No-op if it was already NULL or a string.
        return(order.by)
    }
}

#' @export
#' @rdname scoreMarkers.se
#' @importFrom S4Vectors List make_zero_col_DFrame cbind
#' @importClassesFrom S4Vectors DataFrame
formatScoreMarkersResult <- function(marker.res, extra.columns = NULL, order.by = TRUE) {
    effect.sizes <- c("cohens.d", "auc", "delta.mean", "delta.detected")
    summaries <- c("min", "mean", "median", "max", "quantile", "min.rank")
    output <- List()
    check.order.by <- FALSE

    for (group in as.character(marker.res$group.ids)) {
        current <- make_zero_col_DFrame(marker.res$nrow)
        rownames(current) <- marker.res$row.names
        if (!is.null(extra.columns)) {
            current <- cbind(current, extra.columns)
        }

        if ("mean" %in% names(marker.res)) {
            current$mean <- marker.res$mean[,group]
        }
        if ("detected" %in% names(marker.res)) {
            current$detected <- marker.res$detected[,group]
        }

        for (eff in effect.sizes) {
            eff.all <- marker.res[[eff]]
            if (is.null(eff.all)) {
                next
            }

            eff.df <- eff.all[[group]]
            for (summ in summaries) {
                eff.summ <- eff.df[[summ]]
                if (is.null(eff.summ)) {
                    next
                }

                if (is(eff.summ, "DataFrame")) {
                    colnames(eff.summ) <- sprintf("%s.%s.%s", eff, summ, colnames(eff.summ))
                    current <- cbind(current, eff.summ)
                } else {
                    current[[paste0(eff, ".", summ)]] <- eff.summ
                }
            }
        }

        if (!check.order.by) {
            order.by <- .findOrderBy(current, order.by)
            check.order.by <- TRUE
        }
        if (!is.null(order.by)) {
            dec <- !endsWith(order.by, ".min.rank")
            current <- current[order(current[[order.by]], decreasing=dec),,drop=FALSE]
        }

        output[[group]] <- current
    }

    output
}

#' @export
#' @rdname scoreMarkers.se
#' @importFrom utils head
previewMarkers <- function(
    marker.df, 
    columns = c("mean", "detected", lfc="delta.mean.mean"),
    pre.columns = NULL,
    post.columns = NULL,
    rows = 10,
    order.by = NULL,
    include.order.by = !is.null(order.by)
) {
    if (is.null(columns)) {
        columns <- as.character(0)
    }
    columns <- c(pre.columns, columns, post.columns)

    order.by <- .findOrderBy(marker.df, order.by)
    if (!is.null(order.by) && !isFALSE(include.order.by)) {
        columns <- c(columns, order.by)
        if (is.character(include.order.by)) {
            names(columns)[length(columns)] <- include.order.by
        }
    }

    columns <- columns[columns %in% colnames(marker.df)]
    df <- marker.df[,columns,drop=FALSE]
    if (!is.null(names(columns))) {
        replace <- names(columns) != ""
        colnames(df)[replace] <- names(columns)[replace]
    }

    if (!is.null(order.by)) {
        dec <- !endsWith(order.by, ".min.rank")
        o <- order(marker.df[[order.by]], decreasing=dec) # use marker.df in case the ordering statistic was renamed.
        if (!is.null(rows)) {
            o <- head(o, rows)
        }
        df <- df[o,,drop=FALSE]
    } else {
        if (!is.null(rows)) {
            df <- head(df, rows)
        }
    }

    df
}

# library(testthat); library(scrapper); source("test-se_scoreMarkers.R")

library(S4Vectors)

test_that("finding an ordering statistic works as expected", {
    expect_identical(scrapper:::.findOrderBy(DataFrame(cohens.d.mean=1, auc.max=2), TRUE), "cohens.d.mean")
    expect_identical(scrapper:::.findOrderBy(DataFrame(cohens.d.max=1, auc.median=2), TRUE), "auc.median")
    expect_null(scrapper:::.findOrderBy(DataFrame(cohens.d.max=1, auc.median=2), FALSE))
    expect_null(scrapper:::.findOrderBy(DataFrame(cohens.d.max=1, auc.median=2), NULL))
    expect_null(scrapper:::.findOrderBy(DataFrame(stuff=2), TRUE))
    expect_identical(scrapper:::.findOrderBy(DataFrame(cohens.d.max=1, auc.median=2), "foobar"), "foobar")
})

set.seed(6900)
library(SummarizedExperiment)
mat <- matrix(rpois(1000, 10), ncol=20)
rownames(mat) <- sprintf("gene%s", seq_len(nrow(mat)))
se <- SummarizedExperiment(list(counts=mat))
se <- normalizeRnaCounts.se(se)
groups <- rep(LETTERS[1:4], 5)

test_that("scoreMarkers.se works as expected", {
    out <- scoreMarkers.se(se, groups)
    expect_identical(names(out), LETTERS[1:4])

    for (g in LETTERS[1:4]) {
        df <- out[[g]]
        expect_identical(nrow(df), nrow(se))
        expect_true(all(rownames(mat) %in% rownames(df)))

        expect_type(df$cohens.d.mean, "double")
        expect_type(df$auc.median, "double")
        expect_type(df$delta.mean.min.rank, "integer")
        expect_type(df$mean, "double")
        expect_type(df$detected, "double")

        expect_false(is.unsorted(-df$cohens.d.mean)) # i.e., the default order-by choice.
    }

    # Also works with non-character groupings.
    out2 <- scoreMarkers.se(se, as.integer(factor(groups)))
    expect_identical(names(out2), as.character(1:4))
    for (g in 1:4) {
        expect_identical(out2[[as.character(g)]], out[[LETTERS[g]]])
    }
})

test_that("scoreMarkers.se works with extra columns", {
    rowData(se)$symbol <- sprintf("SYMBOL-%s", seq_len(nrow(mat)))
    out <- scoreMarkers.se(se, groups, extra.columns="symbol")
    for (g in LETTERS[1:4]) {
        df <- out[[g]]
        expect_identical(df$symbol, rowData(se)$symbol[match(rownames(df), rownames(se))])
    }
})

test_that("scoreMarkers.se works with quantiles", {
    out <- scoreMarkers.se(se, groups, more.marker.args=list(compute.summary.quantiles=c(0, 0.5, 1)))
    for (g in LETTERS[1:4]) {
        df <- out[[g]]
        expect_equal(df$auc.quantile.0.5, df$auc.median)
        expect_equal(df$cohens.d.quantile.0, df$cohens.d.min)
        expect_equal(df$delta.detected.quantile.1, df$delta.detected.max)
    }
})

test_that("scoreMarkers.se behaves without any metrics", {
    out <- scoreMarkers.se(se, groups, more.marker.args=list(compute.group.mean=FALSE, compute.group.detected=FALSE, compute.cohens.d=FALSE))
    for (g in LETTERS[1:4]) {
        df <- out[[g]]
        expect_identical(nrow(df), nrow(se))
        expect_true(all(rownames(mat) %in% rownames(df)))

        expect_null(df$cohens.d.mean)
        expect_null(df$mean)
        expect_null(df$detected)
        expect_type(df$auc.mean, "double")

        expect_false(is.unsorted(-df$auc.mean)) # i.e., the next default order-by choice.
    }
})

test_that("scoreMarkers.se sorts by min.rank correctly", {
    out <- scoreMarkers.se(se, groups, order.by="cohens.d.min.rank")
    for (g in LETTERS[1:4]) {
        df <- out[[g]]
        expect_false(is.unsorted(df$cohens.d.min.rank))
    }
})

test_that("scoreMarkers.se can still make DFs when no statistics are computed", {
    out <- scoreMarkers.se(se, groups, more.marker.args=list(
        compute.cohens.d=FALSE,
        compute.auc=FALSE,
        compute.group.mean=FALSE,
        compute.group.detected=FALSE,
        compute.delta.mean=FALSE,
        compute.delta.detected=FALSE
    ))

    for (g in LETTERS[1:4]) {
        expect_identical(nrow(out[[g]]), nrow(se))
        expect_identical(rownames(out[[g]]), rownames(se))
        expect_identical(ncol(out[[g]]), 0L)
    }
})

test_that("previewMarkers works as expected", {
    out <- scoreMarkers.se(se, groups)

    preview <- previewMarkers(out[[1]])
    expect_identical(colnames(preview), c("mean", "detected", "lfc"))
    expect_identical(nrow(preview), 10L)

    order_preview <- previewMarkers(out[[1]], order.by=TRUE)
    expect_identical(colnames(order_preview), c("mean", "detected", "lfc", "cohens.d.mean"))
    expect_identical(nrow(order_preview), 10L)

    noninc_preview <- previewMarkers(out[[1]], columns=NULL, order.by=TRUE, include.order.by=FALSE)
    expect_identical(colnames(noninc_preview), character(0))
    expect_identical(nrow(noninc_preview), 10L)
    expect_identical(rownames(order_preview), rownames(noninc_preview))

    eff_preview <- previewMarkers(out[[1]], order.by=TRUE, include.order.by="effect")
    expect_identical(eff_preview$effect, order_preview$cohens.d.mean)
    expect_identical(nrow(eff_preview), 10L)
    expect_identical(rownames(order_preview), rownames(eff_preview))

    more_preview <- previewMarkers(out[[1]], pre.columns=c("auc.median"), post.columns=c(DD="delta.detected.min"))
    expect_identical(more_preview, previewMarkers(out[[1]], columns=c("auc.median", "mean", "detected", lfc="delta.mean.mean", DD="delta.detected.min")))

    preview <- previewMarkers(out[[1]], rows=NULL)
    expect_identical(rownames(preview), rownames(out[[1]]))

    preview <- previewMarkers(out[[1]], order.by="auc.median")
    expect_identical(nrow(preview), 10L)

    preview <- previewMarkers(out[[1]], order.by="auc.median", rows=NULL)
    expect_identical(rownames(preview), rownames(out[[1]])[order(out[[1]]$auc.median, decreasing=TRUE)])

    preview <- previewMarkers(out[[1]], order.by="auc.min.rank", rows=NULL)
    expect_identical(rownames(preview), rownames(out[[1]])[order(out[[1]]$auc.min.rank)])
})

# library(testthat); library(scrapper); source("test-se_normalizeRnaCounts.R")

set.seed(12341234)
library(SummarizedExperiment)
mat <- matrix(rpois(1000, 5), ncol=10)
se <- SummarizedExperiment(list(counts=mat))

test_that("normalizeRnaCounts.se works as expected", {
    out <- normalizeRnaCounts.se(se)
    expect_type(out$sizeFactor, "double")
    expect_equal(mean(out$sizeFactor), 1)
    expect_s4_class(assay(out, "logcounts"), "DelayedArray")

    libs <- colSums(mat)
    out2 <- normalizeRnaCounts.se(se, size.factors=libs)
    expect_equal(out, out2)

    sf <- libs/mean(libs)
    out3 <- normalizeRnaCounts.se(se, size.factors=sf, center=FALSE)
    expect_equal(out, out3)

    out4 <- normalizeRnaCounts.se(se, factor.name=NULL)
    expect_null(out4$sizeFactor)
})

---
title: Using scrapper to analyze single-cell data
author: 
- name: Aaron Lun
  email: infinite.monkeys.with.keyboards@gmail.com
date: "Revised: December 26, 2025"
output:
  BiocStyle::html_document
package: scrapper
vignette: >
  %\VignetteIndexEntry{Using scrapper to analyze single-cell data}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}    
---

```{r, echo=FALSE, results="hide", message=FALSE}
require(knitr)
opts_chunk$set(error=FALSE, message=FALSE, warning=FALSE)

library(BiocStyle)
self <- Biocpkg("scrapper")
```

# Overview

`r self` implements bindings to C++ code for analyzing single-cell data, mostly from the [**libscran**](https://github.com/libscran) libraries.
Each function performs an individual analysis step ranging from quality control to clustering and marker detection.
Users can also conveniently perform end-to-end analyses with a single call to the `analyze.se()` function. 

# Basic analysis

Let's fetch a small single-cell RNA-seq dataset for demonstration purposes:

```{r}
library(scRNAseq)
sce.z <- ZeiselBrainData()
sce.z
```

We run it through the `r self` analysis pipeline.
This performs the usual steps - quality control, normalization, selection of highly variable genes, PCA, clustering, visualization and marker detection,
see the [OSCA book](https://bioconductor.org/books/release/OSCA/) for more theoretical details. 

```{r}
library(scrapper)
is.mito <- grep("^mt-", rownames(sce.z))
results <- analyze.se(
    sce.z,
    rna.qc.subsets=list(Mito=is.mito),
    num.threads=2 # limit number of threads to keep R CMD check happy.
)
results$x # most outputs are stored in a copy of the input SingleCellExperiment. 
```

Now we can have a look at some of the results.
For example, we can make a t-SNE plot:

```{r}
library(scater)
plotReducedDim(results$x, "TSNE", colour_by="graph.cluster")
```

We can also look at the markers defining each cluster:

```{r}
# Ordering by the median AUC across pairwise comparisons involving the cluster.
previewMarkers(results$markers$rna[[1]], order.by="auc.median")
```

`analyze.se()` can also be broken down to its component functions for further customization:

```{r}
test <- sce.z
test <- quickRnaQc.se(test, subsets=list(mt=is.mito), num.threads=2)
test <- test[,test$keep]
test <- normalizeRnaCounts.se(test, size.factors=test$sum)
test <- chooseRnaHvgs.se(test, num.threads=2)
test <- runPca.se(test, features=rowData(test)$hvg, num.threads=2)
test <- runAllNeighborSteps.se(test, num.threads=2)
markers <- scoreMarkers.se(test, groups=test$clusters, num.threads=2)
```

# Blocking on batches 

Let's fetch another brain dataset and combine it with the previous one.

```{r}
sce.t <- TasicBrainData()

# Only using the common genes in both datasets.
common <- intersect(rownames(sce.z), rownames(sce.t))
combined <- combineCols(sce.t[common,], sce.z[common,])
block <- rep(c("tasic", "zeisel"), c(ncol(sce.t), ncol(sce.z)))
```

We set `block=` in each step to account for the batch structure.
This ensures that the various calculations are not affected by inter-block differences.
It also uses MNN correction to batch effects in the low-dimensional space prior to further analyses.

```{r}
# No mitochondrial genes in the combined set, so we'll just skip it.
blocked_res <- analyze.se(combined, block=block, num.threads=2)
```

Hopefully, the correction is able to get rid of the batch effect and merge the datasets together. 

```{r}
plotReducedDim(blocked_res$x, "TSNE", colour_by="block")
```

We can also check the distribution of clusters across batches.
The presence of batch-specific clusters might indicate that the batch effect was not fully removed;
or they might represent cell types that are unique to one of the studies, who knows.

```{r}
table(blocked_res$x$graph.cluster, blocked_res$x$block)
```

Finally, some markers for another cluster:

```{r}
previewMarkers(blocked_res$markers$rna[[1]])
```

We can also compute pseudo-bulk profiles for each cluster-dataset combination, e.g., for differential expression analyses.

```{r}
aggregates <- aggregateAcrossCells.se(
    blocked_res$x,
    list(cluster=blocked_res$x$graph.cluster, dataset=blocked_res$x$block)
)
aggregates
```

# Combining multiple modalities

Let's fetch a single-cell dataset with both RNA-seq and CITE-seq data.
To keep the run-time short, we'll only consider the first 5000 cells.

```{r}
sce.s <- StoeckiusHashingData(mode=c("human", "adt1", "adt2"))
sce.s <- sce.s[,1:5000]

# Combining the various ADT-related alternative experiments into a single object.
altExp(sce.s, "all.adts") <- rbind(altExp(sce.s, "adt1"), altExp(sce.s, "adt2"))
altExp(sce.s, "adt1") <- altExp(sce.s, "adt2") <- NULL

sce.s
```

We specify `adt.altexp=` to indicate that one of the alternative experiments contains ADT data.
This instructs the analysis to use data from both modalities during clustering and visualization.

```{r}
is.mito <- grepl("^MT-", rownames(sce.s))
is.igg <- grepl("^IgG", rownames(altExp(sce.s, 'all.adts')))
multi_res <- analyze.se(
    sce.s,
    adt.altexp="all.adts",
    rna.qc.subsets=list(mito=is.mito),
    adt.qc.subsets=list(igg=is.igg),
    num.threads=2
)
```

Here's an obligatory plot:

```{r}
plotReducedDim(multi_res$x, "UMAP", colour_by="graph.cluster")
```

We can inspect the top markers among the ADTs for a cluster:

```{r}
# By default, markers are sorted by the mean Cohen's d.
# We just replace the column name when printing the dataframe here.
previewMarkers(multi_res$markers$adt[[5]], c(effect="cohens.d.mean"))
```

... along with the RNA markers, as before.

```{r}
# Using the delta-detected to focus on silent-to-expressed genes.
previewMarkers(multi_res$markers$rna[[5]], order.by="delta.detected.mean")
```

# Other useful functions

The `runAllNeighborSteps()` fucntion will run `runUmap()`, `runTsne()`, `buildSnnGraph()` and `clusterGraph()` in a single call.
This runs the UMAP/t-SNE iterations and the clustering in parallel to maximize use of multiple threads.

The `scoreGeneSet()` function will compute a gene set score based on the input expression matrix.
This can be used to summarize the activity of pathways into a single per-cell score for visualization.

The `subsampleByNeighbors()` function will deterministically select a representative subset of cells based on their local neighborhood.
This can be used to reduce the compute time of the various steps downstream of the PCA.

For CRISPR data, quality control can be performed using `computeCrisprQcMetrics()`, `suggestCrisprQcThresholds()` and `filterCrisprQcMetrics()`.
To normalize, we use size factors defined by centering the total sum of guide counts for each cell.

# Session information {-}

```{r}
sessionInfo()
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scoreMarkers.R
\name{reportGroupMarkerStatistics}
\alias{reportGroupMarkerStatistics}
\title{Report marker statistics for a single group}
\usage{
reportGroupMarkerStatistics(
  results,
  group,
  effect.sizes = NULL,
  summaries = NULL,
  include.mean = TRUE,
  include.detected = TRUE
)
}
\arguments{
\item{results}{Named list of marker statistics, typically generated by \code{\link{scoreMarkers}} with \code{all.pairwise=FALSE}.}

\item{group}{String or integer scalar specifying the group of interest.}

\item{effect.sizes}{Character vector specifying the effect sizes of interest.
If \code{NULL}, all effect sizes are reported in the returned data frame.}

\item{summaries}{Character vector specifying the summary statistics of interest.
If \code{NULL}, all summaries are reported in the returned data frame.}

\item{include.mean}{Logical scalar indicating whether the mean expression should be reported in the returned data frame.}

\item{include.detected}{Logical scalar indicating whether the proportion of detected cells should be reported in the returned data frame.}
}
\value{
Data frame where each row corresponds to a gene.
Each column contains the requested statistics for \code{group}.
Effect size summary columns are named as \code{<EFFECT>.<SUMMARY>}.
}
\description{
Combine all marker statistics for a single group into a data frame for easy inspection.
Users can pick one of the columns for sorting potential marker genes.
}
\seealso{
\code{\link{scoreMarkers}}, to generate \code{results}.

\code{\link{summarizeEffects}}, for the trade-offs between effect size summaries.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scrapper-package.R
\docType{package}
\name{scrapper-package}
\alias{scrapper}
\alias{scrapper-package}
\title{scrapper: Bindings to C++ Libraries for Single-Cell Analysis}
\description{
Implements R bindings to C++ code for analyzing single-cell (expression) data, mostly from various libscran libraries. Each function performs an individual step in the single-cell analysis workflow, ranging from quality control to clustering and marker detection. Additional wrappers are provided for easy construction of end-to-end workflows involving Bioconductor objects like SingleCellExperiments.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/libscran/scrapper}
  \item Report bugs at \url{https://github.com/libscran/scrapper/issues}
}

}
\author{
\strong{Maintainer}: Aaron Lun \email{infinite.monkeys.with.keyboards@gmail.com}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/se_clusterGraph.R
\name{clusterGraph.se}
\alias{clusterGraph.se}
\title{Graph-based clustering of cells in a SingleCellExperiment}
\usage{
clusterGraph.se(
  x,
  num.neighbors = 10,
  num.threads = 1,
  more.build.args = list(),
  method = "multilevel",
  resolution = NULL,
  more.cluster.args = list(),
  reddim.type = "PCA",
  output.name = "clusters",
  meta.name = NULL,
  graph.name = NULL
)
}
\arguments{
\item{x}{A \link[SingleCellExperiment]{SingleCellExperiment} object or one of its subclasses.
Rows correspond to genomic features and columns correspond to cells.}

\item{num.neighbors}{Number of neighbors for constructing the graph, passed to \code{\link{buildSnnGraph}}.}

\item{num.threads}{Number of threads for graph construction, passed to \code{\link{buildSnnGraph}}.}

\item{more.build.args}{Named list of further arguments to be passed to \code{\link{buildSnnGraph}}.}

\item{method}{Clustering method to use, passed to \code{\link{clusterGraph}}.}

\item{resolution}{Resolution for the community detection method in \code{\link{clusterGraph}}.
This is either passed to \code{multilevel.resolution} or \code{leiden.resolution} depending on \code{method}.}

\item{more.cluster.args}{Named list of further arguments to be passed to \code{\link{clusterGraph}}.}

\item{reddim.type}{Integer or string specifying the existing embedding in the \code{\link[SingleCellExperiment]{reducedDim}} of \code{x}.
Alternatively, a named integer or character vector of length 1,
where the name specifies an alternative experiment of \code{x} and the value is the name/index of a \code{reducedDim} entry in that alternative experiment.}

\item{output.name}{String containing the name of the column of the \code{\link[SummarizedExperiment]{colData}} in which to store the cluster assignments.}

\item{meta.name}{String containing the name of the \code{\link[S4Vectors]{metadata}} entry in which to store extra clustering output.
If \code{NULL}, no extra clustering output is stored.}

\item{graph.name}{String containing the name of the \code{\link[S4Vectors]{metadata}} entry in which to store the SNN graph.
If \code{NULL}, the SNN graph is not stored.}
}
\value{
\code{x} is returned with the cluster assignment for each cell stored in the \code{colData}.
Additional clustering output is stored in the \code{metadata}.
}
\description{
Construct a shared-nearest neighbor (SNN) graph from an existing low-dimensional embedding
by calling \code{\link{buildSnnGraph}} on a reduced dimension entry in a \link[SingleCellExperiment]{SingleCellExperiment}.
Then, apply community detection algorithms to obtain clusters of cells with \code{\link{clusterGraph}}.
}
\examples{
sce <- getTestRnaData.se("pca")
sce <- clusterGraph.se(sce)
table(sce$clusters)

}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modelGeneVariances.R
\name{modelGeneVariances}
\alias{modelGeneVariances}
\title{Model per-gene variances in expression}
\usage{
modelGeneVariances(
  x,
  block = NULL,
  block.average.policy = c("mean", "quantile"),
  block.weight.policy = c("variable", "equal", "none"),
  variable.block.weight = c(0, 1000),
  block.quantile = 0.5,
  mean.filter = TRUE,
  min.mean = 0.1,
  transform = TRUE,
  span = 0.3,
  use.min.width = FALSE,
  min.width = 1,
  min.window.count = 200,
  num.threads = 1
)
}
\arguments{
\item{x}{A matrix-like object where rows correspond to genes or genomic features and columns correspond to cells.
It is typically expected to contain log-expression values, e.g., from \code{\link{normalizeCounts}}.}

\item{block}{Factor specifying the block of origin (e.g., batch, sample) for each cell in \code{x}.
If provided, calculation of means/variances and trend fitting are performed within each block to ensure that block effects do not confound the estimates.
The weighted average of each statistic across all blocks is reported for each gene.
Alternatively \code{NULL}, if all cells are from the same block.}

\item{block.average.policy}{String specifying the policy to use for average statistics across blocks.
This can either be a (weighted) \code{"mean"} or a \code{"quantile"}.
Only used if \code{block} is not \code{NULL}.}

\item{block.weight.policy}{String specifying the policy to use for weighting different blocks when computing the average for each statistic.
See the argument of the same name in \code{\link{computeBlockWeights}} for more detail.
Only used if \code{block} is not \code{NULL} and \code{block.average.policy="mean"}.}

\item{variable.block.weight}{Numeric vector of length 2, specifying the parameters for variable block weighting.
See the argument of the same name in \code{\link{computeBlockWeights}} for more detail.
Only used if \code{block} is not \code{NULL}, \code{block.average.policy="mean"} and \code{block.weight.policy = "variable"}.}

\item{block.quantile}{Number specifying the probability of the quantile of statistics across blocks. 
Defaults to 0.5, i.e., the median of per-block statistics.
Only used if \code{block} is not \code{NULL} and \code{block.average.policy="quantile"}.}

\item{mean.filter}{Logical scalar indicating whether to filter on the means before trend fitting.
The assumption is that there is a bulk of low-abundance genes that are uninteresting and should be removed to avoid skewing the windows of the LOWESS smoother.}

\item{min.mean}{Numeric scalar specifying the minimum mean of genes to use in trend fitting.
Genes with lower means do not participate in the LOWESS fit, to ensure that windows are not skewed towards the majority of low-abundance genes.
Instead, the fitted values for these genes are defined by extrapolating the left edge of the fitted trend is extrapolated to the origin.
The default value is chosen based on the typical distribution of means of log-expression values across genes.
Only used if \code{mean.filter=TRUE}.}

\item{transform}{Logical scalar indicating whether a quarter-root transformation should be applied before trend fitting.
This transformation is copied from \code{limma::voom} and shrinks all values towards 1, flattening any sharp gradients in the trend for an easier fit.
The default of \code{TRUE} assumes that the variances are computed from log-expression values, in which case there is typically a strong \dQuote{hump} in the mean-variance relationship.}

\item{span}{Numeric scalar specifying the span of the LOWESS smoother, as a proportion of the total number of points.
Larger values improve stability at the cost of sensitivity to changes in low-density regions.
Ignored if \code{use.min.width=TRUE}.}

\item{use.min.width}{Logical scalar indicating whether a minimum width constraint should be applied to the LOWESS smoother.
This replaces the proportion-based span for defining each window.
Instead, the window for each point must be of a minimum width and is extended until it contains a minimum number of points. 
Setting this to `TRUE` ensures that sensitivity is maintained in the trend fit at low-density regions for the distribution of means, e.g., at high abundances.
It also avoids overfitting from very small windows in high-density intervals.}

\item{min.width}{Minimum width of the window to use when \code{use.min.width=TRUE}.
The default value is chosen based on the typical range of means in single-cell RNA-seq data.}

\item{min.window.count}{Minimum number of observations in each window.
This ensures that each window contains at least a given number of observations for a stable fit.
If the minimum width window contains fewer observations, it is extended using the standard LOWESS logic until the minimum number is achieved.
Only used if \code{use.min.width=TRUE}.}

\item{num.threads}{Integer scalar specifying the number of threads to use.}
}
\value{
A list containing \code{statistics}, a \link[S4Vectors]{DataFrame} with number of rows equal to the number of genes.
This contains the columns \code{means}, \code{variances}, \code{fitted} and \code{residuals},
each of which is a numeric vector containing the statistic of the same name across all genes.

If \code{block} is supplied, each of the column vectors described above contains the average across all blocks.
The list will also contain \code{per.block}, a list of DataFrames containing the equivalent statistics for each block;
and \code{block.ids}, a vector containing the identities of the unique blocks in the same order as \code{per.block}.
}
\description{
Model the per-gene variances as a function of the mean in single-cell expression data.
Highly variable genes can then be selected for downstream analyses.
}
\details{
We compute the mean and variance for each gene and fit a trend to the variances with respect to the means using \code{\link{fitVarianceTrend}}.
We assume that most genes at any given abundance are not highly variable, such that the fitted value of the trend is interpreted as the \dQuote{uninteresting} variance - 
this is mostly attributed to technical variation like sequencing noise, but can also represent constitutive biological noise like transcriptional bursting.
Under this assumption, the residual can be treated as a measure of biologically interesting variation.
Genes with large residuals can then be selected for downstream analyses, e.g., with \code{\link{chooseHighlyVariableGenes}}.
}
\examples{
library(Matrix)
x <- abs(rsparsematrix(1000, 100, 0.1) * 10)
out <- modelGeneVariances(x)
out

# Throwing in some blocking.
block <- sample(letters[1:4], ncol(x), replace=TRUE)
out <- modelGeneVariances(x, block=block)
out

}
\seealso{
The \code{model_gene_variances} function in \url{https://libscran.github.io/scran_variances/}.

\code{\link{chooseRnaHvgs.se}}, which computes the variances and trend from a \link[SummarizedExperiment]{SummarizedExperiment}.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aggregateAcrossGenes.R
\name{aggregateAcrossGenes}
\alias{aggregateAcrossGenes}
\title{Aggregate expression across genes}
\usage{
aggregateAcrossGenes(x, sets, average = FALSE, convert = TRUE, num.threads = 1)
}
\arguments{
\item{x}{A matrix-like object where rows correspond to genes or genomic features and columns correspond to cells.
Values are usually normalized expression values, possibly log-transformed depending on the application.}

\item{sets}{List of vectors where each entry corresponds to a gene set.
Each entry may be an integer vector of row indices, a logical vector of length equal to the number of rows, or a character vector of row names.
For integer and character vectors, duplicate elements are ignored.
For character vectors, any strings not present in \code{rownames(x)} are ignored.

Alternatively, each entry may be a list of two vectors.
The first vector should be either integer (row indices) or character (row names), specifying the genes in the set. 
The second vector should be numeric and of the same length as the first vector, specifying the weight associated with each gene.
If duplicate genes are present, only the first occurrence is used.
If the first vector contains gene names not present in \code{x}, those genes are ignored.}

\item{average}{Logical scalar indicating whether to compute the average rather than the sum.}

\item{convert}{Logical scalar indicating whether to convert gene identities to non-duplicate row indices in each entry of \code{sets}.
Can be set to \code{FALSE} for greater efficiency if the \code{sets} already contains non-duplicated integer vectors.}

\item{num.threads}{Integer specifying the number of threads to be used for aggregation.}
}
\value{
A list of length equal to that of \code{sets}.
Each entry is a numeric vector of length equal to the number of columns in \code{x}, 
containing the (weighted) sum/mean of expression values for the corresponding set across all cells.
}
\description{
Aggregate expression values across genes, potentially with weights.
This is typically used to summarize expression values for gene sets into a single per-cell score.
}
\examples{
# Mocking a matrix:
library(Matrix)
x <- round(abs(rsparsematrix(1000, 100, 0.1) * 100))

# Unweighted aggregation:
sets <- list(
   foo = sample(nrow(x), 20),
   bar = sample(nrow(x), 10)
)
agg <- aggregateAcrossGenes(x, sets)
str(agg)

# Weighted aggregation:
sets <- list(
   foo = list(sample(nrow(x), 20), runif(20)),
   bar = list(sample(nrow(x), 10), runif(10))
)
agg2 <- aggregateAcrossGenes(x, sets, average = TRUE)
str(agg2)

}
\seealso{
The \code{aggregate_across_genes} function in \url{https://libscran.github.io/scran_aggregate/}. 

\code{\link{aggregateAcrossGenes.se}}, to perform aggregation on a \link[SummarizedExperiment]{SummarizedExperiment}.

\code{\link{aggregateAcrossCells}}, to aggregate expression values across groups of cells.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/LogNormalizedMatrix.R
\name{LogNormalizedMatrix}
\alias{LogNormalizedMatrix}
\alias{LogNormalizedMatrix-class}
\alias{LogNormalizedMatrixSeed-class}
\alias{dim,LogNormalizedMatrixSeed-method}
\alias{dimnames,LogNormalizedMatrixSeed-method}
\alias{type,LogNormalizedMatrixSeed-method}
\alias{is_sparse,LogNormalizedMatrixSeed-method}
\alias{extract_array,LogNormalizedMatrixSeed-method}
\alias{extract_sparse_array,LogNormalizedMatrixSeed-method}
\alias{matrixClass,LogNormalizedMatrixSeed-method}
\alias{DelayedArray,LogNormalizedMatrixSeed-method}
\alias{initializeCpp,LogNormalizedMatrixSeed-method}
\alias{LogNormalizedMatrixSeed}
\title{Delayed log-normalization of a matrix}
\usage{
LogNormalizedMatrix(x, size.factors, pseudo.count = 1, log.base = 2)

LogNormalizedMatrixSeed(x, size.factors, pseudo.count = 1, log.base = 2)
}
\arguments{
\item{x}{Count matrix to be normalized.}

\item{size.factors}{Numeric vector of size factors, of length equal to the number of columns of \code{x}.}

\item{pseudo.count}{Number specifying the pseudo-count to add prior to log-transformation.}

\item{log.base}{Number specifying the base of the log-transformation.}
}
\value{
An instance of a LogNormalizedMatrix(Seed).
}
\description{
Delayed calculation of log-normalized expression values, typically returned by \code{\link{normalizeCounts}}.
}
\details{
This is based on the \pkg{DelayedArray} framework and
}
\examples{
mat <- matrix(rpois(1000, lambda=2), ncol=10)
sf <- centerSizeFactors(colSums(mat))
norm <- LogNormalizedMatrix(mat, sf)
norm

# Also works with sparse matrices.
library(Matrix)
smat <- abs(rsparsematrix(50, 20, density=0.21))
ssf <- centerSizeFactors(colSums(smat))
snorm <- LogNormalizedMatrix(smat, ssf)
snorm

}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/se_correctMnn.R
\name{correctMnn.se}
\alias{correctMnn.se}
\title{MNN correction on a SingleCellExperiment}
\usage{
correctMnn.se(
  x,
  block,
  BNPARAM = AnnoyParam(),
  num.threads = 1,
  more.mnn.args = list(),
  reddim.type = "PCA",
  output.name = "MNN",
  delayed.transpose = FALSE
)
}
\arguments{
\item{x}{A \link[SingleCellExperiment]{SingleCellExperiment} object or one of its subclasses.
Rows correspond to genomic features and columns correspond to cells.}

\item{block}{Block assignment for each cell, passed to \code{\link{correctMnn}}.}

\item{BNPARAM}{Algorithm for the nearest neighbor search, passed to \code{\link{correctMnn}}.}

\item{num.threads}{Number of threads, passed to \code{\link{correctMnn}}.}

\item{more.mnn.args}{Named list of additional arguments to pass to \code{\link{correctMnn}}.}

\item{reddim.type}{String or integer specifying the \code{\link[SingleCellExperiment]{reducedDim}} entry on which to perform MNN correction.
Alternatively, a named integer or character vector of length 1,
where the name specifies an alternative experiment of \code{x} and the value is the name/index of a \code{reducedDim} entry in that alternative experiment.}

\item{output.name}{String containing the name of the \code{\link[SingleCellExperiment]{reducedDim}} entry in which to store the corrected embedding.}

\item{delayed.transpose}{Logical scalar indicating whether to delay the transposition when storing coordinates in the \code{\link[SingleCellExperiment]{reducedDims}}.}
}
\value{
\code{x} is returned with the corrected embedding stored as a \code{reducedDim} entry.
}
\description{
Correct batch effects from an existing embedding with mutual nearest neighbors (MNNs),
by calling \code{\link{correctMnn}} on a reduced dimension entry of a \link[SingleCellExperiment]{SingleCellExperiment}.
}
\examples{
library(SingleCellExperiment)
sce <- getTestRnaData.se("pca")
# Treating the tissue of origin as the batch.
sce <- correctMnn.se(sce, sce$tissue)
reducedDimNames(sce)

}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/correctMnn.R
\name{correctMnn}
\alias{correctMnn}
\title{Batch correction with mutual nearest neighbors}
\usage{
correctMnn(
  x,
  block,
  num.neighbors = 15,
  num.steps = 1,
  merge.policy = c("rss", "size", "variance", "input"),
  num.mads = NULL,
  robust.iterations = NULL,
  robust.trim = NULL,
  mass.cap = NULL,
  order = NULL,
  reference.policy = NULL,
  BNPARAM = AnnoyParam(),
  num.threads = 1
)
}
\arguments{
\item{x}{Numeric matrix where rows are dimensions and columns are cells,
typically containing coordinates in a low-dimensional embedding (e.g., from \code{\link{runPca}}).}

\item{block}{Factor specifying the block of origin (e.g., batch, sample) for each cell in \code{x}.}

\item{num.neighbors}{Integer scalar specifying the number of neighbors in the various search steps.
Larger values improve the stability of the correction by increasing the number of MNN pairs and including more observations in each center of mass.
However, this comes at the cost of reduced resolution when matching subpopulations across batches.}

\item{num.steps}{Integer scalar specifying the number of steps for the recursive neighbor search to compute the center of mass.
Larger values mitigate the kissing effect but increase the risk of including inappropriately distant subpopulations into the center of mass.}

\item{merge.policy}{String specifying the policy to use to choose the order of batches to merge.
\itemize{
\item \code{"input"} will use the input order of the batches.
Observations in the last batch are corrected first, and then the second-last batch, and so on.
This allows users to control the merge order by simply changing the inputs.
\item \code{"size"} will merge batches in order of increasing size (i.e., the number of observations).
So, the smallest batch is corrected first while the largest batch is unchanged.
The aim is to lower compute time by reducing the number of observations that need to be reprocessed in later merge steps.
\item \code{"variance"} will merge batches in order of increasing variance between observations. 
So, the batch with the lowest variance is corrected first while the batch with the highest variance is unchanged.
The aim is to lower compute time by encouraging more observations to be corrected to the most variable batch, thus avoid reprocessing in later merge steps.
\item \code{"rss"} will merge batches in order of increasing residual sum of squares (RSS).
This is effectively a compromise between \code{"variance"} and \code{"size"}.
}}

\item{num.mads}{Deprecated and ignored.}

\item{robust.iterations}{Deprecated and ignored.}

\item{robust.trim}{Deprecated and ignored.}

\item{mass.cap}{Deprecated and ignored.}

\item{order}{Deprecated and ignored, the merge order is now always automatically determined.}

\item{reference.policy}{Deprecated, use \code{merge.policy} instead.}

\item{BNPARAM}{A \link[BiocNeighbors]{BiocNeighborParam} object specifying the nearest-neighbor algorithm to use.}

\item{num.threads}{Integer scalar specifying the number of threads to use.}
}
\value{
List containing \code{corrected}, a numeric matrix of the same dimensions as \code{x}, containing the corrected values.
}
\description{
Apply mutual nearest neighbor (MNN) correction to remove batch effects from a low-dimensional embedding.
}
\examples{
# Mocking up a dataset with multiple batches.
x <- matrix(rnorm(10000), nrow=10)
b <- sample(3, ncol(x), replace=TRUE)
x[,b==2] <- x[,b==2] + 3
x[,b==3] <- x[,b==3] + 5
lapply(split(colMeans(x), b), mean) # indeed the means differ...

corrected <- correctMnn(x, b)
str(corrected)
lapply(split(colMeans(corrected$corrected), b), mean) # now merged.

}
\references{
Haghverdi L, Lun ATL, Morgan MD, Marioni JC (2018).
Batch effects in single-cell RNA-sequencing data are corrected by matching mutual nearest neighbors.
\emph{Nat. Biotechnol.} 36(5):421-427
}
\seealso{
The \code{compute} function in \url{https://libscran.github.io/mnncorrect/}.

\code{\link{correctMnn.se}}, to perform MNN correction on a \link[SingleCellExperiment]{SingleCellExperiment}.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/computeBlockWeights.R
\name{computeBlockWeights}
\alias{computeBlockWeights}
\title{Compute block weights}
\usage{
computeBlockWeights(
  sizes,
  block.weight.policy = c("variable", "equal", "size", "none"),
  variable.block.weight = c(0, 1000)
)
}
\arguments{
\item{sizes}{Numeric vector containing the size of (i.e., number of cells in) each block.}

\item{block.weight.policy}{String specifying the policy to use for weighting different blocks. 
This should be one of:
\itemize{
\item \code{"size"}: the contribution of each block is proportional to its size.
\code{"none"} is also a deprecated alias for \code{"size"}. 
\item \code{"equal"}: blocks are equally weighted regardless of their size.
The exception is that of empty blocks with no cells, which receive zero weight.
\item \code{"variable"}: blocks are equally weighted past a certain threshold size.
Below that size, the contribution of each block is proportional to its size.
This avoids outsized contributions from very large blocks.
}}

\item{variable.block.weight}{Numeric vector of length 2, specifying the parameters for variable block weighting.
The first and second values are used as the lower and upper bounds, respectively, for the variable weight calculation.
Only used if \code{block.weight.policy = "variable"}.}
}
\value{
Numeric vector containing the relative block weights.
}
\description{
Compute a weight for each block based on the number of cells in each block.
This is typically used to aggregate statistics across blocks, e.g., with weighted sums/averages.
}
\examples{
computeBlockWeights(c(1, 10, 100, 1000, 10000))
computeBlockWeights(c(1, 10, 100, 1000, 10000), block.weight.policy="equal")
computeBlockWeights(c(1, 10, 100, 1000, 10000), variable.block.weight=c(50, 5000))

}
\seealso{
The \code{compute_weights} function from \url{https://libscran.github.io/scran_blocks/}.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/combineFactors.R
\name{combineFactors}
\alias{combineFactors}
\title{Combine multiple factors}
\usage{
combineFactors(factors, keep.unused = FALSE)
}
\arguments{
\item{factors}{An ordinary list or \link[S4Vectors]{List} of vectors or factors of the same length.
Corresponding elements across all vectors/factors represent the combination of levels for a single observation.
For factors, any existing levels are respected.
For other vectors, the sorted and unique values are used as levels.

Alternatively, a data frame or \link[S4Vectors]{DataFrame} where each column is a vector or factor and each row corresponds to an observation.}

\item{keep.unused}{Logical scalar indicating whether to report unused combinations of levels.}
}
\value{
List containing:
\itemize{
\item \code{levels}, a \link[S4Vectors]{DataFrame} containing the sorted and unique combinations of levels from \code{factors}.
Each column corresponds to a factor in \code{factors} while each row corresponds to a unqiue combination.
\item \code{index}, an integer vector specifying the index into \code{levels} for each observation.
}
For observation \code{i} and factor \code{j}, \code{levels[[[j]][index[i]]} will recover \code{factors[[j]][i]}.
}
\description{
Combine multiple factors into a single factor where each level of the latter is a unique combination of levels from the former.
}
\examples{
combineFactors(list(
    sample(LETTERS[1:5], 100, replace=TRUE),
    sample(3, 100, replace=TRUE)
))

combineFactors(list(
    factor(sample(LETTERS[1:5], 10, replace=TRUE), LETTERS[1:5]),
    factor(sample(5, 10, replace=TRUE), 1:5)
), keep.unused=TRUE)

}
\seealso{
The \code{combine_to_factor} function in \url{https://libscran.github.io/factorize/}.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scoreMarkers.R
\name{scoreMarkers}
\alias{scoreMarkers}
\title{Score marker genes}
\usage{
scoreMarkers(
  x,
  groups,
  block = NULL,
  block.average.policy = c("mean", "quantile"),
  block.weight.policy = c("variable", "equal", "none"),
  variable.block.weight = c(0, 1000),
  block.quantile = 0.5,
  compute.group.mean = TRUE,
  compute.group.detected = TRUE,
  compute.delta.mean = TRUE,
  compute.delta.detected = TRUE,
  compute.cohens.d = TRUE,
  compute.auc = TRUE,
  compute.summary.min = TRUE,
  compute.summary.mean = TRUE,
  compute.summary.median = TRUE,
  compute.summary.max = TRUE,
  compute.summary.quantiles = NULL,
  compute.summary.min.rank = TRUE,
  threshold = 0,
  all.pairwise = FALSE,
  top.index.only = FALSE,
  min.rank.limit = 500,
  num.threads = 1
)
}
\arguments{
\item{x}{A matrix-like object where rows correspond to genes or genomic features and columns correspond to cells.
It is typically expected to contain log-expression values, e.g., from \code{\link{normalizeCounts}}.}

\item{groups}{A vector specifying the group assignment for each cell in \code{x}.}

\item{block}{Factor specifying the block of origin (e.g., batch, sample) for each cell in \code{x}.
If provided, comparisons are performed within each block to ensure that block effects do not confound the estimates.
The weighted average of the effect sizes across all blocks is reported for each gene.
Alternatively \code{NULL}, if all cells are from the same block.}

\item{block.average.policy}{String specifying the policy to use for average statistics across blocks.
This can either be a (weighted) \code{"mean"} or a \code{"quantile"}.
Only used if \code{block} is not \code{NULL}.}

\item{block.weight.policy}{String specifying the policy to use for weighting different blocks when computing the average for each statistic.
See the argument of the same name in \code{\link{computeBlockWeights}} for more detail.
Only used if \code{block} is not \code{NULL}.}

\item{variable.block.weight}{Numeric vector of length 2, specifying the parameters for variable block weighting.
See the argument of the same name in \code{\link{computeBlockWeights}} for more detail.
Only used if \code{block} is not \code{NULL} and \code{block.weight.policy = "variable"}.}

\item{block.quantile}{Number specifying the probability of the quantile of statistics across blocks. 
Defaults to 0.5, i.e., the median of per-block statistics.
Only used if \code{block} is not \code{NULL} and \code{block.average.policy="quantile"}.}

\item{compute.group.mean}{Logical scalar indicating whether to compute the group-wise mean expression for each gene.}

\item{compute.group.detected}{Logical scalar indicating whether to compute the group-wise proportion of detected cells for each gene.}

\item{compute.delta.mean}{Logical scalar indicating whether to compute the delta-means, i.e., the log-fold change when \code{x} contains log-expression values.}

\item{compute.delta.detected}{Logical scalar indicating whether to compute the delta-detected, i.e., differences in the proportion of cells with detected expression.}

\item{compute.cohens.d}{Logical scalar indicating whether to compute Cohen's d.}

\item{compute.auc}{Logical scalar indicating whether to compute the AUC.
Setting this to \code{FALSE} can improve speed and memory efficiency.}

\item{compute.summary.min}{Boolean specifying whether to compute the minimum as a summary statistic for each effect size.
Only used if \code{all.pairwise=FALSE}.}

\item{compute.summary.mean}{Boolean specifying whether to compute the mean as a summary statistic for each effect size.
Only used if \code{all.pairwise=FALSE}.}

\item{compute.summary.median}{Boolean specifying whether to compute the median as a summary statistic for each effect size.
Only used if \code{all.pairwise=FALSE}.}

\item{compute.summary.max}{Boolean specifying whether to compute the maximum as a summary statistic for each effect size.
Only used if \code{all.pairwise=FALSE}.}

\item{compute.summary.quantiles}{Numeric scalars containing the probabilities of quantiles to compute as summary statistics for each effect size.
If \code{NULL}, no quantiles are computed.
Only used if \code{all.pairwise=FALSE}.}

\item{compute.summary.min.rank}{Boolean specifying whether to compute the mininum rank as a summary statistic for each effect size.
If \code{NULL}, no quantiles are computed.
Only used if \code{all.pairwise=FALSE}.}

\item{threshold}{Non-negative numeric scalar specifying the minimum threshold on the differences in means (i.e., the log-fold change, if \code{x} contains log-expression values). 
This is incorporated into the effect sizes for Cohen's d and the AUC.
Larger thresholds will favor genes with large differences at the expense of genes with low variance that would otherwise have comparable effect sizes.}

\item{all.pairwise}{Logical scalar indicating whether to report the effect sizes for every pairwise comparison between groups.
Alternatively, an integer scalar indicating the number of top markers to report from each pairwise comparison between groups.
If \code{FALSE}, only the summary statistics are reported.}

\item{top.index.only}{Boolean indicating whether to only report the indices of the top genes when \code{all.pairwise} is an integer.
This is more efficient when the effect sizes of the top genes are not required.
Ignored for all other values of \code{all.pairwise}.}

\item{min.rank.limit}{Integer scalar specifying the maximum value of the min-rank to report.
Lower values improve memory efficiency at the cost of discarding information about lower-ranked genes.
Only used if \code{all.pairwise=FALSE}.}

\item{num.threads}{Integer scalar specifying the number of threads to use.}
}
\value{
A named list containing:
\itemize{
\item \code{nrow}, integer specifying the number of rows in \code{x}.
\item \code{row.names}, character vector or \code{NULL} containing the row names of \code{x}.
\item \code{group.ids}, vector contaning the identities of the unique groups.
\item \code{mean}, a numeric matrix containing the mean expression for each group.
Each row is a gene and each column is a group in \code{group.ids}.
Omitted if \code{compute.group.mean=FALSE}.
\item \code{detected}, a numeric matrix containing the proportion of detected cells in each group.
Each row is a gene and each column is a group in \code{group.ids}.
Omitted if \code{compute.group.detected=FALSE}.
}

If \code{all.pairwise=FALSE}, the list also contains:
\itemize{
\item \code{cohens.d}, a list of \link[S4Vectors]{DataFrame}s where each DataFrame corresponds to a group in \code{group.ids}.
Each row of a DataFrame represents a gene, while each column contains a summary of Cohen's d from pairwise comparisons to all other groups.
This includes \code{min}, \code{mean}, \code{median}, \code{max}, \code{quantile.*} and \code{min.rank} - check out \code{?\link{summarizeEffects}} for details.
Omitted if \code{compute.cohens.d=FALSE}.
\item \code{auc}, a list like \code{cohens.d} but containing the summaries of the AUCs from each pairwise comparison.
Omitted if \code{compute.auc=FALSE}.
\item \code{delta.mean}, a list like \code{cohens.d} but containing the summaries of the delta-mean from each pairwise comparison.
Omitted if \code{compute.delta.mean=FALSE}.
\item \code{delta.detected}, a list like \code{cohens.d} but containing the summaries of the delta-detected from each pairwise comparison.
Omitted if \code{compute.delta.detected=FALSE}.
}

If \code{all.pairwise=TRUE}, the list also contains:
\itemize{
\item \code{cohens.d}, a 3-dimensional numeric array containing the Cohen's d from each pairwise comparison between groups.
The extents of the first two dimensions are equal to the number of groups in \code{group.ids}, while the extent of the final dimension is equal to the number of genes.
The entry \code{cohens.d[i, j, k]} represents Cohen's d from the comparison of group \code{group.ids[j]} over group \code{group.ids[i]} for gene \code{k}.
Omitted if \code{compute.cohens.d=FALSE}.
\item \code{auc}, an array like \code{cohens.d} but containing the AUCs from each pairwise comparison.
Omitted if \code{compute.auc=FALSE}.
\item \code{delta.mean}, an array like \code{cohens.d} but containing the delta-mean from each pairwise comparison.
Omitted if \code{compute.delta.mean=FALSE}.
\item \code{delta.detected}, an array like \code{cohens.d} but containing the delta-detected from each pairwise comparison.
Omitted if \code{compute.delta.detected=FALSE}.
}

If \code{all.pairwise} is an integer and \code{top.index.only=FALSE}, the list also contains:
\itemize{
\item \code{cohens.d}, a list of list of \link[S4Vectors]{DataFrame}s containing the top genes with the largest Cohen's d for each pairwise comparison.
Specifically, \code{cohens.d[[i]][[j]]} is a DataFrame that contains the top \code{all.pairwise} genes from the comparison of group \code{group.ids[i]} over group \code{group.ids[j]}. 
Each DataFrame contains an \code{index} column, the row index of the gene; and an \code{effect} column, the Cohen's d for that gene.
Omitted if \code{compute.cohens.d=FALSE}.
\item \code{auc}, a list of list of DataFrames like \code{cohens.d} but containing the AUCs from each pairwise comparison.
Omitted if \code{compute.auc=FALSE}.
\item \code{delta.mean}, a list of list of DataFrames like \code{cohens.d} but containing the delta-mean from each pairwise comparison.
Omitted if \code{compute.delta.mean=FALSE}.
\item \code{delta.detected}, a list of list of DataFrames like \code{cohens.d} but containing the delta-detected from each pairwise comparison.
Omitted if \code{compute.delta.detected=FALSE}.
}

If \code{all.pairwise} is an integer and \code{top.index.only=TRUE}, the list also contains:
\itemize{
\item \code{cohens.d}, a list of list of integer vectors containing the row indices of the top genes with the largest Cohen's d for each pairwise comparison.
Specifically, \code{cohens.d[[i]][[j]]} is a vector that contains the top \code{all.pairwise} genes from the comparison of group \code{group.ids[i]} over group \code{group.ids[j]}. 
Omitted if \code{compute.cohens.d=FALSE}.
\item \code{auc}, a list of list of DataFrames like \code{cohens.d} but containing the AUCs from each pairwise comparison.
Omitted if \code{compute.auc=FALSE}.
\item \code{delta.mean}, a list of list of DataFrames like \code{cohens.d} but containing the delta-mean from each pairwise comparison.
Omitted if \code{compute.delta.mean=FALSE}.
\item \code{delta.detected}, a list of list of DataFrames like \code{cohens.d} but containing the delta-detected from each pairwise comparison.
Omitted if \code{compute.delta.detected=FALSE}.
}
}
\description{
Score marker genes for each group using a variety of effect sizes from pairwise comparisons between groups.
This includes Cohen's d, the area under the curve (AUC), the difference in the means (delta-mean) and the difference in the proportion of detected cells (delta-detected).
For each group, the strongest markers are those genes with the largest effect sizes (i.e., upregulated) when compared to all other groups.
}
\section{Choice of effect size}{

The delta-mean is the difference in the mean expression between groups.
This is fairly straightforward to interpret - a positive delta-mean corresponds to increased expression in the first group compared to the second. 
The delta-mean can also be treated as the log-fold change if the input matrix contains log-transformed normalized expression values.

The delta-detected is the difference in the proportion of cells with detected expression between groups.
This lies between 1 and -1, with the extremes occurring when a gene is silent in one group and detected in all cells of the other group.
For this interpretation, we assume that the input matrix contains non-negative expression values, where a value of zero corresponds to lack of detectable expression.

Cohen's d is the standardized difference between two groups.
This is defined as the difference in the mean for each group scaled by the average standard deviation across the two groups.
(Technically, we should use the pooled variance; however, this introduces some unintuitive asymmetry depending on the variance of the larger group, so we take a simple average instead.)
A positive value indicates that the gene has increased expression in the first group compared to the second.
Cohen's d is analogous to the t-statistic in a two-sample t-test and avoids spuriously large effect sizes from comparisons between highly variable groups.
We can also interpret Cohen's d as the number of standard deviations between the two group means.

The area under the curve (AUC) is the probability that a randomly chosen observation in one group is greater than a randomly chosen observation in the other group. 
Values greater than 0.5 indicate that a gene is upregulated in the first group.
The AUC is closely related to the U-statistic used in the Wilcoxon rank sum test. 
The key difference between the AUC and Cohen's d is that the former is less sensitive to the variance within each group, e.g.,
if two distributions exhibit no overlap, the AUC is the same regardless of the variance of each distribution. 
This may or may not be desirable as it improves robustness to outliers but reduces the information available to obtain a fine-grained ranking.
}

\section{With a minimum change threshold}{

Setting a minimum change threshold (i.e., \code{threshold}) prioritizes genes with large shifts in expression instead of those with low variances.
Currently, only positive thresholds are supported, which focuses on genes that are upregulated in the first group compared to the second.
The effect size definitions are generalized when testing against a non-zero threshold:
\itemize{
\item Cohen's d is redefined as the standardized difference between the difference in means and the specified threshold,
analogous to the TREAT method from the \pkg{limma} package.
Large positive values are only obtained when the observed difference in means is significantly greater than the threshold.
For example, if we had a threshold of 2 and we obtained a Cohen's d of 3, this means that the observed difference in means was 3 standard deviations greater than 2.
Note that a negative Cohen's d cannot be intepreted as downregulation, as the difference in means may still be positive but less than the threshold.
\item The AUC is generalized to the probability of obtaining a random observation in one group that is greater than a random observation plus the threshold in the other group.
For example, if we had a threshold of 2 and we obtained an AUC of 0.8, this means that, 80% of the time,
the random observation from the first group would be greater than a random observation from the second group by 2 or more.
Again, AUCs below 0.5 cannot be interpreted as downregulation, as it may be caused by a positive shift that is less than the threshold.
}
}

\examples{
# Mocking a matrix:
library(Matrix)
x <- round(abs(rsparsematrix(1000, 100, 0.1) * 100))
normed <- normalizeCounts(x, size.factors=centerSizeFactors(colSums(x)))

# Compute marker summaries for each group:
g <- sample(letters[1:4], ncol(x), replace=TRUE)
scores <- scoreMarkers(normed, g)
names(scores)
head(scores$mean)
head(scores$cohens.d[["a"]])

# Report marker statistics for a single group:
reportGroupMarkerStatistics(scores, "b")

}
\seealso{
The \code{score_markers_summary}, \code{score_markers_pairwise} and \code{score_markers_best} functions in \url{https://libscran.github.io/scran_markers/}.
See their blocked equivalents (e.g., \code{score_markers_summary_blocked}) when \code{block} is specified.

\code{\link{summarizeEffects}}, to summarize the pairwise effects returned when \code{all.pairwise=TRUE}.

\code{\link{reportGroupMarkerStatistics}}, to consolidate the statistics for a single group into its own data frame.

\code{\link{scoreMarkers.se}}, to score markers from a \link[SummarizedExperiment]{SummarizedExperiment}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/se_aggregateAcrossCells.R
\name{aggregateAcrossCells.se}
\alias{aggregateAcrossCells.se}
\alias{aggregateColData}
\title{Aggregate expression across cells in a SummarizedExperiment}
\usage{
aggregateAcrossCells.se(
  x,
  factors,
  num.threads = 1,
  more.aggr.args = list(),
  assay.type = "counts",
  output.prefix = "factor.",
  counts.name = "counts",
  meta.name = "aggregated",
  include.coldata = TRUE,
  more.coldata.args = list(),
  altexps = NULL,
  copy.altexps = FALSE
)

aggregateColData(coldata, index, number, only.atomic = TRUE, placeholder = NA)
}
\arguments{
\item{x}{A \link[SummarizedExperiment]{SummarizedExperiment} object or one of its subclasses.
Rows correspond to genes and columns correspond to cells.}

\item{factors}{List or data frame (or their equivalents from \pkg{S4Vectors}) containing grouping factors, see \code{\link{aggregateAcrossCells}} for more details.

Alternatively, an atomic vector or factor representing a single variable.}

\item{num.threads}{Number of threads, passed to \code{\link{aggregateAcrossCells}}.}

\item{more.aggr.args}{Named list of additional arguments to pass to \code{\link{aggregateAcrossCells}}.}

\item{assay.type}{Integer or string specifying the assay of \code{x} to be aggregated.}

\item{output.prefix}{String specifying a prefix to add to the names of the \code{link[SummarizedExperiment]{colData}} columns storing the factor combinations. 
If \code{NULL}, no prefix is added.}

\item{counts.name}{String specifying the name of the \code{\link[SummarizedExperiment]{colData}} column in which to store the cell count for each factor combination. 
If \code{NULL}, the cell counts are not reported.}

\item{meta.name}{String specifying the name of the \code{\link[S4Vectors]{metadata}} entry in which to store additional outputs like the combination indices.
If \code{NULL}, additional outputs are not reported.}

\item{include.coldata}{Logical scalar indicating whether to add the aggregated \code{colData} from \code{x} to the output.}

\item{more.coldata.args}{Named list of additional arguments to pass to \code{aggregateColData}.
Only relevant if \code{include.coldata=TRUE}.}

\item{altexps}{Unnamed integer or character vector specifying the indices/names of alternative experiments of \code{x} to aggregate.
The aggregated assay from each alternative experiment is determined by \code{assay.type}.

Alternatively, this may be a named integer or character vector.
Each name specifies an alternative experiment while each value is the index/name of the assay to aggregate from that experiment.

Only relevant if \code{x} is a \link[SingleCellExperiment]{SingleCellExperiment}.}

\item{copy.altexps}{Logical scalar indicating whether to copy the \code{colData} and \code{metadata} of the output SingleCellExperiment into each of its alternative experiments.}

\item{coldata}{\link[S4Vectors]{DataFrame} of column data, containing one row for each cell.}

\item{index}{Integer vector containing the index of the factor combination to which each cell in \code{coldata} was assigned.}

\item{number}{Integer specifying the total number of unique factor combinations.
All elements of \code{index} should be less than or equal to \code{number}.}

\item{only.atomic}{Logical scalar specifying whether to skip non-atomic, non-factor columns.}

\item{placeholder}{Placeholder value to store in the output column when a factor combination does not have a single unique value.}
}
\value{
For \code{aggregateAcrossCells.se}, a SummarizedExperiment is returned where each column corresponds to a factor combination.
Each row corresponds to a gene in \code{x} and the \code{\link[SummarizedExperiment]{rowData}} is taken from \code{x}.
The assays contain the sum of counts (\code{"sums"}) and the number of detected cells (\code{"detected"}) in each combination for each gene.
The \code{colData} contains:
\itemize{
\item The factor combinations, with column names prefixed by \code{output.prefix}.
\item The cell count for each combination, named by \code{counts.name}.
\item Additional \code{colData} from \code{x} if \code{include.coldata=TRUE}.
This is aggregated with \code{aggregateColData} on the combination indices.
}
The metadata contains a list named as \code{meta.name}, containing a \code{index} integer vector of length equal to the number of cells in \code{x}.
Each entry of this vector is an index of the factor combination (i.e., column of the output object) to which the corresponding cell was assigned.

If \code{altexps} is specified, a SingleCellExperiment is returned instead.
The same aggregation for the main experiment is applied to each alternative experiment.
If \code{copy.altexps=TRUE}, the \code{colData} for each alternative experiment will contain a copy of the factor combinations and cell counts,
and the \code{metadata} will contain a copy of the index vector.

For \code{aggregateColData}, a \link[S4Vectors]{DataFrame} is returned with number of rows equal to \code{number}.
Each atomic or factor column in \code{coldata} is represented by a column in the output DataFrame.
In each column, the \code{j}-th entry is equal to the unique value of all rows where \code{index == j},
or \code{placeholder} if there is not exactly one unique value.
If \code{only.atomic=FALSE}, any non-atomic/non-factor columns of \code{coldata} are represented in the output DataFrame by a vector of \code{placeholder} values.
If \code{only.atomic=TRUE}, any non-atomic/non-factor columns of \code{coldata} are skipped.
}
\description{
Aggregate expression values across groups of cells for each gene,
by calling \code{\link{aggregateAcrossCells}} on an assay in a \link[SummarizedExperiment]{SummarizedExperiment}.
}
\examples{
library(SingleCellExperiment)
sce <- getTestRnaData.se("start")
aggr <- aggregateAcrossCells.se(sce, sce$level1class)
head(assay(aggr))
colData(aggr)

# We can also aggregate within alternative experiments.
aggr2 <- aggregateAcrossCells.se(sce, sce$level1class, altexps="ERCC")
head(assay(altExp(aggr2, "ERCC")))

}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aggregateAcrossCells.R
\name{aggregateAcrossCells}
\alias{aggregateAcrossCells}
\title{Aggregate expression across cells}
\usage{
aggregateAcrossCells(
  x,
  factors,
  compute.sum = TRUE,
  compute.detected = TRUE,
  compute.median = TRUE,
  num.threads = 1
)
}
\arguments{
\item{x}{A matrix-like object where rows correspond to genes or genomic features and columns correspond to cells.
Values are typically expected to be counts.}

\item{factors}{A list or data frame (or their equivalents from \pkg{S4Vectors}) containing one or more grouping factors, see \code{\link{combineFactors}}.}

\item{compute.sum}{Boolean indicating whether to compute the sum in each group.}

\item{compute.detected}{Boolean indicating whether to compute the number of cells with detected expression in each group.}

\item{compute.median}{Boolean indicating whether to compute the median in each group.}

\item{num.threads}{Integer specifying the number of threads to be used for aggregation.}
}
\value{
A list containing:
\itemize{
\item \code{sums}, a numeric matrix where each row corresponds to a gene and each column corresponds to a unique combination of levels from \code{factors}.
Each entry contains the summed expression across all cells with that combination. 
Omitted if \code{compute.sum=FALSE}.
\item \code{detected}, an integer matrix where each row corresponds to a gene and each column corresponds to a unique combination of levels from \code{factors}.
Each entry contains the number of cells with detected (i.e., positive) expression in that combination.
Omitted if \code{compute.detected=FALSE}.
\item \code{medians}, a numeric matrix where each row corresponds to a gene and each column corresponds to a unique combination of levels from \code{factors}.
Each entry contains the median expression across all cells with that combination. 
Omitted if \code{compute.median=FALSE}.
\item \code{combinations}, a \link[S4Vectors]{DataFrame} containing the unique combination of levels from \code{factors}.
Rows correspond to columns of \code{sums} and \code{detected}, while columns correspond to the factors in \code{factors}.
\item \code{counts}, the number of cells associated with each combination.
Each entry corresponds to a row of \code{combinations}.
\item \code{index}, an integer vector of length equal to the number of cells in \code{x}.
This specifies the combination in \code{combinations} to which each cell was assigned.
}
}
\description{
Aggregate expression values across cells based on one or more grouping factors.
This is usually applied to a count matrix to create pseudo-bulk profiles for each cluster/sample combination.
These profiles can then be used as if they were counts from bulk data, e.g., for differential analyses with \pkg{edgeR}.
}
\examples{
# Mocking a matrix:
library(Matrix)
x <- round(abs(rsparsematrix(1000, 100, 0.1) * 100))

# Simple aggregation:
clusters <- sample(LETTERS, 100, replace=TRUE)
agg <- aggregateAcrossCells(x, list(cluster=clusters))
str(agg)

# Multi-factor aggregation
samples <- sample(1:5, 100, replace=TRUE)
agg2 <- aggregateAcrossCells(x, list(cluster=clusters, sample=samples))
str(agg2)

}
\seealso{
The \code{aggregate_across_cells} function in \url{https://libscran.github.io/scran_aggregate/}.

\code{\link{aggregateAcrossCells.se}}, to perform aggregation on a \link[SummarizedExperiment]{SummarizedExperiment}.

\code{\link{aggregateAcrossGenes}}, to aggregate expression values across gene sets.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sanitizeSizeFactors.R
\name{sanitizeSizeFactors}
\alias{sanitizeSizeFactors}
\title{Sanitize size factors}
\usage{
sanitizeSizeFactors(
  size.factors,
  replace.zero = TRUE,
  replace.negative = TRUE,
  replace.infinite = TRUE,
  replace.nan = TRUE
)
}
\arguments{
\item{size.factors}{Numeric vector of size factors across cells.}

\item{replace.zero}{Logical scalar indicating whether to replace size factors of zero with the lowest positive factor in \code{size.factors}.
This ensures that the normalized values will be large to reflect the extremity of the scaling, but still finite for sensible downstream processing. 
If \code{FALSE}, zeros are retained.}

\item{replace.negative}{Logical scalar indicating whether to replace negative size factors with the lowest positive factor in \code{size.factors}.
This ensures that the normalized values will be large to reflect the extremity of the scaling, but still finite for sensible downstream processing. 
If \code{FALSE}, negative values are retained.}

\item{replace.infinite}{Logical scalar indicating whether to replace infinite size factors with the largest positive factor in \code{size.factors}.
This ensures that any normalized values will be, at least, finite; the choice of a relatively large replacement value reflects the extremity of the scaling.
If \code{FALSE}, infinite values are retained.}

\item{replace.nan}{Logical scalar indicating whether to replace NaN size factors with unity, e.g., scaling normalization is a no-op.
If \code{FALSE}, NaN values are retained.}
}
\value{
Numeric vector of length equal to \code{size.factors}, containing the sanitized size factors.
}
\description{
Replace invalid size factors, i.e., zero, negative, infinite or NaN values.
Such size factors can occasionally arise if, e.g., insufficient quality control was performed upstream.
Removing them ensures that the normalized values from \code{\link{normalizeCounts}} remain finite for sensible downstream processing.
}
\examples{
sf <- 2^rnorm(100)
sf[1] <- 0
sf[2] <- -1
sf[3] <- Inf
sf[4] <- NaN
sanitizeSizeFactors(sf)

}
\seealso{
The \code{sanitize_size_factors} function in \url{https://libscran.github.io/scran_norm/}.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/runPca.R
\name{runPca}
\alias{runPca}
\title{Principal components analysis}
\usage{
runPca(
  x,
  number = 25,
  scale = FALSE,
  block = NULL,
  block.weight.policy = c("variable", "equal", "none"),
  variable.block.weight = c(0, 1000),
  components.from.residuals = FALSE,
  subset = NULL,
  extra.work = 7,
  iterations = 1000,
  seed = 5489,
  realized = TRUE,
  warn = TRUE,
  num.threads = 1
)
}
\arguments{
\item{x}{A matrix-like object where rows correspond to genes or genomic features and columns correspond to cells.
Typically, the matrix is expected to contain log-expression values (see \code{\link{normalizeCounts}}) for \dQuote{interesting} genes (see \code{\link{chooseHighlyVariableGenes}}).}

\item{number}{Integer scalar specifying the number of top PCs to retain.
More PCs will capture more biological signal at the cost of increasing noise and compute time.
If this is greater than the maximum number of PCs (i.e., the smaller dimension of \code{x}), only the maximum number of PCs will be reported in the results.}

\item{scale}{Logical scalar indicating whether to scale all genes to have the same variance.
This ensures that each gene contributes equally to the PCA, favoring consistent variation across many genes rather than large variation in a few genes.
If \code{block} is specified, each gene's variance is calculated as a weighted sum of the variances from each block. 
Genes with zero variance are ignored.}

\item{block}{Factor specifying the block of origin (e.g., batch, sample) for each cell in \code{x}.
The PCA will be performed on the residuals after regressing out the block effect, ensuring that differences between block do not dominate the variation in the dataset.
Alternatively \code{NULL} if all cells are from the same block.}

\item{block.weight.policy}{String specifying the policy to use for weighting the contribution of different blocks to the PCA.
See the argument of the same name in \code{\link{computeBlockWeights}} for more detail.
Only used if \code{block} is not \code{NULL}.}

\item{variable.block.weight}{Numeric vector of length 2, specifying the parameters for variable block weighting.
See the argument of the same name in \code{\link{computeBlockWeights}} for more detail.
Only used if \code{block} is not \code{NULL} and \code{block.weight.policy = "variable"}.}

\item{components.from.residuals}{Logical scalar indicating whether to compute the PC scores from the residuals in the presence of a blocking factor.
By default, the residuals are only used to compute the rotation matrix, and the original expression values of the cells are projected onto this new space (see Details).
Only used if \code{block} is not \code{NULL}.}

\item{subset}{Integer, logical or character vector specifying the rows of \code{x} to use for the PCA.
This yields the same results as \code{runPca} on \code{x[subset,]}, except that entries of the rotation matrix will also be computed for rows outside of the subset.
If \code{NULL}, all rows of \code{x} are used.}

\item{extra.work}{Integer scalar specifying the extra dimensions for the IRLBA workspace.
Larger values improve accuracy at the cost of compute time.}

\item{iterations}{Integer scalar specifying the maximum number of restart iterations for IRLBA.
Larger values improve accuracy at the cost of compute time.}

\item{seed}{Integer scalar specifying the seed for the initial random vector in IRLBA.}

\item{realized}{Logical scalar indicating whether to realize \code{x} into an optimal memory layout for IRLBA.
This speeds up computation at the cost of increased memory usage.}

\item{warn}{Boolean specifying whether a warning should be emitted if IRLBA failed to converge.}

\item{num.threads}{Number of threads to use.}
}
\value{
List containing:
\itemize{
\item \code{components}, a matrix of PC scores.
Rows are dimensions (i.e., PCs) and columns are cells.
\item \code{rotation}, the rotation matrix.
Rows are genes and columns are dimensions.
\item \code{variance.explained}, the vector of variances explained by each PC.
\item \code{total.variance}, the total variance in the dataset.
This can be used to divide \code{variance.explained} to obtain the proportion of variance explained by each PC.
\item \code{center}, a numeric vector containing the mean for each gene.
If \code{block} is provided, this is instead a matrix containing the mean for each gene (column) in each block (row).
\item \code{block.ids}, a vector containing the identities of the unique blocks in the same order as the rows of \code{center}.
Only reported if \code{block} is provided. 
\item \code{scale}, a numeric vector containing the scaling for each gene.
Only reported if \code{scale=TRUE}.
\item \code{metrics}, a named list containing IRLBA metrics.
This includes \code{converged}, whether the algorithm converged successfully;
\code{iterations}, the number of IRLBA restart iterations;
and \code{multiplications}, the number of matrix multiplications.
}
}
\description{
Run a PCA on the gene-by-cell log-expression matrix and extract the top principal components (PCs).
This yields a low-dimensional representation that reduces noise and compute time in downstream analyses.
For efficiency, the PCA itself is approximated using IRLBA.
}
\details{
When \code{block} is specified, the nature of the reported PC scores depends on the choice of \code{components.from.residuals}:
\itemize{
\item If \code{TRUE}, the PC scores are computed from the matrix of residuals.
This yields a low-dimensional space where inter-block differences have been removed,
assuming that all blocks have the same subpopulation composition and the inter-block differences are consistent for all cell subpopulations.
Under these assumptions, we could use these components for downstream analysis without any concern for block-wise effects.
\item If \code{FALSE}, the rotation vectors are first computed from the matrix of residuals.
To obtain PC scores, each cell is then projected onto the associated subspace using its original expression values.
This approach ensures that inter-block differences do not contribute to the PCA but does not attempt to explicitly remove them.
}
In complex datasets, the assumptions mentioned for \code{TRUE} not hold and more sophisticated batch correction methods like MNN correction are required.
Functions like \code{\link{correctMnn}} will accept a low-dimensional embedding of cells that can be created as described above with \code{FALSE}.
}
\examples{
library(Matrix)
x <- abs(rsparsematrix(1000, 100, 0.1) * 10)
y <- normalizeCounts(x, size.factors=centerSizeFactors(colSums(x)))

# A simple PCA:
out <- runPca(y)
str(out)

# Blocking on uninteresting factors:
block <- sample(LETTERS[1:3], ncol(y), replace=TRUE)
bout <- runPca(y, block=block)
str(bout)

}
\seealso{
The \code{simple_pca} and \code{blocked_pca} functions for \url{https://libscran.github.io/scran_pca/}. 

\code{\link{runPca.se}}, to run a PCA on a \link[SummarizedExperiment]{SummarizedExperiment}.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/runAllNeighborSteps.R
\name{runAllNeighborSteps}
\alias{runAllNeighborSteps}
\title{Run all neighbor-related steps}
\usage{
runAllNeighborSteps(
  x,
  runUmap.args = list(),
  runTsne.args = list(),
  buildSnnGraph.args = list(),
  clusterGraph.args = list(),
  BNPARAM = AnnoyParam(),
  return.graph = FALSE,
  collapse.search = TRUE,
  num.threads = 3
)
}
\arguments{
\item{x}{Numeric matrix where rows are dimensions and columns are cells,
typically containing a low-dimensional representation from, e.g., \code{\link{runPca}}.

Alternatively, an index constructed by \code{\link[BiocNeighbors]{buildIndex}}.}

\item{runUmap.args}{Named list of further arguments to pass to \code{\link{runUmap}}.
This can be set to \code{NULL} to omit the UMAP.}

\item{runTsne.args}{Named list of further arguments to pass to \code{\link{runTsne}}.
This can be set to \code{NULL} to omit the t-SNE.}

\item{buildSnnGraph.args}{Named list of further arguments to pass to \code{\link{buildSnnGraph}}.
Ignored if \code{clusterGraph.args=NULL}.}

\item{clusterGraph.args}{Named list of further arguments to pass to \code{\link{clusterGraph}}.
This can be set to \code{NULL} to omit the clustering.}

\item{BNPARAM}{A \link[BiocNeighbors]{BiocNeighborParam} instance specifying the nearest-neighbor search algorithm to use.}

\item{return.graph}{Logical scalar indicating whether to return the output of \code{\link{buildSnnGraph}}.
By default, only the output of \code{\link{clusterGraph}} is returned.}

\item{collapse.search}{Logical scalar indicating whether to collapse the nearest-neighbor search for each step into a single search.
Steps that need fewer neighbors will take a subset of the neighbors from the collapsed search.
Setting this to \code{TRUE} is faster but may not give the same results as separate searches for some nearest-neighbor algorithms (e.g., approximate methods).}

\item{num.threads}{Integer scalar specifying the number of threads to use.
At least one thread should be available for each step.}
}
\value{
A named list containing the results of each step.
See each individual function for the format of the results.
}
\description{
Run all steps that require a nearest-neighbor search.
This includs \code{\link{runUmap}}, \code{\link{runTsne}} and \code{\link{buildSnnGraph}} with \code{\link{clusterGraph}}.
The idea is to build the index once, perform the neighbor search, and run each task in parallel to save time.
}
\examples{
x <- t(as.matrix(iris[,1:4]))
# (Turning down the number of threads so that R CMD check is happy.)
res <- runAllNeighborSteps(x, num.threads=2)
str(res)

}
\seealso{
\code{\link{runAllNeighborSteps.se}}, to run each neighbor-related step on a \link[SingleCellExperiment]{SingleCellExperiment}.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summarizeEffects.R
\name{summarizeEffects}
\alias{summarizeEffects}
\title{Summarize pairwise effect sizes for each group}
\usage{
summarizeEffects(
  effects,
  compute.summary.min = TRUE,
  compute.summary.mean = TRUE,
  compute.summary.median = TRUE,
  compute.summary.max = TRUE,
  compute.summary.quantiles = NULL,
  compute.summary.min.rank = TRUE,
  num.threads = 1
)
}
\arguments{
\item{effects}{A 3-dimensional numeric containing the effect sizes from each pairwise comparison between groups.
The extents of the first two dimensions are equal to the number of groups, while the extent of the final dimension is equal to the number of genes.
The entry \code{[i, j, k]} represents the effect size from the comparison of group \code{j} against group \code{i} for gene \code{k}.
See also the output of \code{\link{scoreMarkers}} with \code{all.pairwise=TRUE}.}

\item{compute.summary.min}{Boolean specifying whether to compute the minimum as a summary statistic.}

\item{compute.summary.mean}{Boolean specifying whether to compute the mean as a summary statistic.}

\item{compute.summary.median}{Boolean specifying whether to compute the median as a summary statistic.}

\item{compute.summary.max}{Boolean specifying whether to compute the maximum as a summary statistic.}

\item{compute.summary.quantiles}{Numeric scalars containing the probabilities of quantiles to compute as summary statistics. 
If \code{NULL}, no quantiles are computed.}

\item{compute.summary.min.rank}{Boolean specifying whether to compute the mininum rank as a summary statistic.}

\item{num.threads}{Integer scalar specifying the number of threads to use.}
}
\value{
List of \link[S4Vectors]{DataFrame}s containing summary statistics for the effect sizes.
Each DataFrame corresponds to a group, each row corresponds to a gene, and each column contains a summary statistic.
If \code{compute.summary.quantiles} is provided, the \code{"quantile"} column is a nested DataFrame where each column coresponds to a probability in \code{compute.summary.quantiles}.
}
\description{
For each group, summarize the effect sizes for all pairwise comparisons to other groups.
This yields a set of summary statistics that can be used to rank marker genes for each group.
}
\details{
Each summary statistic can be used to prioritize different sets of marker genes for the group of interest, by ranking them in decreasing order according to said statistic:
\itemize{
\item \code{min} contains the minimum effect size across all comparisons involving the group of interest.
Genes with large values are upregulated in all comparisons.
As such, it is the most stringent summary as markers will only have large values if they are uniquely upregulated in the group of interest compared to every other group.
\item \code{mean} contains the mean effect size across all comparisons involving the group of interest.
Genes with large values are upregulated on average compared to the other groups.
This is a good general-purpose summary statistic.
\item \code{median} contains the median effect size across all comparisons involving the group of interest.
Genes with large values are upregulated compared to most (i.e., at least 50% of) other groups.
Compared to the mean, this is more robust to outlier effects but less sensitive to strong effects in a minority of comparisons.
\item \code{max} contains the maximum effect size across all comparisons involving the group of interest.
Using this to define markers will focus on genes that are upregulated in at least one comparison.
As such, it is the least stringent summary as markers can achieve large values if they are upregulated in the group of interest compared to any one other group.
\item \code{quantile[[P]]} contains the quantile P across all comparisons involving the group of interest.
This is a generalization of the minimum, median and maximum for arbitrary quantile probabilities.
For example, a large \code{quantile[["20"]]} would mean that the gene is upregulated in the group of interest compared to 80% of other groups.
}

The exact definition of \dQuote{large} depends on the choice of effect size.
For signed effects like Cohen's d, delta-mean and delta-detected, the value must be positive to be considered \dQuote{large}.
For the AUC, a value greater than 0.5 is considered \dQuote{large}.
This interpretation is also affected by the choice of \code{threshold=} used to compute each effect size in \code{\link{scoreMarkers}},
e.g., a negative Cohen's d cannot be interpreted as downregulation when the threshold is positive. 

The \code{min.rank} is a more exotic summary statistic, containing the minimum rank for each gene across all comparisons involving the group of interest.
This is defined by ranking the effect sizes across genes within each comparison, and then taking the minimum of these ranks across comparisons.
Taking all genes with \code{min.rank <= T} will yield a set containing the top \code{T} genes from each comparison.
The idea is to ensure that there are at least \code{T} genes that can distinguish the group of interest from any other group.

\code{NaN} effect sizes are allowed, e.g., if two groups do not exist in the same block for a blocked analysis in \code{\link{scoreMarkers}} with \code{block=}. 
This function will ignore \code{NaN} values when computing each summary.
If all effects are NaN for a particular group, the summary statistic will also be \code{NaN}.
}
\examples{
# Mocking a matrix:
library(Matrix)
x <- round(abs(rsparsematrix(1000, 100, 0.1) * 100))
normed <- normalizeCounts(x, size.factors=centerSizeFactors(colSums(x)))

g <- sample(letters[1:4], ncol(x), replace=TRUE)
effects <- scoreMarkers(normed, g, all.pairwise=TRUE)

summarized <- summarizeEffects(effects$cohens.d)
str(summarized)

}
\seealso{
The \code{summarize_effects} function in \url{https://libscran.github.io/scran_markers/}.

\code{\link{scoreMarkers}}, to compute the pairwise effects in the first place.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/normalizeCounts.R
\name{normalizeCounts}
\alias{normalizeCounts}
\title{Normalize the count matrix}
\usage{
normalizeCounts(
  x,
  size.factors,
  log = TRUE,
  pseudo.count = 1,
  log.base = 2,
  preserve.sparsity = FALSE,
  delayed = TRUE
)
}
\arguments{
\item{x}{A matrix-like object where rows correspond to genes or genomic features and columns correspond to cells.
Values are expected to be non-negative counts.
Alternatively, an external pointer created by \code{\link[beachmat]{initializeCpp}}.}

\item{size.factors}{A numeric vector of length equal to the number of cells in \code{x}, containing positive size factors for all cells.
Any invalid values should be replaced with \code{\link{sanitizeSizeFactors}}.
For most applications, these size factors should also be centered with \code{\link{centerSizeFactors}}.}

\item{log}{Logical scalar indicating whether log-transformation should be performed.
This ensures that downstream analyses like t-tests and distance calculations focus on relative fold-changes rather than absolute differences.
The log-transformation also provides some measure of variance stabilization so that the downstream analyses are not dominated by sampling noise at large counts.}

\item{pseudo.count}{Numeric scalar specifying the positive pseudo-count to add before any log-transformation.
Larger values shrink the differences between cells towards zero, reducing spurious differences (but also signal) at low counts - see \code{\link{choosePseudoCount}} for comments.
Ignored if \code{log=FALSE}.}

\item{log.base}{Numeric scalar specifying the base of the log-transformation.
Ignored if \code{log=FALSE}.}

\item{preserve.sparsity}{Logical scalar indicating whether to preserve sparsity for \code{pseudo.count != 1}.
If \code{TRUE}, users should manually add \code{log(pseudo.count, log.base)} to the returned matrix to obtain the desired log-transformed expression values.
Ignored if \code{log = FALSE} or \code{pseudo.count = 1}.}

\item{delayed}{Logical scalar indicating whether operations on a matrix-like \code{x} should be delayed.
This improves memory efficiency at the cost of some speed in downstream operations.}
}
\value{
If \code{x} is a matrix-like object and \code{delayed=TRUE}, a \link[DelayedArray]{DelayedArray} is returned containing the (log-transformed) normalized expression matrix.
If \code{delayed=FALSE}, the type of the (log-)normalized matrix will depend on the operations applied to \code{x}.

If \code{x} is an external pointer produced by \code{\link[beachmat]{initializeCpp}}, a new external pointer is returned containing the normalized expression matrix.
}
\description{
Apply scaling normalization and log-transformation to a count matrix.
This yields normalized expression values that can be used in downstream procedures like PCA.
}
\examples{
# Mocking a matrix:
library(Matrix)
x <- round(abs(rsparsematrix(1000, 100, 0.1) * 100))
sf <- centerSizeFactors(colSums(x))
normed <- normalizeCounts(x, size.factors=sf)
normed

# Passing a pointer.
ptr <- beachmat::initializeCpp(x)
optr <- normalizeCounts(ptr, sf)
optr

}
\seealso{
The \code{normalize_counts} function in \url{https://libscran.github.io/scran_norm/}.

\code{\link{normalizeRnaCounts.se}} and related functions, which compute normalized expression values from a \link[SummarizedExperiment]{SummarizedExperiment}.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/runUmap.R
\name{runUmap}
\alias{runUmap}
\title{Uniform manifold approximation and projection}
\usage{
runUmap(
  x,
  num.dim = 2,
  local.connectivity = 1,
  bandwidth = 1,
  mix.ratio = 1,
  spread = 1,
  min.dist = 0.1,
  a = NULL,
  b = NULL,
  repulsion.strength = 1,
  initialize.method = c("spectral", "random", "none"),
  initial.coordinates = NULL,
  initialize.random.on.spectral.fail = TRUE,
  initialize.spectral.scale = 10,
  initialize.spectral.jitter = FALSE,
  initialize.spectral.jitter.sd = 1e-04,
  initialize.random.scale = 10,
  initialize.seed = 9876543210,
  num.epochs = NULL,
  learning.rate = 1,
  negative.sample.rate = 5,
  num.neighbors = 15,
  optimize.seed = 1234567890,
  num.threads = 1,
  parallel.optimization = FALSE,
  BNPARAM = AnnoyParam()
)
}
\arguments{
\item{x}{Numeric matrix where rows are dimensions and columns are cells,
typically containing a low-dimensional representation from, e.g., \code{\link{runPca}}.

Alternatively, a named list of nearest-neighbor search results like that returned by \code{\link[BiocNeighbors]{findKNN}}.
This should contain \code{index}, an integer matrix where rows are neighbors and columns are cells;
and \code{distance}, a numeric matrix of the same dimensions containing the distances to each neighbor.
Each column contains 1-based indices for the nearest neighbors of the corresponding cell, ordered by increasing distance.
The number of neighbors should be the same as \code{num.neighbors}, otherwise a warning is raised.

Alternatively, an index constructed by \code{\link[BiocNeighbors]{buildIndex}}.}

\item{num.dim}{Integer scalar specifying the number of dimensions of the output embedding.}

\item{local.connectivity}{Numeric scalar specifying the number of nearest neighbors that are assumed to be always connected, with maximum membership confidence.
Larger values increase the connectivity of the embedding and reduce the focus on local structure.
This may be a fractional number of neighbors, in which case interpolation is performed when computing the membership confidence.}

\item{bandwidth}{Numeric scalar specifying the effective bandwidth of the kernel when converting the distance to a neighbor into a fuzzy set membership confidence.
Larger values reduce the decay in confidence with respect to distance, increasing connectivity and favoring global structure.}

\item{mix.ratio}{Numeric scalar between 0 and 1 specifying the mixing ratio when combining fuzzy sets.
A mixing ratio of 1 will take the union of confidences, a ratio of 0 will take the intersection, and intermediate values will interpolate between them.
Larger values favor connectivity and more global structure.}

\item{spread}{Numeric scalar specifying the scale of the coordinates of the final low-dimensional embedding.
Ignored if \code{a} and \code{b} are provided.}

\item{min.dist}{Numeric scalar specifying the minimum distance between observations in the final low-dimensional embedding.
Smaller values will increase local clustering while larger values favor a more even distribution of observations throughout the low-dimensional space.
This is interpreted relative to \code{spread}.
Ignored if \code{a} and \code{b} are provided.}

\item{a}{Numeric scalar specifying the \eqn{a} parameter for the fuzzy set membership strength calculations.
Larger values yield a sharper decay in membership strength with increasing distance between observations.
If this or \code{b} are \code{NULL}, a suitable value for this parameter is automatically determined from \code{spread} and \code{min.dist}.}

\item{b}{Numeric scalar specifying the \eqn{b} parameter for the fuzzy set membership strength calculations.
Larger values yield an earlier decay in membership strength with increasing distance between observations.
If this or \code{a} are \code{NULL}, a suitable value for this parameter is automatically determined from \code{spread} and \code{min.dist}.}

\item{repulsion.strength}{Numeric scalar specifying the modifier for the repulsive force.
Larger values increase repulsion and favor local structure.}

\item{initialize.method}{String specifying how to initialize the embedding.
This should be one of:
\itemize{
\item \code{spectral}: spectral decomposition of the normalized graph Laplacian.
Specifically, the initial coordinates are defined from the eigenvectors corresponding to the smallest non-zero eigenvalues.
This fails in the presence of multiple graph components or if the approximate SVD fails to converge.
\item \code{random}: fills the embedding with random draws from a normal distribution.
\item \code{none}: uses initial values from \code{initial.coordinates}.
}}

\item{initial.coordinates}{Numeric matrix of initial coordinates, with number of rows equal to the number of observations and number of columns equal to \code{num.dim}.
Only relevant if \code{initialize.method = "none"};
or \code{initialize.method = "spectral"} and spectral initialization fails and \code{initialize.random.on.spectral.fail = FALSE}.}

\item{initialize.random.on.spectral.fail}{Logical scalar indicating whether to fall back to random sampling (i.e., same as \code{random})
if spectral initialization fails due to the presence of multiple components in the graph.
If \code{FALSE}, the values in \code{initial.coordinates} will be used instead, i.e., same as \code{none}.
Only relevant if \code{initialize.method = "spectral"} and spectral initialization fails.}

\item{initialize.spectral.scale}{Numeric scalar specifying the maximum absolute magnitude of the coordinates after spectral initialization.
All initial coordinates are scaled such that the maximum of the absolute values is equal to \code{initialize.spectral.scale}.
This ensures that outlier observations will not have large absolute distances that may interfere with optimization.
Only relevant if \code{initialize.method = "spectral"} and spectral initialization does not fail.}

\item{initialize.spectral.jitter}{Logical scalar indicating whether to jitter coordinates after spectral initialization to separate duplicate observations (e.g., to avoid overplotting).
This is done using normally-distributed noise of mean zero and standard deviation of \code{initialize.spectral.jitter.sd}.
Only relevant if \code{initialize.method = "spectral"} and spectral initialization does not fail.}

\item{initialize.spectral.jitter.sd}{Numeric scalar specifying the standard deviation of the jitter to apply after spectral initialization.
Only relevant if \code{initialize.method = "spectral"} and spectral initialization does not fail and \code{initialize.spectral.jitter = TRUE}.}

\item{initialize.random.scale}{Numeric scalar specifying the scale of the randomly generated initial coordinates.
Coordinates are sampled from a uniform distribution from \eqn{[-x, x)} where \eqn{x} is \code{initialize.random.scale}.
Only relevant if \code{initialize.method = "random"},
or \code{initialize.method = "spectral"} and spectral initialization fails and \code{initialize.random.on.spectral.fail = TRUE}.}

\item{initialize.seed}{Numeric scalar specifying the seed for the random number generation during initialization.
Only relevant if \code{initialize.method = "random"},
or \code{initialize.method = "spectral"} and \code{initialize.spectral.jitter = TRUE};
or \code{initialize.method = "spectral"} and spectral initialization fails and \code{initialize.random.on.spectral.fail = TRUE}.}

\item{num.epochs}{Integer scalar specifying the number of epochs for the gradient descent, i.e., optimization iterations. 
Larger values improve accuracy at the cost of increased compute time.
If \code{NULL}, a value is automatically chosen based on the size of the dataset:
\itemize{
\item For datasets with no more than 10000 observations, the default number of epochs is set to 500.
\item For larger datasets, the number of epochs is inversely proportional to the number of cells, starting from 500 and decreasing asymptotically to a lower limit of 200.
This choice aims to reduce computational work for very large datasets. 
}}

\item{learning.rate}{Numeric scalar specifying the initial learning rate used in the gradient descent.
Larger values can accelerate convergence but at the risk of skipping over suitable local optima.}

\item{negative.sample.rate}{Numeric scalar specifying the rate of sampling negative observations to compute repulsive forces.
Greater values will improve accuracy but increase compute time.}

\item{num.neighbors}{Integer scalar specifying the number of neighbors to use to define the fuzzy sets.
Larger values improve connectivity and favor preservation of global structure, at the cost of increased compute time.
Ignored if \code{x} contains pre-computed neighbor search results.}

\item{optimize.seed}{Numeric scalar specifying the seed to use for the optimization epochs.}

\item{num.threads}{Integer scalar specifying the number of threads to use.}

\item{parallel.optimization}{Logical scalar specifying whether to parallelize the optimization step.}

\item{BNPARAM}{A \link[BiocNeighbors]{BiocNeighborParam} object specifying the algorithm to use.
Only used if \code{x} is not a prebuilt index or a list of existing nearest-neighbor search results.}
}
\value{
A numeric matrix where rows are cells and columns are the two dimensions of the embedding.
}
\description{
Compute UMAP coordinates to visualize similarities between cells.
}
\examples{
x <- t(as.matrix(iris[,1:4]))
embedding <- runUmap(x)
plot(embedding[,1], embedding[,2], col=iris[,5])

}
\references{
McInnes L, Healy J, Melville J (2020).
UMAP: Uniform Manifold Approximation and Projection for Dimension Reduction.
\emph{arXiv}, \url{https://arxiv.org/abs/1802.03426}
}
\seealso{
\url{https://libscran.github.io/umappp/}, for details on the underlying implementation.

\code{\link{runUmap.se}}, to run UMAP on a \link[SingleCellExperiment]{SingleCellExperiment}.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/se_runTsne.R
\name{runTsne.se}
\alias{runTsne.se}
\title{t-SNE on a SummarizedExperiment}
\usage{
runTsne.se(
  x,
  perplexity = 30,
  num.threads = 1,
  more.tsne.args = list(),
  reddim.type = "PCA",
  output.name = "TSNE"
)
}
\arguments{
\item{x}{A \link[SingleCellExperiment]{SingleCellExperiment} object or one of its subclasses.
Rows correspond to genomic features and columns correspond to cells.}

\item{perplexity}{Perplexity to use in the t-SNE algorithm, passed to \code{\link{runTsne}}.}

\item{num.threads}{Number of threads for the neighbor search and optimization, passed to \code{\link{runTsne}}.}

\item{more.tsne.args}{Named list of further arguments to pass to \code{\link{runTsne}}.}

\item{reddim.type}{Integer or string specifying the existing embedding in the \code{\link[SingleCellExperiment]{reducedDim}} of \code{x}.
Alternatively, a named integer or character vector of length 1,
where the name specifies an alternative experiment of \code{x} and the value is the name/index of a \code{reducedDim} entry in that alternative experiment.}

\item{output.name}{String containing the name of the output \code{\link[SingleCellExperiment]{reducedDim}}.}
}
\value{
\code{x} is returned with the t-SNE coordinates stored in the \code{reducedDim}.
}
\description{
Generate a t-SNE visualization from an existing embedding,
by calling \code{\link{runUmap}} on a reduced dimension entry in \link[SingleCellExperiment]{SingleCellExperiment}.
}
\examples{
library(SingleCellExperiment)
sce <- getTestRnaData.se("pca")
# Using fewer iterations for a faster-running example.
sce <- runTsne.se(sce, more.tsne.args=list(max.iterations=50))
head(reducedDim(sce, "TSNE"))

}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/se_aggregateAcrossGenes.R
\name{aggregateAcrossGenes.se}
\alias{aggregateAcrossGenes.se}
\title{Aggregate expression across gene sets in a SummarizedExperiment}
\usage{
aggregateAcrossGenes.se(
  x,
  sets,
  num.threads = 1,
  more.aggr.args = list(),
  assay.type = "logcounts",
  output.name = NULL
)
}
\arguments{
\item{x}{A \link[SummarizedExperiment]{SummarizedExperiment} object or one of its subclasses.
Rows correspond to genes and columns correspond to cells.}

\item{sets}{List of gene sets, see \code{\link{aggregateAcrossGenes}} for more details.

Alternatively, \code{sets} may be a \link[S4Vectors]{List} subclass,
in which case the \code{\link[S4Vectors]{mcols}} are used as the \code{\link[SummarizedExperiment]{rowData}} of the output object.
Weighted gene sets can be represented by a list of \pkg{DataFrames} or a \link[IRanges]{DataFrameList},
where each DataFrame contains two columns, i.e., the gene identities and the associated weights.}

\item{num.threads}{Number of threads, passed to \code{\link{aggregateAcrossGenes}}.}

\item{more.aggr.args}{Named list of additional arguments to pass to \code{\link{aggregateAcrossGenes}}.}

\item{assay.type}{Integer or string specifying the assay of \code{x} to be aggregated.}

\item{output.name}{String specifying the assay name in the output object.
Defaults to \code{assay.type} if it is a string, otherwise \code{"aggregated"}.}
}
\value{
A \link[SummarizedExperiment]{SummarizedExperiment} with number of rows equal to the number of gene sets.
The lone assay contains the aggregated values for each gene set for all cells.
The \code{\link[SummarizedExperiment]{colData}} is the same as that of \code{x}.
}
\description{
Aggregate expression values across sets of genes for each cell,
by calling \code{\link{aggregateAcrossGenes}} on an assay in a \link[SummarizedExperiment]{SummarizedExperiment}.
}
\examples{
library(SingleCellExperiment)
sce <- getTestRnaData.se("norm")

library(org.Mm.eg.db)
set.df <- select(
    org.Mm.eg.db,
    keytype="GO",
    keys=c(
        "GO:0048709", # oligodendrocyte differentiation 
        "GO:0048699", # neuron development
        "GO:0048143"  # astrocyte activation 
    ),
    columns="SYMBOL"
)
sets <- splitAsList(set.df$SYMBOL, set.df$GO)

aggregated <- aggregateAcrossGenes.se(sce, sets)
aggregated
assay(aggregated)[,1:10]

}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/analyze.R
\name{convertAnalyzeResults}
\alias{convertAnalyzeResults}
\title{Convert analysis results into a SingleCellExperiment}
\usage{
convertAnalyzeResults(
  results,
  main.modality = NULL,
  flatten.qc.subsets = TRUE,
  include.per.block.variances = FALSE
)
}
\arguments{
\item{results}{List of results produced by \code{\link{analyze}}.}

\item{main.modality}{String specifying the modality to use as the main experiment of a \link[SingleCellExperiment]{SingleCellExperiment}.}

\item{flatten.qc.subsets}{Logical scalar indicating whether QC metrics for subsets should be flattened in the column data.
If \code{FALSE}, subset metrics are reported as a nested \link[S4Vectors]{DataFrame}.}

\item{include.per.block.variances}{Logical scalar indicating whether the per-block variances should be reported as a nested \link[S4Vectors]{DataFrame} in the row data.}
}
\value{
A \link[SingleCellExperiment]{SingleCellExperiment} containing most of the analysis results.
Filtered and normalized matrices are stored in the assays.
QC metrics, size factors and clusterings are stored in the column data.
Gene variances are stored in the row data.
PCA, t-SNE and UMAP results are stored in the reduced dimensions.
Further modalities are stored as alternative experiments.
}
\description{
Convert results from \code{\link{analyze}} into a \link[SingleCellExperiment]{SingleCellExperiment} for further analysis with Bioconductor packages.
}
\seealso{
\code{\link{analyze}}, to generate \code{results}.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clusterGraph.R
\name{clusterGraph}
\alias{clusterGraph}
\title{Graph-based clustering of cells}
\usage{
clusterGraph(
  x,
  method = c("multilevel", "leiden", "walktrap"),
  multilevel.resolution = 1,
  leiden.resolution = 1,
  leiden.objective = c("modularity", "cpm", "er"),
  walktrap.steps = 4,
  seed = 42
)
}
\arguments{
\item{x}{List containing graph information or an external pointer to a graph, as returned by \code{\link{buildSnnGraph}}.
Alternatively, an \link[igraph]{igraph} object with edge weights.}

\item{method}{String specifying the algorithm to use.
\itemize{
\item \code{"multilevel"} uses multi-level modularity optimization, also known as the Louvain algorithm,
see \url{https://igraph.org/c/doc/igraph-Community.html#igraph_community_multilevel} for details.
\item \code{"walktrap"} uses the Walktrap community finding algorithm,
see \url{https://igraph.org/c/doc/igraph-Community.html#igraph_community_walktrap} for details.
\item \code{"leiden"} uses the Leiden algorithm,
see \url{https://igraph.org/c/doc/igraph-Community.html#igraph_community_leiden} for details.
}}

\item{multilevel.resolution}{Numeric scalar specifying the resolution when \code{method="multilevel"}.
Lower values favor fewer, larger communities; higher values favor more, smaller communities.}

\item{leiden.resolution}{Numeric scalar specifying the resolution when \code{method="leiden"}.
Lower values favor fewer, larger communities; higher values favor more, smaller communities.}

\item{leiden.objective}{String specifying the objective function when \code{method="leiden"}.
\code{"modularity"} uses the generalized modularity, \code{"cpm"} uses the Constant Potts Model, and \code{"er"} uses the Erd\"os-R\'enyi G(n, p) model.
The CPM typically yields more fine-grained clusters than the modularity at the same \code{leiden.resolution}.}

\item{walktrap.steps}{Integer scalar specifying the number of steps to use when \code{method="walktrap"}.
This determines the ability of the Walktrap algorithm to distinguish highly interconnected communities from the rest of the graph.}

\item{seed}{Integer scalar specifying the random seed to use for \code{method="multilevel"} or \code{"leiden"}.}
}
\value{
A list containing \code{membership}, a factor containing the cluster assignment for each cell.
Additional fields may be present depending on the \code{method}:
\itemize{
\item For \code{method="multilevel"}, the \code{levels} list contains the clustering result at each level of the algorithm.
A \code{modularity} numeric vector also contains the modularity at each level, the highest of which corresponds to the reported \code{membership}.
\item For \code{method="leiden"}, a \code{quality} numeric scalar containg the quality of the partitioning.
\item For \code{method="walktrap"}, a \code{merges} matrix specifies the pair of cells or clusters that were merged at each step of the algorithm.
A \code{modularity} numeric scalar also contains the modularity of the final partitioning.
}
}
\description{
Identify clusters by applying community detection algorithms to a graph.
This assumes that that the nodes on the graph represent cells and weighted edges are formed between related cells.
}
\examples{
data <- matrix(rnorm(10000), ncol=1000)
gout <- buildSnnGraph(data)
str(gout)

str(clusterGraph(gout))
str(clusterGraph(gout, method="leiden"))
str(clusterGraph(gout, method="walktrap"))

}
\seealso{
The various \code{cluster_*} functions in \url{https://libscran.github.io/scran_graph_cluster/}. 

\code{\link{clusterGraph.se}}, which performs clustering on graph constructed from a \link[SingleCellExperiment]{SingleCellExperiment}.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/adt_quality_control.R
\name{adt_quality_control}
\alias{adt_quality_control}
\alias{computeAdtQcMetrics}
\alias{suggestAdtQcThresholds}
\alias{filterAdtQcMetrics}
\title{Quality control for ADT count data}
\usage{
computeAdtQcMetrics(x, subsets, num.threads = 1)

suggestAdtQcThresholds(
  metrics,
  block = NULL,
  min.detected.drop = 0.1,
  num.mads = 3
)

filterAdtQcMetrics(thresholds, metrics, block = NULL)
}
\arguments{
\item{x}{A matrix-like object where rows are ADTs and columns are cells.
Values are expected to be counts.}

\item{subsets}{Named list of vectors specifying tag subsets of interest, typically control tags like IgGs.
Each vector may be logical (whether to keep each row), integer (row indices) or character (row names).
For character vectors, strings not present in \code{rownames(x)} are ignored.}

\item{num.threads}{Integer scalar specifying the number of threads to use.}

\item{metrics}{\link[S4Vectors]{DataFrame} of per-cell QC metrics.
This should have the same structure as the return value of \code{computeAdtQcMetrics}.}

\item{block}{Factor specifying the block of origin (e.g., batch, sample) for each cell in \code{metrics}.
Alternatively \code{NULL} if all cells are from the same block.

For \code{filterAdtQcMetrics}, a blocking factor should be provided if \code{block} was used to construct \code{thresholds}.}

\item{min.detected.drop}{Minimum drop in the number of detected features from the median, in order to consider a cell to be of low quality.}

\item{num.mads}{Number of median from the median, to define the threshold for outliers in each metric.}

\item{thresholds}{List with the same structure as produced by \code{suggestAdtQcThresholds}.}
}
\value{
For \code{computeAdtQcMetrics}, a \link[S4Vectors]{DataFrame} is returned with one row per cell in \code{x}.
This contains the following columns:
\itemize{
\item \code{sum}, a numeric vector containing the total ADT count for each cell.
In theory, this represents the efficiency of library preparation and sequencing.
Compared to RNA, the sum is less useful as a QC metric for ADT data as it is strongly influenced by biological variation in the abundance of the targeted features.
Nonetheless, we compute it for diagnostic purposes.
\item \code{detected}, an integer vector containing the number of detected tags per cell.
Even though ADTs are typically used in situations where few features are highly abundant (e.g., cell type-specific markers), 
we still expect detectable coverage of most features due to ambient contamination, non-specific binding or some background expression.
Low numbers of detected tags indicates that library preparation or sequencing depth was suboptimal.
\item \code{subsets}, a nested DataFrame where each column corresponds to a control subset and is a numeric vector containing the total count in that subset. 
The exact interpretation depends on the nature of the feature subset but the most common use case involves isotype control (IgG) features.
IgG antibodies should not bind to anything so a high subset sum suggests that non-specific binding is a problem, e.g., due to antibody conjugates.
(Unlike RNA quality control, we do not use proportions here as it is entirely possible for a cell to have low counts for other tags due to the absence of their targeted features;
this would result in a high proportion even if the cell has a "normal" level of non-specific binding.)
}
Each vector is of length equal to the number of cells.

For \code{suggestAdtQcThresholds}, a named list is returned:
\itemize{
\item If \code{block=NULL}, the list contains:
\itemize{
\item \code{detected}, a numeric scalar containing the lower bound on the number of detected tags. 
This is defined as the lower of (i) \code{num.mads} MADs below the median for the log-transformed values across all cells,
and (ii) the product of \code{1 - min.detected.drop} and the median across all cells.
The latter avoids overly aggressive filtering when the MAD is zero.
\item \code{subsets}, a numeric vector containing the upper bound on the sum of counts in each control subset. 
This is defined as \code{num.mads} MADs above the median of the log-transformed metrics across all cells.
}
\item Otherwise, if \code{block} is supplied, the list contains:
\itemize{
\item \code{detected}, a numeric vector containing the lower bound on the number of detected tags for each blocking level.
Here, the threshold is computed independently for each block, using the same method as the unblocked case.
\item \code{subsets}, a list of numeric vectors containing the upper bound on the sum of counts in each control subset for each blocking level.
Here, the threshold is computed independently for each block, using the same method as the unblocked case.
\item \code{block.ids}, a vector containing the identities of the unique blocks.
}
Each vector is of length equal to the number of levels in \code{block} and is named accordingly.
}

For \code{filterAdtQcMetrics}, a logical vector of length \code{ncol(x)} is returned indicating which cells are of high quality. 
High-quality cells are defined as those with numbers of detected tags above the \code{detected} threshold and control subset sums below the \code{subsets} threshold.
}
\description{
Compute per-cell QC metrics from an initialized matrix of ADT counts,
and use the metrics to suggest filter thresholds to retain high-quality cells.
}
\examples{
# Mocking a matrix:
library(Matrix)
x <- round(abs(rsparsematrix(1000, 100, 0.1) * 100))

# Mocking up a control set.
sub <- list(IgG=rbinom(nrow(x), 1, 0.1) > 0)

qc <- computeAdtQcMetrics(x, sub)
qc 

filt <- suggestAdtQcThresholds(qc)
str(filt)

keep <- filterAdtQcMetrics(filt, qc)
summary(keep)

}
\seealso{
The \code{compute_adt_qc_metrics}, \code{compute_adt_qc_filters} and \code{compute_adt_qc_filters_blocked} functions in \url{https://libscran.github.io/scran_qc/}.

\code{\link{quickAdtQc.se}}, to run all of the ADT-related QC functions on a \link[SummarizedExperiment]{SummarizedExperiment}.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/se_getTestData.R
\name{getTestData.se}
\alias{getTestData.se}
\alias{getTestRnaData.se}
\alias{getTestAdtData.se}
\alias{getTestCrisprData.se}
\title{Get datasets for testing}
\usage{
getTestRnaData.se(at = c("start", "qc", "norm", "hvg", "pca", "cluster"))

getTestAdtData.se(at = c("start", "qc", "norm", "hvg", "pca"))

getTestCrisprData.se(at = c("start", "qc"))
}
\arguments{
\item{at}{String specifying the level of processing.
For \code{"start"}, no processing was performed.
Otherwise, the dataset is returned after quality control (\code{"qc"}),
normalization (\code{"norm"}),
feature selection (\code{"hvg"}),
PCA (\code{"PCA"})
or graph-based clustering (\code{"cluster"}).}
}
\value{
A \link[SingleCellExperiment]{SingleCellExperiment} containing a dataset at the specified level of processing.
}
\description{
Get single-cell datasets from the \pkg{scRNAseq} package with varying levels of processing.
This is primarily intended for testing other \pkg{scrapper} functions, e.g., in their Examples section.
}
\details{
For \code{getTestRnaData}, this is a scRNA-seq dataset of the mouse brain,
where the main experiment contains RNA counts and the alternative experiments contain ERCC and repeat element counts.
This is obtained with \code{\link[scRNAseq]{fetchDataset}("zeisel-brain-2015", "2023-12-14")}.

For \code{getTestAdtData}, this is a CITE-seq dataset of human PBMCs,
where the main experiment contains RNA counts and the alternative experiment contains ADT counts.
This is obtained with \code{\link[scRNAseq]{fetchDataset}("kotliarov-pbmc-2020", "2024-04-18")}.
Only the first 5000 cells are loaded for speed.

For \code{getTestCrisprData}, this is a Perturb-seq dataset of a pancreatic beta cell line,
where the main experiment contains RNA counts and the alternative experiment contains CRISPR guide counts.
This is obtained with \code{\link[scRNAseq]{fetchDataset}("cao-pancreas-2025", "2025-10-10", "rqc")}.
Only the first 5000 cells are loaded for speed.
}
\examples{
getTestRnaData.se()
getTestAdtData.se()
getTestCrisprData.se()

}
\seealso{
\code{\link[scRNAseq]{fetchDataset}}, used to obtain each dataset.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/se_chooseRnaHvgs.R
\name{chooseRnaHvgs.se}
\alias{chooseRnaHvgs.se}
\alias{formatModelGeneVariancesResult}
\title{Choose highly variable genes from a SummarizedExperiment}
\usage{
chooseRnaHvgs.se(
  x,
  block = NULL,
  num.threads = 1,
  more.var.args = list(),
  top = 4000,
  more.choose.args = list(),
  assay.type = "logcounts",
  output.prefix = NULL,
  include.per.block = FALSE
)

formatModelGeneVariancesResult(
  model.res,
  choose.res = NULL,
  include.per.block = FALSE
)
}
\arguments{
\item{x}{A \link[SummarizedExperiment]{SummarizedExperiment} object or one of its subclasses.
Rows correspond to genes and columns correspond to cells.}

\item{block}{Block assignment for each cell, to pass to \code{\link{modelGeneVariances}}.}

\item{num.threads}{Number of threads, to pass to \code{\link{modelGeneVariances}}.}

\item{more.var.args}{Named list of arguments to pass to \code{\link{modelGeneVariances}}.}

\item{top}{Number of HVGs to choose, to pass to \code{\link{chooseHighlyVariableGenes}}.}

\item{more.choose.args}{Named list of arguments to pass to \code{\link{chooseHighlyVariableGenes}}.}

\item{assay.type}{Integer or string specifying the assay of \code{x} containing the log-normalized expression matrix for the RNA data.}

\item{output.prefix}{String containing a prefix to add to the names of the \code{link[SummarizedExperiment]{rowData}} columns containing the output statistics.}

\item{include.per.block}{Logical scalar indicating whether the per-block statistics should be stored in the output \code{rowData}.
Only relevant if \code{block} is specified.}

\item{model.res}{List returned by \code{\link{modelGeneVariances}}.}

\item{choose.res}{Integer vector returned by \code{\link{chooseHighlyVariableGenes}}.
This may be \code{NULL}, in which case the identities of the HVGs will not be stored.}
}
\value{
For \code{chooseRnaHvgs.se}, \code{x} is returned with the per-gene variance modelling statistics added to its \code{rowData}.
The \code{hvg} column in the \code{rowData} indicates whether a gene was chosen as a HVG.
If \code{include.per.block=TRUE} and \code{block} is specified, the per-block statistics are stored as a nested DataFrame in the \code{per.block} column.

For \code{formatModelGeneVariancesResult}, a \link[S4Vectors]{DataFrame} is returned with the per-gene variance modelling statistics.
If \code{choose.res} is provided, a \code{hvg} column is also stored that indicates whether a gene was chosen as a HVG.
}
\description{
Model the mean-variance relationship across genes and choose highly variable genes (HVGs) based on the residuals of the fitted trend.
This calls \code{\link{modelGeneVariances}} on an assay of a \link[SummarizedExperiment]{SummarizedExperiment},
and then calls \code{\link{chooseHighlyVariableGenes}} on the residuals.
}
\examples{
library(SingleCellExperiment)
sce <- getTestRnaData.se("norm")
sce <- chooseRnaHvgs.se(sce, more.var.args=list(use.min.width=TRUE))
summary(rowData(sce)$hvg)

plot(rowData(sce)$means, rowData(sce)$variances, col=factor(rowData(sce)$hvg))
curve(approxfun(rowData(sce)$means, rowData(sce)$fitted)(x), col="dodgerblue", add=TRUE)

}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/se_clusterKmeans.R
\name{clusterKmeans.se}
\alias{clusterKmeans.se}
\title{k-means clustering of cells in a SingleCellExperiment}
\usage{
clusterKmeans.se(
  x,
  k,
  num.threads = 1,
  more.kmeans.args = list(),
  reddim.type = "PCA",
  output.name = "clusters",
  meta.name = NULL
)
}
\arguments{
\item{x}{A \link[SingleCellExperiment]{SingleCellExperiment} object or one of its subclasses.
Rows correspond to genomic features and columns correspond to cells.}

\item{k}{Number of clusters, passed to \code{\link{clusterKmeans}}.}

\item{num.threads}{Number of threads, passed to \code{\link{clusterKmeans}}.}

\item{more.kmeans.args}{Named list of further arguments to be passed to \code{\link{clusterKmeans}}.}

\item{reddim.type}{Integer or string specifying the existing embedding in the \code{\link[SingleCellExperiment]{reducedDim}} of \code{x}.
Alternatively, a named integer or character vector of length 1,
where the name specifies an alternative experiment of \code{x} and the value is the name/index of a \code{reducedDim} entry in that alternative experiment.}

\item{output.name}{String containing the name of the \code{\link[SummarizedExperiment]{colData}} column in which to store the cluster assignments.}

\item{meta.name}{String containing the name of the \code{\link[S4Vectors]{metadata}} entry in which to store extra clustering output.
If \code{NULL}, no extra clustering output is stored.}
}
\value{
\code{x} is returned with the cluster assignment for each cell stored in the \code{colData}.
Additional clustering output is stored in the \code{metadata}.
}
\description{
Perform k-means clustering on an existing low-dimensional embedding
by calling \code{\link{clusterKmeans}} on a reduced dimension entry in a \link[SingleCellExperiment]{SingleCellExperiment}.
}
\examples{
sce <- getTestRnaData.se("pca")
sce <- clusterKmeans.se(sce, k=10)
table(sce$clusters)

}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/se_scoreGeneSet.R
\name{scoreGeneSet.se}
\alias{scoreGeneSet.se}
\title{Score a gene set in a SummarizedExperiment}
\usage{
scoreGeneSet.se(
  x,
  set,
  block = NULL,
  num.threads = 1,
  more.score.args = list(),
  assay.type = "logcounts"
)
}
\arguments{
\item{x}{A \link[SummarizedExperiment]{SummarizedExperiment} object or one of its subclasses.
Rows correspond to genes and columns correspond to cells.}

\item{set}{Vector containing the gene set, see \code{?\link{scoreGeneSet}} for details.}

\item{block}{Block assignment for each cell, passed to \code{\link{scoreGeneSet}}.}

\item{num.threads}{Number of threads for \code{\link{scoreGeneSet}}.}

\item{more.score.args}{Named list of further arguments to pass to \code{\link[scrapper]{scoreGeneSet}}.}

\item{assay.type}{Integer or string specifying the relevant assay in \code{x}, usually containing log-normalized expression values.}
}
\value{
List containing \code{scores}, a numeric vector of the gene set scores across all cells in \code{x};
and \code{weights}, a numeric vector of weights for all genes in \code{set}.
}
\description{
Compute a gene set activity score for each cell based on the expression values of the genes in the set,
by calling \code{\link{scoreGeneSet}} on an assay of a \link[SummarizedExperiment]{SummarizedExperiment}.
}
\examples{
# Defining a gene set of oligodendrocyte genes.
library(org.Mm.eg.db)
oligo.set <- select(org.Mm.eg.db, keytype="GO", keys="GO:0048709", columns="SYMBOL")
oligo.set <- unique(oligo.set$SYMBOL)

sce <- getTestRnaData.se("norm")
oligo.scores <- scoreGeneSet.se(sce, oligo.set)
summary(oligo.scores$scores)

}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/se_quickRnaQc.R
\name{quickRnaQc.se}
\alias{quickRnaQc.se}
\alias{computeRnaQcMetricsWithAltExps}
\alias{formatComputeRnaQcMetricsResult}
\title{Quick quality control for RNA data in a SummarizedExperiment}
\usage{
quickRnaQc.se(
  x,
  subsets,
  num.threads = 1,
  thresholds = NULL,
  block = NULL,
  more.suggest.args = list(),
  altexp.proportions = NULL,
  assay.type = "counts",
  output.prefix = NULL,
  meta.name = "qc",
  flatten = TRUE
)

computeRnaQcMetricsWithAltExps(
  x,
  subsets,
  altexp.proportions,
  num.threads = 1,
  assay.type = "counts"
)

formatComputeRnaQcMetricsResult(compute.res, flatten = TRUE)
}
\arguments{
\item{x}{A \link[SummarizedExperiment]{SummarizedExperiment} object or one of its subclasses.
Rows correspond to genes and columns correspond to cells.}

\item{subsets}{List of subsets of control genes, see \code{?\link{computeRnaQcMetrics}} for more details.}

\item{num.threads}{Number of threads, to pass to \code{\link{computeRnaQcMetrics}}.}

\item{thresholds}{List containing pre-defined thresholds for each QC metric,
see the return value of \code{\link{suggestRnaQcThresholds}} for the expected format.}

\item{block}{Block assignment for each cell, to pass to \code{\link{suggestRnaQcThresholds}} and \code{\link{filterRnaQcMetrics}}.}

\item{more.suggest.args}{Named list of additional arguments to pass to \code{\link{suggestRnaQcThresholds}}.}

\item{altexp.proportions}{Alternative experiments for which to compute QC metrics.
This is typically used to refer to alternative experiments holding spike-in data.
For each alternative experiment, the proportion is defined as \eqn{X/(X+Y)} where \eqn{X} is the alternative experiment's total and \eqn{Y} is the RNA total.
These proportions will be used for filtering in the same manner as the proportions computed from \code{subsets}.

More specifically, \code{altexp.proportions} should be an unnamed integer or character vector containing the names/indices of the alternative experiments of interest.
The assay to use from each alternative experiment is determined by \code{assay.type}.

Alternatively, \code{altexp.proportions} may be a named integer or character vector.
Each name specifies an alternative experiment while each value is the index/name of the assay to use from that experiment.

Only relevant if \code{x} is a \link[SingleCellExperiment]{SingleCellExperiment}.}

\item{assay.type}{Integer or string specifying the assay of \code{x} containing the RNA count matrix.}

\item{output.prefix}{String containing a prefix to add to the names of the \code{link[SummarizedExperiment]{colData}} columns containing the output statistics.}

\item{meta.name}{String containing the name of the \code{\link[S4Vectors]{metadata}} entry containing the additional outputs such as the filtering thresholds.
If \code{NULL}, additional outputs are not reported.}

\item{flatten}{Logical scalar indicating whether to flatten the subset proportions into separate columns of the \code{link[SummarizedExperiment]{colData}}.
If \code{FALSE}, the subset proportions are stored in a nested \link[S4Vectors]{DataFrame}.}

\item{compute.res}{\link[S4Vectors]{DataFrame} returned by \code{\link{computeRnaQcMetrics}}.}
}
\value{
For \code{quickRnaQc.se}, \code{x} is returned with additional columns added to its \code{\link[SummarizedExperiment]{colData}}.
Each column contains per-cell values for one of the QC metrics, see \code{\link{computeRnaQcMetrics}} for details.
The suggested thresholds are stored as a list in \code{\link[S4Vectors]{metadata}}.
The \code{colData} also contains a \code{keep} column, specifying which cells are to be retained.
If \code{altexp.proportions} is provided, QC metrics are added to the \code{colData} of the specified alternative experiments in the output object.

For \code{computeRnaQcMetricsWithAltExps}, a list is returned containing:
\itemize{
\item \code{main}, the result of calling \code{\link{computeRnaQcMetrics}} on the RNA count matrix in \code{x}.
The proportion of counts in each alternative experiment is added to the \code{subsets}.
\item \code{altexp}, a named list of length equal to \code{altexp.proportions}.
Each inner list is the result of calling \code{computeRnaQcMetrics} on the RNA count matrix of the corresponding alternative experiment of \code{x}.
}

For \code{formatComputeRnaQcMetricsResult}, a \link[S4Vectors]{DataFrame} is returned containing the per-cell QC metrics.
}
\description{
Quickly compute quality control (QC) metrics, thresholds and filters from RNA data in a \link[SummarizedExperiment]{SummarizedExperiment}.
This calls \code{\link{computeRnaQcMetrics}} on an assay in a SummarizedExperiment,
followed by \code{\link{suggestRnaQcThresholds}} and \code{\link{filterRnaQcMetrics}} to identify high-quality cells.
}
\examples{
library(SingleCellExperiment)
sce <- getTestRnaData.se()
sce <- quickRnaQc.se(sce, subsets=list(mito=grepl("^mt", rownames(sce))))
colData(sce)[,c("sum", "detected", "subset.proportion.mito")]
metadata(sce)$qc$thresholds
summary(sce$keep)

# Computing spike-in proportions, if available.
sce <- getTestRnaData.se()
sce <- quickRnaQc.se(
   sce,
   subsets=list(mito=grepl("^mt", rownames(sce))),
   altexp.proportions="ERCC"
)
colData(sce)[,c("sum", "detected", "subset.proportion.mito", "subset.proportion.ERCC")]
colData(altExp(sce, "ERCC"))[,c("sum", "detected")]

}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scaleByNeighbors.R
\name{scaleByNeighbors}
\alias{scaleByNeighbors}
\title{Scale and combine multiple embeddings}
\usage{
scaleByNeighbors(
  x,
  num.neighbors = 20,
  block = NULL,
  block.weight.policy = c("variable", "equal", "none"),
  variable.block.weight = c(0, 1000),
  num.threads = 1,
  weights = NULL,
  BNPARAM = AnnoyParam()
)
}
\arguments{
\item{x}{List of numeric matrices of principal components or other embeddings, one for each modality.
For each entry, rows are dimensions and columns are cells.
All entries should have the same number of columns but may have different numbers of rows.}

\item{num.neighbors}{Integer scalar specifying the number of neighbors to use to define the scaling factor.}

\item{block}{Factor specifying the block of origin (e.g., batch, sample) for each cell in \code{x}.
If provided, the scaling factor is computed as a weighted average across blocks to ensure that block effects do not inflate the within-population variance.
Alternatively \code{NULL}, if all cells are from the same block.}

\item{block.weight.policy}{String specifying the policy to use for weighting different blocks when computing the average scaling factor.
See the argument of the same name in \code{\link{computeBlockWeights}} for more detail.
Only used if \code{block} is not \code{NULL}.}

\item{variable.block.weight}{Numeric vector of length 2, specifying the parameters for variable block weighting.
See the argument of the same name in \code{\link{computeBlockWeights}} for more detail.
Only used if \code{block} is not \code{NULL} and \code{block.weight.policy = "variable"}.}

\item{num.threads}{Integer scalar specifying the number of threads to use.}

\item{weights}{Numeric vector of length equal to that of \code{x}, specifying the weights to apply to each modality.
Each value represents a multiplier of the within-population variance of its modality, i.e., larger values increase the contribution of that modality in the combined output matrix.
\code{NULL} is equivalent to an all-1 vector, i.e., all modalities are scaled to have the same within-population variance.}

\item{BNPARAM}{A \link[BiocNeighbors]{BiocNeighborParam} object specifying how to perform the neighbor search.}
}
\value{
List containing \code{scaling}, a vector of scaling factors to be aplied to each embedding;
and \code{combined}, a numeric matrix creating by scaling each entry of \code{x} by \code{scaling} and then \code{rbind}ing them together.
}
\description{
Scale multiple embeddings (usually derived from different modalities for the same cells) so that their within-population variances are comparable,
and then combine them into a single embedding matrix for further analyses like clustering, t-SNE, etc.
The aim is to equalize uninteresting variance across modalities so that high technical variance in one modality does not drown out interesting biology in another modality.
}
\examples{
pcs <- list(
    gene = matrix(rnorm(10000), ncol=200),
    protein = matrix(rnorm(1000, sd=3), ncol=200),
    guide = matrix(rnorm(2000, sd=5), ncol=200)
)

out <- scaleByNeighbors(pcs)
out$scaling
dim(out$combined)
}
\seealso{
\url{https://libscran.github.io/mumosa/}, for the basis and caveats of this approach.

\code{\link{scaleByNeighbors.se}}, to combine embeddings in a \link[SingleCellExperiment]{SingleCellExperiment}.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/se_runPca.R
\name{runPca.se}
\alias{runPca.se}
\title{Principal components analysis of a Summarizedexperiment}
\usage{
runPca.se(
  x,
  features,
  number = 25,
  block = NULL,
  num.threads = 1,
  more.pca.args = list(),
  assay.type = "logcounts",
  output.name = "PCA",
  meta.name = "PCA",
  delayed.transpose = FALSE
)
}
\arguments{
\item{x}{A \link[SummarizedExperiment]{SummarizedExperiment} object or one of its subclasses.
Rows correspond to genomic features and columns correspond to cells.}

\item{features}{Integer, logical or character vector containing the features of interest to use in the PCA.
For RNA data, this is typically the \code{hvg} vector added by \code{\link{chooseRnaHvgs.se}}.
If \code{NULL}, all available features are used.}

\item{number}{Number of PCs to retain, passed to \code{\link{runPca}}.}

\item{block}{Block assignment for each cell, passed to \code{\link{runPca}}.}

\item{num.threads}{Number of threads for the PCA, passed to \code{\link{runPca}}.}

\item{more.pca.args}{Named list of additional arguments to pass to \code{\link{runPca}}.}

\item{assay.type}{Integer or string specifying the assay of \code{x} to be used for PCA.
This is typically the log-normalized expression matrix created by \code{\link{normalizeRnaCounts.se}}.}

\item{output.name}{String containing the name of the \code{\link[SingleCellExperiment]{reducedDim}} entry in which to store the PC scores.}

\item{meta.name}{String containing the name of the \code{link[S4Vectors]{metadata}} entry in which to store other PCA statistics.}

\item{delayed.transpose}{Logical scalar indicating whether to delay the transposition when storing coordinates in the \code{\link[SingleCellExperiment]{reducedDims}}.}
}
\value{
\code{x} is returned with the principal component scores in the \code{reducedDim}.
(This is converted to a \link[SingleCellExperiment]{SingleCellExperiment} if it wasn't one already.)
Additional outputs (e.g., rotation matrix, variance explained) are stored in the \code{\link[S4Vectors]{metadata}}.
}
\description{
Compact and denoise the dataset by performing PCA on the (log-)normalized expression matrix,
by calling \code{\link{runPca}} on an assay of a \link[SummarizedExperiment]{SummarizedExperiment}.
}
\examples{
library(SingleCellExperiment)
sce <- getTestRnaData.se("hvg")
sce <- runPca.se(sce, rowData(sce)$hvg)
dim(reducedDim(sce, "PCA"))
plot(metadata(sce)$PCA$variance.explained / metadata(sce)$PCA$total.variance)

}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/buildSnnGraph.R
\name{buildSnnGraph}
\alias{buildSnnGraph}
\title{Build a shared nearest neighbor graph}
\usage{
buildSnnGraph(
  x,
  num.neighbors = 10,
  weight.scheme = "ranked",
  num.threads = 1,
  BNPARAM = AnnoyParam(),
  as.pointer = FALSE
)
}
\arguments{
\item{x}{For \code{buildSnnGraph}, a numeric matrix where rows are dimensions and columns are cells,
typically containing a low-dimensional representation from, e.g., \code{\link{runPca}}.

Alternatively, a named list of nearest-neighbor search results.
This should contain \code{index}, an integer matrix where rows are neighbors and columns are cells.
Each column contains 1-based indices for the nearest neighbors of the corresponding cell, ordered by increasing distance.
The number of neighbors for each cell should be equal to \code{num.neighbors}, otherwise a warning is raised.

Alternatively, an index constructed by \code{\link[BiocNeighbors]{buildIndex}}.}

\item{num.neighbors}{Integer scalar specifying the number of neighbors to use to construct the graph.
Larger values increase the connectivity of the graph and reduce the granularity of subsequent community detection steps, at the cost of speed.
Ignored if \code{x} contains pre-computed neighbor search results.}

\item{weight.scheme}{String specifying the weighting scheme to use for constructing the SNN graph.
This can be one of:
\itemize{
\item \code{"ranked"}, where the weight of the edge is defined by the smallest sum of ranks across all shared neighbors.
More shared neighbors, or shared neighbors that are close to both observations, will generally yield larger weights.
\item \code{"number"}, where the weight of the edge is the number of shared nearest neighbors between them. 
This is a simpler scheme that is also slightly faster but does not account for the ranking of neighbors within each set.
\item \code{"jaccard"}, where the weight of the edge is the Jaccard index of their neighbor sets,
This is a monotonic transformation of the weight used in \code{"number"}.
}}

\item{num.threads}{Integer scalar specifying the number of threads to use.
Only used if \code{x} is not a list of existing nearest-neighbor search results.}

\item{BNPARAM}{A \link[BiocNeighbors]{BiocNeighborParam} object specifying the algorithm to use.
Only used if \code{x} is not a list of existing nearest-neighbor search results.}

\item{as.pointer}{Logical scalar indicating whether to return an external pointer for direct use in \code{\link{clusterGraph}}.
This avoids the extra memory usage caused by conversion to/from an R list.}
}
\value{
If \code{as.pointer=FALSE}, a list is returned containing:
\itemize{
\item \code{vertices}, an integer scalar specifying the number of vertices in the graph (i.e., cells in \code{x}).
\item \code{edges}, an integer vector of 1-based indices for graph edges.
Pairs of values represent the endpoints of an (undirected) edge,
i.e., \code{edges[1:2]} form the first edge, \code{edges[3:4]} form the second edge and so on.
\item \code{weights}, a numeric vector of weights for each edge.
This has length equal to half the length of \code{edges}.
}

If \code{as.pointer=TRUE}, an external pointer to the graph is returned that can be directly used in \code{\link{clusterGraph}}.
}
\description{
Build a shared nearest neighbor (SNN) graph where each node is a cell.
Edges are formed between cells that share one or more nearest neighbors, weighted by the number or ranking of those shared neighbors.
If two cells are close together but have distinct sets of neighbors, the corresponding edge is downweighted as the two cells are unlikely to be part of the same neighborhood.
In this manner, strongly weighted edges will only form within highly interconnected neighborhoods where many cells share the same neighbors.
This provides a more sophisticated definition of similarity between cells compared to a simpler (unweighted) nearest neighbor graph that just focuses on immediate proximity.
}
\examples{
data <- matrix(rnorm(10000), ncol=1000)
out <- buildSnnGraph(data)
str(out)

# We can use this to make an igraph::graph.
g <- igraph::make_undirected_graph(out$edges, n = out$vertices)
igraph::E(g)$weight <- out$weight

}
\seealso{
The \code{build_snn_graph} function in \url{https://libscran.github.io/scran_graph_cluster/}.

\code{\link{clusterGraph}}, to define clusters (i.e., communities) from the graph.

\code{\link{clusterGraph.se}}, which builds an SNN graph from a \link[SingleCellExperiment]{SingleCellExperiment}.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/choosePseudoCount.R
\name{choosePseudoCount}
\alias{choosePseudoCount}
\title{Choose a suitable pseudo-count}
\usage{
choosePseudoCount(size.factors, quantile = 0.05, max.bias = 1, min.value = 1)
}
\arguments{
\item{size.factors}{Numeric vector of size factors for all cells.
It is expected that these have already been centered, e.g., with \code{\link{centerSizeFactors}}.
Invalid size factors (e.g., non-positive, non-finite) will be ignored.}

\item{quantile}{Numeric scalar specifying the quantile to use for finding the smallest/largest size factors.
Setting this to zero will use the observed minimum and maximum, though in practice, this is usually too sensitive to outliers.
The default is to take the 5th and 95th percentile to obtain a range that captures most of the distribution.}

\item{max.bias}{Numeric scalar specifying the maximum allowed bias.
This is the maximum absolute value of any spurious log2-fold change between the cells with the smallest and largest size factors.}

\item{min.value}{Numeric scalar specifying the minimum value for the pseudo-count.
Defaults to 1 to stabilize near-zero normalized expression values, otherwise these manifest as avoid large negative values.}
}
\value{
A choice of pseudo-count for \code{\link{normalizeCounts}}.
}
\description{
Choose a suitable pseudo-count to control the bias introduced by log-transformation of normalized counts from \code{\link{normalizeCounts}}.
Larger pseudo-counts shrink log-expression values towards the zero-expression baseline,
reducing the impact of the transformation bias at the cost of some sensitivity.
}
\examples{
sf <- centerSizeFactors(runif(100))
choosePseudoCount(sf)
choosePseudoCount(sf, quantile=0.01)
choosePseudoCount(sf, max.bias=0.5)

}
\references{
Lun ATL (2018).
Overcoming systematic errors caused by log-transformation of normalized single-cell RNA sequencing data.
\emph{biorXiv} doi:10.1101/404962
}
\seealso{
\code{choose_pseudo_count} in \url{https://libscran.github.io/scran_norm/}.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/se_runAllNeighborSteps.R
\name{runAllNeighborSteps.se}
\alias{runAllNeighborSteps.se}
\title{Run all nearest neighbor steps on a SummarizedExperiment}
\usage{
runAllNeighborSteps.se(
  x,
  umap.output.name = "UMAP",
  more.umap.args = list(),
  tsne.output.name = "TSNE",
  more.tsne.args = list(),
  build.graph.name = NULL,
  more.build.graph.args = list(),
  cluster.output.name = "clusters",
  cluster.meta.name = NULL,
  more.cluster.graph.args = list(),
  BNPARAM = AnnoyParam(),
  num.threads = 3,
  more.neighbor.args = list(),
  reddim.type = "PCA"
)
}
\arguments{
\item{x}{A \link[SingleCellExperiment]{SingleCellExperiment} object or one of its subclasses.
Rows correspond to genomic features and columns correspond to cells.}

\item{umap.output.name}{String containing the name of the \code{\link[SingleCellExperiment]{reducedDim}} entry to store the UMAP coordinates.
If \code{NULL}, the UMAP is not computed.}

\item{more.umap.args}{Named list of additional arguments to pass to \code{\link{runAllNeighborSteps}} as \code{runUmap.args}.}

\item{tsne.output.name}{String containing the name of the \code{\link[SingleCellExperiment]{reducedDim}} entry to store the t-SNE coordinates.
If \code{NULL}, the t-SNE is not computed.}

\item{more.tsne.args}{Named list of additional arguments to pass to \code{\link{runAllNeighborSteps}} as \code{runTsne.args}.}

\item{build.graph.name}{String containing the name of the \code{\link[S4Vectors]{metadata}} entry in which to store the nearest neighbor graph.
If \code{NULL}, the graph is not stored.}

\item{more.build.graph.args}{Named list of additional arguments to pass to \code{\link{runAllNeighborSteps}} as \code{buildSnnGraph.args}.}

\item{cluster.output.name}{String containing the name of the \code{\link[SummarizedExperiment]{colData}} column in which to store the cluster assignments.
If \code{NULL}, graph-based clustering is not performed.}

\item{cluster.meta.name}{String containing the name of the \code{\link[S4Vectors]{metadata}} entry in which to store additional clustering outputs.
If \code{NULL}, these additional outputs are not stored.}

\item{more.cluster.graph.args}{Named list of additional arguments to pass to \code{\link{runAllNeighborSteps}} as \code{clusterGraph.args}.}

\item{BNPARAM, num.threads}{Arguments to pass to \code{\link{runAllNeighborSteps}}.}

\item{more.neighbor.args}{Named list of additional arguments to pass to \code{\link{runAllNeighborSteps}}.}

\item{reddim.type}{String or integer specifying the \code{\link[SingleCellExperiment]{reducedDim}} entry on which to perform a nearest neighbor search.
Alternatively, a named integer or character vector of length 1,
where the name specifies an alternative experiment of \code{x} and the value is the name/index of a \code{reducedDim} entry in that alternative experiment.}
}
\value{
\code{x} is returned with additional coordinates stored in its \code{\link[SingleCellExperiment]{reducedDims}}
and clustering output in its \code{\link[SummarizedExperiment]{colData}}.
Additional information may also be stored in its \code{\link[S4Vectors]{metadata}}.
}
\description{
Concurrently run all steps involving a nearest-neighbor search (t-SNE, UMAP and graph-based clustering) using the same nearest-neighbor index,
by calling \code{\link{runAllNeighborSteps}} on a reduced dimension entry of a \link[SingleCellExperiment]{SingleCellExperiment}.
}
\examples{
library(SingleCellExperiment)
sce <- getTestRnaData.se("pca")
sce <- runAllNeighborSteps.se(
   sce,
   more.tsne.args=list(max.iterations=50),
   more.umap.args=list(num.epochs=50),
   num.threads=2 # to keep R CMD check happy
)
reducedDimNames(sce)
table(sce$clusters)

}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/se_quickAdtQc.R
\name{quickAdtQc.se}
\alias{quickAdtQc.se}
\alias{formatComputeAdtQcMetricsResult}
\title{Quick quality control for ADT data in a SummarizedExperiment}
\usage{
quickAdtQc.se(
  x,
  subsets,
  num.threads = 1,
  thresholds = NULL,
  block = NULL,
  more.suggest.args = list(),
  assay.type = "counts",
  output.prefix = NULL,
  meta.name = "qc",
  flatten = TRUE
)

formatComputeAdtQcMetricsResult(compute.res, flatten = TRUE)
}
\arguments{
\item{x}{A \link[SummarizedExperiment]{SummarizedExperiment} object or one of its subclasses.
Rows correspond to antibody-derived tags (ADTs) and columns correspond to cells.}

\item{subsets}{List of subsets of control tags, see \code{?\link{computeAdtQcMetrics}} for more details.}

\item{num.threads}{Number of threads, to pass to \code{\link{computeAdtQcMetrics}}.}

\item{thresholds}{List containing pre-defined thresholds for each QC metric,
see the return value of \code{\link{suggestAdtQcThresholds}} for the expected format.}

\item{block}{Block assignment for each cell, to pass to \code{\link{suggestAdtQcThresholds}} and \code{\link{filterAdtQcMetrics}}.}

\item{more.suggest.args}{Named list of additional arguments to pass to \code{\link{suggestAdtQcThresholds}}.}

\item{assay.type}{Integer or string specifying the assay of \code{x} containing the ADT count matrix.}

\item{output.prefix}{String containing a prefix to add to the names of the \code{link[SummarizedExperiment]{colData}} columns containing the output statistics.}

\item{meta.name}{String containing the name of the \code{\link[S4Vectors]{metadata}} entry containing additional outputs like the filtering thresholds.
If \code{NULL}, additional outputs are not reported.}

\item{flatten}{Logical scalar indicating whether to flatten the subset proportions into separate columns of the \code{link[SummarizedExperiment]{colData}}.
If \code{FALSE}, the subset proportions are stored in a nested \link[S4Vectors]{DataFrame}.}

\item{compute.res}{\link[S4Vectors]{DataFrame} returned by \code{\link{computeAdtQcMetrics}}.}
}
\value{
For \code{quickAdtQc.se}, \code{x} is returned with additional columns added to its \code{\link[SummarizedExperiment]{colData}}.
Each column contains per-cell values for one of the QC metrics, see \code{\link{computeAdtQcMetrics}} for details.
The suggested thresholds are stored as a list in \code{\link[S4Vectors]{metadata}}.
The \code{colData} also contains a \code{keep} column, specifying which cells are to be retained.

For \code{formatComputeAdtQcMetricsResult}, a \link[S4Vectors]{DataFrame} is returned with the per-cell QC metrics.
}
\description{
Quickly compute quality control (QC) metrics, thresholds and filters from ADT data in a \link[SummarizedExperiment]{SummarizedExperiment}.
This calls \code{\link{computeAdtQcMetrics}} on an assay in a SummarizedExperiment,
followed by \code{\link{suggestAdtQcThresholds}} and \code{\link{filterAdtQcMetrics}} to identify high-quality cells.
}
\examples{
library(SingleCellExperiment)
sce <- altExp(getTestAdtData.se(), "ADT")
sce <- quickAdtQc.se(sce, subsets=list(igg=grepl("IgG", rownames(sce))))
colData(sce)[,c("sum", "detected", "subset.sum.igg")]
metadata(sce)$qc$thresholds
summary(sce$keep)

}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/se_quickCrisprQc.R
\name{quickCrisprQc.se}
\alias{quickCrisprQc.se}
\alias{formatComputeCrisprQcMetricsResult}
\title{Quick quality control for CRISPR data in a SummarizedExperiment}
\usage{
quickCrisprQc.se(
  x,
  num.threads = 1,
  thresholds = NULL,
  block = NULL,
  more.suggest.args = list(),
  assay.type = "counts",
  output.prefix = NULL,
  meta.name = "qc"
)

formatComputeCrisprQcMetricsResult(compute.res)
}
\arguments{
\item{x}{A \link[SummarizedExperiment]{SummarizedExperiment} object or one of its subclasses.
Rows correspond to CRISPR guides and columns correspond to cells.}

\item{num.threads}{Number of threads, to pass to \code{\link{computeCrisprQcMetrics}}.}

\item{thresholds}{List containing pre-defined thresholds for each QC metric,
see the return value of \code{\link{suggestRnaQcThresholds}} for the expected format.}

\item{block}{Block assignment for each cell, to pass to \code{\link{suggestCrisprQcThresholds}} and \code{\link{filterCrisprQcMetrics}}.}

\item{more.suggest.args}{Named list of additional arguments to pass to \code{\link{suggestCrisprQcThresholds}}.}

\item{assay.type}{Integer or string specifying the assay of \code{x} containing the CRISPR count matrix.}

\item{output.prefix}{String containing a prefix to add to the names of the \code{link[SummarizedExperiment]{colData}} columns containing the output statistics.}

\item{meta.name}{String containing the name of the \code{\link[S4Vectors]{metadata}} entry containing additional outputs like the filtering thresholds.
If \code{NULL}, additional outputs are not reported.}

\item{compute.res}{\link[S4Vectors]{DataFrame} returned by \code{\link[scrapper]{computeCrisprQcMetrics}}.}
}
\value{
For \code{quickCrisprQc.se}, \code{x} is returned with additional columns added to its \code{\link[SummarizedExperiment]{colData}}.
Each column contains per-cell values for one of the QC metrics, see \code{\link[scrapper]{computeCrisprQcMetrics}} for details.
The suggested thresholds are stored as a list in \code{\link[S4Vectors]{metadata}}.
The \code{colData} also contains a \code{keep} column, specifying which cells are to be retained.

For \code{formatComputeCrisprQcMetricsResult}, a \link[S4Vectors]{DataFrame} is returned with the per-cell QC metrics.
}
\description{
Quickly compute quality control (QC) metrics, thresholds and filters from CRISPR data in a \link[SummarizedExperiment]{SummarizedExperiment}.
This calls \code{\link{computeCrisprQcMetrics}} on an assay in a SummarizedExperiment,
followed by \code{\link{suggestCrisprQcThresholds}} and \code{\link{filterCrisprQcMetrics}} to identify high-quality cells.
}
\examples{
library(SingleCellExperiment)
sce <- altExp(getTestCrisprData.se(), "CRISPR Guide Capture")
sce <- quickCrisprQc.se(sce)
colData(sce)[,c("sum", "detected", "max.value", "max.index")]
metadata(sce)$qc$thresholds
summary(sce$keep)

}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/testEnrichment.R
\name{testEnrichment}
\alias{testEnrichment}
\title{Test for gene set enrichment}
\usage{
testEnrichment(x, sets, universe = NULL, log = FALSE, num.threads = 1)
}
\arguments{
\item{x}{Vector of identifiers for some interesting genes, e.g., symbols or Ensembl IDs.
This is usually derived from a selection of top markers, e.g., from \code{\link{scoreMarkers}}.}

\item{sets}{List of vectors of identifiers for the pre-defined gene sets.
Each inner vector corresponds to a gene set and should contain the same type of identifiers as \code{x}.}

\item{universe}{Vector of identifiers for the universe of genes in the dataset.
\code{x} and each vector in \code{sets} will be subsetted to only include those genes in \code{universe}.
If \code{NULL}, the universe is defined as the union of all genes in \code{x} and \code{sets}. 

Alternatively, an integer scalar specifying the number of genes in the universe.
This is assumed to be greater than or equal to the number of unique genes in \code{x} and \code{sets}.}

\item{log}{Logical scalar indicating whether to report log-transformed p-values.
This may be desirable to avoid underflow at near-zero p-values.}

\item{num.threads}{Integer scalar specifying the number of threads to use.}
}
\value{
\link[S4Vectors]{DataFrame} with one row per gene set and the following columns:
\itemize{
\item \code{overlap}, the overlap between \code{x} and each entry of \code{sets}, i.e., the number of genes in the intersection.
\item \code{size}, the set of each entry of \code{sets}.
\item \code{p.value}, the (possibly log-transformed) p-value for overrepresentation of the gene set in \code{x}.
}
}
\description{
Perform a hypergeometric test for enrichment of gene sets in a list of interesting genes (e.g., markers).
}
\examples{
testEnrichment(
    x=LETTERS[1:5], 
    sets=list(
        first=LETTERS[1:10],
        second=LETTERS[1:5 * 2],
        third=LETTERS[10:20]
    ),
    universe=LETTERS
)

}
\seealso{
\code{\link{phyper}} and \url{https://libscran.github.io/phyper/},
which is the basis for the underlying calculation.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/runTsne.R
\name{runTsne}
\alias{runTsne}
\alias{tsnePerplexityToNeighbors}
\title{t-stochastic neighbor embedding}
\usage{
runTsne(
  x,
  perplexity = 30,
  num.neighbors = tsnePerplexityToNeighbors(perplexity),
  theta = 1,
  early.exaggeration.iterations = 250,
  exaggeration.factor = 12,
  momentum.switch.iterations = 250,
  start.momentum = 0.5,
  final.momentum = 0.8,
  eta = 200,
  max.depth = 7,
  leaf.approximation = FALSE,
  max.iterations = 500,
  seed = 42,
  num.threads = 1,
  BNPARAM = AnnoyParam()
)

tsnePerplexityToNeighbors(perplexity)
}
\arguments{
\item{x}{Numeric matrix where rows are dimensions and columns are cells,
typically containing a low-dimensional representation from, e.g., \code{\link{runPca}}.

Alternatively, a named list of nearest-neighbor search results like that returned by \code{\link[BiocNeighbors]{findKNN}}.
This should contain \code{index}, an integer matrix where rows are neighbors and columns are cells;
and \code{distance}, a numeric matrix of the same dimensions containing the distances to each neighbor.
Each column contains 1-based indices for the nearest neighbors of the corresponding cell, ordered by increasing distance.
The number of neighbors should be the same as \code{num.neighbors}, otherwise a warning is raised.

Alternatively, an index constructed by \code{\link[BiocNeighbors]{buildIndex}}.}

\item{perplexity}{Numeric scalar specifying the perplexity to use in the t-SNE algorithm.
Higher perplexities will focus on global structure, at the cost of increased runtime and decreased local resolution.}

\item{num.neighbors}{Integer scalar specifying the number of neighbors, typically derived from \code{perplexity}.
If \code{x} contains pre-computed neighbor search results with a different number of neighbors than \code{num.neighbors}, an error is thrown;
this can be suppressed by setting \code{num.neighbors = NULL}.}

\item{theta}{Numeric scalar specifying the approximation level for the Barnes-Hut calculation of repulsive forces.
Lower values increase accuracy at the cost of increased compute time.
All values should be non-negative.}

\item{early.exaggeration.iterations}{Integer scalar specifying the number of iterations of the early exaggeration phase,
where clusters are artificially compacted to leave more empty space so that cells can easily relocate to find a good global organization.
Larger values improve convergence within this phase at the cost of reducing the remaining iterations in \code{max.iterations}.}

\item{exaggeration.factor}{Numeric scalar containing the exaggeration factor for the early exaggeration phase (see \code{early.exaggeration.iterations}).
Larger values increase the attraction between nearest neighbors to favor local structure.}

\item{momentum.switch.iterations}{Integer scalar specifying the number of iterations to perform before switching from the starting momentum to the final momentum.
Higher momentums can improve convergence by increasing the step size and smoothing over local oscillations, at the risk of potentially skipping over relevant minima.}

\item{start.momentum}{Numeric scalar containing the starting momentum, to be used in the iterations before the momentum switch at \code{momentum.switch.iterations}.
This is usually lower than \code{final.momentum} to avoid skipping over suitable local minima.}

\item{final.momentum}{Numeric scalar containing the final momentum, to be used in the iterations after the momentum switch at \code{momentum.switch.iterations}.
This is usually higher than \code{start.momentum} to accelerate convergence to the local minima once the observations are moderately well-organized.}

\item{eta}{Numeric scalar containing the learning rate, used to scale the updates for each cell.
Larger values can speed up convergence at the cost of skipping over local minima.}

\item{max.depth}{Integer scalar specifying the maximum depth of the Barnes-Hut quadtree.
If neighboring cells cannot be separated before the maximum depth is reached, they will be assigned to the same leaf node of the quadtree.
Smaller values (7-10) improve speed by bounding the recursion depth at the cost of accuracy.}

\item{leaf.approximation}{Logical scalar indicating whether to use the \dQuote{leaf approximation}.
If \code{TRUE}, repulsive forces are computed between leaf nodes and re-used across all cells assigned to that leaf node. 
This sacrifices some accuracy for greater speed, assuming that \code{max.depth} is small enough for multiple cells to be assigned to the same leaf.}

\item{max.iterations}{Integer scalar specifying the maximum number of iterations to perform.
Larger values improve convergence at the cost of compute time.}

\item{seed}{Integer scalar specifying the seed to use for generating the initial coordinates.}

\item{num.threads}{Integer scalar specifying the number of threads to use.}

\item{BNPARAM}{A \link[BiocNeighbors]{BiocNeighborParam} object specifying the algorithm to use.
Only used if \code{x} is not a prebuilt index or a list of existing nearest-neighbor search results.}
}
\value{
For \code{runTsne}, a numeric matrix where rows are cells and columns are the two dimensions of the embedding.

For \code{tsnePerplexityToNeighbors}, an integer scalar specifying the number of neighbors to use for a given \code{perplexity}.
}
\description{
Compute t-SNE coordinates to visualize similarities between cells.
}
\examples{
x <- t(as.matrix(iris[,1:4]))
embedding <- runTsne(x)
plot(embedding[,1], embedding[,2], col=iris[,5])

}
\references{
van der Maaten LJP and Hinton GE (2008). 
Visualizing high-dimensional data using t-SNE. 
\emph{Journal of Machine Learning Research_} 9, 2579-2605.

van der Maaten LJP (2014). 
Accelerating t-SNE using tree-based algorithms. 
\emph{Journal of Machine Learning Research} 15, 3221-3245.
}
\seealso{
\url{https://libscran.github.io/qdtsne/}, for an explanation of the approximations.

\code{\link{runTsne.se}}, to run t-SNE on a \link[SingleCellExperiment]{SingleCellExperiment}.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chooseHighlyVariableGenes.R
\name{chooseHighlyVariableGenes}
\alias{chooseHighlyVariableGenes}
\title{Choose highly variable genes}
\usage{
chooseHighlyVariableGenes(
  stats,
  top = 4000,
  larger = TRUE,
  keep.ties = TRUE,
  bound = 0
)
}
\arguments{
\item{stats}{Numeric vector of variances (or a related statistic) across all genes.
Typically, the residuals from \code{\link{modelGeneVariances}} are used here.}

\item{top}{Integer specifying the number of top genes to retain.
Note that the actual number of retained genes may not be equal to \code{top}, depending on the other options.}

\item{larger}{Logical scalar indicating whether larger values of \code{stats} correspond to more variable genes.
If \code{TRUE}, HVGs are defined as those with the largest values of \code{stats}.
This is typically the case for variances or related statistics, e.g., residuals.}

\item{keep.ties}{Logical scalar indicating whether to keep tied values of \code{stats}, even if \code{top} may be exceeded.}

\item{bound}{Numeric scalar specifying the lower bound (if \code{larger=TRUE}) or upper bound (otherwise) to be applied to \code{stats}.
Genes are not considered to be HVGs if they do not satisfy this bound, even if they are within the \code{top} genes.
For example, residuals from the fitted trend should be positive, which can be enforced by setting \code{bound} to zero.
Ignored if \code{NULL}.}
}
\value{
Integer vector containing the indices of genes in \code{stats} that are considered to be highly variable.
}
\description{
Choose highly variable genes (HVGs) based on a variance-related statistic.
This is typically used to subset the gene-cell matrix prior to calling \code{\link{runPca}}.
}
\examples{
resids <- rexp(10000)
str(chooseHighlyVariableGenes(resids))

}
\seealso{
The \code{choose_highly_variable_genes} function in \url{https://libscran.github.io/scran_variances/}. 

\code{\link{chooseRnaHvgs.se}}, which choose the HVGs from the residuals computed from a \link[SummarizedExperiment]{SummarizedExperiment}.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scoreGeneSet.R
\name{scoreGeneSet}
\alias{scoreGeneSet}
\title{Score gene set activity for each cell}
\usage{
scoreGeneSet(
  x,
  set,
  rank = 1,
  scale = FALSE,
  block = NULL,
  block.weight.policy = c("variable", "equal", "none"),
  variable.block.weight = c(0, 1000),
  extra.work = 7,
  iterations = 1000,
  seed = 5489,
  realized = TRUE,
  num.threads = 1
)
}
\arguments{
\item{x}{A matrix-like object where rows correspond to genes or genomic features and columns correspond to cells.
Typically, the matrix is expected to contain log-expression values.}

\item{set}{Vector specifying the rows of \code{x} that belong to the gene set.
This may be an integer vector of row indices, a logical vector of length equal to the number of rows, or a character vector of row names.
For integer and character vectors, duplicate entries are ignored.
For a character vector, any string not present in \code{rownames(x)} is ignored.}

\item{rank}{Integer scalar specifying the rank of the approximation.
The default value of 1 assumes that each gene set only describes a single coordinated biological function.}

\item{scale}{Logical scalar indicating whether to scale all genes to have the same variance.
This ensures that each gene contributes equally to the PCA, favoring consistent variation across many genes rather than large variation in a few genes.
If \code{block} is specified, each gene's variance is calculated as a weighted sum of the variances from each block. 
Genes with zero variance are ignored.}

\item{block}{Factor specifying the block of origin (e.g., batch, sample) for each cell in \code{x}.
The PCA will be performed on the residuals after regressing out the block effect, ensuring that differences between block do not dominate the variation in the dataset.
Alternatively \code{NULL} if all cells are from the same block.}

\item{block.weight.policy}{String specifying the policy to use for weighting the contribution of different blocks to the PCA.
See the argument of the same name in \code{\link{computeBlockWeights}} for more detail.
Only used if \code{block} is not \code{NULL}.}

\item{variable.block.weight}{Numeric vector of length 2, specifying the parameters for variable block weighting.
See the argument of the same name in \code{\link{computeBlockWeights}} for more detail.
Only used if \code{block} is not \code{NULL} and \code{block.weight.policy = "variable"}.}

\item{extra.work}{Integer scalar specifying the extra dimensions for the IRLBA workspace.
Larger values improve accuracy at the cost of compute time.}

\item{iterations}{Integer scalar specifying the maximum number of restart iterations for IRLBA.
Larger values improve accuracy at the cost of compute time.}

\item{seed}{Integer scalar specifying the seed for the initial random vector in IRLBA.}

\item{realized}{Logical scalar indicating whether to realize \code{x} into an optimal memory layout for IRLBA.
This speeds up computation at the cost of increased memory usage.}

\item{num.threads}{Number of threads to use.}
}
\value{
List containing:
\itemize{
\item \code{scores}, a numeric vector of per-cell scores for each column in \code{x}.
\item \code{weights}, a \link[S4Vectors]{DataFrame} containing \code{row}, an integer vector of ordered and unique row indices corresponding to the genes in \code{set};
and \code{weight}, a numeric vector of per-gene weights for each gene in \code{row}. 
}
}
\description{
Compute per-cell scores for a gene set, defined as the column sums of a rank-1 approximation to the submatrix for the gene set.
This uses the same approach as the \pkg{GSDecon} package by Jason Hackney, adapted to use an approximate PCA (via IRLBA) and to support blocking.
}
\examples{
library(Matrix)
x <- round(abs(rsparsematrix(1000, 100, 0.1) * 100))
normed <- normalizeCounts(x, size.factors=centerSizeFactors(colSums(x)))
scoreGeneSet(normed, set=c(1,3,5,10,20,100))

}
\seealso{
The \code{compute} and \code{compute_blocked} functions in \url{https://libscran.github.io/gsdecon/}.

\code{\link{scoreGeneSet.se}}, to compute gene set scores from a \link[SummarizedExperiment]{SummarizedExperiment}.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/analyze.R
\name{analyze}
\alias{analyze}
\title{Analyze single-cell data}
\usage{
analyze(
  rna.x,
  adt.x = NULL,
  crispr.x = NULL,
  block = NULL,
  rna.subsets = list(),
  adt.subsets = list(),
  suggestRnaQcThresholds.args = list(),
  suggestAdtQcThresholds.args = list(),
  suggestCrisprQcThresholds.args = list(),
  filter.cells = TRUE,
  centerSizeFactors.args = list(),
  computeClrm1Factors.args = list(),
  normalizeCounts.args = list(),
  modelGeneVariances.args = list(),
  chooseHighlyVariableGenes.args = list(),
  runPca.args = list(),
  use.rna.pcs = TRUE,
  use.adt.pcs = TRUE,
  use.crispr.pcs = TRUE,
  scaleByNeighbors.args = list(),
  correctMnn.args = list(),
  runUmap.args = list(),
  runTsne.args = list(),
  buildSnnGraph.args = list(),
  clusterGraph.args = list(),
  runAllNeighborSteps.args = list(),
  kmeans.clusters = NULL,
  clusterKmeans.args = list(),
  clusters.for.markers = c("graph", "kmeans"),
  scoreMarkers.args = list(),
  BNPARAM = AnnoyParam(),
  rna.assay = 1L,
  adt.assay = 1L,
  crispr.assay = 1L,
  num.threads = 3L
)
}
\arguments{
\item{rna.x}{Matrix-like object containing RNA counts.
This should have the same number of columns as the other \code{*.x} arguments.

Alternatively, a \link[SummarizedExperiment]{SummarizedExperiment} instance containing such a matrix in its \code{rna.assay}.

Alternatively \code{NULL}, if no RNA counts are available.}

\item{adt.x}{Matrix-like object containing ADT counts.
This should have the same number of columns as the other \code{*.x} arguments.

Alternatively, a \link[SummarizedExperiment]{SummarizedExperiment} instance containing such a matrix in its \code{adt.assay}.

Alternatively \code{NULL}, if no ADT counts are available.}

\item{crispr.x}{Matrix-like object containing ADT counts.
This should have the same number of columns as the other \code{*.x} arguments.

Alternatively, a \link[SummarizedExperiment]{SummarizedExperiment} instance containing such a matrix in its \code{crispr.assay}.

Alternatively \code{NULL}, if no ADT counts are available.}

\item{block}{Factor specifying the block of origin (e.g., batch, sample) for each cell in the \code{*_x} matrices.
Alternatively \code{NULL}, if all cells are from the same block.}

\item{rna.subsets}{Gene subsets for quality control, typically used for mitochondrial genes.
See the \code{subsets} arguments in \code{\link{computeRnaQcMetrics}} for details.}

\item{adt.subsets}{ADT subsets for quality control, typically used for IgG controls.
See the \code{subsets} arguments in \code{\link{computeAdtQcMetrics}} for details.}

\item{suggestRnaQcThresholds.args}{Named list of arguments to pass to \code{\link{suggestRnaQcThresholds}}.}

\item{suggestAdtQcThresholds.args}{Named list of arguments to pass to \code{\link{suggestAdtQcThresholds}}.}

\item{suggestCrisprQcThresholds.args}{Named list of arguments to pass to \code{\link{suggestCrisprQcThresholds}}.}

\item{filter.cells}{Logical scalar indicating whether to filter the count matrices to only retain high-quality cells in all modalities.
If \code{FALSE}, QC metrics and thresholds are still computed but are not used to filter the count matrices.}

\item{centerSizeFactors.args}{Named list of arguments to pass to \code{\link{centerSizeFactors}}.}

\item{computeClrm1Factors.args}{Named list of arguments to pass to \code{\link{computeClrm1Factors}}.
Only used if \code{adt.x} is provided.}

\item{normalizeCounts.args}{Named list of arguments to pass to \code{\link{normalizeCounts}}.}

\item{modelGeneVariances.args}{Named list of arguments to pass to \code{\link{modelGeneVariances}}.
Only used if \code{rna.x} is provided.}

\item{chooseHighlyVariableGenes.args}{Named list of arguments to pass to \code{\link{chooseHighlyVariableGenes}}.
Only used if \code{rna.x} is provided.}

\item{runPca.args}{Named list of arguments to pass to \code{\link{runPca}}.}

\item{use.rna.pcs}{Logical scalar indicating whether to use the RNA-derived PCs for downstream steps (i.e., clustering, visualization).
Only used if \code{rna.x} is provided.}

\item{use.adt.pcs}{Logical scalar indicating whether to use the ADT-derived PCs for downstream steps (i.e., clustering, visualization).
Only used if \code{adt.x} is provided.}

\item{use.crispr.pcs}{Logical scalar indicating whether to use the CRISPR-derived PCs for downstream steps (i.e., clustering, visualization).
Only used if \code{crispr.x} is provided.}

\item{scaleByNeighbors.args}{Named list of arguments to pass to \code{\link{scaleByNeighbors}}.
Only used if multiple modalities are available and their corresponding \code{use.*.pcs} arguments are \code{TRUE}.}

\item{correctMnn.args}{Named list of arguments to pass to \code{\link{correctMnn}}.
Only used if \code{block} is supplied.}

\item{runUmap.args}{Named list of arguments to pass to \code{\link{runUmap}}.
If \code{NULL}, UMAP is not performed.}

\item{runTsne.args}{Named list of arguments to pass to \code{\link{runTsne}}.
If \code{NULL}, t-SNE is not performed.}

\item{buildSnnGraph.args}{Named list of arguments to pass to \code{\link{buildSnnGraph}}.
Ignored if \code{clusterGraph.args = NULL}.}

\item{clusterGraph.args}{Named list of arguments to pass to \code{\link{clusterGraph}}.
If \code{NULL}, graph-based clustering is not performed.}

\item{runAllNeighborSteps.args}{Named list of arguments to pass to \code{\link{runAllNeighborSteps}}.}

\item{kmeans.clusters}{Integer scalar specifying the number of clusters to use in k-means clustering.
If \code{NULL}, k-means clustering is not performed.}

\item{clusterKmeans.args}{Named list of arguments to pass to \code{\link{clusterKmeans}}.
Ignored if \code{kmeans.clusters = NULL}.}

\item{clusters.for.markers}{Character vector of clustering algorithms (either \code{"graph"} or \code{"kmeans"}, specifying the clustering to be used for marker detection.
The first available clustering will be chosen.}

\item{scoreMarkers.args}{Named list of arguments to pass to \code{\link{scoreMarkers}}.
Ignored if no suitable clusterings are available.}

\item{BNPARAM}{A \link[BiocNeighbors]{BiocNeighborParam} instance specifying the nearest-neighbor search algorithm to use.}

\item{rna.assay}{Integer scalar or string specifying the assay to use if \code{rna.x} is a \link[SummarizedExperiment]{SummarizedExperiment}.}

\item{adt.assay}{Integer scalar or string specifying the assay to use if \code{adt.x} is a \link[SummarizedExperiment]{SummarizedExperiment}.}

\item{crispr.assay}{Integer scalar or string specifying the assay to use if \code{crispr.x} is a \link[SummarizedExperiment]{SummarizedExperiment}.}

\item{num.threads}{Integer scalar specifying the number of threads to use in each step.}
}
\value{
List containing the results of the entire analysis:
\describe{
\item{\code{rna.qc.metrics}:}{Results of \code{\link{computeRnaQcMetrics}}.
If RNA data is not available, this is set to \code{NULL} instead.}
\item{\code{rna.qc.thresholds}:}{Results of \code{\link{suggestRnaQcThresholds}}.
If RNA data is not available, this is set to \code{NULL} instead.}
\item{\code{rna.qc.filter}:}{Results of \code{\link{filterRnaQcMetrics}}.
If RNA data is not available, this is set to \code{NULL} instead.}
\item{\code{adt.qc.metrics}:}{Results of \code{\link{computeAdtQcMetrics}}.
If ADT data is not available, this is set to \code{NULL} instead.}
\item{\code{adt.qc.thresholds}:}{Results of \code{\link{suggestAdtQcThresholds}}.
If ADT data is not available, this is set to \code{NULL} instead.}
\item{\code{adt.qc.filter}:}{Results of \code{\link{filterAdtQcMetrics}}.
If ADT data is not available, this is set to \code{NULL} instead.}
\item{\code{crispr.qc.metrics}:}{Results of \code{\link{computeCrisprQcMetrics}}.
If CRISPR data is not available, this is set to \code{NULL} instead.}
\item{\code{crispr.qc.thresholds}:}{Results of \code{\link{suggestCrisprQcThresholds}}.
If CRISPR data is not available, this is set to \code{NULL} instead.}
\item{\code{crispr.qc.filter}:}{Results of \code{\link{filterCrisprQcMetrics}}.
If CRISPR data is not available, this is set to \code{NULL} instead.}
\item{\code{combined.qc.filter}:}{Logical vector indicating which cells are of high quality and should be retained for downstream analyses.}
\item{\code{rna.filtered}:}{Matrix of RNA counts that has been filtered to only contain the high-quality cells in \code{combined.qc.filter}.
If RNA data is not available, this is set to \code{NULL} instead.}
\item{\code{adt.filtered}:}{Matrix of ADT counts that has been filtered to only contain the high-quality cells in \code{combined.qc.filter}.
If ADT data is not available, this is set to \code{NULL} instead.}
\item{\code{crispr.filtered}:}{Matrix of CRISPR counts that has been filtered to only contain the high-quality cells in \code{combined.qc.filter}.
If CRISPR data is not available, this is set to \code{NULL} instead.}
\item{\code{rna.size.factors}:}{Size factors for the RNA count matrix, derived from the sum of counts for each cell and centered with \code{\link{centerSizeFactors}}.
If RNA data is not available, this is set to \code{NULL} instead.}
\item{\code{rna.normalized}:}{Matrix of (log-)normalized expression values derived from RNA counts, as computed by \code{\link{normalizeCounts}} using \code{rna.size.factors}.
If RNA data is not available, this is set to \code{NULL} instead.}
\item{\code{adt.size.factors}:}{Size factors for the ADT count matrix, computed by \code{\link{computeClrm1Factors}} and centered with \code{\link{centerSizeFactors}}.
If ADT data is not available, this is set to \code{NULL} instead.}
\item{\code{adt.normalized}:}{Matrix of (log-)normalized expression values derived from ADT counts, as computed by \code{\link{normalizeCounts}} using \code{adt.size.factors}.
If ADT data is not available, this is set to \code{NULL} instead.}
\item{\code{crispr.size.factors}:}{Size factors for the CRISPR count matrix, derived from the sum of counts for each cell and centered with \code{\link{centerSizeFactors}}.
If CRISPR data is not available, this is set to \code{NULL} instead.}
\item{\code{crispr.normalized}:}{Matrix of (log-)normalized expression values derived from CRISPR counts, as computed by \code{\link{normalizeCounts}} using \code{crispr.size.factors}.
If CRISPR data is not available, this is set to \code{NULL} instead.}
\item{\code{rna.gene.variances}:}{Results of \code{\link{modelGeneVariances}}.
If RNA data is not available, this is set to \code{NULL} instead.}
\item{\code{rna.highly.variable.genes}:}{Results of \code{\link{chooseHighlyVariableGenes}}.
If RNA data is not available, this is set to \code{NULL} instead.}
\item{\code{rna.pca}:}{Results of calling \code{\link{runPca}} on \code{rna.normalized} with the \code{rna.highly.variable.genes} subset.
If RNA data is not available, this is set to \code{NULL} instead.}
\item{\code{adt.pca}:}{Results of calling \code{\link{runPca}} on \code{adt.normalized}.
If ADT data is not available, this is set to \code{NULL} instead.}
\item{\code{crispr.pca}:}{Results of calling \code{\link{runPca}} on \code{crispr.normalized}.
If CRISPR data is not available, this is set to \code{NULL} instead.}
\item{\code{combined.pca}:}{If only one modality is used for the downstream analysis, this is a string specifying the list element containing the components to be used, e.g., \code{"rna.pca"}.
If multiple modalities are to be combined for downstream analysis, this contains the results of \code{\link{scaleByNeighbors}} on the PCs of those modalities.}
\item{\code{block}:}{Vector or factor containing the blocking factor for all cells (after filtering, if \code{filter.cells = TRUE}).
This is set to \code{NULL} if no blocking factor was supplied.}
\item{\code{mnn.corrected}:}{Results of \code{\link{correctMnn}} on the PCs in or referenced by \code{combined.pca}.
If no blocking factor is supplied, this is set to \code{NULL} instead.}
\item{\code{tsne}:}{Results of \code{\link{runTsne}}.
This is \code{NULL} if t-SNE was not performed.}
\item{\code{umap}:}{Results of \code{\link{runUmap}}.
This is \code{NULL} if UMAP was not performed.}
\item{\code{snn.graph}:}{Results of \code{\link{buildSnnGraph}}.
This is \code{NULL} if graph-based clustering was not performed, or if \code{return.graph=FALSE} in \code{\link{runAllNeighborSteps}}.}
\item{\code{graph.clusters}:}{Results of \code{\link{clusterGraph}}.
This is \code{NULL} if graph-based clustering was not performed.}
\item{\code{kmeans.clusters}:}{Results of \code{\link{clusterKmeans}}.
This is \code{NULL} if k-means clustering was not performed.}
\item{\code{clusters}:}{Integer vector containing the cluster assignment for each cell (after filtering, if \code{filter.cells = TRUE}).
This may be derived from \code{graph.clusters} or \code{kmeans.clusters} depending on the choice of \code{clusters.for.markers}.
If no suitable clusterings are available, this is set to NULL.}
\item{\code{rna.markers}:}{Results of calling \code{\link{scoreMarkers}} on \code{rna.normalized}.
This is \code{NULL} if RNA data is not available or no suitable clusterings are available.}
\item{\code{adt.markers}:}{Results of calling \code{\link{scoreMarkers}} on \code{adt.normalized}.
This is \code{NULL} if ADT data is not available or no suitable clusterings are available.}
\item{\code{crispr.markers}:}{Results of calling \code{\link{scoreMarkers}} on \code{crispr.normalized}.
This is \code{NULL} if CRISPR data is not available or no suitable clusterings are available.}
}
}
\description{
Execute a simple single-cell analysis pipeline, starting from a count matrix and ending with clusters, visualizations and markers.
This also supports integration of multiple modalities and correction of batch effects.
}
\examples{
sce <- getTestRnaData.se()
sce <- sce[,1:500] # smaller dataset for a faster runtime for R CMD check. 
res <- analyze(
    sce, 
    rna.subsets=list(mito=grep("^mt-", rownames(sce))),
    num.threads=2 # keep R CMD check happy
)
str(res)
convertAnalyzeResults(res)

}
\seealso{
\code{\link{convertAnalyzeResults}}, to convert the results into a \link[SingleCellExperiment]{SingleCellExperiment}.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/se_scoreMarkers.R
\name{scoreMarkers.se}
\alias{scoreMarkers.se}
\alias{formatScoreMarkersResult}
\alias{previewMarkers}
\title{Score marker genes in a SummarizedExperiment}
\usage{
scoreMarkers.se(
  x,
  groups,
  block = NULL,
  num.threads = 1,
  more.marker.args = list(),
  assay.type = "logcounts",
  extra.columns = NULL,
  order.by = TRUE
)

formatScoreMarkersResult(marker.res, extra.columns = NULL, order.by = TRUE)

previewMarkers(
  marker.df,
  columns = c("mean", "detected", lfc = "delta.mean.mean"),
  pre.columns = NULL,
  post.columns = NULL,
  rows = 10,
  order.by = NULL,
  include.order.by = !is.null(order.by)
)
}
\arguments{
\item{x}{A \link[SummarizedExperiment]{SummarizedExperiment} object or one of its subclasses.
Rows correspond to genes and columns correspond to cells.}

\item{groups}{Group assignment for each cell, passed to \code{\link{scoreMarkers}}.}

\item{block}{Block assignment for each cell, passed to \code{\link{scoreMarkers}}.}

\item{num.threads}{Number of threads for marker scoring, passed to \code{\link{scoreMarkers}}.}

\item{more.marker.args}{Named list of additional arguments to pass to \code{\link{scoreMarkers}}.}

\item{assay.type}{Integer or string specifying the assay to use for differential comparisons, usually containing log-normalized expression values.}

\item{extra.columns}{\link[S4Vectors]{DataFrame} containing extra columns to add each DataFrame.
This should have the same number of rows as \code{x}.
For \code{scoreMarkers.se}, this may also be a character vector specifying the columns of \code{\link[SummarizedExperiment]{rowData}} to be added.}

\item{order.by}{String specifying the column to order each DataFrame by.
Alternatively \code{TRUE}, a column is automatically chosen from the effect size summaries.
If \code{NULL} or \code{FALSE}, no ordering is performed.}

\item{marker.res}{List containing the result of \code{\link{scoreMarkers}}.}

\item{marker.df}{DataFrame containing the marker statistics for a single group.}

\item{columns}{Character vector of the names of columns to retain in the preview.
This may be named, in which the names are used as the column names.}

\item{pre.columns, post.columns}{Character vector of the names of additional columns to retain in the preview.
These are added before or after the columns in \code{columns}, for \code{pre.columns} and \code{post.columns} respectively.}

\item{rows}{Integer specifying the number of rows to show.
If \code{NULL}, all rows are returned.}

\item{include.order.by}{Boolean indicating whether the column specified by \code{order.by} should be included in the output DataFrame.
A string may also be supplied and will be treated as \code{TRUE}; the value of the string will be used as the column name in the output DataFrame.}
}
\value{
For \code{scoreMarkers.se} and \code{formatScoreMarkersResult}, a \link[S4Vectors]{List} of \link[S4Vectors]{DataFrame}s is returned.
Each DataFrame corresponds to a unique group in \code{groups}.
Each row contains statistics for a gene in \code{x}, with the following columns:
\itemize{
\item \code{mean}, the mean expression in the current group.
\item \code{detected}, the proportion of cells with detected expression in the current group.
\item \code{<effect>.<summary>}, a summary statistic for an effect size,
e.g., \code{cohens.d.mean} contains the mean Cohen's d across comparisons involving the current group.
}

For \code{previewMarkers}, a DataFrame is returned containing the specified columns and rows.
}
\description{
Identify candidate marker genes based on effect sizes from pairwise comparisons between groups of cells,
by calling \code{\link{scoreMarkers}} on an assay of a \link[SummarizedExperiment]{SummarizedExperiment}.
}
\examples{
sce <- getTestRnaData.se("cluster")
markers <- scoreMarkers.se(sce, sce$clusters)
previewMarkers(markers[["1"]])

}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/se_normalizeCrisprCounts.R
\name{normalizeCrisprCounts.se}
\alias{normalizeCrisprCounts.se}
\title{Normalize CRISPR counts in a SummarizedExperiment}
\usage{
normalizeCrisprCounts.se(
  x,
  size.factors = NULL,
  center = TRUE,
  block = NULL,
  mode = "lowest",
  log = TRUE,
  pseudo.count = 1,
  more.norm.args = list(),
  assay.type = "counts",
  output.name = "logcounts",
  factor.name = "sizeFactor",
  num.threads = 1
)
}
\arguments{
\item{x}{A \link[SummarizedExperiment]{SummarizedExperiment} object or one of its subclasses.
Rows correspond to CRISPR guides and columns correspond to cells.}

\item{size.factors}{Numeric vector of length equal to the number of columns of \code{x},
containing the size factor for each cell in \code{x}.
If \code{NULL}, this defaults to the column sums of the count matrix in \code{x}.}

\item{center}{Logical scalar indicating whether to center the \code{size.factors},
see \code{?\link{centerSizeFactors}} for more details.}

\item{block}{Block assignments for each cell, passed to \code{\link{centerSizeFactors}}.}

\item{mode}{How to center size factors in different blocks, see \code{?\link{centerSizeFactors}} for more details.}

\item{log}{Whether to log-transform the normalized expression values, see \code{?\link{normalizeCounts}} for more details.}

\item{pseudo.count}{The pseudo-count for log-transformation, see \code{?\link{normalizeCounts}} for more details.}

\item{more.norm.args}{Named list of additional arguments to pass to \code{\link{normalizeCounts}}.}

\item{assay.type}{Integer or string specifying the assay of \code{x} with the count matrix.}

\item{output.name}{String containing the name of the assay to store the normalized matrix.}

\item{factor.name}{String containing the name of the \code{\link[SummarizedExperiment]{colData}} column in which to store the size factors in the output object.
If \code{NULL}, the size factors are not stored.}

\item{num.threads}{Integer specifying the number of threads for computing column sums.
Only used if \code{size.factors=NULL}.}
}
\value{
\code{x} is returned with a new assay containing the (log-)normalized matrix.
Size factors are also stored in the \code{\link[SummarizedExperiment]{colData}}.
}
\description{
Compute (log-)normalized expression values after performing scaling normalization of an CRISPR count matrix.
This calls \code{\link{normalizeCounts}} on an assay of a \link[SummarizedExperiment]{SummarizedExperiment},
after centering the size factors with \code{\link{centerSizeFactors}}.
}
\examples{
library(SingleCellExperiment)
sce <- altExp(getTestCrisprData.se("qc"), "CRISPR Guide Capture")
sce <- normalizeCrisprCounts.se(sce, size.factors=sce$sum)
assayNames(sce)
summary(sizeFactors(sce))

}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/se_analyze.R
\name{analyze.se}
\alias{analyze.se}
\title{Analyze single-cell data from a SummarizedExperiment}
\usage{
analyze.se(
  x,
  rna.altexp = NA,
  adt.altexp = NULL,
  crispr.altexp = NULL,
  rna.assay.type = "counts",
  adt.assay.type = "counts",
  crispr.assay.type = "counts",
  block = NULL,
  block.name = "block",
  rna.qc.subsets = list(),
  rna.qc.output.prefix = NULL,
  more.rna.qc.args = list(),
  adt.qc.subsets = list(),
  adt.qc.output.prefix = NULL,
  more.adt.qc.args = list(),
  crispr.qc.output.prefix = NULL,
  more.crispr.qc.args = list(),
  filter.cells = TRUE,
  rna.norm.output.name = "logcounts",
  more.rna.norm.args = list(),
  adt.norm.output.name = "logcounts",
  more.adt.norm.args = list(),
  crispr.norm.output.name = "logcounts",
  more.crispr.norm.args = list(),
  rna.hvg.output.prefix = NULL,
  more.rna.hvg.args = list(),
  rna.pca.output.name = "PCA",
  more.rna.pca.args = list(),
  adt.pca.output.name = "PCA",
  more.adt.pca.args = list(),
  use.rna.pcs = TRUE,
  use.adt.pcs = TRUE,
  scale.output.name = "combined",
  more.scale.args = list(),
  mnn.output.name = "MNN",
  more.mnn.args = list(),
  more.umap.args = list(),
  more.tsne.args = list(),
  cluster.graph.output.name = "graph.cluster",
  more.build.graph.args = list(),
  more.cluster.graph.args = list(),
  more.neighbor.args = list(),
  kmeans.clusters = NULL,
  kmeans.clusters.output.name = "kmeans.cluster",
  more.kmeans.args = list(),
  clusters.for.markers = c("graph", "kmeans"),
  more.rna.marker.args = list(),
  more.adt.marker.args = list(),
  more.crispr.marker.args = list(),
  BNPARAM = AnnoyParam(),
  num.threads = 3L
)
}
\arguments{
\item{x}{A \link[SummarizedExperiment]{SummarizedExperiment} object or one of its subclasses.
Rows correspond to genomic features (genes, ADTs or CRISPR guides) and columns correspond to cells.}

\item{rna.altexp}{String or integer specifying the alternative experiment of \code{x} containing the RNA data.
If \code{NA}, the main experiment is assumed to contain the RNA data.
If \code{NULL}, it is assumed that no RNA data is available.}

\item{adt.altexp}{String or integer specifying the alternative experiment of \code{x} containing the ADT data.
If \code{NA}, the main experiment is assumed to contain the ADT data.
If \code{NULL}, it is assumed that no ADT data is available.}

\item{crispr.altexp}{String or integer specifying the alternative experiment of \code{x} containing the CRISPR data.
If \code{NA}, the main experiment is assumed to contain the CRISPR data.
If \code{NULL}, it is assumed that no CRISPR data is available.}

\item{rna.assay.type}{String or integer specifying the assay containing the RNA count data.
Only used if \code{rna.altexp} is not \code{NULL}.}

\item{adt.assay.type}{String or integer specifying the assay containing the ADT count data.
Only used if \code{adt.altexp} is not \code{NULL}.}

\item{crispr.assay.type}{String or integer specifying the assay containing the CRISPR count data.
Only used if \code{crispr.altexp} is not \code{NULL}.}

\item{block}{Vector or factor specifying the block of origin (e.g., batch, sample) for each cell in \code{x}.
Alternatively \code{NULL}, if all cells are from the same block.}

\item{block.name}{String specifying the name of the \code{\link[SummarizedExperiment]{colData}} column in which to store the blocking factor.
Only used if \code{block} is not \code{NULL}.
If \code{NULL}, the blocking factor is not stored in the \code{colData}.}

\item{rna.qc.subsets}{Passed to \code{\link{quickRnaQc.se}} as the \code{subsets} argument.
Only used if \code{rna.altexp} is not \code{NULL}.}

\item{rna.qc.output.prefix}{Passed to \code{\link{quickRnaQc.se}} as the \code{output.prefix} argument.
Only used if \code{rna.altexp} is not \code{NULL}.}

\item{more.rna.qc.args}{Named list of additional arguments to pass to \code{\link{quickRnaQc.se}}.
Only used if \code{rna.altexp} is not \code{NULL}.}

\item{adt.qc.subsets}{Passed to \code{\link{quickAdtQc.se}} as the \code{subsets} argument.
Only used if \code{adt.altexp} is not \code{NULL}.}

\item{adt.qc.output.prefix}{Passed to \code{\link{quickAdtQc.se}} as the \code{output.prefix} argument.
Only used if \code{adt.altexp} is not \code{NULL}.}

\item{more.adt.qc.args}{Named list of additional arguments to pass to \code{\link{quickAdtQc.se}}.
Only used if \code{adt.altexp} is not \code{NULL}.}

\item{crispr.qc.output.prefix}{Passed to \code{\link{quickCrisprQc.se}} as the \code{output.prefix} argument.
Only used if \code{crispr.altexp} is not \code{NULL}.}

\item{more.crispr.qc.args}{Named list of additional arguments to pass to \code{\link{quickCrisprQc.se}}.
Only used if \code{crispr.altexp} is not \code{NULL}.}

\item{filter.cells}{Logical scalar indicating whether to filter \code{x} to only retain high-quality cells in all modalities.
If \code{FALSE}, QC metrics and thresholds are still computed but are not used to filter the count matrices.}

\item{rna.norm.output.name}{Passed to \code{\link{normalizeRnaCounts.se}} as the \code{output.name} argument.
Only used if \code{rna.altexp} is not \code{NULL}.}

\item{more.rna.norm.args}{Named list of arguments to pass to \code{\link{normalizeRnaCounts.se}}.
Only used if \code{rna.altexp} is not \code{NULL}.}

\item{adt.norm.output.name}{Passed to \code{\link{normalizeAdtCounts.se}} as the \code{output.name} argument.
Only used if \code{adt.altexp} is not \code{NULL}.}

\item{more.adt.norm.args}{Named list of arguments to pass to \code{\link{normalizeAdtCounts.se}}.
Only used if \code{adt.altexp} is not \code{NULL}.}

\item{crispr.norm.output.name}{Passed to \code{\link{normalizeCrisprCounts.se}} as the \code{output.name} argument.
Only used if \code{crispr.altexp} is not \code{NULL}.}

\item{more.crispr.norm.args}{Named list of arguments to pass to \code{\link{normalizeCrisprCounts.se}}.
Only used if \code{crispr.altexp} is not \code{NULL}.}

\item{rna.hvg.output.prefix}{Passed to \code{\link{chooseRnaHvgs.se}} as the \code{output.prefix} argument.
Only used if \code{rna.altexp} is provided.}

\item{more.rna.hvg.args}{Named list of arguments to pass to \code{\link{chooseRnaHvgs.se}}.
Only used if \code{rna.altexp} is provided.}

\item{rna.pca.output.name}{Passed to \code{\link{runPca.se}} as the \code{output.name} argument.
Only used if \code{rna.altexp} is provided.}

\item{more.rna.pca.args}{Named list of arguments to pass to \code{\link{runPca.se}}.
Only used if \code{rna.altexp} is provided.}

\item{adt.pca.output.name}{Passed to \code{\link{runPca.se}} as the \code{output.name} argument.
Only used if \code{adt.altexp} is provided.}

\item{more.adt.pca.args}{Named list of arguments to pass to \code{\link{runPca.se}}.
Only used if \code{adt.altexp} is provided.}

\item{use.rna.pcs}{Logical scalar indicating whether to use the RNA-derived PCs for downstream steps (i.e., clustering, visualization).
Only used if \code{rna.altexp} is provided.}

\item{use.adt.pcs}{Logical scalar indicating whether to use the ADT-derived PCs for downstream steps (i.e., clustering, visualization).
Only used if \code{adt.altexp} is provided.}

\item{scale.output.name}{Passed to \code{\link{scaleByNeighbors.se}} as the \code{output.name} argument.
Only used if multiple modalities are available and their corresponding \code{use.*.pcs} arguments are \code{TRUE}.}

\item{more.scale.args}{Named list of arguments to pass to \code{\link{scaleByNeighbors.se}}.
Only used if multiple modalities are available and their corresponding \code{use.*.pcs} arguments are \code{TRUE}.}

\item{mnn.output.name}{Passed to \code{\link{correctMnn.se}} as the \code{output.name} argument.
Only used if \code{block} is supplied.}

\item{more.mnn.args}{Named list of arguments to pass to \code{\link{correctMnn.se}}.
Only used if \code{block} is supplied.}

\item{more.umap.args}{Passed to \code{\link{runAllNeighborSteps.se}}.}

\item{more.tsne.args}{Passed to \code{\link{runAllNeighborSteps.se}}.}

\item{cluster.graph.output.name}{Passed to \code{\link{runAllNeighborSteps.se}} as \code{cluster.output.name}.}

\item{more.build.graph.args}{Passed to \code{\link{runAllNeighborSteps.se}}.}

\item{more.cluster.graph.args}{Passed to \code{\link{runAllNeighborSteps.se}}.}

\item{more.neighbor.args}{Passed to \code{\link{runAllNeighborSteps.se}}.}

\item{kmeans.clusters}{Passed to \code{\link{clusterKmeans.se}} as the \code{k} argument.
If \code{NULL}, k-means clustering is not performed.}

\item{kmeans.clusters.output.name}{Passed to \code{\link{clusterKmeans.se}} as the \code{output.name} argument.
Ignored if \code{kmeans.clusters = NULL}.}

\item{more.kmeans.args}{Named list of arguments to pass to \code{\link{clusterKmeans.se}}.
Ignored if \code{kmeans.clusters = NULL}.}

\item{clusters.for.markers}{Character vector of clustering algorithms (either \code{"graph"} or \code{"kmeans"}), specifying the clustering to be used for marker detection.
The first available clustering will be chosen.
If no clustering is available from the list, markers will not be computed.}

\item{more.rna.marker.args}{Named list of arguments to pass to \code{\link{scoreMarkers.se}} for the RNA data.
Ignored if no suitable clusterings are available or if \code{rna.altexp=NULL}.}

\item{more.adt.marker.args}{Named list of arguments to pass to \code{\link{scoreMarkers.se}} for the ADT data.
Ignored if no suitable clusterings are available or if \code{adt.altexp=NULL}.}

\item{more.crispr.marker.args}{Named list of arguments to pass to \code{\link{scoreMarkers.se}} for the CRISPR data.
Ignored if no suitable clusterings are available or if \code{crispr.altexp=NULL}.}

\item{BNPARAM}{A \link[BiocNeighbors]{BiocNeighborParam} instance specifying the nearest-neighbor search algorithm to use.}

\item{num.threads}{Integer scalar specifying the number of threads to use in each step.}
}
\value{
List containing:
\itemize{
\item \code{x}, a \link[SingleCellExperiment]{SingleCellExperiment} that is a copy of the input \code{x}.
It is also decorated with the results of each analysis step - see Details.
\item \code{markers}, a list of list of \link[S4Vectors]{DataFrame}s containing the marker statistics for each modality.
Each inner list corresponds to a modality (RNA, ADT, etc.) while each DataFrame corresponds to a cluster.
If no clusterings were generated, this is set to \code{NULL}.
}
}
\description{
Execute a simple single-cell analysis pipeline, starting from a count matrix and ending with clusters, visualizations and markers.
This also supports integration of multiple modalities and correction of batch effects.
}
\details{
This function is equivalent to:
\itemize{
\item Running \code{\link{quickRnaQc.se}}, \code{\link{quickAdtQc.se}} and/or \code{\link{quickCrisprQc.se}}, for quality control.
\item Subsetting \code{x} to only retain the high-quality cells in all modalities, based on \code{filter.cells}.
\item Running \code{\link{normalizeRnaCounts.se}}, \code{\link{normalizeAdtCounts.se}} and/or \code{\link{normalizeCrisprCounts.se}}, for normalization.
\item Running \code{\link{chooseRnaHvgs.se}} to identify highly variable genes.
\item Running \code{\link{runPca.se}} on the RNA and/or ADT data.
\item Running \code{\link{scaleByNeighbors.se}} if multiple modalities are present.
\item Running \code{\link{correctMnn.se}} if multiple batches are present in \code{block}.
\item Running \code{\link{runAllNeighborSteps.se}} to obtain t-SNE and UMAP coordinates, and to perform graph-based clustering.
\item Running \code{\link{clusterKmeans.se}} to perform k-means clustering, depending on \code{kmeans.clusters}. 
\item Running \code{\link{scoreMarkers.se}} to compute markers for each modality based on one of the clusterings. 
}
}
\examples{
library(SingleCellExperiment)
sce <- getTestRnaData.se("start")
res <- analyze.se(
    sce, 
    rna.qc.subsets=list(mito=grep("^mt-", rownames(sce))),
    num.threads=2 # keep R CMD check happy
)
assayNames(res$x)
reducedDimNames(res$x)
colData(res$x)
previewMarkers(res$markers$rna[[1]], "cohens.d.mean")

}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fitVarianceTrend.R
\name{fitVarianceTrend}
\alias{fitVarianceTrend}
\title{Fit a mean-variance trend}
\usage{
fitVarianceTrend(
  means,
  variances,
  mean.filter = TRUE,
  min.mean = 0.1,
  transform = TRUE,
  span = 0.3,
  use.min.width = FALSE,
  min.width = 1,
  min.window.count = 200,
  num.threads = 1
)
}
\arguments{
\item{means}{Numeric vector containing the mean (log-)expression for each gene.}

\item{variances}{Numeric vector containing the variance in the (log-)expression for each gene.}

\item{mean.filter}{Logical scalar indicating whether to filter on the means before trend fitting.
The assumption is that there is a bulk of low-abundance genes that are uninteresting and should be removed to avoid skewing the windows of the LOWESS smoother.}

\item{min.mean}{Numeric scalar specifying the minimum mean of genes to use in trend fitting.
Genes with lower means do not participate in the LOWESS fit, to ensure that windows are not skewed towards the majority of low-abundance genes.
Instead, the fitted values for these genes are defined by extrapolating the left edge of the fitted trend is extrapolated to the origin.
The default value is chosen based on the typical distribution of means of log-expression values across genes.
Only used if \code{mean.filter=TRUE}.}

\item{transform}{Logical scalar indicating whether a quarter-root transformation should be applied before trend fitting.
This transformation is copied from \code{limma::voom} and shrinks all values towards 1, flattening any sharp gradients in the trend for an easier fit.
The default of \code{TRUE} assumes that the variances are computed from log-expression values, in which case there is typically a strong \dQuote{hump} in the mean-variance relationship.}

\item{span}{Numeric scalar specifying the span of the LOWESS smoother, as a proportion of the total number of points.
Larger values improve stability at the cost of sensitivity to changes in low-density regions.
Ignored if \code{use.min.width=TRUE}.}

\item{use.min.width}{Logical scalar indicating whether a minimum width constraint should be applied to the LOWESS smoother.
This replaces the proportion-based span for defining each window.
Instead, the window for each point must be of a minimum width and is extended until it contains a minimum number of points. 
Setting this to `TRUE` ensures that sensitivity is maintained in the trend fit at low-density regions for the distribution of means, e.g., at high abundances.
It also avoids overfitting from very small windows in high-density intervals.}

\item{min.width}{Minimum width of the window to use when \code{use.min.width=TRUE}.
The default value is chosen based on the typical range of means in single-cell RNA-seq data.}

\item{min.window.count}{Minimum number of observations in each window.
This ensures that each window contains at least a given number of observations for a stable fit.
If the minimum width window contains fewer observations, it is extended using the standard LOWESS logic until the minimum number is achieved.
Only used if \code{use.min.width=TRUE}.}

\item{num.threads}{Number of threads to use.}
}
\value{
List containing \code{fitted}, a numeric vector containing the fitted values of the trend for each gene;
and \code{residuals}, a numeric vector containing the residuals from the trend.
}
\description{
Fit a trend to the per-gene variances with respect to their means, typically from normalized and log-transformed expression values.
}
\examples{
# Setting up some single-cell-like data.
mu <- 2^runif(1000, -10, 10)
counts <- matrix(rpois(20 * length(mu), lambda=mu), ncol=20)

sf <- centerSizeFactors(colSums(counts))
normalized <- normalizeCounts(counts, size.factors=sf)
stats <- modelGeneVariances(normalized)

out <- fitVarianceTrend(stats$statistics$means, stats$statistics$variances)
plot(stats$statistics$means, stats$statistics$variances)
curve(approxfun(stats$statistics$means, out$fitted)(x), col="red", add=TRUE)

}
\seealso{
\code{\link{modelGeneVariances}}, to compute the means and variances on which the trend is fitted.

The \code{fit_variance_trend} function in \url{https://libscran.github.io/scran_variances/}.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/countGroupsByBlock.R
\name{countGroupsByBlock}
\alias{countGroupsByBlock}
\title{Count cells in groups and blocks}
\usage{
countGroupsByBlock(
  groups,
  block,
  normalize.block = FALSE,
  normalize.groups = FALSE
)
}
\arguments{
\item{groups}{Factor specifying the group to which each cell was assigned.
This is typically used for clusters.}

\item{block}{Factor specifying the block to which each cell was assigned.
This is typically used for batches or samples.}

\item{normalize.block}{Boolean indicating whether to normalize the number of cells across blocks.
If \code{TRUE}, frequencies are divided by the column sums.}

\item{normalize.groups}{Boolean indicating whether to normalize the number of cells across groups.
If \code{TRUE}, frequencies are divided by the row sums.
This is performed after normalization of the block counts if \code{normalize.block=TRUE}.}
}
\value{
Matrix of (normalized) frequencies.
Each row corresponds to a group and each column corresponds to a block.
}
\description{
Tabulate the frequency of cells in each combination of groups and blocks.
This is typically used to examine the distribution of cells across batches for each cluster -
the presence of a batch-specific cluster may be indicative of a batch effect.
}
\examples{
groups <- sample(10, 100, replace=TRUE)
block <- sample(LETTERS[1:6], 100, replace=TRUE)

countGroupsByBlock(groups, block)
countGroupsByBlock(groups, block, normalize.block=TRUE)
countGroupsByBlock(groups, block, normalize.groups=TRUE)
countGroupsByBlock(groups, block, normalize.block=TRUE, normalize.groups=TRUE)

}
\seealso{
\code{\link{table}}, which is used internally by this function.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/subsampleByNeighbors.R
\name{subsampleByNeighbors}
\alias{subsampleByNeighbors}
\title{Subsample cells based on their neighbors}
\usage{
subsampleByNeighbors(
  x,
  num.neighbors = 20,
  min.remaining = 10,
  num.threads = 1,
  BNPARAM = AnnoyParam()
)
}
\arguments{
\item{x}{A numeric matrix where rows are dimensions and columns are cells,
typically containing a low-dimensional representation from, e.g., \code{\link{runPca}}.

Alternatively, an index constructed by \code{\link[BiocNeighbors]{buildIndex}}.

Alternatively, a list containing existing nearest-neighbor search results.
This should contain:
\itemize{
\item \code{index}, an integer matrix where rows are neighbors and columns are cells.
Each column contains 1-based indices for the nearest neighbors of the corresponding cell, ordered by increasing distance.
\item \code{distance}, a numeric matrix of the same dimensions as \code{index},
containing the distances to each of the nearest neighbors.
}
The number of neighbors should be equal to \code{num.neighbors}, otherwise a warning is raised.}

\item{num.neighbors}{Integer scalar specifying the number of neighbors to use.
Larger values result in stronger downsampling. 
Ignored if \code{x} contains pre-computed neighbor search results.}

\item{min.remaining}{Integer scalar specifying the minimum number of remaining neighbors that a cell must have in order to be considered for selection.
This should be less than or equal to \code{num.neighbors}.
Larger values result in stronger downsampling.}

\item{num.threads}{Integer scalar specifying the number of threads to use for the nearest-neighbor search.
Only used if \code{x} does not contain existing nearest-neighbor results.}

\item{BNPARAM}{A \link[BiocNeighbors]{BiocNeighborParam} object specifying the algorithm to use.
Only used if \code{x} does not contain existing nearest-neighbor results.}
}
\value{
Integer vector with the indices of the selected cells in the subsample.
}
\description{
Subsample a dataset by selecting cells to represent all of their nearest neighbors.
The aim is to preserve the relative density of the original dataset while guaranteeing representation of low-frequency subpopulations.
}
\details{
Starting from the densest region in the high-dimensional space, we select an observation for inclusion into the subsampled dataset.
Every time we select an observation, we remove it and all of its nearest neighbors from the dataset.
We then select the next observation with the most remaining neighbors, with ties broken by density; this is repeated until there are no more observations.

The premise is that each selected observation serves as a representative for its nearest neighbors.
This ensures that the subsampled points are well-distributed across the original dataset.
Low-frequency subpopulations will always have at least a few representatives if they are sufficiently distant from other subpopulations.
We also preserve the relative density of the original dataset as more representatives will be generated from high-density regions.
}
\examples{
x <- matrix(rnorm(10000), nrow=2)
keep <- subsampleByNeighbors(x, 10)
plot(x[1,], x[2,])
points(x[1,keep], x[2,keep], col="red")
legend('topright', col=c('black', 'red'), legend=c('all', 'subsample'), pch=1)

}
\seealso{
\url{https://libscran.github.io/nenesub/}, for more details on the underlying algorithm.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/se_scaleByNeighbors.R
\name{scaleByNeighbors.se}
\alias{scaleByNeighbors.se}
\title{Scale and combine multiple embeddings in a SingleCellExperiment}
\usage{
scaleByNeighbors.se(
  x,
  altexp.reddims,
  main.reddims = "PCA",
  num.neighbors = 20,
  block = NULL,
  BNPARAM = AnnoyParam(),
  num.threads = 1,
  more.scale.args = list(),
  output.name = "combined",
  meta.name = "combined",
  delayed.transpose = FALSE
)
}
\arguments{
\item{x}{A \link[SingleCellExperiment]{SingleCellExperiment} object or one of its subclasses.
Rows correspond to genomic features and columns correspond to cells.}

\item{altexp.reddims}{Named list of character or integer vectors.
Each entry is named after an alternative experiment.
Each vector contains the names/indices of the \code{\link[SingleCellExperiment]{reducedDim}} embeddings from that experiment to be combined.}

\item{main.reddims}{Character or integer vector specifying the names/indices of the \code{\link[SingleCellExperiment]{reducedDim}} entries from \code{x} to be combined.}

\item{num.neighbors}{Number of neighbors used to define the scaling factor, passed to \code{\link{scaleByNeighbors}}.}

\item{block}{Block assignment for each cell, passed to \code{\link{scaleByNeighbors}}.}

\item{BNPARAM}{Algorithm for the nearest neighbor search, passed to \code{\link{scaleByNeighbors}}.}

\item{num.threads}{Number of threads for the neighbor search, passed to \code{\link{scaleByNeighbors}}.}

\item{more.scale.args}{Named list of additional arguments to pass to \code{\link[scrapper]{scaleByNeighbors}}.}

\item{output.name}{String containing the name of the \code{\link[SingleCellExperiment]{reducedDim}} entry in which to store the combined embeddings.}

\item{meta.name}{String containing the name of the \code{\link[S4Vectors]{metadata}} entry in which to store additional metrics.
If \code{NULL}, additional metrics are not stored.}

\item{delayed.transpose}{Logical scalar indicating whether to delay the transposition when storing coordinates in the \code{\link[SingleCellExperiment]{reducedDims}}.}
}
\value{
\code{x} is returned with the combined embeddings stored in its \code{reducedDims}.
The scaling factors for all embeddings are stored in the \code{metadata}.
}
\description{
Scale embeddings for different modalities to equalize their intra-population variance, and combine them into a single embedding for downstream analysis.
This calls \code{\link{scaleByNeighbors}} on the reduced dimensions of the main/alternative experiments in a \link[SingleCellExperiment]{SingleCellExperiment}.
}
\examples{
library(SingleCellExperiment)
sce <- getTestAdtData.se("pca")
sce <- scaleByNeighbors.se(sce, altexp.reddims=list(ADT="PCA"))
reducedDimNames(sce) 
metadata(sce)$combined

}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clusterKmeans.R
\name{clusterKmeans}
\alias{clusterKmeans}
\title{K-means clustering}
\usage{
clusterKmeans(
  x,
  k,
  init.method = c("var-part", "kmeans++", "random"),
  refine.method = c("hartigan-wong", "lloyd"),
  var.part.optimize.partition = TRUE,
  var.part.size.adjustment = 1,
  lloyd.iterations = 100,
  hartigan.wong.iterations = 10,
  hartigan.wong.quick.transfer.iterations = 50,
  hartigan.wong.quit.quick.transfer.failure = FALSE,
  seed = 5489L,
  warn = TRUE,
  num.threads = 1
)
}
\arguments{
\item{x}{Matrix-like object where rows are dimensions and columns are cells.
This is typically a dense double-precision matrix containing a low-dimensional representation from, e.g., \code{\link{runPca}}.
However, any matrix representation supported by \code{\link[beachmat]{initializeCpp}} can also be used.}

\item{k}{Integer scalar specifying the number of clusters.}

\item{init.method}{String specifying the initialization method for the centers:
\itemize{
\item \code{"var-part"} uses variance partitioning as described by Su and Dy (2007).
The dataset is repeatedly split along the dimension of greatest variance until \code{k} partitions are formed, the centroids of which form the initial clusters.
This approach is slower than the others but fully deterministic.
\item \code{"kmeans++"} uses the weighted sampling method described by Arthur and Vassilvitskii (2007).
\code{k} points are sampled with probability based on the smallest distance to any previously sampled point.
This improves the likelihood of choosing initial centroids that are far apart from each other.
\item \code{"random"} initialization involves choosing \code{k} random points as the initial centers.
This is the simplest and fastest method but may not yield good starting points. 
}}

\item{refine.method}{String specifying the refinement method.
\itemize{
\item \code{"lloyd"} uses Lloyd's algorithm, which performs a batch update in each iteration.
This is simple and amenable to parallelization but may not converge.
\item \code{"hartigan-wong"} uses the Hartigan-Wong algorithm, which transfers points between clusters to optimize the drop in the within-cluster sum of squares.
This is slower but has a greater chance of convergence.
}}

\item{var.part.optimize.partition}{Logical scalar indicating whether each partition boundary should be optimized to reduce the sum of squares in the child partitions.
This is slower but improves the quality of the partition.
Only used if \code{init.method = "var.part"}.}

\item{var.part.size.adjustment}{Numeric scalar between 0 and 1, specifying the adjustment to the cluster size when selecting the next cluster to partition.
Setting this to 0 or 1 will select the cluster with the highest variance or sum of squares, respectively, for partitioning.
In other words, a value of 0 will ignore the cluster size while setting a value of 1 will generally cause larger clusters to be selected. 
Only used if \code{init.method = "var.part"}.}

\item{lloyd.iterations}{Integer scalar specifying the maximum number of iterations for the Lloyd algorithm.
Larger values increase the chance of convergence at the cost of increasing compute time.
Only used if \code{refine.method = "lloyd"}.}

\item{hartigan.wong.iterations}{Integer scalar specifying the maximum number of iterations for the Hartigan-Wong algorithm.
Larger values increase the chance of convergence at the cost of increasing compute time.
Only used if \code{refine.method = "hartigan-wong"}.}

\item{hartigan.wong.quick.transfer.iterations}{Integer scalar specifying the maximum number of quick transfer iterations for the Hartigan-Wong algorithm.
Larger values increase the chance of convergence at the cost of increasing compute time.
Only used if \code{refine.method = "hartigan-wong"}.}

\item{hartigan.wong.quit.quick.transfer.failure}{Logical scalar indicating whether to quit the Hartigan-Wong algorithm upon convergence failure during quick transfer iterations.
Setting this to \code{FALSE} gives the algorithm another chance to converge by attempting another optimal transfer iteration, at the cost of more compute time.
If \code{TRUE}, the function follows the same behavior as R's \code{\link{kmeans}}.
Only used if \code{refine.method = "hartigan-wong"}.}

\item{seed}{Integer scalar specifying the seed for random number generation.
Only used if \code{init.method = "random"} or \code{"kmeans++"}.}

\item{warn}{Boolean specifying whether a warning should be emitted if the k-means algorithm failed to converge.}

\item{num.threads}{Integer scalar specifying the number of threads to use.}
}
\value{
By default, a list is returned containing:
\itemize{
\item \code{clusters}, a factor containing the cluster assignment for each cell.
The number of levels is no greater than \code{k}, where each level is an integer that refer to a column of \code{centers}.
\item \code{centers}, a numeric matrix with the coordinates of the cluster centroids (dimensions in rows, centers in columns).
The number of columns is no greater than \code{k}.
Empty clusters are automatically removed.
\item \code{iterations}, an integer scalar specifying the number of refinement iterations that were performed.
\item \code{status}, an integer scalar specifying the completion status of the algorithm.
A value of zero indicates success while the meaning of any non-zero value depends on the choice of \code{refine.method}:
\itemize{
\item For Lloyd, a value of 2 indicates convergence failure.
\item For Hartigan-Wong, a value of 2 indicates convergence failure in the optimal transfer iterations.
A value of 4 indicates convergence failure in the quick transfer iterations when \code{hartigan.wong.quit.quick.transfer.failure = TRUE}.
}
}
}
\description{
Perform k-means clustering with a variety of different initialization and refinement algorithms.
}
\examples{
x <- t(as.matrix(iris[,1:4]))
clustering <- clusterKmeans(x, k=3)
table(clustering$clusters, iris[,"Species"])

}
\references{
Hartigan JA. and Wong MA (1979).
Algorithm AS 136: A K-means clustering algorithm.
\emph{Applied Statistics} 28, 100-108.

Arthur D and Vassilvitskii S (2007).
k-means++: the advantages of careful seeding.
\emph{Proceedings of the eighteenth annual ACM-SIAM symposium on Discrete algorithms} 1027-1035.

Su T and Dy JG (2007).
In Search of Deterministic Methods for Initializing K-Means and Gaussian Mixture Clustering.
\emph{Intelligent Data Analysis} 11, 319-338.
}
\seealso{
\url{https://libscran.github.io/kmeans/}, which describes the various initialization and refinement algorithms in more detail.

\code{\link{clusterKmeans.se}}, for k-means clustering on a \link[SingleCellExperiment]{SingleCellExperiment}.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/centerSizeFactors.R
\name{centerSizeFactors}
\alias{centerSizeFactors}
\title{Center size factors}
\usage{
centerSizeFactors(size.factors, block = NULL, mode = c("lowest", "per-block"))
}
\arguments{
\item{size.factors}{Numeric vector of size factors across cells.
Invalid size factors (e.g., non-positive, non-finite) will be ignored.}

\item{block}{Vector or factor of length equal to \code{size.factors}, specifying the block of origin for each cell.
Alternatively \code{NULL}, in which case all cells are assumed to be in the same block.}

\item{mode}{String specifying how to scale size factors across blocks.
This can be either \code{"lowest"} or \code{"per-block"}, see Details.
Only used if \code{block} is provided.}
}
\value{
Numeric vector of length equal to \code{size.factors}, containing the centered size factors.
}
\description{
Scale the size factors so they are centered at unity.
This ensures that the original scale of the counts is preserved in the normalized values from \code{\link{normalizeCounts}},
which simplifies interpretation and ensures that any pseudo-count added prior to log-transformation has a predictable shrinkage effect.
}
\details{
\code{"lowest"} will compute the average size factor in each block, identify the lowest average across all blocks, and then scale all size factors by that value.
Here, our normalization strategy involves downscaling all blocks to match the coverage of the lowest-coverage block.
This is useful for datasets with big differences in coverage between blocks as it avoids egregious upscaling of low-coverage blocks.
Specifically, strong upscaling allows the log-transformation to ignore any shrinkage from the pseudo-count.
This is problematic as it inflates differences between cells at log-values derived from low counts, increasing noise and overstating log-fold changes. 
Downscaling is safer as it allows the pseudo-count to shrink the log-differences between cells towards zero at low counts,
effectively sacrificing some information in the higher-coverage batches so that they can be compared to the low-coverage batches
(which is preferable to exaggerating the informativeness of the latter for comparison to the former).

\code{"per-block"} will compute the average size factor in each block, and then scale each size factor by the average of block to which it belongs.
The scaled size factors are identical to those obtained by separate invocations of `center_size_factors()` on the size factors for each block.
This can be desirable to ensure consistency with independent analyses of each block - otherwise, the centering would depend on the size factors in other blocks.
However, any systematic differences in the size factors between blocks are lost, i.e., systematic changes in coverage between blocks will not be normalized.
}
\examples{
centerSizeFactors(runif(100))

centerSizeFactors(runif(100), block=sample(3, 100, replace=TRUE))

}
\seealso{
The \code{center_size_factors} and \code{center_size_factors_blocked} functions in \url{https://libscran.github.io/scran_norm/}.

\code{\link{normalizeRnaCounts.se}} and related functions, which center the size factors prior to normalization of a \link[SummarizedExperiment]{SummarizedExperiment}.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/computeClrm1Factors.R
\name{computeClrm1Factors}
\alias{computeClrm1Factors}
\title{Compute size factors for ADT counts}
\usage{
computeClrm1Factors(x, num.threads = 1)
}
\arguments{
\item{x}{A matrix-like object containing ADT count data.
Rows correspond to tags and columns correspond to cells.}

\item{num.threads}{Number of threads to use.}
}
\value{
Numeric vector containing the CLRm1 size factor for each cell.
Note that these size factors are not centered and should be passed through, e.g., \code{\link{centerSizeFactors}} before normalization.
}
\description{
Compute size factors from an ADT count matrix using the CLRm1 method.
This is a variant of the centered log-ratio (CLR) method, where the size factors are defined from the geometric mean of counts within each cell.
}
\examples{
library(Matrix)
x <- abs(rsparsematrix(1000, 100, 0.1) * 10)
head(computeClrm1Factors(x))

}
\seealso{
\url{https://github.com/libscran/clrm1}, for a description of the CLRm1 method.

\code{\link{normalizeAdtCounts.se}}, which computes CLRm1 factors prior to normalization.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rna_quality_control.R
\name{rna_quality_control}
\alias{rna_quality_control}
\alias{computeRnaQcMetrics}
\alias{suggestRnaQcThresholds}
\alias{filterRnaQcMetrics}
\title{Quality control for RNA count data}
\usage{
computeRnaQcMetrics(x, subsets, num.threads = 1)

suggestRnaQcThresholds(metrics, block = NULL, num.mads = 3)

filterRnaQcMetrics(thresholds, metrics, block = NULL)
}
\arguments{
\item{x}{A matrix-like object where rows are genes and columns are cells.
Values are expected to be counts.}

\item{subsets}{Named list of vectors specifying gene subsets of interest, typically for control-like features like mitochondrial genes or spike-in transcripts.
Each vector may be logical (whether to keep each row), integer (row indices) or character (row names).
For character vectors, strings not present in \code{rownames(x)} are ignored.}

\item{num.threads}{Integer scalar specifying the number of threads to use.}

\item{metrics}{\link[S4Vectors]{DataFrame} of per-cell QC metrics.
This should have the same structure as the return value of \code{computeRnaQcMetrics}.}

\item{block}{Factor specifying the block of origin (e.g., batch, sample) for each cell in \code{metrics}.
Alternatively \code{NULL} if all cells are from the same block.

For \code{filterRnaQcMetrics}, a blocking factor should be provided if \code{block} was used to construct \code{thresholds}.}

\item{num.mads}{Number of median from the median, to define the threshold for outliers in each metric.}

\item{thresholds}{List with the same structure as produced by \code{suggestRnaQcThresholds}.}
}
\value{
For \code{computeRnaQcMetrics}, a \link[S4Vectors]{DataFrame} is returned with one row per cell in \code{x}.
This contains the following columns:
\itemize{
\item \code{sum}, a numeric vector containing the total RNA count for each cell.
This represents the efficiency of library preparation and sequencing.
Low totals indicate that the library was not successfully captured.
\item \code{detected}, an integer vector containing the number of detected genes per cell.
This also quantifies library preparation efficiency but with greater focus on capturing transcriptional complexity.
\item \code{subsets}, a nested DataFrame where each column corresponds to a feature subset and is a numeric vector containing the proportion of counts in that subset.
The exact interpretation of which depends on the nature of the subset.
For example, if one subset contains all genes on the mitochondrial chromosome, higher proportions are representative of cell damage;
the assumption is that cytoplasmic transcripts leak through tears in the cell membrane while the mitochondria are still trapped inside.
The proportion of spike-in transcripts can be interpreted in a similar manner, where the loss of endogenous transcripts results in higher spike-in proportions.
}
Each vector is of length equal to the number of cells.

For \code{suggestRnaQcThresholds}, a named list is returned.
\itemize{
\item If \code{block=NULL}, the list contains:
\itemize{
\item \code{sum}, a numeric scalar containing the lower bound on the sum.
This is defined as \code{num.mads} MADs below the median of the log-transformed metrics across all cells.
\item \code{detected}, a numeric scalar containing the lower bound on the number of detected genes. 
This is defined as \code{num.mads} MADs below the median of the log-transformed metrics across all cells.
\item \code{subsets}, a numeric vector containing the upper bound on the sum of counts in each feature subset. 
This is defined as \code{num.mads} MADs above the median across all cells.
}
\item Otherwise, if \code{block} is supplied, the list contains:
\itemize{
\item \code{sum}, a numeric vector containing the lower bound on the sum for each blocking level.
Here, the threshold is computed independently for each block, using the same method as the unblocked case.
\item \code{detected}, a numeric vector containing the lower bound on the number of detected genes for each blocking level.
Here, the threshold is computed independently for each block, using the same method as the unblocked case.
\item \code{subsets}, a list of numeric vectors containing the upper bound on the sum of counts in each feature subset for each blocking level.
Here, the threshold is computed independently for each block, using the same method as the unblocked case.
\item \code{block.ids}, a vector containing the identities of the unique blocks.
}
Each vector is of length equal to the number of levels in \code{block} and is named accordingly.
}

For \code{filterRnaQcMetrics}, a logical vector of length \code{ncol(x)} is returned indicating which cells are of high quality. 
High-quality cells are defined as those with sums and detected genes above their respective thresholds and subset proportions below the \code{subsets} threshold.
}
\description{
Compute per-cell QC metrics from an initialized matrix of RNA counts,
and use the metrics to suggest filter thresholds to retain high-quality cells.
}
\examples{
# Mocking a matrix:
library(Matrix)
x <- round(abs(rsparsematrix(1000, 100, 0.1) * 100))

# Mocking up a control set.
sub <- list(mito=rbinom(nrow(x), 1, 0.1) > 0)

qc <- computeRnaQcMetrics(x, sub)
qc

filt <- suggestRnaQcThresholds(qc)
str(filt)

keep <- filterRnaQcMetrics(filt, qc)
summary(keep)

}
\seealso{
The \code{compute_rna_qc_metrics}, \code{compute_rna_qc_filters} and \code{compute_rna_qc_filters_blocked} functions in \url{https://libscran.github.io/scran_qc/}.

\code{\link{quickRnaQc.se}}, to run all of the RNA-related QC functions on a \link[SummarizedExperiment]{SummarizedExperiment}.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/crispr_quality_control.R
\name{crispr_quality_control}
\alias{crispr_quality_control}
\alias{computeCrisprQcMetrics}
\alias{suggestCrisprQcThresholds}
\alias{filterCrisprQcMetrics}
\title{Quality control for CRISPR count data}
\usage{
computeCrisprQcMetrics(x, num.threads = 1)

suggestCrisprQcThresholds(metrics, block = NULL, num.mads = 3)

filterCrisprQcMetrics(thresholds, metrics, block = NULL)
}
\arguments{
\item{x}{A matrix-like object where rows are CRISPRs and columns are cells.
Values are expected to be counts.}

\item{num.threads}{Integer scalar specifying the number of threads to use.}

\item{metrics}{\link[S4Vectors]{DataFrame} of per-cell QC metrics.
This should have the same structure as the return value of \code{computeCrisprQcMetrics}.}

\item{block}{Factor specifying the block of origin (e.g., batch, sample) for each cell in \code{metrics}.
Alternatively \code{NULL} if all cells are from the same block.

For \code{filterCrisprQcMetrics}, a blocking factor should be provided if \code{block} was used to construct \code{thresholds}.}

\item{num.mads}{Number of median from the median, to define the threshold for outliers in each metric.}

\item{thresholds}{List with the same structure as produced by \code{suggestCrisprQcThresholds}.}
}
\value{
For \code{computeCrisprQcMetrics}, a \link[S4Vectors]{DataFrame} is returned with one row per cell in \code{x}.
This contains the following columns:
\itemize{
\item \code{sum}, a numeric vector containing the total CRISPR count for each cell.
Low counts indicate that the cell was not successfully transfected with a construct or that library preparation and sequencing failed.
\item \code{detected}, an integer vector containing the number of detected guides per cell.
In theory, this should be 1, as each cell should express no more than one guide construct.
However, ambient contamination may introduce non-zero counts for multiple guides, without necessarily interfering with downstream analyses.
As such, this metric is less useful for guide data, though we compute it anyway.
\item \code{max.value}, a numeric vector containing the count for the most abundant guide in cell.
Low values indicate that the cell was not successfully transfected or that library preparation and sequencing failed.
\item \code{max.index}, an integer vector containing the row index of the most abundant guide in cell.
}
Each vector is of length equal to the number of cells.

For \code{suggestCrisprQcThresholds}, a named list is returned.
\itemize{
\item If \code{block=NULL}, the list contains:
\itemize{
\item \code{max.value}, a numeric scalar containing the lower bound on the maximum count. 
This is defined as \code{num.mads} MADs below the median of the log-transformed metrics across cells with high maximum proportions (see Details).
}
\item Otherwise, if \code{block} is supplied, the list contains:
\itemize{
\item \code{max.value}, a numeric vector containing the lower bound on the maximum counts for each blocking level.
Here, the threshold is computed independently for each block, using the same method as the unblocked case.
\item \code{block.ids}, a vector containing the identities of the unique blocks.
}
Each vector is of length equal to the number of levels in \code{block} and is named accordingly.
}

For \code{filterCrisprQcMetrics}, a logical vector of length \code{ncol(x)} is returned indicating which cells are of high quality. 
High-quality cells are defined as those with maximum counts above the \code{max.value} threshold.
}
\description{
Compute per-cell QC metrics from an initialized matrix of CRISPR counts,
and use the metrics to suggest filter thresholds to retain high-quality cells.
}
\details{
In CRISPR data, a cell is considered to be of low quality if it has a low count for its most abundant guide.
However, directly defining a MAD-based outlier threshold on the maximum count is somewhat tricky as unsuccessful transfection can be common.
This often results in a large subpopulation with low maximum counts, inflating the MAD and compromising the threshold calculation.
Instead, we use the following approach:
\itemize{
\item Compute the proportion of counts in the most abundant guide (i.e., the maximum proportion) in each cell.
Cells that were successfully transfected should have high maximum proportions.
In contrast, unsuccessfully transfected cells will be dominated by ambient contamination and have low proportions.
\item Subset the dataset to only retain those cells with maximum proportions above the median.
This assumes that at least 50% of cells were successfully transfected.
Thus, we remove all of the unsucessful transfections and enrich for mostly-high-quality cells.
\item Define a MAD-based threshold for low outliers on the log-transformed maximum count within the subset (see `choose_filter_thresholds()` for details).
This is now possible as we can assume that most of the remaining cells are of high quality.
}
Note that the maximum proportion is only used to define the subset for threshold calculation.
Once the maximum count threshold is computed, it is applied to all cells regardless of their maximum proportions.
This ensures that we correctly remove cells with low coverage, even if the proportion is high.
It also allows us to retain cells transfected with multiple guides, as long as the maximum is high enough -
such cells are not necessarily uninteresting, e.g., for examining interaction effects, so we will err on the side of caution and leave them in.
}
\examples{
# Mocking a matrix:
library(Matrix)
x <- round(abs(rsparsematrix(100, 100, 0.1) * 100))

qc <- computeCrisprQcMetrics(x)
qc

filt <- suggestCrisprQcThresholds(qc)
str(filt)

keep <- filterCrisprQcMetrics(filt, qc)
summary(keep)

}
\seealso{
The \code{compute_crispr_qc_metrics}, \code{compute_crispr_qc_filters} and \code{compute_crispr_qc_filters_blocked} functions in \url{https://libscran.github.io/scran_qc/}.

\code{\link{quickCrisprQc.se}}, to run all of the CRISPR-related QC functions on a \link[SummarizedExperiment]{SummarizedExperiment}.
}
\author{
Aaron Lun
}
